package bingo.base;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Vector;
import java.net.URL;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextField;
import javax.swing.JTree;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeSelectionModel;

import bingo.crawler.handler.UrlVerifier;
import bingo.crawler.frontier.url2resolve;
import bingo.language.LanguageConstants;
import bingo.util.BINGODesktop;
import bingo.util.BingoDocument;
import bingo.util.BingoTreeNode;
import bingo.util.SessionBuffer;


/**
 *  Die Klasse dient zum manuellen Einfgen von Trainingsdaten in die
 *  Trainingsdatentabelle (Klasse BaseFrame). Die URL wird in ein Textfenster
 *  eingegeben. Die Kategorie kann aus einem graphisch dargestellten
 *  Hierarchiebaum ausgewhlt werden.
 *
 * @author     T.
 * @created    15. Juni 2001
 */
public class BaseDialog extends JDialog implements ActionListener
{
   
   private static LanguageConstants lc = LanguageConstants.getInstance();
   
   private SessionBuffer session = SessionBuffer.getInstance();
   private JButton ok = new JButton( lc.get( "insert" ) );
   private JButton cancel = new JButton( lc.get( "cancel" ) );
   private JTextField text1 = new JTextField( 30 );
   private JScrollPane treePane;
   private JButton label1 = new JButton( " " + lc.get( "baseDialogURL" ) + " " );
   private JButton label2 = new JButton( "<" + lc.get( "baseDialogCategory" ) + ">" );
   private JPanel panel1 = new JPanel();
   private JTree tree;
   private DefaultTreeModel treeModel;
   private BaseFrame parentFrame;
   private BingoTreeNode node;
   
   private static String dialogTitle = lc.get( "baseDialogTitle" );
   
   
   /**
    *  Konstruktor fr das BaseDialog Objekt
    *
    * @param  parentFrame  Description of Parameter
    */
   public BaseDialog( BaseFrame parentFrame )
   {
      super( BINGODesktop.getInstance(), dialogTitle, false );
      this.parentFrame = parentFrame;
      this.setSize(500, 300);
      Container content = this.getContentPane();
      
      content.setLayout( new BorderLayout() );
      
      treeModel = parentFrame.getTreeModel();
      tree = new JTree(treeModel);
      
      tree.getSelectionModel().setSelectionMode( TreeSelectionModel.SINGLE_TREE_SELECTION );
      tree.addTreeSelectionListener(
      new TreeSelectionListener()
      {
         public void valueChanged( TreeSelectionEvent e )
         {
            node = (BingoTreeNode) tree.getLastSelectedPathComponent();
            
            if ( node == null )
            {
               return;
            }
         }
      } );
      
      treePane = new JScrollPane( tree );
      
      JSplitPane splitPane = new JSplitPane( JSplitPane.VERTICAL_SPLIT );
      
      splitPane.setDividerSize( 5 );
      splitPane.setContinuousLayout( true );
      splitPane.setTopComponent( text1 );
      splitPane.setBottomComponent( treePane );
      splitPane.setDividerLocation( 35 );
      
      text1.setBackground( Color.gray );
      text1.setForeground( Color.white );
      
      treePane.setBackground( Color.gray );
      treePane.setForeground( Color.white );
      
      JSplitPane header = new JSplitPane( JSplitPane.VERTICAL_SPLIT );
      
      header.setDividerSize( 5 );
      header.setContinuousLayout( true );
      header.setTopComponent( label1 );
      header.setBottomComponent( label2 );
      header.setDividerLocation( 35 );
      
      JSplitPane mainPane = new JSplitPane( JSplitPane.HORIZONTAL_SPLIT );
      
      mainPane.setDividerSize( 5 );
      mainPane.setContinuousLayout( true );
      mainPane.setLeftComponent( header );
      mainPane.setRightComponent( splitPane );
      mainPane.setDividerLocation( 120 );
      
      panel1.setLayout( new FlowLayout() );
      panel1.add( ok );
      panel1.add( cancel );
      
      content.add( mainPane, BorderLayout.CENTER );
      content.add( panel1, BorderLayout.SOUTH );
      
      ok.addActionListener( this );
      label1.addActionListener( this );
      label2.addActionListener( this );
      cancel.addActionListener( this );
      
      setLocationRelativeTo( parentFrame );
      
   }
   
   
   /**
    *  Liefert die eingegebene URL als String.
    *
    * @return    Die Kategorie.
    */
   public String getURL()
   {
      return text1.getText();
   }
   
   
   public void setSelectedIndex( int i )
   {
   }
   
   /**
    *  Description of the Method
    *
    * @param  e  Description of Parameter
    */
   public void actionPerformed( ActionEvent e )
   {
      String cmd = e.getActionCommand();
      
      try
      {
         if ( cmd == ( lc.get( "cancel" ) ) )
         {
            text1.setText( "" );
            super.dispose();
         }
         else if ( cmd == ( lc.get( "insert" ) ) )
         {
            insertURL();
         }/*
            else if ( cmd.equals( "<" + lc.get( "baseDialogCategory" ) + ">" ) ) {
           
            }
            else if ( cmd.equals( " " + lc.get( "baseDialogURL" ) + " " ) ) {
           
            }*/
      }
      catch ( Exception any )
      {
         session.log(any);
      }
   }
   
   public boolean insertURL()
   {
      if (session == null)
         return false;
      
      String url = text1.getText();
      if (url == null || url.length() <= 0)
         return false;
      
      BingoDocument bd = UrlVerifier.string2BingoDocument(url);
      url = null;
      if (bd == null)
         return false;
      
      bd.setNode(node);
      bd.setStatus('B');
      bd.setContentType(UrlVerifier.guessContentTypeFromName(bd.getUrl2r().getURL().getFile()));
      bd.setTitle(url);
      node.addDocument(bd);
      
      Vector newURLs = new Vector();
      newURLs.addElement(bd);
      FeatureDialog fd = new FeatureDialog( BINGODesktop.getInstance(), newURLs );
      fd.setRemoveFlag(false);
      fd.setVisible(true);
      this.dispose();
      
      parentFrame.updateTree(session.getRootNode());
      this.dispose();
      return true;
   }
   
}

