
package bingo.base;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.JTree;
import javax.swing.ListSelectionModel;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.table.TableColumn;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeSelectionModel;

import bingo.crawler.CrawlerFrame;
import bingo.language.LanguageConstants;
import bingo.util.BINGODesktop;
import bingo.util.BingoDocument;
import bingo.util.BingoTreeNode;
import bingo.util.BookmarkParser;
import bingo.util.ParserFrame;
import bingo.util.SessionBuffer;
import bingo.util.TreeUpdateListener;
import bingo.util.URLClassTableModel;

/**
 *  Zentrale Klasse zur Verwaltung der Trainingsdatenbasis des
 *  SVM-Klassifikators und graphische Benutzerschnittstelle in Form einer
 *  Tabelle. Es werden jeweils die URL, die Kategorie, und ob ein Featurevektor
 *  berechnet wurde, dargestellt. Weiterhin knnen von diesem Frame aus
 *  Featurevektoren berechnet werden. Es knnen Trainigsdaten in Dateien
 *  gespeichert, neu geladen und manipuliert werden. Es kann ein Crawler ber
 *  der aktuellen Datenbasis gestartet werden.
 */

public class BaseFrame extends JInternalFrame implements TreeUpdateListener, ActionListener
{
   
   private static LanguageConstants lc = LanguageConstants.getInstance();
   
   private SessionBuffer session = SessionBuffer.getInstance();
   
   private JMenuBar menuBar;
   private JPanel panel1 = new JPanel();
   private JPanel panel2 = new JPanel();
   private JButton bCrawl = new JButton( lc.get( "crawlerButton" ) );
   private JButton bNew = new JButton( lc.get( "newButton" ) );
   private JButton bRemove = new JButton( lc.get( "delButton" ) );
   private JButton bFeatures = new JButton( lc.get( "computeButton" ) );
   private JButton bView = new JButton( lc.get( "sourceButton" ) );
   private JButton bBrowser = new JButton( lc.get( "browserButton" ) );
   private JButton bShowFeat = new JButton( lc.get( "featureButton" ) );
   private JButton bShowLinks = new JButton( lc.get( "linkButton" ) );
   
   private static final String[] columnNames =
   {"URL", "STATUS", "TITLE", "MIME", "FEATURES"};
   public URLClassTableModel baseTable;
   public final JTable table;
   
   private Vector vURLs = null;
   
   private BingoTreeNode top;
   private DefaultTreeModel treeModel;
   private final JTree tree;
   
   private JLabel label3 = new JLabel();
   
   private JSplitPane splitPane = new JSplitPane( JSplitPane.HORIZONTAL_SPLIT );
   private Font f = new Font( "Serif", Font.ITALIC, 16 );
   
   
   protected BaseFrame(String myTitle)
   {
      super( myTitle,
      true,
      //resizable
      true,
      //closable
      true,
      //maximizable
      true );
      //iconifiable
      
      setDefaultCloseOperation( HIDE_ON_CLOSE );
      
      session.registerForTreeUpdating( this );
      
      // Aufbau der GUI
      Container content = this.getContentPane();
      
      content.setLayout( new BorderLayout() );
      baseTable =
      new URLClassTableModel()
      {
         public boolean isCellEditable( int x, int y )
         {
            return false;
         }
      };
      baseTable.setColumnNames( columnNames );
      table = new JTable( baseTable );
      
      table.setPreferredScrollableViewportSize( new Dimension( 625, 620 ) );
      table.setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
      
      TableColumn column = null;
      
      for ( int i = 0; i < 5; i++ )
      {
         column = table.getColumnModel().getColumn( i );
         if ( i == 0 )
         {
            column.setPreferredWidth( 175 );
         } else if ( i == 2 )
         {
            column.setPreferredWidth( 175 );
         } else
         {
            column.setPreferredWidth( 50 );
         }
      }
      
      table.setBackground( Color.gray );
      table.setForeground( Color.white );
      
      panel1.setLayout( new BorderLayout() );
      panel2.setLayout( new FlowLayout() );
      
      panel2.add( bCrawl );
      panel2.add( bNew );
      panel2.add( bRemove );
      panel2.add( bView );
      panel2.add( bBrowser );
      panel2.add( bShowFeat );
      panel2.add( bShowLinks );
      panel2.add( bFeatures );
      
      JScrollPane tableScrollPane = new JScrollPane( table );
      
      // erstelle Baum
      top = new BingoTreeNode( "ROOT" );
      treeModel = new DefaultTreeModel( top );
      
      updateTree( top );
      
      tree = new JTree( treeModel );
      tree.getSelectionModel().setSelectionMode( TreeSelectionModel.SINGLE_TREE_SELECTION );
      tree.addTreeSelectionListener(
      new TreeSelectionListener()
      {
         public void valueChanged( TreeSelectionEvent e )
         {
            BingoTreeNode node = (BingoTreeNode) tree.getLastSelectedPathComponent();
            
            if ( node == null )
            {
               return;
            }
            vURLs =  new Vector(node.getNodeDocuments());
            ArrayList junk = node.getJunk();
            Iterator junkIter = junk.iterator();
            while(junkIter.hasNext())
            {
               Object o = junkIter.next();
               if (!vURLs.contains(o)) vURLs.addElement(o);
            }
            
            buildTable(vURLs);
         }
      }
      );
      
      JScrollPane treeScrollPane = new JScrollPane();
      
      treeScrollPane.setViewportView( tree );
      
      splitPane.setDividerSize( 10 );
      splitPane.setContinuousLayout( true );
      splitPane.setLeftComponent( treeScrollPane );
      splitPane.setRightComponent( tableScrollPane );
      splitPane.setDividerLocation( 240 );
      
      label3.setFont( f );
      
      content.add( label3, BorderLayout.NORTH );
      content.add( splitPane, BorderLayout.CENTER );
      content.add( panel2, BorderLayout.SOUTH );
      
      // MenLeiste
      menuBar = new JMenuBar();
      setJMenuBar( menuBar );
      
      // Men Trainingsdaten
      JMenu mFile = new JMenu( lc.get( "td" ) );
      
      menuBar.add( mFile );
      
      JMenuItem miNew = addMenuItem( mFile, lc.get( "newTd" ) );
      
      mFile.addSeparator();
      
      JMenuItem miFile = addMenuItem( mFile, lc.get( "readTd" ) );
      JMenuItem miSave = addMenuItem( mFile, lc.get( "saveTd" ) );
      
      mFile.addSeparator();
      
      JMenuItem miFeatures = addMenuItem( mFile, lc.get( "computeFeatures" ) );
      
      mFile.addSeparator();
      
      JMenuItem miCrawl = addMenuItem( mFile, lc.get( "send2Crawler" ) );
      
      mFile.addSeparator();
      
      JMenu mImport = new JMenu( lc.get( "imports" ) );
      
      mFile.add( mImport );
      
      JMenuItem miNetscape = addMenuItem( mImport, lc.get( "nsBm" ) );
      
      JMenuItem miClose = addMenuItem( mFile, lc.get( "close" ) );
      
      // Bearbeiten
      JMenu mEdit = new JMenu( lc.get( "edit" ) );
      
      menuBar.add( mEdit );
      
      JMenuItem miCut = addMenuItem( mEdit, lc.get( "cut" ) );
      JMenuItem miCopy = addMenuItem( mEdit, lc.get( "copy" ) );
      JMenuItem miPaste = addMenuItem( mEdit, lc.get( "paste" ) );
      
      mEdit.addSeparator();
      
      JMenuItem miSelectAll = addMenuItem( mEdit, lc.get( "selectAll" ) );
      
      mEdit.addSeparator();
      
      JMenuItem miClear = addMenuItem( mEdit, lc.get( "clearSelection" ) );
      JMenuItem miClearAll = addMenuItem( mEdit, lc.get( "deleteTable" ) );
      JMenuItem miRemoveFalse = addMenuItem( mEdit, lc.get( "deleteFalse" ) );
      
      mEdit.addSeparator();
      
      JMenuItem miSearch = addMenuItem( mEdit, lc.get( "search" ) );
      
      // ActionListener fr alle Mens registrieren
      bView.addActionListener( this );
      bBrowser.addActionListener( this );
      bNew.addActionListener( this );
      bFeatures.addActionListener( this );
      bRemove.addActionListener( this );
      miNew.addActionListener( this );
      miNetscape.addActionListener( this );
      miSave.addActionListener( this );
      miSearch.addActionListener( this );
      miFile.addActionListener( this );
      miClose.addActionListener( this );
      miClear.addActionListener( this );
      miClearAll.addActionListener( this );
      miSelectAll.addActionListener( this );
      miCut.addActionListener( this );
      miCopy.addActionListener( this );
      miPaste.addActionListener( this );
      miRemoveFalse.addActionListener( this );
      miCrawl.addActionListener( this );
      miFeatures.addActionListener( this );
      bShowFeat.addActionListener( this );
      bShowLinks.addActionListener( this );
      bCrawl.addActionListener( this );
      
      // KeyListener fr Desktop-Shortcuts
      bView.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      bNew.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      bFeatures.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      bRemove.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miNew.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miNetscape.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miSave.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miSearch.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miFile.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miClose.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miClear.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miClearAll.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miSelectAll.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miCut.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miCopy.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miPaste.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miRemoveFalse.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miCrawl.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      miFeatures.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      bShowFeat.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      bShowLinks.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      bCrawl.addKeyListener(BINGODesktop.getInstance().keyAdapter );
      tree.addKeyListener(BINGODesktop.getInstance().keyAdapter);
      
      buildTable( vURLs );
      
      this.setSize( session.defaultFrameWidth, session.defaultFrameHeight );
      this.setVisible( true );
   }
   
   
   /**
    *  berwacht die Buttons und Menleiste.
    *
    * @param  e  Das Event.
    */
   public void actionPerformed( ActionEvent e )
   {
      String cmd = e.getActionCommand();
      
      try
      {
         if ( cmd == ( lc.get( "close" ) ) )
         {
            setVisible( false );
         }
         else if ( cmd == ( lc.get( "deleteFalse" ) ) )
         {
            ArrayList docs = session.getRootNode().getPathDocuments();
            Iterator it = docs.iterator();
            while(it.hasNext())
            {
               BingoDocument tmp = (BingoDocument)it.next();
               if (! tmp.hasFeatureV())
                  (tmp.getNode()).removeDocument(tmp);
            }
            it = null;
            updateTree(session.getRootNode());
         }
         else if ( cmd == ( lc.get( "clearSelection" ) ) || cmd == ( lc.get( "delButton" ) ) )
         {
            clearTable( table.getSelectedRows() );
            table.clearSelection();
         } else if ( cmd == ( lc.get( "cut" ) ) )
         {
            selectionOk();
            
            int[] iRows = table.getSelectedRows();
            
            session.clearURLBuffer();
            for ( int i = 0; i < iRows.length; i++ )
            {
               session.addURLtoBuffer( vURLs.elementAt( iRows[i] ) );
            }
            clearTable( table.getSelectedRows() );
            table.clearSelection();
         } else if ( cmd == ( lc.get( "copy" ) ) )
         {
            selectionOk();
            
            int[] iRows = table.getSelectedRows();
            
            session.clearURLBuffer();
            for ( int i = 0; i < iRows.length; i++ )
            {
               session.addURLtoBuffer( vURLs.elementAt( iRows[i] ) );
            }
            table.clearSelection();
         } else if ( cmd == ( lc.get( "paste" ) ) )
         {
            Vector v = session.getURLBuffer();
            int iRow = table.getSelectedRow();
            
            for ( Enumeration enum = v.elements(); enum.hasMoreElements();  )
            {
               if ( ( iRow > -1 ) && ( iRow < vURLs.size() ) )
               {
                  vURLs.insertElementAt( enum.nextElement(), table.getSelectedRow() );
               } else
               {
                  vURLs.addElement( enum.nextElement() );
               }
            }
            buildTable( vURLs );
         }
         else if ( cmd == ( lc.get( "deleteTable" ) ) )
         {
            buildTable( new Vector() );
         }
         else if ( cmd == ( lc.get( "selectAll" ) ) )
         {
            table.selectAll();
         }
         else if (cmd == ( lc.get( "featureButton" ) ) )
         {
            BingoDocument bd = (BingoDocument) vURLs.elementAt( table.getSelectedRow() );
            if (bd != null)
               new FeatureViewer( this, bd ).setVisible( true );
         }
         else if (cmd == ( lc.get( "linkButton" ) ) )
         {
            BingoDocument bd = (BingoDocument) vURLs.elementAt( table.getSelectedRow() );
            if (bd != null)
               new LinksViewer( this, bd ).setVisible( true );
         }
         else if ( cmd == ( lc.get( "sourceButton" ) ) )
         {
            int row = table.getSelectedRow();
            if (row >= 0)
            {
               BingoDocument selectedURL = (BingoDocument) vURLs.elementAt( row ) ;
               new ParserFrame( selectedURL );
            }
         }
         else if ( cmd == ( lc.get( "browserButton" ) ) )
         {
            int row = table.getSelectedRow();
            if (row >= 0)
            {
               String target = null;
               BingoDocument selectedURL = (BingoDocument) vURLs.elementAt( row ) ;
               if (selectedURL.getUrl2r() != null && selectedURL.getUrl2r().getURL() != null)
               {
                  target = selectedURL.getUrl2r().getURL().toString();
               }
               else
               {
                  target = "http://" + selectedURL.getUrl();
               }
               
               if (target != null)
               {
                  try
                  {
                     bingo.util.BrowserLauncher.openURL(target);
                  }
                  catch (Exception ble)
                  {  }
               }
               target = null;
            }
         }
         
         else if ( (cmd == ( lc.get( "crawlerButton" ) )) || (cmd == ( lc.get("send2Crawler")) ))
         {
            selectionOk();
            
            int[] rows = table.getSelectedRows();
            CrawlerFrame cf = CrawlerFrame.getInstance();
            
            for ( int i = 0; i < rows.length; i++ )
            {
               BingoDocument tmp = (BingoDocument) vURLs.elementAt( rows[i] );
               
               cf.addLink( tmp );
            }
         } else if ( cmd == ( lc.get( "nsBm" ) ) )
         {
            FileDialog fileDialog = new FileDialog( BINGODesktop.getInstance(), lc.get( "openBookmarks" ), FileDialog.LOAD );
            
            fileDialog.setVisible( true );
            if ( fileDialog.getFile() != null )
            {
               String file = fileDialog.getDirectory() + fileDialog.getFile();
               
               table.clearSelection();
               importBookmarks( file );
               //SeedCheck.storeSeed(); // debug, speichert die Crawlsaat
               this.setTitle( lc.get( "bookmarksParsed" ) + " " + file );
            }
         } else if (cmd == (lc.get("computeButton")) || cmd == (lc.get("computeFeatures")))
         {
            selectionOk();
            new FeatureDialog( BINGODesktop.getInstance(), vURLs, table.getSelectedRows() ).setVisible( true );
            buildTable( vURLs );
         } else if (cmd == (lc.get("newButton")) || cmd == lc.get("newTd"))
         {
            BaseDialog mybd = new BaseDialog(BaseFrame.getInstance());
            mybd.setVisible(true);
         }
         
      }
      catch ( Exception any )
      {
         session.log(any);
      }
   }
   
   
   private JMenuItem addMenuItem( JMenu target, String text )
   {
      JMenuItem mi = new JMenuItem( text );
      
      target.add( mi );
      return mi;
   }
   
   
   public void buildTable( Vector vURLs )
   {
      this.vURLs = vURLs;
      baseTable.removeAllRows();
      if (vURLs == null)
         return;
      for ( int i = 0; i < vURLs.size(); i++ )
      {
         String[] sRow =
         {
            ( (BingoDocument) vURLs.elementAt( i ) ).getUrl(),
            String.valueOf(((BingoDocument) vURLs.elementAt( i ) ).getStatus()),
            ( (BingoDocument) vURLs.elementAt( i ) ).getTitle(),
            ( (BingoDocument) vURLs.elementAt( i ) ).getContentType(),
            String.valueOf(( (BingoDocument) vURLs.elementAt( i ) ).hasFeatureV())
         };
         
         baseTable.addRow( sRow );
      }
      baseTable.fireTableDataChanged();
   }
   
   
   /**
    *  Aktualisiert die Tabelle (z.B. nach dem Einfgen neuer Traingsdaten). Dies
    *  fhrt ebenso zu einem Update des Hierarchiebaums in der Klasse SVMC.
    */
   public void refreshTable()
   {
      buildTable( this.vURLs );
   }
   
   
   /**
    *  Der Hierarchiebaum wird aktualisiert bzw. durch einen neuen Baum mit
    *  referenzfreien Knoten ersetzt. Nach Anwendung dieser Methode wird evtl.
    *  die Neuberechnung des SVM-Modells zum Klassifizieren notwendig, da eine
    *  nderung der Baumstruktur durch eine nderung der Trainingsdaten ausgelst
    *  wird, was sich auch auf das zu betrachtende SVM-Modell auswirkt
    *
    * @param  vURLs  Vector mit Trainingsdaten in Form von BaseURL-Eintrgen.
    */
   
   
   public void updateTree( BingoTreeNode node )
   {
      treeModel.setRoot( node );
      treeModel.nodeStructureChanged( node );
      ArrayList tmp = node.getNodeDocuments();
      tmp.addAll(node.getJunk());
      Vector tmp1 = new Vector();
      if (tmp != null)
         tmp1.addAll(tmp);
      buildTable(tmp1);
      
      tmp1=null; tmp=null;
      label3.setText( node.getPathDocuments().size() + " " + lc.get( "baseFrameInfo" ) );
   }
   
   /**
    *  bernimmt eine URL zusammen mit der ihr zugeordneten Kategorie in die
    *  Trainingsdatentabelle. Schreibt die URL in die Datenbank (optional).
    *
    * @param  temp  Einzutragende BaseURL.
    */
   public void addNewTableEntry( BingoDocument temp )
   {
      //        boolean bFound = false;
      
      for ( Enumeration enum = vURLs.elements(); enum.hasMoreElements();  )
      {
         String nextURL = ( (BingoDocument) enum.nextElement() ).getUrl();
         
         if ( temp.getUrl().equals( nextURL ) )
            return;
      }
      vURLs.addElement( temp );
      
      String[] sRow =
      {
         temp.getUrl(),
         String.valueOf(temp.getStatus()),
         temp.getTitle(),
         temp.getContentType(),
         String.valueOf(temp.hasFeatureV())
      };
      baseTable.addRow( sRow );
   }
   
   private void importBookmarks( String sFileName )
   {
      BookmarkParser bp = new BookmarkParser( sFileName );
      BingoTreeNode newRootNode = bp.parseBookmarks();
      
      if (newRootNode == null)
         return;
      
      session.setRootNode(newRootNode);
   }
/*
    private void crawlBookmarks() {
        FeatureDialog fd = new FeatureDialog( BINGODesktop.getInstance(), new Vector(session.getRootNode().getPathDocuments()) );
        fd.setVisible( true );
        updateTree(session.getRootNode());
        fd = null;
    }
 */
   private void clearTable( int[] iRows )
   {
      Vector tempURLs = new Vector();
      
      for ( int i = 0; i < vURLs.size(); i++ )
      {
         BingoDocument temp = (BingoDocument) vURLs.elementAt( i );
         
         if ( !isInArray( i, iRows ) )
            tempURLs.addElement( temp );
      }
      vURLs = tempURLs;
      buildTable( vURLs );
   }
   
   
   private boolean isInArray( int i, int[] ii )
   {
      for ( int k = 0; k < ii.length; k++ )
      {
         if ( i == ii[k] )
            return true;
      }
      return false;
   }
   
   
   private void selectionOk()
   {
      int[] rows = table.getSelectedRows();
      
      if ( rows.length == 0 )
      {
         table.setRowSelectionInterval( 0, vURLs.size() - 1 );
         return;
      }
   }
   
   // Singleton
   private static BaseFrame baseFrame = null;
   
   
   /**
    *  Die Methode dient dem Aufruf der (einzigen) Instanz der Klasse.
    *
    * @return    The Instance value
    */
   public static BaseFrame getInstance()
   {
      if ( baseFrame == null )
      {
         baseFrame = new BaseFrame(lc.get("baseFrameTitle"));
         BINGODesktop.getInstance().add( baseFrame );
      }
      return baseFrame;
   }
   
   /**
    *  Selektion einer Tabellenzeile.
    *
    * @param  i  Nummer der Zeile.
    */
   public void addSelection( int i )
   {
      try
      {
         table.addRowSelectionInterval( i, i );
      }
      catch ( Exception any )
      {
         session.log(any);
      }
   }
   
   
   public void setSelectedIndex( int i )
   {
      try
      {
         table.setRowSelectionInterval( i, i );
      }
      catch ( Exception any )
      {
         session.log(any);
      }
   }
   
   
   /**
    *  Interfacemethode zu BaseTreeCaller. ndert die Kategorie eines
    *  Tabelleneintrags
    *
    * @param  s     Die Kategorie als String
    * @param  dmtn  Die Kategorie als Baumstruktur
    * @param  rank  The new path value
    */
   public void setPath( String s, DefaultMutableTreeNode dmtn, Double rank )
   {
      try
      {
         BingoTreeNode node = (BingoTreeNode) dmtn;
         table.setValueAt( s, table.getSelectedRow(), 1 );
         BingoDocument temp = (BingoDocument) vURLs.elementAt( table.getSelectedRow() );
         temp.setNode( node );
         temp.setSvmRank( rank );
      }
      catch ( Exception any )
      {
         session.log(any);
      }
   }
   
   public DefaultTreeModel getTreeModel()
   {
      return this.treeModel;
   }
}