package bingo.base;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.net.HttpURLConnection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JProgressBar;
import javax.swing.JSplitPane;

import bingo.crawler.LinkEvent;
import bingo.crawler.LinkListener;
import bingo.crawler.URL_Crawler;
import bingo.util.BingoDocument;
import bingo.util.BingoTreeNode;
import bingo.util.SessionBuffer;

/**
 *  Multi-Threaded Steuerung der Feature-Berechnung.
 */

public class FeatureDialog extends JDialog implements ActionListener {
    private SessionBuffer session = SessionBuffer.getInstance();
//    private Thread featureThread;
    private JButton cancel = new JButton( "Abbruch" );
    private JProgressBar pBar1 = new JProgressBar();
    private JProgressBar pBar2 = new JProgressBar();
    
    private static Vector pDialogs = new Vector();
    
    public Enumeration tasks;
    public int num_tasks = 0;
    private boolean removeTrainingData = true;
    
    private Vector bingoDocuments;
    
    boolean importNew;
    int progress = 1;
    Loader loader;
    
    
    /**
4     *  Konstruktor fr das FeatureDialog Objekt.
     *
     *@param  parentFrame  Der Start-Frame.
     *@param  vURLs        Der Vektor mit den URLs.
     *@param  rows         Die markierten Eintrge.
     */
    public FeatureDialog( JFrame parentFrame, Vector vURLs, int[] rows ) {
        super( parentFrame, true );
        this.setSize( 400, 115 );
        initGUI();
        removeTrainingData = true;
        setLocationRelativeTo( parentFrame );
        
        if ( rows == null ) {
            rows = new int[vURLs.size()];
            for ( int i = 0; i < vURLs.size(); i++ ) {
                rows[i] = i;
            }
        }
        
        bingoDocuments = new Vector();
        
        session.setShowParserDialog( true );
        session.setParserBreak( false );
        pBar2.setMaximum( rows.length );
        
        for ( int i = 0; i < rows.length; i++ ) {
            if ( i >= vURLs.size() ) {
                break;
            }
            BingoDocument bd = (BingoDocument) vURLs.elementAt( rows[i] );
            bingoDocuments.addElement( bd );
        }
        
        //urls = baseURLs.elements();
        tasks = bingoDocuments.elements();
        num_tasks = bingoDocuments.size();
        
        this.setTitle( "Berechne Feature-Vektoren" );
        
        loader = new Loader( this );
        new Thread( loader ).start();
    }
    
    
    /**
     *  Konstruktor fr das FeatureDialog Objekt.
     *
     *@param  parentFrame  Der Start-Frame.
     *@param  vURLs        Der Vektor mit den URLs.
     */
    public FeatureDialog( JFrame parentFrame, Vector vURLs ) {
        super( parentFrame, true );
        this.setTitle( "Berechne Feature-Vektoren" );
        this.setSize( 400, 115 );
        initGUI();
        setLocationRelativeTo( parentFrame );
        
        bingoDocuments = vURLs;
        
        session.setShowParserDialog( true );
        session.setParserBreak( false );
        pBar2.setMaximum( vURLs.size() );
        
        //urls = bingoDocuments.elements();
        tasks = bingoDocuments.elements();
        num_tasks = bingoDocuments.size();
        
        this.setTitle( "Berechne Feature-Vektoren" );
        
        loader = new Loader( this );
        new Thread( loader ).start();
    }
    
    
    private void initGUI() {
        Container content = this.getContentPane();
        content.setLayout( new BorderLayout() );
        
        pBar1.setString( " URL " );
        pBar2.setString( " TOTAL " );
        pBar1.setMaximum( 100 );
        pBar1.setMinimum( 0 );
        pBar2.setMinimum( 0 );
        pBar1.setStringPainted( true );
        pBar2.setStringPainted( true );
        pBar1.setValue( 100 );
        
        JSplitPane splitPane = new JSplitPane( JSplitPane.VERTICAL_SPLIT );
        splitPane.setDividerSize( 5 );
        splitPane.setContinuousLayout( true );
        splitPane.setTopComponent( pBar1 );
        splitPane.setBottomComponent( pBar2 );
        splitPane.setDividerLocation( 30 );
        
        content.add( splitPane, BorderLayout.CENTER );
        content.add( cancel, BorderLayout.SOUTH );
        
        cancel.addActionListener( this );
    }
    
    
    public void actionPerformed( ActionEvent e ) {
        String cmd = e.getActionCommand();
        if ( cmd.equals( "Abbruch" ) ) {
            loader.abbruch();
        }
    }
    
    
    public static void closeDialogs() {
        for ( Enumeration enum = pDialogs.elements(); enum.hasMoreElements();  ) {
            JDialog pDialog = (JDialog) enum.nextElement();
            pDialog.dispose();
        }
        pDialogs = new Vector();
    }
    
    public Vector getbingoDocuments()
    {return bingoDocuments;}
    
    //=======================================================
    
    class Loader implements Runnable, LinkListener {
        SessionBuffer session;
        URL_Crawler crawler;
        FeatureDialog fd;
        
        public Loader( FeatureDialog fd ) {
            this.fd = fd;
            session = SessionBuffer.getInstance();
            crawler = session.getCrawler();
        }
        
        
        public void abbruch() {
            crawler.stop();
        }
        
        
        //===============================================
        // Diese Methode gehrt in den Thread loader,
        // um wirklich parallel zu sein !!!
        // Sergej
        
        public void run() {
            int sessionDepth = session.maxCrawlingDepth;
            long crawlerLifetime = crawler.getLifetime();
            crawler.setLifetime(5*1000);  // kleine Wartezeit um interaktiv zu sein
            
            session.maxCrawlingDepth = 0;
            String dbType = session.dbType;
            session.dbType = null;
            boolean connected = session.dbConnect;
            session.dbConnect = false;
            
            BingoDocument bd;
            int progress = 0;
            
            int linksSent = 0;
            
            while ( fd.tasks.hasMoreElements() && !fd.session.bParserBreak ) {
                bd = (BingoDocument) fd.tasks.nextElement();
                bd.setPriority((float) 2.0);
                // System.out.println ("Added: " + bd.getStatus());
                if ( bd.getUrl2r().isBad() )
                    continue;
                
                //bd.setRootNode( bd.getNode() );
                //bd.setClassifiedNode( bd.getNode() );
//                link.setStatus( baseURL.getStatus() );
//                link.setDepth(0);
                crawler.addLink( bd.getUrl2r() );
                linksSent++;
            }
            crawler.addLinkListener(this);
            bingoDocuments = new Vector();
            crawler.go();
            
            boolean fertig = false;
            fd.pBar2.setMaximum( linksSent );
            
            while ( !fertig ) {
                try {
                    Thread.sleep( 1000 );
                } catch ( Exception e ) {}
                
                synchronized ( fd ) {
                    progress = crawler.countCrawled() + crawler.countSuspended();
                    fd.pBar2.setValue( progress );
                    fd.pBar2.setString( crawler.countActiveThreads() + " running" );
                    if (crawler.countActiveThreads() == 0)
                        //          if ( crawler.isIdle())
                        fertig = true;
                }
            }
            crawler.removeLinkListener(this);
            crawler.setLifetime(crawlerLifetime);
            crawler.reset();
            session.maxCrawlingDepth = sessionDepth;
            session.dbType = dbType;
            session.dbConnect = connected;
            
//            Enumeration en = fd.bingoDocuments.elements();
            fd.pBar2.setMaximum( fd.bingoDocuments.size() );
            progress = 0;
            fd.pBar2.setValue( progress );
            
            fd.setVisible( false );
            fd.pBar2.setString( "Features initialisiert" );
            
            session.log( "UPDATING bingoDocuments", 3 );
            BingoTreeNode rootNode = session.getRootNode();
            if (removeTrainingData)
                rootNode.removeTrainingData();
            
            Iterator it = bingoDocuments.iterator();
            while (it.hasNext()) {
                BingoDocument tmp = (BingoDocument)it.next();
                if (tmp.getText() == null || tmp.getText().size() <= 0)
                    continue;
                BingoTreeNode node = tmp.getNode();
                if (node != null) {
                    if (tmp.getStatus() == 'B')
                        node.addDocument(tmp);
                    else if (tmp.getStatus() == 'J')
                        node.addJunk(tmp);
                }
            }
            it=null;
            bingoDocuments=null;
        }
        
        // ====================================================================
        
        public void crawled(LinkEvent event) {
            if (event.getID() == LinkEvent.VISITED ) {
                BingoDocument bd = event.getLink();
                if (bd == null)
                    return;
                
                if ((bd.getStatus() == 'B' 
                	|| bd.getStatus() == 'J') 
					&& !bd.isBroken() 
					&& bd.getResponseCode() == HttpURLConnection.HTTP_OK)
                {
                    //BaseURL bu = new BaseURL(link.url2r.getURL().toString(), link.getClassifiedNode());
//                    try {
                        //bu.setStatus(link.getStatus());
                        //bu.id = link.id;
                        //bu.setText( page.getTerms() );
                        //bu.preview = page.preview;
//                        bu.title = page.title;
//                        bu.bulk = page.bulk;
//                        bu.bulkSize = page.bulkSize;
//                        bu.setFeatures( new TFBuilder().getTFFeatures( page.getTerms() ) );
//                        bu.setLinks( page.getLinks() );
//                        bu.setContentType( page.getContentType() );
                    	bingoDocuments.add(bd);
//                    }
//                    catch ( Exception e )
//                    {session.log(e);}
                }
            }
        }
        // ====================================================================
        
    }
    
    public void setRemoveFlag(boolean flag) {
        this.removeTrainingData = flag;
    }
}