package bingo.base;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.*;
import java.text.*;
import java.net.*;

import javax.swing.JButton;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.tree.DefaultMutableTreeNode;

import bingo.language.LanguageConstants;
import bingo.svmlight.Feature;
import bingo.util.*;

/**
 *  Klasse zur Ausgabe eines Featurevektors zu einer URL.
 *  Es werden zu den Dokumenten jeweils die Terme und die jeweiligen relativen Hufigkeiten des Auftretens
 *  der Terme ausgegeben.
 */
public class LinksViewer extends JFrame implements ActionListener
{
   
   private static LanguageConstants lc = LanguageConstants.getInstance();
   
   private SessionBuffer session;
   // private JLabel urlLabel;
   private JButton ok = new JButton( lc.get( "ok" ) );
   private JButton browser = new JButton( lc.get( "browserButton" ) );
   private URLClassTableModel featureTable;
   private final JTable table;
   
   private ArrayList links = null;
   
   Font f = new Font( "Serif", Font.PLAIN, 14 );
   Font ff = new Font( "Serif", Font.ITALIC, 18 );
   DefaultMutableTreeNode parent = null;
   
   
   /**
    *  Konstruktor fr das FeatureViewer Objekt eines BingoDocuments
    */
   public LinksViewer(JInternalFrame parent, BingoDocument bd)
   {
      if (bd == null)
      {
         session.log("BingoDocument ist null", 1);
      }
      
      this.setTitle( lc.get( "linksViewerTitle" ) + ": " );
      this.setSize( 600, 450 );
      this.setLocationRelativeTo(parent);
      addWindowListener( new CloseWindow( this ) );
      Container content = this.getContentPane();
      
      String[] columnNames =
      {bd.getUrl()};
      
      links = new ArrayList (bd.getLinks());
      
      //urlLabel = new JLabel( bd.getUrl() );
      //urlLabel.setFont( f );
      featureTable =
      new URLClassTableModel()
      {
         public boolean isCellEditable( int x, int y )
         {
            return false;
         }
      };
      
      featureTable.setColumnNames( columnNames );
      table = new JTable( featureTable );
      
      // featureTable.addMouseListenerToHeaderInTable( table );
      
      table.setBackground( Color.gray );
      table.setForeground( Color.white );
      
      JScrollPane scrollpane = new JScrollPane( table );
      
      //content.add( urlLabel, BorderLayout.NORTH );
      content.add( scrollpane, BorderLayout.CENTER );
      content.add( browser, BorderLayout.NORTH );
      content.add( ok, BorderLayout.SOUTH );
      
      // ActionListener fr alle Mens registrieren
      ok.addActionListener( this );
      browser.addActionListener( this );
      buildTable();
   }
   
   private void buildTable( )
   {
      if (links == null || links.size() <= 0)
         return;
      
      Iterator it = links.iterator();
      while (it.hasNext())
      {
         URL url = (URL)it.next();
         Object[] sRow = new Object[1];
         sRow[0] = url.toString();
         featureTable.addRow( sRow );
      }
      it=null;
      
      this.setTitle( lc.get( "linksViewerTitle" ) + ": " + links.size());
   }
   
   /**
    *  berwacht den Button.
    *
    * @param  e  Das Event.
    */
   public void actionPerformed( ActionEvent e )
   {
      String cmd = e.getActionCommand();
      
      if ( cmd == ( lc.get( "ok" ) ) )
      {
         links = null;
         super.dispose();
      }
      else if ( cmd == ( lc.get( "browserButton" ) ) )
      {
         int row = table.getSelectedRow();
         if (row >= 0)
         {
            String target = null;
            URL url = (URL)links.get(row);
            if (url != null)
            {
               try
               {
                  bingo.util.BrowserLauncher.openURL(url.toString());
               }
               catch (Exception ble)
               {  }
            }
            target = null;
         }
      }
      
   }
}
