package bingo.crawler;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.net.HttpURLConnection;
import java.util.Enumeration;
import java.util.*;
import bingo.crawler.frontier.URL_Queue;
import bingo.crawler.frontier.url2resolve;
import bingo.crawler.handler.HandlerManager;
import bingo.db.DBInterface;
import bingo.util.BingoDocument;
import bingo.util.SessionBuffer;

//=======================================================================================

public class BINGOCrawler implements URL_Crawler, Runnable
{
   private SessionBuffer session = SessionBuffer.getInstance();
   short HTTP_ERROR = 10;
   short HTTP_TIMEOUT = 20;
   public static int crawled = 0;
   public static int garbageBatch = 200;
   public static URL_Queue queue;
   static Random rd;
   public static boolean stopped = false;
   public static ArrayList worms;
   static int max_worms = 50;
   int attempts = 0;
   static int max_attempts = 2;
   public static long lifetime = 30 * 1000;
   // Wartezeit des Threads (falls Leerlauf)
   static long waitingtime = 90 * 1000;
   // max Wartezeit auf Download der URL
   private static Vector crawlListeners;
   private static Vector linkListeners;
   private url2resolve ziel = null;
   public boolean ready = false;
   public DBInterface dbinterface = null;
   public HandlerManager handlerManager;
   private HttpURLConnection conn;
   private int state = CrawlEvent.CLEARED;
   private static final byte ERROR = 0;
   private static final byte TIMEOUT = 1;
   private static final byte DONE = 2;
   private static final byte SUSPEND = 3;
   private static final byte BAD = 4;
   //=====================================================
   public BINGOCrawler()
   {
      synchronized (this.getClass())
      {
         if (queue == null)
            queue = new URL_Queue();
         
         if (worms == null)
         {
            worms = new ArrayList();
            
            System.setProperty("sun.net.spi.nameservice.nameservers","134.96.100.105,134.96.7.7");
            
            java.security.Security.setProperty("networkaddress.cache.ttl", "500");
            java.security.Security.setProperty("networkaddress.cache.negative.ttl", "100");
            System.setProperty("networkaddress.cache.ttl", "500");
            System.setProperty("networkaddress.cache.negative.ttl", "100");
            
            System.setProperty("http.agent", "Mozilla/4.0 (compatible; MSIE 6.01; Windows NT 5.0)");
            System.setProperty("http.keepAlive", "true");
            System.setProperty("http.maxConnections", "4");
            
            System.setProperty("sun.net.client.defaultConnectTimeout", "10000");
            System.setProperty("sun.net.client.defaultReadTimeout", "60000");
            
            System.setProperty("sun.net.inetaddr.ttl", "500");
            System.setProperty("sun.net.inetaddr.negative.ttl", "100");
            
         }
         
         if (linkListeners == null)
            linkListeners = new Vector();
         if (crawlListeners == null)
            crawlListeners = new Vector();
      }
      
      handlerManager = new HandlerManager();
      if (session.dbType != null)
      {
         dbinterface = DBInterface.getInstance();
      }
      waitingtime = session.downloadTimeout;
      max_worms = session.maxCrawlerThreads;
   }
   
   public BINGOCrawler(url2resolve initial)
   {
      this();
      this.ziel = initial;
      waitingtime = session.downloadTimeout;
      max_worms = session.maxCrawlerThreads;
   }
   //=====================================================
   public void run()
   {
      ready = false;
      attempts = 0;
      if (session.dbType != null
      && (dbinterface == null || !dbinterface.isValid()))
      {
         synchronized (worms)
         {
            worms.remove(this);
         }
         return;
      }
      synchronized (worms)
      {
         if (!worms.contains(this))
         {
            worms.add(this);
         }
      }
      state = CrawlEvent.STARTED;
      sendCrawlEvent(state);
      while (true)
      {
         if (stopped)
         {
            if (dbinterface != null)
            {
               dbinterface.close();
            }
            synchronized (worms)
            {
               worms.remove(this);
               worms.notifyAll();
            }
            return;
         }
         if (ziel == null)
         {
            ziel = queue.getBest();
         }
         if (ziel == null)
         {
            ready = true;
            if (attempts >= max_attempts)
            {
               if (dbinterface != null)
               {
                  dbinterface.close();
               }
               dbinterface = null;
               handlerManager = null;
               conn = null;
               if (worms != null)
                  synchronized (worms)
                  {
                     worms.remove(this);
                     worms.notifyAll();
                  }
                  return;
            } else
            {
               attempts++;
            }
         } else
         {
            attempts = 0;
            ready = false;
            synchronized (worms)
            {
               if (worms.size() < max_worms)
               {
                  // ggf. einen neuen Thread starten
                  url2resolve tempZiel = queue.getBest();
                  if (tempZiel != null)
                  {
                     BINGOCrawler tmp = new BINGOCrawler(tempZiel);
                     worms.add(tmp);
                     new Thread(tmp).start();
                     tempZiel = null;
                  }
               }
            }
         }
         //===== Inhaltlich interessante Arbeit des Crawlers
         // =========================
         if (!ready)
         {
            // Checken, ob die connection noch OK ist
            if (dbinterface != null)
            {
               dbinterface.verifyConnection();
               if (dbinterface == null || !dbinterface.isValid())
               {
                  synchronized (worms)
                  {
                     worms.remove(this);
                     worms.notifyAll();
                  }
                  return;
               }
            }
            workOnIt();
            // Die Bearbeitung der geholten URL
         } else
         {
            try
            {
               synchronized (queue.semaphore)
               {
                  queue.semaphore.wait(lifetime);
               }
            } catch (InterruptedException e)
            {
            }
         }
      }
   }
   //== Diese Methode bearbeitet die URL = ==================================
   public synchronized void workOnIt()
   {
      boolean OK = ((ziel != null) && (!ziel.isBad()));
      
      if (OK)
      {
         OK = (!queue.hasSlowHost(ziel));
         if (OK)
         {
            try
            {
               ziel.parentDoc.setEventStatus(LinkEvent.RETRIEVING);
               sendLinkEvent(ziel.parentDoc, LinkEvent.RETRIEVING);
               conn = ziel.getConnection();
               WormThread myWorm = new WormThread(this, ziel, conn);
               new Thread(myWorm).start();
               synchronized (this)
               {
                  try
                  {
                     this.wait(waitingtime);
                  } catch (Exception e)
                  {
                  }
               }
               // Fertig? wenn nein, verbindung kappen und nochmals warten
               // (die Verarbeitung+Speicherung geht normalerweise weiter)
               if (!myWorm.isReady() && conn != null)
               {
                  try
                  {
                     myWorm.abortConnection();
                     conn = null;
                  } catch (Exception e)
                  {
                  }
                  synchronized (this)
                  {
                     try
                     {
                        this.wait(waitingtime);
                     } catch (Exception e)
                     {
                     }
                  }
               }
               if (myWorm.isReady())
               {
                  BingoDocument bd = ziel.parentDoc;
                  if (bd != null && !bd.isBroken())
                  {
                     finish(DONE);
                  } else
                  {
                     finish(ERROR);
                  }
               } else
               {
                  finish(TIMEOUT);
               }
            } catch (InterruptedIOException e)
            {
               ziel.parentDoc.setEventStatus(LinkEvent.ERROR);
               sendLinkEvent(ziel.parentDoc, LinkEvent.ERROR,
               new IOException("INTERRUPTED!"));
               session.log(e);
               finish(TIMEOUT);
            } catch (Exception e)
            {
               ziel.parentDoc.setEventStatus(LinkEvent.ERROR);
               sendLinkEvent(ziel.parentDoc, LinkEvent.ERROR, e);
               session.log(e);
               finish(ERROR);
            }
         }
         else
         {
            ziel.parentDoc.setEventStatus(LinkEvent.ERROR);
            sendLinkEvent(ziel.parentDoc, LinkEvent.ERROR, new IOException(
            "SUSPENDED!"));
            finish(SUSPEND);
         }
      } else
      {
         finish(BAD);
         // NXDOMAIN -- do nothing
      }
   }
   public void finish(byte how)
   {
      try
      {
         if (conn != null)
            conn.disconnect();
      } catch (Exception e)
      {
      }
      queue.unsetLock(ziel);
      if (how == DONE)
      {
         if (dbinterface != null)
            dbinterface.commit();
         ziel.parentDoc.setEventStatus(LinkEvent.VISITED);
         sendLinkEvent(ziel.parentDoc, LinkEvent.VISITED);
         if (ziel.parentDoc.getStatus() != 'B'
         && ziel.parentDoc.getStatus() != 'J')
            destroy();
      } else if (how == TIMEOUT)
      {
         if (dbinterface != null)
            dbinterface.rollback();
         synchronized (queue)
         {
            queue.putSuspended(ziel);
         }
      } else if (how == SUSPEND)
      {
         synchronized (queue)
         {
            queue.putSuspended(ziel);
         }
      } else if (how == ERROR)
      {
         if (dbinterface != null)
            dbinterface.rollback();
         if (ziel.parentDoc.getStatus() != 'B'
         && ziel.parentDoc.getStatus() != 'J')
         {
            destroy();
         }
      } else if (how == BAD)
      {
         if (ziel.parentDoc.getStatus() != 'B'
         && ziel.parentDoc.getStatus() != 'J')
         {
            destroy();
         }
      }
      synchronized (this.getClass())
      {
         crawled++;
      }
      ziel = null;
      ready = true;
   }
   //=============================================================
   public void destroy()
   {
      // Nur wenn es keine listener gibt, sonst werden Docs noch fuer die Anzeige benoetigt
      if ((linkListeners == null || linkListeners.size()<=0)&&(crawlListeners == null || crawlListeners.size()<=0))
      {
         if (ziel != null && ziel.parentDoc != null)
         {
            BingoDocument bd = ziel.parentDoc;
            bd.setAnchors(null);
            bd.setText(null);
            bd.setGoogle(null);
            bd.setSvmRank(null);
            bd.setFeatureVector(null);
            bd.setContentType(null);
            bd.setContentEncoding(null);
            // bd.setRootNode(null);
            // bd.setClassifiedNode(null);
            bd.setNode(null);
            bd.setSourceLink(null);
            bd.setUrl2r(null);
            bd.setHost(null);
            bd.setFilename(null);
            bd.setDirectory(null);
            bd.setText(null);
            bd = null;
         }
         if (ziel != null)
         {
            ziel.parentDoc = null;
            ziel.IP = null;
            ziel.url = null;
         }
      }
      
      ziel = null;
      conn = null;
      
      if (crawled++ % garbageBatch == 0)
      {
         session.log("GARBAGE initiated", 3);
         System.gc();
      }
   }
   //========================================================================
   // Links in die queue hinzufgen
   public synchronized void addLink(url2resolve u)
   {
      //session.log("BingoCrawler.addLink(" + u + ")!", 9);
      queue.put(u);
      //sendLinkEvent(u.parentDoc, LinkEvent.QUEUED);
   }
   public synchronized void putSuspended(url2resolve u)
   {
      if (queue != null)
      {
         synchronized (queue)
         {
            queue.putSuspended(u);
         }
      }
   }
   //========================================================================
   // Statusabfragen, starten, stoppen des Crawlers
   public boolean isIdle()
   {
      int countRunning = worms.size();
      session.log("--------------------------------------------", 5);
      session.log("Running Worms: " + countRunning, 5);
      session.log("--------------------------------------------", 5);
      return (countRunning == 0);
   }
   
   public synchronized void reset()
   {
      worms = new ArrayList();
      queue = new URL_Queue();
      crawlListeners = new Vector();
      linkListeners = new Vector();
      crawled = 0;
      rd = new Random();
      state = CrawlEvent.CLEARED;
      sendCrawlEvent(state);
      session.log("Crawler reset: done", 3);
   }
   public int countCrawled()
   {
      return crawled;
   }
   public int countSuspended()
   {
      return queue.getNumSuspended();
   }
   public int countSlowHosts()
   {
      return queue.getNumSlowHosts();
   }
   
   public int countActiveThreads()
   {
      int res = 0;
      if (worms == null)
         return 0;
      
      synchronized (worms)
      {
         Iterator it = worms.iterator();
         while (it.hasNext())
         {
            if (!((BINGOCrawler) it.next()).ready)
            {
               res++;
            }
         }
         it=null;
      }
      return res;
   }
   public boolean isStopped()
   {
      return stopped;
   }
   public synchronized void go()
   {
      stopped = false;
      new Thread(new BINGOCrawler()).start();
   }
   public synchronized void stop()
   {
      stopped = true;
      synchronized (queue.semaphore)
      {
         // Alle wartenden Threads benachrichtigen
         queue.semaphore.notifyAll();
      }
      session.log("Shutdown all worms.. ", 3);
      while (worms.size() > 0)
      {
         try
         {
            synchronized (worms)
            {
               if (worms.size() > 0)
                  worms.wait();
            }
         } catch (InterruptedException ie)
         {
            session.log(ie);
         }
      }
      session.log("Shutdown done.", 3);
                /*
                 * try { Thread.sleep( waitingtime + 500 ); } catch ( Exception e ) {}
                 */
      state = CrawlEvent.STOPPED;
      sendCrawlEvent(state);
      session.log("Crawler stopped, " + worms.size() + " worms running", 3);
   }
   // ========================================================================
   // ************************************************************************
   // Die nachfolgenden Methoden haben mit der Funktion des Crawlers
   // nichts zu tun, sie sind fr Animationen und hnlichen Unsinn zustndig
   //=========================================================================
   /**
    * Liefert den Status des Crawlers.
    *
    * @return eins aus CrawlEvent.STARTED, PAUSED, STOPPED, CLEARED.
    */
   public int getState()
   {
      return state;
   }
   /**
    * Registriert einen CrawlListener.
    *
    * @param listen
    *            Der CrawlListener.
    */
   public void addCrawlListener(CrawlListener listen)
   {
      if (!crawlListeners.contains(listen))
      {
         crawlListeners.addElement(listen);
      }
   }
   /**
    * Lscht einen registrierten CrawlListeners.
    *
    * @param listen
    *            Der CrawlListener.
    */
   public void removeCrawlListener(CrawlListener listen)
   {
      crawlListeners.removeElement(listen);
   }
   /**
    * Registriert einen LinkListener.
    *
    * @param listen
    *            Der LinkListener.
    */
   public void addLinkListener(LinkListener listen)
   {
      if (!linkListeners.contains(listen))
      {
         linkListeners.addElement(listen);
      }
   }
   /**
    * Lscht einen registrierten LinkListener.
    *
    * @param listen
    *            LinkListener.
    */
   public void removeLinkListener(LinkListener listen)
   {
      linkListeners.removeElement(listen);
   }
   /**
    * Sendet ein CrawlEvent an alle registrierten CrawlListener.
    *
    * @param id
    *            Die Event ID.
    */
   protected void sendCrawlEvent(int id)
   {
      CrawlEvent evt = new CrawlEvent(this, id);
      for (int j = 0, len = crawlListeners.size(); j < len; ++j)
      {
         CrawlListener listen = (CrawlListener) crawlListeners.elementAt(j);
         switch (id)
         {
            case CrawlEvent.STARTED :
               listen.started(evt);
               break;
            case CrawlEvent.STOPPED :
               listen.stopped(evt);
               break;
            case CrawlEvent.CLEARED :
               listen.cleared(evt);
               break;
            case CrawlEvent.TIMED_OUT :
               listen.timedOut(evt);
               break;
            case CrawlEvent.PAUSED :
               listen.paused(evt);
               break;
         }
      }
   }
   /**
    * Sendet ein LinkEvent an alle registrierten LinkListener.
    *
    * @param l
    *            Der Link.
    * @param id
    *            Die Event ID.
    */
   protected void sendLinkEvent(BingoDocument bd, int id)
   {
      try
      {
         LinkEvent evt = new LinkEvent(this, id, bd);
         bd.setEventStatus(id);
         for (int j = 0, len = linkListeners.size(); j < len; ++j)
         {
            LinkListener listen = (LinkListener) linkListeners.elementAt(j);
            listen.crawled(evt);
            session.log(listen + " crawled:" + bd.getUrl() + "@"
            + LinkEvent.eventName[id], 5);
         }
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   /**
    * Sendet ein LinkEvent an alle registrierten LinkListener, falls whrend
    * der Abarbeitung eine Exception aufgetreten ist.
    *
    * @param l
    *            Der Link.
    * @param id
    *            Die Event ID.
    * @param exception
    *            Die Exception.
    */
   protected void sendLinkEvent(BingoDocument bd, int id, Throwable exception)
   {
      try
      {
         if (state == CrawlEvent.STOPPED || state == CrawlEvent.CLEARED
         || state == CrawlEvent.TIMED_OUT)
         {
            return;
         }
         LinkEvent evt = new LinkEvent(this, id, bd, exception);
         bd.setEventStatus(id);
         bd.setStatus('L');
         //bd.setLabel("exception", exception.toString());
         for (int j = 0, len = linkListeners.size(); j < len; ++j)
         {
            LinkListener listen = (LinkListener) linkListeners.elementAt(j);
            listen.crawled(evt);
         }
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   public DBInterface getDBInterface()
   {
      return dbinterface;
   }
   public long getLifetime()
   {
      return lifetime;
   }
   public void setLifetime(long t)
   {
      lifetime = t;
   }
   
   public boolean isAlreadyParsed(BingoDocument bd)
   {
      return queue.isAlreadyParsed(bd);
   }
   
   public void resetLinkListeners()
   {
     this.linkListeners = new Vector();
   }
   
   public void resetCrawlListeners()
   {
     this.crawlListeners = new Vector();
   }   
}