package bingo.crawler;

/**
 *  Ein Crawling-Ereignis. CrawlEvent-Objekte werden erzeugt, wenn der Crawler
 *  startet, stoppt, oder gelscht wird ( clear() ).
 *
 * @author     T
 * @created    15. Juni 2001
 */
public class CrawlEvent {
  private URL_Crawler crawler;
  private int id;

  /**
   *  Crawler gestartet.
   */
  public final static int STARTED = 0;

  /**
   *  Crawler angehalten (wenn keine Links mehr brig).
   */
  public final static int STOPPED = 1;

  /**
   *  Crawler gelscht.
   */
  public final static int CLEARED = 2;

  /**
   *  Crawler Timeout-Zeit berschritten.
   */
  public final static int TIMED_OUT = 3;

  /**
   *  Crawler in Pause-Zustand.
   */
  public final static int PAUSED = 4;


  /**
   *  Generiert ein neues CrawlEvent.
   *
   * @param  crawler  Der Crawler, der dieses Ereignis-Objekt generiert hat.
   * @param  id       Ereignis-ID (eins aus STARTED, STOPPED, etc.)
   */
  public CrawlEvent( URL_Crawler crawler, int id ) {
    this.crawler = crawler;
    this.id = id;
  }


  /**
   *  Liefert den Crawler, der dieses Ereignis-Objekt generiert hat.
   *
   * @return    Der Crawler.
   */
  public URL_Crawler getCrawler() {
    return crawler;
  }


  /**
   *  Liefert die Ereignis-ID.
   *
   * @return    eins aus STARTED, STOPPED, CLEARED, TIMED_OUT, oder PAUSED.
   */
  public int getID() {
    return id;
  }

}
