package bingo.crawler;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Locale;
import java.util.Vector;
import java.net.URL;

import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.table.TableColumn;
import javax.swing.tree.DefaultMutableTreeNode;

import bingo.base.*;
import bingo.crawler.handler.*;
import bingo.crawler.frontier.*;
import bingo.db.DBInterface;
import bingo.language.LanguageConstants;
import bingo.svmlight.ClassifierDialog;
import bingo.svmlight.SVMTreeListener;
import bingo.util.BINGODesktop;
import bingo.util.BingoDocument;
import bingo.util.BingoTreeNode;
import bingo.util.ParserFrame;
import bingo.util.SessionBuffer;
import bingo.util.URLClassTableModel;
/**
 * Die Klasse steuert den Crawler und dient zur Verwaltung der Crawlerdaten.
 * Die URLs werden zusammen mit einem Kategorievorschlag in eine Tabelle
 * bernommen. Die Daten knnen ber verschiedene Optionen manipuliert werden.
 * Es kann vom Benutzer jeweils entschieden werden, ob URLs in die
 * Trainingsdatenbasis oder die Datenbank bernommen werden.
 */
public class CrawlerFrame extends JInternalFrame
implements
CrawlListener,
SVMTreeListener,
Runnable,
ItemListener,
ActionListener
{
   private SessionBuffer session = SessionBuffer.getInstance();
   private static LanguageConstants lc = LanguageConstants.getInstance();
   /**
    * Benutze als Ausgabe von gefundenen Links eine Tabelle. Falls ja werden
    * max. 200 Links angezeigt und im Hauptspeicher verwaltet, was
    * zustzlichen Platz kostet.
    */
   public JCheckBoxMenuItem miTable = new JCheckBoxMenuItem(lc
   .get("cFVOutputOption"), true);
   /**
    * Visualisierung der Linkstruktur.
    */
   public JCheckBoxMenuItem miVis = new JCheckBoxMenuItem(lc
   .get("cFVVisualOption"), false);
   /**
    * Starte den Crawler.
    */
   public JMenuItem miStart;
   /**
    * Crawler-Pause.
    */
   public JMenuItem miReset;
   /**
    * Halte den Crawler an.
    */
   public JMenuItem miStop;
   /**
    * Wiederaufnahme eines unterbrochenen Crawls.
    */
   //  public JMenuItem miResume;
   /**
    * Der Crawler.
    */
   public URL_Crawler crawler;
   /**
    * Der Kontrolldialog zur Steuerung des Crawlers.
    */
   public ControlDialog controlDialog;
   public URLClassTableModel crawlerTable;
   /**
    * Die Tabelle zum Anzeigen und Bearbeiten der beim Crawling erhaltenen
    * URLs und Kategorien.
    */
   public final JTable table;
   /**
    * Der SVM-Klassifikator zur hierarchischen Klassifikation von HTML-Seiten.
    */
   public Classifier svmClassifier;
   private Visualization visualization;
   private JLabel label = new JLabel();
   private JMenuBar menuBar;
   private JPanel panel1 = new JPanel();
   private JPanel panel2 = new JPanel();
   //    private JLabel name = new JLabel();
   private JButton bView = new JButton(lc.get("sourceButton"));
   private JButton bBrowser = new JButton( lc.get( "browserButton" ) );
   //	private JButton bFeatures = new JButton(lc.get("computeButton"));
   private JButton bShowFeat = new JButton(lc.get("featureButton"));
   private JButton bShowLinks = new JButton( lc.get( "linkButton" ) );
   private JButton bTrain = new JButton(lc.get("trainingButton"));
   //	private JButton bClassify = new JButton(lc.get("classifyButton"));
   private JLabel activeThreads = new JLabel();
   private final String[] columnNames =
   {"URL", "CLASS", "DEPTH [MIME]",
    "SVM-RANK", "FEATURES"};
    private Vector vURLs = new Vector();
    private Vector vLabels = new Vector();
    private Font f = new Font("Serif", Font.ITALIC, 16);
    private LinkEventManager linkEventManager;
    private Container content;
    private JMenu mEdit;
    private JScrollPane scrollpane;
    private java.util.Date startDate;
    private int iCrawledURLs;
    // Singleton
    private static CrawlerFrame crawlerFrame = null;
    /**
     * Geschtzter Konstruktor fr das CrawlerFrame Objekt.
     */
    protected CrawlerFrame()
    {
       super("Crawler", true,
       //resizable
       true,
       //closable
       true,
       //maximizable
       true);
       //iconifiable
       setDefaultCloseOperation(HIDE_ON_CLOSE);
       // Aufbau der GUI
       if (miTable.getState())
       {
          this.setSize(session.defaultFrameWidth, session.defaultFrameHeight);
       } else
       {
          this.setSize(session.defaultFrameWidth, 80);
       }
       this.setTitle(lc.get("cFTitle"));
       content = this.getContentPane();
       content.setLayout(new BorderLayout());
       crawlerTable = new URLClassTableModel()
       {
          public boolean isCellEditable(int x, int y)
          {
             return false;
          }
       };
       crawlerTable.setColumnNames(columnNames);
       table = new JTable(crawlerTable);
       MouseAdapter listMouseListener = new MouseAdapter()
       {
          public void mouseClicked(MouseEvent e)
          {
             //                TableColumnModel columnModel = table.getColumnModel();
          }
       };
       table.addMouseListener(listMouseListener);
       table.setPreferredScrollableViewportSize(new Dimension(625, 620));
       TableColumn column = null;
       for (int i = 0; i < 5; i++)
       {
          column = table.getColumnModel().getColumn(i);
          if (i == 0)
          {
             column.setPreferredWidth(170);
          } else if (i == 1)
          {
             column.setPreferredWidth(120);
          } else if (i == 2)
          {
             column.setPreferredWidth(120);
          } else
          {
             column.setPreferredWidth(50);
          }
       }
       label.setFont(f);
       table.setBackground(Color.gray);
       table.setForeground(Color.white);
       panel1.setLayout(new BorderLayout());
       panel1.add(label, BorderLayout.CENTER);
       panel1.add(activeThreads, BorderLayout.EAST);
       panel2.setLayout(new FlowLayout());
       panel2.add(bView);
       panel2.add(bBrowser);
       panel2.add(bShowFeat);
       panel2.add(bShowLinks);
       panel2.add(bTrain);
       //		panel2.add(bFeatures);
       //		panel2.add(bClassify);
       scrollpane = new JScrollPane(table);
       content.add(panel1, BorderLayout.NORTH);
       if (miTable.getState())
       {
          content.add(scrollpane, BorderLayout.CENTER);
          content.add(panel2, BorderLayout.SOUTH);
       }
       // MenLeiste
       menuBar = new JMenuBar();
       setJMenuBar(menuBar);
       // Men Crawler
       JMenu mFile = new JMenu(lc.get("bDCrawler"));
       menuBar.add(mFile);
       mFile.addSeparator();
       miStart = addMenuItem(mFile, lc.get("bDCrawlerStart"));
       miStop = addMenuItem(mFile, lc.get("bDCrawlerStop"));
       miStop.setEnabled(false);
       miReset = addMenuItem(mFile, lc.get("bDCrawlerReset"));
       miReset.setEnabled(true);
       mFile.addSeparator();
       JMenuItem miClose = addMenuItem(mFile, lc.get("close"));
       // Bearbeiten
       mEdit = new JMenu(lc.get("edit"));
       menuBar.add(mEdit);
       JMenuItem miCut = addMenuItem(mEdit, lc.get("cut"));
       JMenuItem miCopy = addMenuItem(mEdit, lc.get("copy"));
       JMenuItem miPaste = addMenuItem(mEdit, lc.get("paste"));
       mEdit.addSeparator();
       JMenuItem miSelectAll = addMenuItem(mEdit, lc.get("selectAll"));
       mEdit.addSeparator();
       JMenuItem miRemove = addMenuItem(mEdit, lc.get("clearSelection"));
       JMenuItem miClear = addMenuItem(mEdit, lc.get("deleteTable"));
       JMenuItem miRemoveFalse = addMenuItem(mEdit, lc.get("deleteFalse"));
       mEdit.addSeparator();
       JMenuItem miSearch = addMenuItem(mEdit, lc.get("search"));
       // Ansicht
       JMenu mView = new JMenu(lc.get("cFView"));
       menuBar.add(mView);
       JMenu mTable = new JMenu(lc.get("cFViewOutput"));
       mTable.add(miTable);
       mView.add(mTable);
       JMenu mVis = new JMenu(lc.get("cFViewVisual"));
       mVis.add(miVis);
       mView.add(mVis);
       
       // ActionListener fr alle Mens registrieren
       bShowFeat.addActionListener(this);
       bShowLinks.addActionListener(this);
       //		bFeatures.addActionListener(this);
       bView.addActionListener(this);
       bTrain.addActionListener(this);
       bBrowser.addActionListener(this);
       //		bClassify.addActionListener(this);
       miStart.addActionListener(this);
       miReset.addActionListener(this);
       miStop.addActionListener(this);
       miClear.addActionListener(this);
       miClose.addActionListener(this);
       miSearch.addActionListener(this);
       miSelectAll.addActionListener(this);
       miCut.addActionListener(this);
       miCopy.addActionListener(this);
       miPaste.addActionListener(this);
       miRemoveFalse.addActionListener(this);
       miRemove.addActionListener(this);
       miTable.addItemListener(this);
       miVis.addItemListener(this);
       if (miTable.getState())
       {
          buildTable(vURLs, vLabels);
       } else
       {
          mEdit.setEnabled(false);
       }
       if (miVis.isSelected())
       {
          visualization = new Visualization();
          visualization.setVisible(true);
       }
       setVisible(true);
    }
    /**
     * Interfacemethode zu BaseTreeCaller. ndert die Kategorie eines
     * Tabelleneintrags.
     *
     * @param s
     *            Die Kategorie als String.
     * @param dmtn
     *            Die Kategorie als Baumstruktur.
     * @param rank
     *            The new path value
     */
    public void setPath(String s, DefaultMutableTreeNode dmtn, Double rank)
    {
                /*
                 * try { table.setValueAt( s, table.getSelectedRow(), 1 );
                 * table.setValueAt( String.valueOf( rank.doubleValue() ),
                 * table.getSelectedRow(), 3 ); ( (BaseURL) vURLs.elementAt(
                 * table.getSelectedRow() ) ).setNode( dmtn ); ( (BaseURL)
                 * vURLs.elementAt( table.getSelectedRow() ) ).setSVMRank( rank ); if (
                 * SessionBuffer.dbConnect ) { Link link = new Link((String)
                 * table.getValueAt( table.getSelectedRow(), 0 ) );
                 * session.dbinterface.updateSVMResult(link.id, s, rank ); link = null; } }
                 * catch ( Exception any ) {}
                 */
    }
    public void setSelectedIndex(int i)
    {
       try
       {
          table.setRowSelectionInterval(i, i);
       } catch (Exception any)
       {
       }
    }
    /**
     * Liefert den Vector mit den gecrawlten BaseURLs.
     *
     * @return BaseURLs aus der Tabelle.
     * @since
     */
    public Vector getCrawlerURLs()
    {
       return vURLs;
    }
    public Vector getLabels()
    {
       return vLabels;
    }
    /**
     * berwacht die Menbox.
     *
     * @param e
     *            Das Event.
     */
    public void itemStateChanged(ItemEvent e)
    {
       if (e.getSource() != miVis)
       {
          if (e.getItem() == miTable)
          {
             setVisible(false);
             if (miTable.getState())
             {
                this.setSize(session.defaultFrameWidth,
                session.defaultFrameHeight);
                content.add(scrollpane, BorderLayout.CENTER);
                content.add(panel2, BorderLayout.SOUTH);
                mEdit.setEnabled(true);
             } else
             {
                this.setSize(session.defaultFrameWidth, 80);
                content.remove(scrollpane);
                content.remove(panel2);
                mEdit.setEnabled(false);
             }
             validate();
             setVisible(true);
          }
       } else if (e.getSource() == miVis)
       {
          if (miVis.isSelected())
          {
             visualization = new Visualization();
             visualization.setVisible(true);
          } else
          {
             if (visualization != null)
             {
                visualization.dispose();
             }
          }
       }
    }
    /**
     * berwacht Buttons und Menleiste.
     *
     * @param e
     *            Das Event.
     */
    public void actionPerformed(ActionEvent e)
    {
       String cmd = e.getActionCommand();
       try
       {
          if (cmd == (lc.get("bDCrawlerStart")))
          {
             initCrawler();
          } else if (cmd == (lc.get("close")))
          {
             disposeFrame();
          } else if (cmd == (lc.get("classifyButton")))
          {
             Vector v = new Vector();
             int[] rows = table.getSelectedRows();
             for (int i = 0; i < rows.length; i++)
             {
                v.add(vURLs.elementAt(rows[i]));
             }
             new ClassifierDialog(v, table.getSelectedRows())
             .setVisible(true);
          } else if (cmd == (lc.get("bDCrawlerStop")))
          {
             crawler.stop();
          } else if (cmd == (lc.get("bDCrawlerReset")))
          {
             if (crawler != null)
             {
                crawler.reset();
             }
             updateTable(new Vector(), new Vector());
             label.setText("");
          } else if (cmd == (lc.get("deleteTable")))
          {
             vURLs = new Vector();
             vLabels = new Vector();
             buildTable(vURLs, vLabels);
          } else if (cmd == (lc.get("deleteFalse")))
          {
             table.clearSelection();
             for (int i = 0; i < vURLs.size(); i++)
             {
                if (((String) table.getValueAt(i, 4)).equals("false"))
                {
                   addSelection(i);
                }
             }
             clearTable(table.getSelectedRows());
             table.clearSelection();
          } else if (cmd == (lc.get("delButton")))
          {
             clearTable(table.getSelectedRows());
             table.clearSelection();
          } else if (cmd == (lc.get("computeButton")))
          {
             selectionOk();
             new FeatureDialog(BINGODesktop.getInstance(), vURLs, table.getSelectedRows()).setVisible(true);
             buildTable(vURLs, vLabels);
          }
          else if (cmd == (lc.get("featureButton")))
          {
             BingoDocument bd = (BingoDocument) vURLs.elementAt(table.getSelectedRow());
             if (bd != null)
               new FeatureViewer(this, bd).setVisible(true);
          }
          else if (cmd == ( lc.get( "linkButton" ) ) )
          {
             BingoDocument bd = (BingoDocument) vURLs.elementAt( table.getSelectedRow() );
             if (bd != null)
                new LinksViewer( this, bd ).setVisible( true );
          }
          else if (cmd == (lc.get("trainingButton")))
          {
             initializeTraining();
          }
          else if (cmd == (lc.get("cut")))
             
          {
             selectionOk();
             int[] iRows = table.getSelectedRows();
             session.clearURLBuffer();
             for (int i = 0; i < iRows.length; i++)
             {
                session.addURLtoBuffer(vURLs.elementAt(iRows[i]));
             }
             clearTable(table.getSelectedRows());
             table.clearSelection();
          } else if (cmd == (lc.get("clearSelection")))
          {
             clearTable(table.getSelectedRows());
             table.clearSelection();
          } else if (cmd == (lc.get("copy")))
          {
             selectionOk();
             int[] iRows = table.getSelectedRows();
             session.clearURLBuffer();
             for (int i = 0; i < iRows.length; i++)
             {
                session.addURLtoBuffer(vURLs.elementAt(iRows[i]));
             }
             table.clearSelection();
          }
          else if (cmd == (lc.get("paste")))
          {
             Vector v = session.getURLBuffer();
             int iRow = table.getSelectedRow();
             for (Enumeration enum = v.elements(); enum.hasMoreElements();)
             {
                Object doc = enum.nextElement();
                if (vURLs.contains(doc))
                   continue;
                
                if ((iRow > -1) && (iRow < vURLs.size()))
                {
                   vURLs.insertElementAt(doc, table.getSelectedRow());
                   vLabels.insertElementAt("ROOT-LINK", table.getSelectedRow());
                }
                else
                {
                   vURLs.addElement(doc);
                   vLabels.addElement("ROOT-LINK");
                }
             }
             buildTable(vURLs, vLabels);
          }
          else if (cmd == (lc.get("selectAll")))
          {
             table.selectAll();
          }
          else if (cmd == (lc.get("sourceButton")))
          {
             BingoDocument selectedURL = (BingoDocument) vURLs.elementAt(table.getSelectedRow());
             new ParserFrame(selectedURL);
          }
          else if ( cmd == ( lc.get( "browserButton" ) ) )
          {
             BingoDocument selectedURL = (BingoDocument) vURLs.elementAt(table.getSelectedRow());
             if (selectedURL != null)
             {
                String target = null;
                if (selectedURL.getUrl2r() != null && selectedURL.getUrl2r().getURL() != null)
                {
                   target = selectedURL.getUrl2r().getURL().toString();
                }
                else
                {
                   target = "http://" + selectedURL.getUrl();
                }
                
                if (target != null)
                {
                   try
                   {
                      bingo.util.BrowserLauncher.openURL(target);
                   }
                   catch (Exception ble)
                   {  }
                }
                target = null;
             }
          }
          else if (cmd == (lc.get("search")))
          {
             new CrawlerSearchDialog().setVisible(true);
          } else if (cmd == (lc.get("cFSettingsLink")))
          {
             ControlDialog.getInstance().setVisible(true);
          }
       } catch (Exception any)
       {
          session.log(any);
       }
    }
    
    public void initializeTraining()
    {
       int[] rows = table.getSelectedRows();
       if (rows== null || rows.length <= 0)
          return;
       
       if (session.getRootNode() == null)
          return;
       
       DBInterface dbi = null;
       if (session.dbConnect)
          dbi = DBInterface.getInstance();
       
       for (int i=0; i<rows.length; i++)
       {
          BingoDocument bd = (BingoDocument) vURLs.elementAt(table.getSelectedRow());
          if (bd == null)
             continue;
          BingoTreeNode hisNode = bd.getNode();
          if (hisNode != null && hisNode != session.getRootNode())
          {
             hisNode.addDocument(bd);
             bd.setStatus('M');
             
             if (dbi != null)
                dbi.updateStatus(bd, 'M');
          }
          bd=null; hisNode=null;
          
       }
       rows = null; dbi=null;
    }
    
    /**
     * Generiert die Tabelle aus neuen gecrawlten URL-Kategorie-Paaren.
     *
     * @param vURLs
     *            Der Vektor mit den BaseURLs.
     * @param vLabels
     *            Description of the Parameter
     */
    public void updateTable(Vector vURLs, Vector vLabels)
    {
       this.vLabels = vLabels;
       this.vURLs = vURLs;
       buildTable(vURLs, vLabels);
    }
    /**
     * Fgt ein neues URL-Kategorie-Paar in die Tabelle ein.
     *
     * @param temp
     *            Einzutragende BaseURL.
     * @param type
     *            Ein Link-Typ. Erhalten durch StandardClassifier.classify(
     *            Page p ).
     */
    public void addNewTableEntry(BingoDocument temp, String type)
    {
       iCrawledURLs++;
       if (!miTable.getState())
          return;
       synchronized (vURLs)
       {
          synchronized (vLabels)
          {
             synchronized (table)
             {
                vURLs.addElement(temp);
                vLabels.addElement(type);
                String url = temp.getUrl();
                BingoTreeNode node = temp.getNode();
                String label = node.label;
                double svmvalue = temp.getSvmRank().doubleValue();
                boolean hasFV = temp.hasFeatureV();
                Object[] sRow =
                {url, label, type,
                 String.valueOf(svmvalue),
                 Boolean.toString(hasFV)};
                 crawlerTable.addRow(sRow);
                 if ((crawlerTable.getDataVector().size() % 1000) == 0)
                 {
                    try
                    {
                       vURLs = new Vector();
                       vLabels = new Vector();
                       crawlerTable.setDataVector(new Vector());
                       updateTable(vURLs, vLabels);
                    } catch (Exception e)
                    {
                    }
                 }
             }
          }
       }
    }
    public void addLink(BingoDocument temp)
    {
       if (!miTable.getState())
       {
          vURLs.addElement(temp);
          vLabels.addElement("ROOT-LINK");
       } else
       {
          addNewTableEntry(temp, "ROOT-LINK");
       }
    }
    /**
     * Erweitert die Tabellenselektion um eine Zeile.
     *
     * @param i
     *            Die Zeile
     */
    public void addSelection(int i)
    {
       try
       {
          table.addRowSelectionInterval(i, i);
       } catch (Exception any)
       {
       }
    }
    /**
     * Thread zum Auffrischen der Threads- und Cache- Anzeige.
     */
    public void run()
    {
       try
       {
          do
          {
             Thread.sleep(1000);
             java.util.Date dist = new java.util.Date(new java.util.Date()
             .getTime()
             - startDate.getTime());
             String text = "";
             text += lc.get("cFTime") + ": ";
             text += new SimpleDateFormat("mm:ss", Locale.ENGLISH)
             .format(dist);
             text += " | ";
             text += String.valueOf(crawler.countActiveThreads());
             text += " " + lc.get("cFThreads");
             activeThreads.setText(text);
             label.setText(" " + lc.get("cFCrawled") + ": "
             + crawler.countCrawled() + "    " + lc.get("cFVisited")
             + ": " + iCrawledURLs + "   " + lc.get("cFSuspended")
             + ": " + crawler.countSuspended() + "   "
             + lc.get("cFSlowHosts") + ": "
             + crawler.countSlowHosts());
          } while (!crawler.isStopped());
       } catch (Exception any)
       {
       }
       activeThreads.setText(lc.get("cFCrawlerStop"));
    }
    /**
     * Initialisiert den Crawler mit einem Vektor von ROOT-URLs.
     *
     */
    public void initCrawler()
    {
       if (!session.useLinkTable)
       {
          session.log("initInteractiveCrawler()", 7);
          initInteractiveCrawler();
       } else
       {
          session.log("initAutoCrawler()", 7);
          initAutoCrawler();
       }
    }
    public void initInteractiveCrawler()
    {
       if (vURLs == null || vURLs.size() == 0) // alle Bookmarks benutzen
       {
          vURLs = new Vector();
          BingoTreeNode node = session.getRootNode();
          if (node == null)
             return;
          Enumeration enum = node.children();
          while (enum.hasMoreElements())
          {
             BingoTreeNode child = (BingoTreeNode) enum.nextElement();
             ArrayList docs = child.getPathDocuments();
             if (docs != null && docs.size() > 0)
             {
                vURLs.addAll(docs);
             }
             child = null;
             docs = null;
          }
          enum = null;
          node = null;
          if (vURLs.size() <= 0)
             return;
       }
       SessionBuffer.log(lc.get("cFCrawlerStart") + ": " + new java.util.Date().toString(), 1);
       session.bShowParserDialog = true;
       label.setText("");
       startDate = new java.util.Date();
       if (session.offlineCrawling)
       {
          crawler = new OfflineCrawler();
       }
       else
       {
          crawler = session.getCrawler();
       }
       
       crawler.resetLinkListeners();
       crawler.resetCrawlListeners();

       // Register Visualization
       if (miVis.getState())
       {
          if (visualization == null)
          {
             visualization = Visualization.getInstance();
          }
          visualization.show();
          visualization.connect(crawler);
       }
       // Register Listener
       linkEventManager = new LinkEventManager(this);
       crawler.addLinkListener(linkEventManager);
       crawler.addCrawlListener(this);
       BingoDocument temp;
       int i = 0;
       LinkHandler lh = new LinkHandler();
       for (Enumeration enum = vURLs.elements(); enum.hasMoreElements(); i++)
       {
          BingoDocument nextRootURL = (BingoDocument) enum.nextElement();
          lh.handle(crawler, nextRootURL);
       }
       lh = null;
       updateTable(new Vector(), new Vector());
       // Starte Anzeige
       iCrawledURLs = 0;
       new Thread(this).start();
       crawler.go();
    }
    public void initAutoCrawler()
    {
       DBInterface dbinterface = DBInterface.getInstance();
       if (dbinterface == null || !dbinterface.isValid())
       {
          SessionBuffer.log("Problem der Datenbank", 2);
          return;
       }
       ArrayList myLinks = dbinterface.getStartLinks();
       if (myLinks == null || myLinks.size() == 0)
       {
          SessionBuffer.log("Keine Start-URLs in der Datenbank", 1);
          return;
       }
       SessionBuffer.log(lc.get("cFCrawlerStart") + ": "
       + new java.util.Date().toString(), 1);
       session.bShowParserDialog = true;
       label.setText("");
       startDate = new java.util.Date();
       if (session.offlineCrawling)
          crawler = new OfflineCrawler();
       else
          crawler = session.getCrawler();
       
       crawler.resetLinkListeners();
       crawler.resetCrawlListeners();

       
       // Register Visualization
       if (miVis.getState())
       {
          if (visualization == null)
             visualization = Visualization.getInstance();
          visualization.show();
          visualization.connect(crawler);
       }
       // Register Listener
       linkEventManager = new LinkEventManager(this);
       crawler.addLinkListener(linkEventManager);
       crawler.addCrawlListener(this);
       Iterator it = myLinks.iterator();
       while (it.hasNext())
       {
          URL tempURL = UrlVerifier.checkAbsoluteLink((String) it.next());
          if (tempURL != null)
          {
             url2resolve tempU = UrlVerifier.verifyURL(tempURL);
             tempURL = null;
             
             if (tempU != null && ! tempU.isBad())
             {
                tempU.setPrio((float) 2.0);
                
                BingoDocument target = new BingoDocument();
                target.setUrl2r(tempU);
                target.setNode(session.getRootNode());
                // target.setStatus(bingoDoc.getStatus());
                target.setStatus('L');
                target.setDepth(0);
                target.setPriority((float)2.0);
                target.setClassDistance(0);
               
                crawler.addLink(tempU);
             }
             tempU = null;
          }
       }
       it = null;
       myLinks = null;
       updateTable(new Vector(), new Vector());
       // Starte Anzeige
       iCrawledURLs = 0;
       new Thread(this).start();
       crawler.go();
    }
    /**
     * Interfacemethode zu CrawlListener: Behandelt das Rcksetzen des Crawlers
     *
     * @param e
     *            Das CrawlEvent
     */
    public void cleared(CrawlEvent e)
    {
       SessionBuffer.log(lc.get("cFCrawlerReset") + ": "
       + new java.util.Date().toString(), 1);
       miStart.setEnabled(true);
       miReset.setEnabled(true);
       miStop.setEnabled(false);
    }
    /**
     * Interfacemethode zu CrawlListener: Behandelt eine Crawlerpause.
     *
     * @param e
     *            Das CrawlEvent
     */
    public void paused(CrawlEvent e)
    {
       miStart.setEnabled(false);
       miReset.setEnabled(false);
       miStop.setEnabled(true);
    }
    /**
     * Interfacemethode zu CrawlListener: Behandelt einen Crawlerstart.
     *
     * @param e
     *            Das CrawlEvent
     */
    public void started(CrawlEvent e)
    {
       miStart.setEnabled(false);
       miReset.setEnabled(false);
       miStop.setEnabled(true);
    }
    /**
     * Interfacemethode zu CrawlListener: Behandelt einen Crawlerstop.
     *
     * @param e
     *            Das CrawlEvent
     */
    public void stopped(CrawlEvent e)
    {
       SessionBuffer.log(lc.get("cFCrawlerStop") + ": "
       + new java.util.Date().toString(), 1);
       miStart.setEnabled(true);
       miReset.setEnabled(true);
       miStop.setEnabled(false);
    }
    /**
     * Interfacemethode zu CrawlListener: Behandelt ein Crawler-Timeout
     *
     * @param e
     *            Das CrawlEvent
     */
    public void timedOut(CrawlEvent e)
    {
       SessionBuffer.log(lc.get("cFCrawlerTimeOut") + ": "
       + new java.util.Date().toString(), 1);
       miStart.setEnabled(true);
       miReset.setEnabled(true);
       miStop.setEnabled(false);
    }
    /**
     * Interfacemethode zu SVMTreeListener. Dokument wurde Kategorieknoten
     * zugeordnet.
     *
     * @param d
     *            Das Klassifikationsergebnis (negativ).
     * @param n
     *            Der Knoten.
     * @param comment
     *            Ein Kommentar.
     */
    public void classified(double d, DefaultMutableTreeNode n, String comment)
    {
    }
    /**
     * Interfacemethode zu SVMTreeListener. Dokument wurde Kategorieknoten
     * nicht zugeordnet.
     *
     * @param d
     *            Das Klassifikationsergebnis (negativ).
     * @param n
     *            Der Knoten.
     * @param comment
     *            Ein Kommentar.
     */
    public void notClassified(double d, DefaultMutableTreeNode n, String comment)
    {
    }
    /**
     * Interfacemethode zu SVMTreeListener. Knotenwechsel in der
     * Baumhierarchie.
     *
     * @param n
     *            Der neue Knoten.
     */
    public void nodeChanged(DefaultMutableTreeNode n)
    {
    }
    /**
     * Interfacemethode zu SVMTreeListener. Erfolgreiche Klassifikation.
     *
     * @param svmRanks
     *            Description of the Parameter
     * @param n
     *            Description of the Parameter
     * @param comment
     *            Description of the Parameter
     */
    public void finished(Vector svmRanks, DefaultMutableTreeNode n,
    String comment)
    {
    }
    /**
     * Interfacemethode zu SVMTreeListener. Behandelt eine Exception, die beim
     * Klassifizieren geworfen wurde.
     *
     * @param e
     *            Die Exception.
     */
    public void handleException(Exception e)
    {
                /*
                 * if ( e.getMessage().equals( SVMTreeClassifier.NO_SPARSE ) ) {
                 * return; } crawler.stop(); if ( session.bShowParserDialog ) { new
                 * NoModelDialog( BINGODesktop.getInstance(), lc.get( "cFClassifierExc" ),
                 * e.getMessage(), crawler ).setVisible( true ); }
                 */
    }
    private void disposeFrame()
    {
       if (crawler != null)
       {
          label.setText("");
          crawler.stop();
       }
       setVisible(false);
    }
    private boolean isInArray(int i, int[] ii)
    {
       for (int k = 0; k < ii.length; k++)
       {
          if (i == ii[k])
          {
             return true;
          }
       }
       return false;
    }
    private JMenuItem addMenuItem(JMenu target, String text)
    {
       JMenuItem mi = new JMenuItem(text);
       target.add(mi);
       return mi;
    }
    private synchronized void buildTable(Vector vURLs, Vector vLabels)
    {
       this.vURLs = vURLs;
       this.vLabels = vLabels;
       crawlerTable.removeAllRows();
       for (int i = 0; i < vURLs.size(); i++)
       {
          BingoDocument temp = (BingoDocument) vURLs.elementAt(i);
          Object[] sRow =
          {temp.getUrl(), temp.getClassString(),
           (String) vLabels.elementAt(i),
           String.valueOf(temp.getSvmRank().doubleValue()),
           String.valueOf(temp.hasFeatureV())};
           crawlerTable.addRow(sRow);
       }
       crawlerTable.fireTableDataChanged();
    }
    private void clearTable(int[] iRows)
    {
       Vector tempURLs = new Vector();
       Vector tempLabels = new Vector();
       for (int i = 0; i < vURLs.size(); i++)
       {
          BingoDocument temp = (BingoDocument) vURLs.elementAt(i);
          if (!isInArray(i, iRows))
          {
             tempURLs.addElement(temp);
             tempLabels.addElement(vLabels.elementAt(i));
          }
       }
       vURLs = tempURLs;
       vLabels = tempLabels;
       buildTable(vURLs, vLabels);
    }
    private void selectionOk()
    {
       int[] rows = table.getSelectedRows();
       if (rows.length == 0)
       {
          table.setRowSelectionInterval(0, vURLs.size() - 1);
          return;
       }
    }
    /**
     * Die Methode liefert die (einzige) Instanz der Klasse.
     *
     * @return Die Instanz.
     */
    public static CrawlerFrame getInstance()
    {
       if (crawlerFrame == null)
       {
          crawlerFrame = new CrawlerFrame();
          BINGODesktop.getInstance().add(crawlerFrame);
       }
       return crawlerFrame;
    }
}
