package bingo.crawler;

//import graph.WebGraph;

import graph.WebGraph;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FileDialog;
import java.awt.GridLayout;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map.Entry;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JColorChooser;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSlider;
import javax.swing.JSplitPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.border.EtchedBorder;
import javax.swing.border.TitledBorder;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.table.TableCellRenderer;

import bingo.language.LanguageConstants;
import bingo.util.BINGODesktop;
import bingo.util.SessionBuffer;
import bingo.util.URLClassTableModel;

/**
 *  Dialog fr den Crawler. Dient zur Festlegung fokussierter Klassen beim
 *  Crawl. Der Crawler akzeptiert nur Links mit entsp	rechender
 *  Kategoriezuordnung und bestimmtem Mindestschwellwert, der dem Link durch den
 *  SVMTreeClassifier zugeordnet wurde.
 *
 * @author     T
 * @created    15. Juni 2001
 */
public class IconDialog extends JDialog implements ActionListener, ChangeListener
{
   static HashSet systemSet;
   
   private static LanguageConstants lc = LanguageConstants.getInstance();
   private JCheckBox automatic;
   
   public Toolkit toolkit;
   
   private JSlider slThr;
   private JSlider slRest;
   private JSlider slSpr;
   private JSlider slChr;
   
   private HashMap myIcons;
   
   private JMenuBar JMenues;
   private JButton ok = new JButton( lc.get( "ok" ) );
   private JButton bNew = new JButton( lc.get( "idNewClass" ) );
   private JButton bRem = new JButton( lc.get( "idDeleteClass" ) );
   private JButton bSave = new JButton( lc.get( "idSave" ) );
   
   private URLClassTableModel eventModel;
   private URLClassTableModel actionModel;
   private final JTable eventTable;
   private final JTable actionTable;
   
   private Visualization visualization;
   
   public final static Dimension RIGID_DIMENSION = new Dimension( 1, 3 );
   
   public void setVisualization(Visualization vis )
   {
      this.visualization = vis;
   }
   
   protected IconDialog( Visualization vis )
   {
      super( vis, lc.get( "idTitle" ), false );
      this.visualization = vis;
      setSize( 400, 500 );
      
      // Aufbau der GUI
      Container content = this.getContentPane();
      
      String[] eventNames =
      {lc.get( "idEvent" ), lc.get( "idIcon" )};
      String[] actionNames =
      {lc.get( "idClass" ), lc.get( "idIcon" )};
      
      final JColorChooser tcc = new JColorChooser( visualization.graph.getBackground() );
      
      tcc.getSelectionModel().addChangeListener(
      new ChangeListener()
      {
         public void stateChanged( ChangeEvent e )
         {
            Color newColor = tcc.getColor();
            
            visualization.graph.setBackground( newColor );
         }
      }
      );
      tcc.setBorder( BorderFactory.createTitledBorder( lc.get( "idBGColor" ) ) );
      
      final JColorChooser ecc = new JColorChooser( visualization.graph.getEdgeColor() );
      
      ecc.getSelectionModel().addChangeListener(
      new ChangeListener()
      {
         public void stateChanged( ChangeEvent e )
         {
            Color newColor = ecc.getColor();
            
            visualization.graph.setEdgeColor( newColor );
            visualization.graph.changedGraph();
         }
      }
      );
      ecc.setBorder( BorderFactory.createTitledBorder( lc.get( "idEdgeColor" ) ) );
      
      final JColorChooser tipcc = new JColorChooser( visualization.graph.getTipColor() );
      
      tipcc.getSelectionModel().addChangeListener(
      new ChangeListener()
      {
         public void stateChanged( ChangeEvent e )
         {
            Color newColor = tipcc.getColor();
            
            visualization.graph.setTipColor( newColor );
         }
      }
      );
      tipcc.setBorder( BorderFactory.createTitledBorder( lc.get( "idHintColor" ) ) );
      
      eventModel =
      new URLClassTableModel()
      {
         public boolean isCellEditable( int x, int y )
         {
            return false;
         }
      };
      
      eventModel.setColumnNames( eventNames );
      eventTable = new JTable( eventModel );
     
      MouseAdapter eventMouseListener =
      new MouseAdapter()
      {
         public void mouseClicked( MouseEvent e )
         {
            if ( eventTable.getSelectedColumn() == 1 )
            {
               try
               {
                  byte[] bulk = loadImage();
                  if (bulk == null || bulk.length<=0)
                     return;
                  myIcons.put(eventTable.getValueAt(eventTable.getSelectedRow(), 0), bulk);
                  eventTable.setValueAt( toolkit.createImage(bulk), eventTable.getSelectedRow(), 1 );
                  bulk = null;
               }
               catch ( Exception any )
               {
                  SessionBuffer.log(any);
               }
            }
         }
      };
      
      eventTable.addMouseListener( eventMouseListener );
      eventTable.setPreferredScrollableViewportSize( new Dimension( 400, 240 ) );
      eventTable.getColumnModel().getColumn( 0 ).setPreferredWidth( 200 );
      
      eventTable.getColumnModel().getColumn( 1 ).setCellRenderer( new CustomCellRenderer() );
      eventTable.setRowHeight( 30 );
      
      JScrollPane eventPane = new JScrollPane( eventTable );
      
      actionModel =
      new URLClassTableModel()
      {
         public boolean isCellEditable( int x, int y )
         {
            return true;
         }
      };
      
      actionModel.setColumnNames( actionNames );
      actionTable = new JTable( actionModel );
      
      MouseAdapter actionMouseListener =
      new MouseAdapter()
      {
         public void mouseClicked( MouseEvent e )
         {
         	if ( actionTable.getSelectedColumn() == 1 )
            {
               try
               {
                  byte[] bulk = loadImage();
                  if (bulk == null || bulk.length<=0)
                     return;
                  myIcons.put(actionTable.getValueAt(actionTable.getSelectedRow(), 0), bulk);
                  actionTable.setValueAt( toolkit.createImage(bulk), actionTable.getSelectedRow(), 1 );
                  bulk = null;
               }
               catch ( Exception any )
               {
                  SessionBuffer.log(any);
               }
            }
         }
      };
      
      actionTable.addMouseListener( actionMouseListener );
      actionTable.setPreferredScrollableViewportSize( new Dimension( 400, 240 ) );
      actionTable.getColumnModel().getColumn( 0 ).setPreferredWidth( 200 );
      
      actionTable.getColumnModel().getColumn( 1 ).setCellRenderer( new CustomCellRenderer() );
      actionTable.setRowHeight( 30 );
      
      JScrollPane actionPane = new JScrollPane( actionTable );
      JSplitPane sp = new JSplitPane( JSplitPane.VERTICAL_SPLIT, eventPane, actionPane );
      
      sp.setDividerLocation( 200 );
      
      Container tables = new Container();
      
      tables.setLayout( new BorderLayout() );
      
      JPanel buttons = new JPanel();
      
      buttons.add( bNew );
      buttons.add( bRem );
      buttons.add(bSave);
      tables.add( sp, BorderLayout.CENTER );
      tables.add( buttons, BorderLayout.SOUTH );
      
      Container constrain = new Container();
      
      constrain.setLayout( new GridLayout( 5, 1, 5, 5 ) );
      
      automatic = new JCheckBox( lc.get( "idAutoOrder" ), true );
      automatic.addChangeListener( this );
      
      JPanel p0 = new JPanel();
      
      p0.setBorder( new TitledBorder( new EtchedBorder(), lc.get( "idGraphLayout" ) ) );
      p0.setLayout( new BoxLayout( p0, BoxLayout.Y_AXIS ) );
      p0.add( Box.createRigidArea( RIGID_DIMENSION ) );
      p0.add( automatic );
      p0.add( Box.createRigidArea( RIGID_DIMENSION ) );
      constrain.add( p0 );
      
      slThr = new JSlider( JSlider.HORIZONTAL, 0, 100, 10 );
      slThr.setPaintLabels( true );
      slThr.setMajorTickSpacing( 10 );
      slThr.setMinorTickSpacing( 5 );
      slThr.setPaintTicks( true );
      slThr.addChangeListener( this );
      
      JPanel p1 = new JPanel();
      
      p1.setBorder( new TitledBorder( new EtchedBorder(), lc.get( "idIteration" ) ) );
      p1.setLayout( new BoxLayout( p1, BoxLayout.Y_AXIS ) );
      p1.add( Box.createRigidArea( RIGID_DIMENSION ) );
      p1.add( slThr );
      p1.add( Box.createRigidArea( RIGID_DIMENSION ) );
      constrain.add( p1 );
      
      slRest = new JSlider( JSlider.HORIZONTAL, 0, 100, 10 );
      slRest.setPaintLabels( true );
      slRest.setMajorTickSpacing( 10 );
      slRest.setMinorTickSpacing( 5 );
      slRest.setPaintTicks( true );
      slRest.addChangeListener( this );
      
      JPanel p2 = new JPanel();
      
      p2.setBorder( new TitledBorder( new EtchedBorder(), lc.get( "idEdgeLength" ) ) );
      p2.setLayout( new BoxLayout( p2, BoxLayout.Y_AXIS ) );
      p2.add( Box.createRigidArea( RIGID_DIMENSION ) );
      p2.add( slRest );
      p2.add( Box.createRigidArea( RIGID_DIMENSION ) );
      constrain.add( p2 );
      
      slSpr = new JSlider( JSlider.HORIZONTAL, 0, 100, 10 );
      slSpr.setPaintLabels( true );
      slSpr.setMajorTickSpacing( 10 );
      slSpr.setMinorTickSpacing( 5 );
      slSpr.setPaintTicks( true );
      slSpr.addChangeListener( this );
      
      JPanel p3 = new JPanel();
      
      p3.setBorder( new TitledBorder( new EtchedBorder(), lc.get( "idNodeAttract" ) ) );
      p3.setLayout( new BoxLayout( p3, BoxLayout.Y_AXIS ) );
      p3.add( Box.createRigidArea( RIGID_DIMENSION ) );
      p3.add( slSpr );
      p3.add( Box.createRigidArea( RIGID_DIMENSION ) );
      constrain.add( p3 );
      
      slChr = new JSlider( JSlider.HORIZONTAL, 0, 100, 10 );
      slChr.setPaintLabels( true );
      slChr.setMajorTickSpacing( 10 );
      slChr.setMinorTickSpacing( 5 );
      slChr.setPaintTicks( true );
      slChr.addChangeListener( this );
      
      JPanel p4 = new JPanel();
      
      p4.setBorder( new TitledBorder( new EtchedBorder(), lc.get( "idNodeTension" ) ) );
      p4.setLayout( new BoxLayout( p4, BoxLayout.Y_AXIS ) );
      p4.add( Box.createRigidArea( RIGID_DIMENSION ) );
      p4.add( slChr );
      p4.add( Box.createRigidArea( RIGID_DIMENSION ) );
      constrain.add( p4 );
      
      JTabbedPane tp = new JTabbedPane();
      
      tp.addTab( lc.get( "idIcons" ), tables );
      tp.addTab( lc.get( "idLayout" ), constrain );
      tp.addTab( lc.get( "idEdges" ), ecc );
      tp.addTab( lc.get( "idBG" ), tcc );
      tp.addTab( lc.get( "idHints" ), tipcc );
      
      content.add( tp, BorderLayout.CENTER );
      content.add( ok, BorderLayout.SOUTH );
      
      // MenLeiste
      JMenues = new JMenuBar();
      setJMenuBar( JMenues );
      
      // Men Dialog
      JMenu mFile = new JMenu( lc.get( "idDialog" ) );
      
      JMenues.add( mFile );
      
      JMenuItem miClose = addMenuItem( mFile, lc.get( "close" ) );
      
      // Hilfe auskommentiert
    /*
    JMenu mHelp = new JMenu( lc.get("help") );
    JMenues.add( mHelp );
    JMenuItem miHelp = addMenuItem( mHelp, lc.get("helpLink") );
     */
      // ActionListener fr alle Mens registrieren
      ok.addActionListener( this );
      bNew.addActionListener( this );
      bRem.addActionListener( this );
      bSave.addActionListener(this);
      miClose.addActionListener( this );
      setLocationRelativeTo( vis );
      
      toolkit = Toolkit.getDefaultToolkit();
      myIcons = readActionModel();
      if (myIcons == null)
        myIcons = new HashMap(50);
      initIcons();
      refreshIcons();
   }
   
   public void refreshIcons()
   {
      if (visualization == null || visualization.graph == null)
         return;
      
      visualization.graph.requestFocus();
      visualization.graph.removeIcons();
      
      Iterator it = myIcons.entrySet().iterator();
      while (it.hasNext())
      {
        Entry entry = (Entry)it.next();
        visualization.graph.setIcon( (String)entry.getKey(), toolkit.createImage((byte[])entry.getValue()) );
        entry=null;
      }   
      it=null;
       
      if (myIcons.get(LinkEvent.eventName[LinkEvent.DOWNLOADED]) != null)
        visualization.graph.setPageIcon( toolkit.createImage((byte[])myIcons.get(LinkEvent.eventName[LinkEvent.DOWNLOADED])));
      if (myIcons.get(LinkEvent.eventName[LinkEvent.QUEUED]) != null)
        visualization.graph.setLinkIcon( toolkit.createImage((byte[])myIcons.get(LinkEvent.eventName[LinkEvent.QUEUED])));
      if (myIcons.get(LinkEvent.eventName[LinkEvent.RETRIEVING]) != null)
        visualization.graph.setRetrievingIcon( toolkit.createImage((byte[])myIcons.get(LinkEvent.eventName[LinkEvent.RETRIEVING])));
      if (myIcons.get(LinkEvent.eventName[LinkEvent.ERROR]) != null)
        visualization.graph.setErrorIcon( toolkit.createImage((byte[])myIcons.get(LinkEvent.eventName[LinkEvent.ERROR])));
   }
   
   /**
    *  berwacht die Buttons.
    *
    * @param  e  Das Event.
    */
   public void actionPerformed( ActionEvent e )
   {
      String cmd = e.getActionCommand();
      
      try
      {
         eventTable.getCellEditor().stopCellEditing();
      }
      catch ( NullPointerException npe )
      {}
      try
      {
         actionTable.getCellEditor().stopCellEditing();
      }
      catch ( NullPointerException npe )
      {}
      
      try
      {
         if ( cmd == ( lc.get( "ok" ) ) || cmd == ( lc.get( "close" ) ) )
         {
            super.dispose();
            refreshIcons();
         }
         else if ( cmd == ( lc.get( "idNewClass" ) ) )
         {
            Object[] row =
            {
               "", 
			   WebGraph.defaultPageIcon};
               actionModel.addRow( row );
               actionTable.clearSelection();
               actionTable.setRowSelectionInterval( actionModel.getRowCount() - 1, actionModel.getRowCount() - 1 );

         } 
         else if ( cmd == ( lc.get( "idDeleteClass" ) ) )
         {
            actionModel.removeRow( actionTable.getSelectedRow() );
         } 
         else if ( cmd == ( lc.get( "idSave" ) ) )
         {
            this.saveActionModel();
         } 
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
   }
   
   
   private JMenuItem addMenuItem( JMenu target, String text )
   {
      JMenuItem mi = new JMenuItem( text );
      
      target.add( mi );
      return mi;
   }
   
   
   public void stateChanged( ChangeEvent e )
   {
      visualization.graph.setAutomaticLayout( automatic.isSelected() );
      visualization.graph.setThreshold( slThr.getValue() * 20 );
      visualization.graph.setRestLength( slRest.getValue() * 10 );
      visualization.graph.setSpringConstant( slSpr.getValue() * 10 );
      visualization.graph.setNodeCharge( slChr.getValue() * 1000 );
   }
   
   
   private void initIcons()
   {
      eventModel.removeAllRows();
      actionModel.removeAllRows();
      
      Toolkit toolkit = Toolkit.getDefaultToolkit();
      
      HashSet tmpSet = new HashSet(LinkEvent.eventName.length);
      for (int i=0; i<LinkEvent.eventName.length; i++)
      {
         tmpSet.add(LinkEvent.eventName[i]);
         Object[] row = new Object[2];
         row[0] = LinkEvent.eventName[i];
         if (myIcons.get(LinkEvent.eventName[i]) == null)
            row[1] = WebGraph.defaultLinkIcon;
         else
            row[1] = toolkit.createImage( (byte[]) (myIcons.get(LinkEvent.eventName[i])) );
         eventModel.addRow( row );
         row = null;
      }
      synchronized (this.getClass())
      {systemSet = tmpSet;}
      tmpSet = null;
      
      Iterator it = myIcons.entrySet().iterator();
      while (it.hasNext())
      {
         Entry entry = (Entry)it.next();
         String imageName = (String)entry.getKey();
         byte[] imageBulk = (byte[])entry.getValue();
         if (systemSet.contains(imageName) || imageBulk == null || imageBulk.length == 0)
            continue;
         
         Object[] row = new Object[2];
         row[0] = imageName;
         row[1] = toolkit.createImage(imageBulk);
         actionModel.addRow(row);
         imageName=null; imageBulk=null;
      }
      it=null;
      toolkit = null;
   }
   
   
   private byte[] loadImage() throws Exception
   {
      FileDialog fileDialog = new FileDialog( BINGODesktop.getInstance(), lc.get( "idLoadIcon" ) );
      
      fileDialog.setVisible( true );
      if ( fileDialog.getFile() == null )
         return null;
      
      byte[] bulk = null;
      try
      {
         String sFileName = fileDialog.getDirectory() + fileDialog.getFile();
         File file = new File(sFileName);
         sFileName=null;
         if (!file.exists())
            return null;
         
         long len = file.length();
         bulk = new byte[(int)len];
         FileInputStream fis = new FileInputStream(file);
         fis.read(bulk);
         fis.close(); fis=null;
         file=null;
      }
      catch (Exception e)
      {SessionBuffer.log(e);}
      
      return bulk;
   }
   
   // Singleton
   private static IconDialog iconDialog = null;
   
   
   /**
    *  Die Methode liefert die (einzige) Instanz der Klasse.
    *
    * @param  vis  Description of the Parameter
    * @return      Die Instanz.
    */
   public static IconDialog getInstance( Visualization vis )
   {
      if ( iconDialog != null )
        iconDialog.setVisualization(vis);
      else
       iconDialog = new IconDialog(vis);

      return iconDialog;
   }
   
   
   /**
    *  Die Methode liefert die (einzige) Instanz der Klasse.
    *
    * @return    Die Instanz.
    */
   public static IconDialog getInstance()
   {
      return getInstance( null );
   }
   
   
   public void setSelectedIndex( int i )
   {
   }
   
   
   class CustomCellRenderer extends JLabel
   implements TableCellRenderer
   {
      
      public Component getTableCellRendererComponent(
      JTable table,
      Object value,
      boolean isSelected,
      boolean hasFocus,
      int row, int column )
      {
         setIcon( new ImageIcon( (Image) value ) );
         return this;
      }
   }
   
   private static HashMap readActionModel()
   {
      HashMap result = null;
      try
      {
         String s = System.getProperty("file.separator");
         String iconfilename = System.getProperty("user.dir") + s + "bingo"
         + s + "data" + s + "iconfile.dat";
         File iconfile = new File(iconfilename);
         if (iconfile.exists())
         {
            ObjectInputStream ois = new ObjectInputStream(new FileInputStream(iconfile));
            result = (HashMap) ois.readObject();
            ois.close(); ois =null;
         }
         iconfile=null; iconfilename=null;s=null;
      }
      catch (Exception e)
      {
         SessionBuffer.log(e);
      }
     
      return result;
   }
   
   private void saveActionModel()
   {
      if (myIcons == null)
         return;
      
      try
      {
         String s = System.getProperty("file.separator");
         String iconfilename = System.getProperty("user.dir") + s + "bingo"
         + s + "data" + s + "iconfile.dat";
         File iconfile = new File(iconfilename);
         ObjectOutputStream oos = new ObjectOutputStream(new FileOutputStream(iconfile));
         oos.writeObject(myIcons);
         oos.flush();oos.close();oos=null;
         iconfile=null; iconfilename=null; s=null;
      }
      catch (Exception e)
      {
         SessionBuffer.log(e);
      }
   }
   
}

