package bingo.crawler;

import bingo.util.BingoDocument;
import bingo.util.SessionBuffer;

/**
 * @author     T
 * @created    15. Juni 2001
 */
public class LinkEventManager extends Thread implements LinkListener
{
   private CrawlerFrame crawlerFrame = null;
   
   private SessionBuffer session = SessionBuffer.getInstance();
   /**
    *  Konstruiert ein neues LinkEventManager Objekt.
    *
    * @param  cf  Der CrawlerFrame.
    */
   public LinkEventManager( CrawlerFrame cf )
   {
      crawlerFrame = cf;
   }
   
   
   /**
    *  Interfacemethode zu LinkListener: Behandelt ein LinkEvent. Falls eine URL
    *  erfolgreich besucht wurde, wird das entsprechende LinkEvent bzw. die
    *  geladene Seite in der Methode LinkEventManager.work() abgearbeitet.
    *
    * @param  e  Das LinkEvent
    */
   public void crawled( LinkEvent e )
   {
      if ( e.getID() == LinkEvent.VISITED )
      {
         work( e );
      }
   }
   
   
   /**
    *  Behandelt ein LinkEvent in einem neuen Thread abhngig von
    *  Programmeinstellungen und DB-Anbindung.
    *
    * @param  linkEvent  Das LinkEvent.
    */
   public synchronized void work( LinkEvent linkEvent )
   {
      if ( linkEvent.getID() != LinkEvent.VISITED )
         return;
      
      BingoDocument bd = linkEvent.getLink();
      
      /*if (bd.getClassifiedNode() == null) 
      	bd.setClassifiedNode(session.getRootNode());
      */
      
      if (bd.getNode() == null) bd.setNode(session.getRootNode());
      
      /*BaseUrl temp = new BaseURL( link.getURL(), link.getClassifiedNode(), page.getContentType() );
      
      temp.setSVMRank( page.getSVMRank() );
      temp.setFeatures( page.getFeatures() );*/
      
      if ( crawlerFrame != null )
         //crawlerFrame.addNewTableEntry( temp, String.valueOf( link.getDepth() + "<" + String.valueOf( link.classDistance ) + "> [" + page.getContentType() + "] " + link.getObjectLabels() ) );
      	crawlerFrame.addNewTableEntry( bd, 
      								   String.valueOf( bd.getDepth() 
      								   	+ "<" 
										+ String.valueOf(bd.getClassDistance()) 
										+ "> [" + bd.getContentType() + "] "));
   }
}

