package bingo.crawler;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.Random;
import java.util.Vector;

import bingo.crawler.frontier.url2resolve;
import bingo.db.DBInterface;
import bingo.util.BingoDocument;
import bingo.util.SessionBuffer;
import bingo.crawler.frontier.OfflineQueue;

//=====================================================================================

public class OfflineCrawler implements URL_Crawler, Runnable
{
   private SessionBuffer session = SessionBuffer.getInstance();
   short HTTP_ERROR = 10;
   short HTTP_TIMEOUT = 20;
   public static int crawled = 0;
   public static int garbageBatch = 200;
   public static OfflineQueue queue;
   static Random rd;
   public static boolean stopped = false;
   public static Vector worms;
   static int max_worms = 50;
   int attempts = 0;
   static int max_attempts = 2;
   public static long lifetime = 30 * 1000;
   // Wartezeit des Threads (falls Leerlauf)
   static long waitingtime = 90 * 1000;
   // max Wartezeit auf Download der URL
   private static Vector crawlListeners;
   private static Vector linkListeners;
   private url2resolve ziel = null;
   public boolean ready = false;
   public DBInterface dbinterface = null;
   private int state = CrawlEvent.CLEARED;
   private static final byte ERROR = 0;
   private static final byte TIMEOUT = 1;
   private static final byte DONE = 2;
   private static final byte SUSPEND = 3;
   private static final byte BAD = 4;
   //=====================================================
   public OfflineCrawler()
   {
      synchronized (this.getClass())
      {
         if (queue == null)
            queue = new OfflineQueue();
         if (worms == null)
            worms = new Vector();
         if (linkListeners == null)
            linkListeners = new Vector();
         if (crawlListeners == null)
            crawlListeners = new Vector();
      }
      
      if (session.dbType != null)
         dbinterface = DBInterface.getInstance();
      
      waitingtime = session.downloadTimeout;
      max_worms = session.maxCrawlerThreads;
   }
   
   public OfflineCrawler(url2resolve initial)
   {
      this();
      this.ziel = initial;
      waitingtime = session.downloadTimeout;
      max_worms = session.maxCrawlerThreads;
   }
   //=====================================================
   public void run()
   {
      ready = false;
      attempts = 0;
      if (session.dbType != null
      && (dbinterface == null || !dbinterface.isValid()))
      {
         synchronized (worms)
         {
            worms.remove(this);
         }
         return;
      }
      synchronized (worms)
      {
         if (!worms.contains(this))
         {
            worms.add(this);
         }
      }
      state = CrawlEvent.STARTED;
      sendCrawlEvent(state);
      while (true)
      {
         if (stopped)
         {
            if (dbinterface != null)
            {
               dbinterface.close();
            }
            synchronized (worms)
            {
               worms.remove(this);
               worms.notifyAll();
            }
            return;
         }
         if (ziel == null)
         {
            ziel = queue.getBest();
         }
         if (ziel == null)
         {
            ready = true;
            if (attempts >= max_attempts)
            {
               if (dbinterface != null)
               {
                  dbinterface.close();
               }
               dbinterface = null;
               if (worms != null)
                  synchronized (worms)
                  {
                     worms.remove(this);
                     worms.notifyAll();
                  }
                  return;
            } else
            {
               attempts++;
            }
         } else
         {
            attempts = 0;
            ready = false;
            synchronized (worms)
            {
               if (worms.size() < max_worms)
               {
                  // ggf. einen neuen Thread starten
                  url2resolve tempZiel = queue.getBest();
                  if (tempZiel != null)
                  {
                     OfflineCrawler tmp = new OfflineCrawler(tempZiel);
                     worms.add(tmp);
                     new Thread(tmp).start();
                     tempZiel = null;
                  }
               }
            }
         }
         //===== Inhaltlich interessante Arbeit des Crawlers
         // =========================
         if (!ready)
         {
            // Checken, ob die connection noch OK ist
            if (dbinterface != null)
            {
               dbinterface.verifyConnection();
               if (dbinterface == null || !dbinterface.isValid())
               {
                  synchronized (worms)
                  {
                     worms.remove(this);
                     worms.notifyAll();
                  }
                  return;
               }
            }
            workOnIt();
            // Die Bearbeitung der geholten URL
         } else
         {
            try
            {
               synchronized (queue.semaphore)
               {
                  queue.semaphore.wait(lifetime);
               }
            } catch (InterruptedException e)
            {
            }
         }
      }
   }
   //== Diese Methode bearbeitet die URL = ==================================
   public synchronized void workOnIt()
   {
      BingoDocument doc = ziel.parentDoc;
      if (doc == null)
      {
         finish(DONE);
         return;
      }
      
      doc.setEventStatus(LinkEvent.RETRIEVING);
      sendLinkEvent(doc, LinkEvent.RETRIEVING);
      
      try
      {
         dbinterface.fillDocument(doc);
      }
      catch (Exception e)
      {
         doc.setEventStatus(LinkEvent.ERROR);
         sendLinkEvent(doc, LinkEvent.ERROR,
         new IOException("Download Problem!"));
         session.log(e);
      }
      
      Collection links = doc.getLinks();
      if (links != null && links.size() > 0)
      {
         Iterator it = links.iterator();
         while (it.hasNext())
         {
            BingoDocument bd = (BingoDocument)it.next();
            url2resolve u2r = bd.getUrl2r();
            if (u2r != null)
               addLink(u2r);
            //System.out.println ("Will Crawlen: " + bd.getUrl2r().getURL().toString());
         }
      }
      
      finish(DONE);
   }
   
   public void finish(byte how)
   {
      queue.unsetLock(ziel);
      if (how == DONE)
      {
         ziel.parentDoc.setEventStatus(LinkEvent.VISITED);
         sendLinkEvent(ziel.parentDoc, LinkEvent.VISITED);
         if (ziel.parentDoc.getStatus() != 'B'
         && ziel.parentDoc.getStatus() != 'J')
            destroy();
      }
      
      synchronized (this.getClass())
      {
         crawled++;
      }
      
      ziel = null;
      ready = true;
   }
   
   //=============================================================
   
   public void destroy()
   {
      BingoDocument bd;
      if (ziel != null && ziel.parentDoc != null)
      {
         bd = ziel.parentDoc;
         bd.setAnchors(null);
         bd.setGoogle(null);
         bd.setSvmRank(null);
         bd.setFeatureVector(null);
         bd.setContentType(null);
         bd.setContentEncoding(null);
         /*bd.setRootNode(null);
         bd.setClassifiedNode(null);*/
         bd.setNode(null);
         bd.setSourceLink(null);
         bd.setUrl2r(null);
         bd.setHost(null);
         bd.setFilename(null);
         bd.setDirectory(null);
      }
      if (ziel != null)
      {
         ziel.parentDoc = null;
         ziel.url = null;
         ziel.IP = null;
      }
      bd = null;
      ziel = null;
      
      if (crawled++ % garbageBatch == 0)
      {
         session.log("GARBAGE initiated", 3);
         System.gc();
      }
   }
   //========================================================================
   // Links in die queue hinzufgen
   public synchronized void addLink(url2resolve u)
   {
      //session.log("BingoCrawler.addLink(" + u + ")!", 9);
      if (u==null)
         return;
      queue.put(u);
      // sendLinkEvent(u.parentDoc, LinkEvent.QUEUED);
   }
   
   //========================================================================
   // Statusabfragen, starten, stoppen des Crawlers
   public boolean isIdle()
   {
      int countRunning = worms.size();
      int countProgress = queue.getNumProgress();
      int countOut = queue.getNumOutgoing();
      session.log("--------------------------------------------", 5);
      session.log("Running Worms: " + countRunning, 5);
      session.log("Progress: " + (countProgress), 5);
      session.log("outgoing: " + countOut, 5);
      session.log("--------------------------------------------", 5);
      if ((countRunning + countProgress + countOut) == 0)
      {
         return true;
      } else
      {
         return false;
      }
   }
   public synchronized void reset()
   {
      worms = new Vector();
      queue = new OfflineQueue();
      crawlListeners = new Vector();
      linkListeners = new Vector();
      crawled = 0;
      rd = new Random();
      state = CrawlEvent.CLEARED;
      sendCrawlEvent(state);
      session.log("Crawler reset: done", 3);
   }
   public int countCrawled()
   {
      return crawled;
   }
   public int countSuspended()
   {
      return queue.getNumSuspended();
   }
   public int countSlowHosts()
   {
      return queue.getNumSlowHosts();
   }
   public int countActiveThreads()
   {
      int res = 0;
      synchronized (worms)
      {
         Enumeration enum = worms.elements();
         while (enum.hasMoreElements())
         {
            if (!((BINGOCrawler) enum.nextElement()).ready)
            {
               res++;
            }
         }
      }
      return res;
   }
   public boolean isStopped()
   {
      return stopped;
   }
   public synchronized void go()
   {
      stopped = false;
      new Thread(new OfflineCrawler()).start();
   }
   public synchronized void stop()
   {
      stopped = true;
      synchronized (queue.semaphore)
      {
         // Alle wartenden Threads benachrichtigen
         queue.semaphore.notifyAll();
      }
      session.log("Shutdown all worms.. ", 3);
      while (worms.size() > 0)
      {
         try
         {
            synchronized (worms)
            {
               if (worms.size() > 0)
                  worms.wait();
            }
         } catch (InterruptedException ie)
         {
            session.log(ie);
         }
      }
      session.log("Shutdown done.", 3);
                /*
                 * try { Thread.sleep( waitingtime + 500 ); } catch ( Exception e ) {}
                 */
      state = CrawlEvent.STOPPED;
      sendCrawlEvent(state);
      session.log("Crawler stopped, " + worms.size() + " worms running", 3);
   }
   // ========================================================================
   // ************************************************************************
   // Die nachfolgenden Methoden haben mit der Funktion des Crawlers
   // nichts zu tun, sie sind fr Animationen und hnlichen Unsinn zustndig
   //=========================================================================
   /**
    * Liefert den Status des Crawlers.
    *
    * @return eins aus CrawlEvent.STARTED, PAUSED, STOPPED, CLEARED.
    */
   public int getState()
   {
      return state;
   }
   /**
    * Registriert einen CrawlListener.
    *
    * @param listen
    *            Der CrawlListener.
    */
   public void addCrawlListener(CrawlListener listen)
   {
      if (!crawlListeners.contains(listen))
      {
         crawlListeners.addElement(listen);
      }
   }
   /**
    * Lscht einen registrierten CrawlListeners.
    *
    * @param listen
    *            Der CrawlListener.
    */
   public void removeCrawlListener(CrawlListener listen)
   {
      crawlListeners.removeElement(listen);
   }
   /**
    * Registriert einen LinkListener.
    *
    * @param listen
    *            Der LinkListener.
    */
   public void addLinkListener(LinkListener listen)
   {
      if (!linkListeners.contains(listen))
      {
         linkListeners.addElement(listen);
      }
   }
   /**
    * Lscht einen registrierten LinkListener.
    *
    * @param listen
    *            LinkListener.
    */
   public void removeLinkListener(LinkListener listen)
   {
      linkListeners.removeElement(listen);
   }
   /**
    * Sendet ein CrawlEvent an alle registrierten CrawlListener.
    *
    * @param id
    *            Die Event ID.
    */
   protected void sendCrawlEvent(int id)
   {
      CrawlEvent evt = new CrawlEvent(this, id);
      for (int j = 0, len = crawlListeners.size(); j < len; ++j)
      {
         CrawlListener listen = (CrawlListener) crawlListeners.elementAt(j);
         switch (id)
         {
            case CrawlEvent.STARTED :
               listen.started(evt);
               break;
            case CrawlEvent.STOPPED :
               listen.stopped(evt);
               break;
            case CrawlEvent.CLEARED :
               listen.cleared(evt);
               break;
            case CrawlEvent.TIMED_OUT :
               listen.timedOut(evt);
               break;
            case CrawlEvent.PAUSED :
               listen.paused(evt);
               break;
         }
      }
   }
   /**
    * Sendet ein LinkEvent an alle registrierten LinkListener.
    *
    * @param l
    *            Der Link.
    * @param id
    *            Die Event ID.
    */
   protected void sendLinkEvent(BingoDocument bd, int id)
   {
      try
      {
         LinkEvent evt = new LinkEvent(this, id, bd);
         bd.setEventStatus(id);
         for (int j = 0, len = linkListeners.size(); j < len; ++j)
         {
            LinkListener listen = (LinkListener) linkListeners.elementAt(j);
            listen.crawled(evt);
            session.log(listen + " crawled:" + bd.getUrl() + "@"
            + LinkEvent.eventName[id], 5);
         }
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   /**
    * Sendet ein LinkEvent an alle registrierten LinkListener, falls whrend
    * der Abarbeitung eine Exception aufgetreten ist.
    *
    * @param l
    *            Der Link.
    * @param id
    *            Die Event ID.
    * @param exception
    *            Die Exception.
    */
   protected void sendLinkEvent(BingoDocument bd, int id, Throwable exception)
   {
      try
      {
         if (state == CrawlEvent.STOPPED || state == CrawlEvent.CLEARED
         || state == CrawlEvent.TIMED_OUT)
         {
            return;
         }
         LinkEvent evt = new LinkEvent(this, id, bd, exception);
         bd.setEventStatus(id);
         bd.setStatus('L');
         //bd.setLabel("exception", exception.toString());
         for (int j = 0, len = linkListeners.size(); j < len; ++j)
         {
            LinkListener listen = (LinkListener) linkListeners.elementAt(j);
            listen.crawled(evt);
         }
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   public DBInterface getDBInterface()
   {
      return dbinterface;
   }
   public long getLifetime()
   {
      return lifetime;
   }
   public void setLifetime(long t)
   {
      lifetime = t;
   }
   
   public void putSuspended(url2resolve u)
   {
      return;
   }
   
   public boolean isAlreadyParsed(BingoDocument bd)
   {
      return queue.isAlreadyParsed(bd);
   }
   
   public void resetLinkListeners()
   {
     this.linkListeners = new Vector();
   }
   
   public void resetCrawlListeners()
   {
     this.crawlListeners = new Vector();
   } 
   
}