package bingo.crawler;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JScrollPane;
import javax.swing.JTable;

import bingo.crawler.handler.UrlVerifier;
import bingo.util.BINGODesktop;
import bingo.util.BingoDocument;
import bingo.util.SessionBuffer;
import bingo.util.URLClassTableModel;

/**
 * Dialog fr den Crawler. Dient zum Anzeigen und ndern von Start-URLs des
 * Crawlings.
 *
 * @author T @created 15. Juni 2001
 */
public class RootDialog extends JDialog implements ActionListener
{
   
   private JMenuBar JMenues;
   
   private JButton ok = new JButton("OK");
   
   private URLClassTableModel tableModel;
   private final JTable table;
   
   private SessionBuffer session =SessionBuffer.getInstance();
   
   protected RootDialog(JFrame parentFrame)
   {
      super(parentFrame, "Root-URLs des Crawlers", false);
      setSize(680, 280);
      
      // Aufbau der GUI
      Container content = this.getContentPane();
      
      String[] columnNames =
      { "ROOT-URL", "Klasse" };
      
      tableModel = new URLClassTableModel()
      {
         public boolean isCellEditable(int x, int y)
         {
            return true;
         }
      };
      
      tableModel.setColumnNames(columnNames);
      table = new JTable(tableModel);
      
                /*
                 * MouseAdapter listMouseListener = new MouseAdapter() { public void
                 * mouseClicked( MouseEvent e ) { } };
                 */
      //table.addMouseListener( listMouseListener );
      table.setPreferredScrollableViewportSize(new Dimension(680, 240));
      
      table.setBackground(Color.gray);
      table.setForeground(Color.white);
      
      JScrollPane scrollpane = new JScrollPane(table);
      
      content.add(scrollpane, BorderLayout.CENTER);
      content.add(ok, BorderLayout.SOUTH);
      
      // MenLeiste
      JMenues = new JMenuBar();
      setJMenuBar(JMenues);
      
      // Men ROOT-URLs
      JMenu mFile = new JMenu("URL");
      
      JMenues.add(mFile);
      
      JMenuItem miNew = addMenuItem(mFile, "Neu");
      JMenuItem miRemove = addMenuItem(mFile, "Lschen");
      
      mFile.addSeparator();
      
      JMenuItem miClose = addMenuItem(mFile, "Schliessen");
      
      // Bearbeiten
      JMenu mEdit = new JMenu("Bearbeiten");
      
      JMenues.add(mEdit);
      
      JMenuItem miClear = addMenuItem(mEdit, "Alle lschen");
      
      // ActionListener fr alle Mens registrieren
      ok.addActionListener(this);
      miNew.addActionListener(this);
      miRemove.addActionListener(this);
      miClear.addActionListener(this);
      miClose.addActionListener(this);
      
      setLocationRelativeTo(BINGODesktop.getInstance());
   }
   
   private JMenuItem addMenuItem(JMenu target, String text)
   {
      JMenuItem mi = new JMenuItem(text);
      
      target.add(mi);
      return mi;
   }
   
   /**
    * berwacht den Button.
    *
    * @param e
    *            Das Event.
    */
   public void actionPerformed(ActionEvent e)
   {
      try
      {
         table.getCellEditor().stopCellEditing();
      } catch (NullPointerException npe)
      {
      }
      
      String cmd = e.getActionCommand();
      
      try
      {
         if (cmd.equals("OK") || cmd.equals("Schliessen"))
         {
            getRoots();
            super.dispose();
         } else if (cmd.equals("Neu"))
         {
            String[] row =
            { "", "" };
            
            tableModel.addRow(row);
         } else if (cmd.equals("Lschen"))
         {
            removeSelected();
         } else if (cmd.equals("Alle lschen"))
         {
            tableModel.removeAllRows();
         }
      } catch (Exception any)
      {
         SessionBuffer.log(any);
      }
   }
   
   public void setRoots(Vector rootURLs)
   {
      tableModel.removeAllRows();
      
      if (rootURLs == null || rootURLs.size() == 0)
      {
         return;
      }
      
      for (Enumeration enum = rootURLs.elements(); enum.hasMoreElements();)
      {
         BingoDocument root = (BingoDocument) enum.nextElement();
         String[] row =
         { root.getUrl(), root.getClassString()};
         
         tableModel.addRow(row);
      }
   }
   
   public void getRoots()
   {
      Vector rootURLs = new Vector();
      Vector labels = new Vector();
      
      Vector vURLs = CrawlerFrame.getInstance().getCrawlerURLs();
      Vector vLabels = CrawlerFrame.getInstance().getLabels();
      
      for (int i = 0; i < tableModel.getRowCount(); i++)
      {
         String sURL = (String) table.getValueAt(i, 0);
         String sClass = (String) table.getValueAt(i, 1);
         
         try
         {
            BingoDocument temp = null;
            
            for (Enumeration enum = vURLs.elements();
            enum.hasMoreElements();
            )
            {
               BingoDocument b = (BingoDocument) enum.nextElement();
               
               if (sURL.equals(b.getUrl()))
               {
                  temp = b;
                  break;
               }
            }
            
            if (temp == null)
            {
               temp = UrlVerifier.string2BingoDocument(sURL);
               if (temp!=null)
                  temp.setNode(session.getRootNode());
            }
            //				if (sClass.startsWith("ROOT/") || sClass.equals("UNKNOWN/")) {
            //					temp.setNode(BaseURL.stringToNode(sClass));
            //				}
            rootURLs.add(temp);
            
            try
            {
               labels.add(vLabels.elementAt(i));
            } catch (ArrayIndexOutOfBoundsException ae)
            {
               labels.add("ROOT-LINK");
            }
            
         } catch (Exception e)
         {
            SessionBuffer.log(e);
         }
         
                        /*
                         * catch ( MalformedURLException e ) { SessionBuffer.log(e);
                         */
         
      }
      CrawlerFrame.getInstance().updateTable(rootURLs, labels);
   }
   
   private void removeSelected()
   {
      tableModel.removeRow(table.getSelectedRow());
   }
   
   // Singleton
   private static RootDialog rootDialog = null;
   
   /**
    * Die Methode liefert die (einzige) Instanz der Klasse.
    *
    * @param frame
    *            Description of the Parameter
    * @return Die Instanz.
    */
   public static RootDialog getInstance(JFrame frame)
   {
      if (rootDialog == null)
      {
         rootDialog = new RootDialog(frame);
      }
      return rootDialog;
   }
   
   /**
    * Die Methode liefert die (einzige) Instanz der Klasse.
    *
    * @return Die Instanz.
    */
   public static RootDialog getInstance()
   {
      if (rootDialog == null)
      {
         rootDialog = new RootDialog(null);
      }
      return rootDialog;
   }
   
   public void setSelectedIndex(int i)
   {
   }
   
   public static void main(String args[])
   {
      RootDialog r = RootDialog.getInstance();
      r.setVisible(true);
   }
   
}
