package bingo.crawler;

//import graph.WebGraph;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.ImageIcon;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JTabbedPane;

import bingo.language.LanguageConstants;
import bingo.util.BINGODesktop;
import bingo.util.SessionBuffer;
import graph.WebGraph;

/**
 *  Visualiserung der Linkstruktur in Form eines Webgraphen und als
 *  Baumstruktur.
 *
 * @author     T
 * @created    15. Juni 2001
 */
public class Visualization extends JFrame implements ActionListener {

  private static LanguageConstants lc = LanguageConstants.getInstance();
  private SessionBuffer session = SessionBuffer.getInstance();
  private JMenuBar JMenues;
  public WebGraph graph;
  private IconDialog iconDialog;

  private Font ff = new Font( "Serif", Font.BOLD, 16 );


  protected Visualization() {
    Container content = this.getContentPane();

    this.setTitle( lc.get( "vTitle" ) );
    this.setSize( session.defaultFrameWidth, session.defaultFrameHeight );

    JTabbedPane container = new JTabbedPane();

    container.setBackground( Color.white );

    graph = new WebGraph();
    graph.setRestLength( 50 );
    graph.setDownloadedIcon( WebGraph.defaultPageIcon );
    graph.setAutomaticLayout( true );
    graph.setDefaultFilter( WebGraph.NO_LINKS );
    graph.setDefaultRendering( WebGraph.ICON );
    graph.setBackground( Color.white );
    graph.setEdgeColor( Color.red );
    
    IconDialog id = IconDialog.getInstance(this);
    id.refreshIcons();
    id=null;

    //outline = new WebOutline();

    JLabel label1 = new JLabel( LinkEvent.eventName[LinkEvent.QUEUED], new ImageIcon( graph.linkIcon ), JLabel.LEFT );
    JLabel label2 = new JLabel( LinkEvent.eventName[LinkEvent.RETRIEVING], new ImageIcon( graph.retrievingIcon ), JLabel.LEFT );
    JLabel label3 = new JLabel( LinkEvent.eventName[LinkEvent.DOWNLOADED], new ImageIcon( graph.pageIcon ), JLabel.LEFT );
    JLabel label4 = new JLabel( LinkEvent.eventName[LinkEvent.ERROR], new ImageIcon( graph.errorIcon ), JLabel.LEFT );

    label1.setFont( ff );
    label2.setFont( ff );
    label3.setFont( ff );
    label4.setFont( ff );

    JPanel panel = new JPanel();

    panel.setLayout( new FlowLayout() );
    panel.add( label1 );
    panel.add( label2 );
    panel.add( label3);
    panel.add( label4 );

    container.addTab( lc.get( "vGraph" ), graph );
    //container.addTab( lc.get( "vOutline" ), outline );
    content.add( container, BorderLayout.CENTER );
    content.add( panel, BorderLayout.SOUTH );

    // MenLeiste
    JMenues = new JMenuBar();
    setJMenuBar( JMenues );

    // Men Vis
    JMenu mFile = new JMenu( lc.get( "vVisual" ) );

    JMenues.add( mFile );

    JMenuItem miClose = addMenuItem( mFile, lc.get( "close" ) );

    // Bearbeiten
    JMenu mEdit = new JMenu( lc.get( "vSettings" ) );

    JMenues.add( mEdit );

    JMenuItem miIcons = addMenuItem( mEdit, lc.get( "vSettingsIconDialog" ) );

    mEdit.add( miIcons );

    /*
    // Hilfe
    JMenu mHelp = new JMenu( lc.get("help") );
    JMenues.add( mHelp );
    JMenuItem miHelp = addMenuItem( mHelp, lc.get("helpLink") );
    */
    // ActionListener fr alle Mens registrieren
    miIcons.addActionListener( this );
    miClose.addActionListener( this );

    setLocationRelativeTo( BINGODesktop.getInstance() );
    setVisible( true );

  }


  private JMenuItem addMenuItem( JMenu target, String text ) {
    JMenuItem mi = new JMenuItem( text );

    target.add( mi );
    return mi;
  }


  /**
   *  berwacht den Button.
   *
   * @param  e  Das Event.
   */
  public void actionPerformed( ActionEvent e ) {
    String cmd = e.getActionCommand();

    try {
      if ( cmd == ( lc.get( "close" ) ) ) {
        disconnect();
        super.dispose();
      } else if ( cmd == ( lc.get( "vSettingsIconDialog" ) ) ) {
        iconDialog = IconDialog.getInstance( this );
        iconDialog.setVisible( true );
      }
    }
    catch ( Exception any ) {}
  }


  /**
   *  Verbindet einen Crawler mit der Visualisierung.
   *
   * @param  crawler  Der Crawler.
   */
  public void connect( URL_Crawler crawler ) {
    graph.clear();
    crawler.addCrawlListener( graph );
    crawler.addLinkListener( graph );

  }


  /**
   *  Beendet die Verbindung zum Crawler.
   */
  public void disconnect() {
    graph.clear();
    //outline.clear();
  }


  private static Visualization vis = null;


  public static Visualization getInstance() {
    if ( vis == null ) {
      vis = new Visualization();
    }
    return vis;
  }
}

