package bingo.crawler.frontier;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.Arrays;

import bingo.util.MyInteger;
import bingo.util.SessionBuffer;

//=======================================================================================

public class Class_Queue implements Comparable, timedListCallback
{
   static float loadFactor = (float)1.0;
   static int max_attempts = 0;
   
   static int queueBatch = 100;
   
   public Object[] sortedQueue;
   
   public int sortedQueueCounter = 0;
   public int batchCounter = 0;
   
   public int whereStartSearch = 0;
   public int accessCount = 0;
   public int accessCountBatch = 0;
   
   public timedListOrdered suspended;
   
   public float access;
   static MyInteger counter;
   public int id;
   
   public HashSet visited;
   public HashSet inserted;
   
   private URL_Queue parent;
   
   public int overflow = 0;
   
   // parameter der Suspended-Schlange
   static long bad_ttl = 6 * 3600 * 1000;
   // 6 Std TTL bei Fehlern (DNS Timeout etc.)
   static long cleaningInterval = 10 * 60 * 1000;
   // 10 Min cleaning-intervall von timedList
   
   // Max suspended URLs
   public static int suspMax = 50;
   
   // max URLs in der Queue
   public static int queueMax = 10000;
   public String ontoClass;
   private SessionBuffer session;
   
   public Class_Queue( String myclass, URL_Queue p )
   {
      session = SessionBuffer.getInstance();
      if (session.maxQueueSize > 0)
         queueMax = session.maxQueueSize;
      
      sortedQueue = new Object[queueMax + queueBatch];
      for (int i=0; i<sortedQueue.length; i++)
         sortedQueue[i] = getDummy();
      sortedQueueCounter = sortedQueue.length - 1;
      batchCounter = 0;
      
      whereStartSearch = sortedQueue.length - 1;
      accessCount = 0;
      accessCountBatch = session.maxCrawlerThreads / 2;
      
      suspended = new timedListOrdered( cleaningInterval, suspMax, this );
      
      parent = p;
      
      synchronized ( this.getClass() )
      {
         if ( counter == null )
            counter = new MyInteger( 1 );
      }
      
      synchronized ( counter )
      {
         counter.increase();
         this.id = counter.intValue();
      }
      
      ontoClass = myclass;
      overflow = 0;
      access = 0;
   }
   
   
   //=============  Methods for suspended Links ===========================
   
   public void addSuspended( url2resolve link )
   {
      link.attempts++;
      if ( link.attempts < max_attempts )
      {
         link.setTTL( bad_ttl );
         link.setCreated( new Date().getTime() );
         synchronized ( suspended )
         {
            suspended.put( link );
         }
      }/*
        else {
            // actually - do nothing (forget this link)
            // bleibt dabei in den locks drin - also aufgegeben
        }*/
   }
   
   
   public void destroy( url2resolve ziel )
   {
      if ( ziel != null )
      {
         ziel.parentDoc = null;
         ziel.url = null;
         ziel.IP = null;
      }
   }
   
   
   public void takeBack( Object u )
   {
      session.log( "Take back", 5 );
      addLink( (url2resolve) u );
   }
   
   //========================================================
   
   public void addLink( url2resolve link )
   {
      if (link == null)
         return;
      if (link.isBad())
         return;
      Object[] toForget = null;
      
      synchronized ( sortedQueue )
      {
         sortedQueue[sortedQueueCounter] = link;
         sortedQueueCounter--;
         
         batchCounter++;
         
         if (batchCounter % queueBatch == 0)  // reorganize Queue
         {
            Arrays.sort(sortedQueue);
            sortedQueueCounter = queueBatch-1;
            toForget = new Object[queueBatch];
            for (int i = sortedQueueCounter; i>=0; i--)
            {
               if ( isDummy(sortedQueue[i]) )
                  break;
               else
               {
                  toForget[i] = sortedQueue[i];
                  sortedQueue[i] = getDummy();
               }
            }
         }
      }
      
      synchronized ( parent.semaphore )
      {
         parent.semaphore.notify();
      }
      
      synchronized (this)
      {this.notifyAll();}
      
      if (toForget!=null && toForget.length > 0)
      {
         for (int i = toForget.length-1; i>=0; i--)
         {
            if (toForget[i] == null)
               break;
            else
            {
               forget(toForget[i]);
               toForget[i] = null;
            }
         }
      }
      toForget=null;
   }
   
   public url2resolve getDummy()
   {
      url2resolve result = new url2resolve();
      result.setPrio(Float.NaN);
      return result;
   }
   
   public boolean isDummy(Object u)
   {
      return Float.isNaN(((url2resolve)u).getPrio() );
   }
   
   //======================================================
   
   public void forget( Object u )
   {
      parent.forget( u );
   }
   
   
   //======================================================
   
   public url2resolve getBest()
   {
      if (sortedQueue == null)
         return null;
      
      url2resolve tmp = null;
      boolean found = false;
      
      accessCount ++;
      if (accessCount > accessCountBatch)
      {
        accessCount = 1;
        whereStartSearch = sortedQueue.length - 1;
      }  
      
      if (whereStartSearch < 0)
         return null;
      
      synchronized ( sortedQueue )
      {
         int pos = whereStartSearch;
         while (!found && pos >= 0)
         {
            if ( isDummy(sortedQueue[pos]) )
            {
               pos--;
               continue;
            }
            
            tmp = (url2resolve) sortedQueue[pos];
            if ( !parent.isLocked( tmp ) )
            {
               access += 0.001;
               sortedQueue[pos] = getDummy();
               found = true;
            }
            else
            {
               pos--;
               tmp = null;
            }
         }
         whereStartSearch = pos - 1;
      }
      
      if ( found )
         synchronized(this)
         {this.notifyAll();}
         
         return tmp;
   }
   
   //===== Comparable Interface ==================================
   
   public int compareTo( Object o )
   {
      int result = 0;
      
      if ( o == null )
         result = +1;
      else if ( access > ( (Class_Queue) o ).access )
         result = +1;
      else if ( access < ( (Class_Queue) o ).access )
         result = -1;
      else if ( this.id < ( (Class_Queue) o ).id )
         result = -1;
      else if ( this.id > ( (Class_Queue) o ).id )
         result = +1;
      
      return result;
   }
   
   //===== Comparable Interface ==================================
   
}
