package bingo.crawler.frontier;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;

import bingo.util.BingoDocument;
import bingo.util.MyInteger;
import bingo.util.SessionBuffer;

//=======================================================================================

public class OfflineQueue
{
   static float loadFactor = (float)0.9;
   
   public HashMap queues;
   static int queuesSize = 200;
   
   static HashSet visitedPath;
   static int visitedPathSize = 10000000;
   
   static int kenn_ich_schon = 0;
   
   static timedList slowHosts;
   static long cleaningInterval = 60 * 1000;
   // 1 Min cleaning-intervall von timedList
   static long bad_ttl = 5 * 60 * 1000;
   // 5 Min Host timeout bei Fehlern
   static int listsize = 7000;
   
   static HashMap locks;
   static int locksSize = 100;
   static int max_locks = 4;
   
   public static Object semaphore = new Object();
   
   public OfflineQueue()
   {
      
      visitedPath =  new HashSet(visitedPathSize, loadFactor);
      
      locks = new HashMap(locksSize, loadFactor);
      queues = new HashMap(queuesSize, loadFactor);
      slowHosts = new timedList( cleaningInterval, listsize );
   }
   
   //=================================================================
   
   public boolean isLocked( url2resolve u )
   {
      if (u == null || u.isBad())
         return false;
      
      String host = u.getHost();
      if (host == null || host.length() <= 0)
         return false;
      
      synchronized ( locks )
      {
         MyInteger x = (MyInteger)locks.get( host);
         host = null;
         return ( (x != null) && (x.intValue() >= max_locks));
      }
   }
   
   public boolean isAlreadyParsed(BingoDocument bd)  // Signaturbasierte Duplikaterkennung
   {
     return false;
   }
   
   //=================================================================
   
   public void setLock( url2resolve u )
   {
      if (u==null || u.isBad())
         return;
      
      String host  = u.getHost();
      if (host == null || host.length() <= 0)
         return;
      
      synchronized ( locks )
      {
         MyInteger x = (MyInteger)locks.get( host );
         
         if ( x == null )
            locks.put( host, new MyInteger( 1 ) );
         else
            x.increase();
         x=null;
      }
      host=null;
   }
   
   
   
   //=================================================================
   
   public void unsetLock( url2resolve u )
   {
      if (u==null)
         return;
      
      String host = u.getHost();
      if (host == null || host.length() <= 0)
         return;
      
      synchronized ( locks )
      {
         MyInteger x = (MyInteger)locks.get( host );
         
         if ( x == null )
         {
            SessionBuffer.log( "Warning: lost IP in locks URL_Queue", 3 );
            SessionBuffer.log( "Host: " + host, 4 );
            SessionBuffer.log(u.getStatus() + " =?= ", 4);
            SessionBuffer.log("URL: " + u.getURL(), 4);
         }
         else
         {
            x.decrease();
            if (x.intValue() <= 0)
               locks.remove(host);
         }
         x=null;
      }
      host = null;
      
      synchronized ( semaphore )
      {
         semaphore.notify();
      }
   }
   
   
   //=================================================================
   //=================================================================
   public void put( BingoDocument bd )
   {
      SessionBuffer.log("URL_Queue.put(bd)", 7);
      put( bd.getUrl2r() );
   }
   
   
   public void put( url2resolve u )
   {
      Long key = new Long(u.parentDoc.getId());
      boolean seen = false;
      synchronized ( visitedPath )
      {
         if (visitedPath.contains(key))
            seen = true;
         else
            visitedPath.add(key);
      }
      key = null;
      
      if (seen)
         return;
      
      if (u.isBad())
         return;
      
      if ( u.getOntoClass() == null)
      {
         SessionBuffer.log("Warnung: url2resolve Ontoclass NULL", 3);
         return;
      }
      
      OfflineClassQueue current = null;
      synchronized(queues)
      {
         current = (OfflineClassQueue)queues.get(u.getOntoClass());
         if (current == null)
         {
            current = new OfflineClassQueue(u.getOntoClass(),this);
            queues.put(u.getOntoClass(), current);
         }
      }
      current.addLink(u);
      SessionBuffer.log("---> Link added: " + u.getURL(), 7);
   }
   
   //=================================================================
   
   public void forget( Object u )
   {
      Long key = new Long(( (url2resolve) u ).parentDoc.getId());
      synchronized ( visitedPath )
      {
         visitedPath.remove( key );
      }
      key = null;
   }
   
   
   //=================================================================
   
   public void addVisitedPath( int k)
   {
      Long key = new Long(k);
      synchronized (visitedPath)
      {
         visitedPath.add( key );
      }
      key = null;
   }
   
   public void addVisitedPath( Object x )
   {
      url2resolve u = (url2resolve) x;
      
      if (u.isBad())
         return;
      
      Long key = new Long(u.parentDoc.getId());
      u = null;
      
      synchronized ( visitedPath )
      {
         if ( visitedPath.contains( key ) )
         {
            SessionBuffer.log( "Alarm! Key bereits drin!", 3 );
            //System.exit( 1 ); // auskommentiert 3.9.03 kaster
         }
         else
            visitedPath.add(key);
      }
      key = null;
   }
   
   
   //==================================================================
   
   public boolean isVisitedPath( Object x )
   {
      url2resolve u = (url2resolve) x;
      boolean result = false;
      
      if (u.isBad())
         return result;
      
      Long key = new Long(( u.parentDoc ).getId());
      u = null;
      
      synchronized ( visitedPath )
      {
         result = visitedPath.contains( key ) ;
      }
      key = null;
      
      return result;
   }
   
   
   //=================================================================
   //=================================================================
   
   public url2resolve getBest()
   {
      url2resolve u_tmp = null;
      if (queues==null || queues.size()<=0)
         return null;
      
      Object[] queueArray = null;
      synchronized (queues)
      {
         queueArray = queues.values().toArray();
      }
      
      Arrays.sort(queueArray);
      for (int i=0; i<queueArray.length; i++)
      {
         u_tmp = ((OfflineClassQueue)queueArray[i]).getBest();
         if (u_tmp != null)
            break;
      }
      queueArray = null;
      
      if ( u_tmp != null )
      {
         setLock( u_tmp );
         if (u_tmp.getStatus() != 0)
            SessionBuffer.log("Warnung: ausgegeben mit status != 0", 3);
      }
      
      return u_tmp;
   }
   
   //===== Debugging methods ==============================================
   
   public int getNumOutgoing()  // unpraezise, da ohne synchronisation
   {
      int a = 0;
      if (queues.size()<=0 || queues==null)
         return a;
      
      Object[] classQueues = queues.values().toArray();
      for (int i=0;i<classQueues.length; i++)
         a+= ((OfflineClassQueue)classQueues[i]).outgoing.size();
      classQueues = null;
      
      return a;
   }
   
   
   public int getNumProgress()
   {
      return 0;
   }
   
   
   public int getNumSuspended()
   {
      return 0;
   }
   
   public int getNumSlowHosts()
   {
      return 0;
   }
}