package bingo.crawler.frontier;

import java.util.*;
import java.net.*;

import bingo.util.*;

//=======================================================================================

public class URL_Queue
{
   static float loadFactor = (float)0.9;
   
   public HashMap queues;
   public Comparable[] queuesArray;
   static int queuesSize = 200;
   
   public static HashSet visitedPath;
   public static HashMap parsedPath;
   static int visitedPathSize = 10000000;    // Locking by URL-signature 
   static int parsedPathSize = 10000000;     // Locking by content-signature
  
   static int maxPerHost = 200;
   static HashMap hostCounter = null;
   
   static int kenn_ich_schon = 0;
   
   static timedList slowHosts;
   static long cleaningInterval = 60 * 1000;
   // 1 Min cleaning-intervall von timedList
   static long bad_ttl = 5 * 60 * 1000;
   // 5 Min Host timeout bei Fehlern
   static int listsize = 7000;
   
   static HashMap locks;
   static int locksSize = 100;
   static int max_locks = 4;
   
   public static Object semaphore = new Object();
   
   public MyInteger sortBatch;
   public static int sortBatchSize = 10;
   
   public URL_Queue()
   {
      sortBatch = new MyInteger(0);
      
      visitedPath =  new HashSet(visitedPathSize, loadFactor);
      parsedPath =  new HashMap(parsedPathSize, loadFactor);
      
      locks = new HashMap(locksSize, loadFactor);
      queues = new HashMap(queuesSize, loadFactor);
      slowHosts = new timedList( cleaningInterval, listsize );
      
      hostCounter = new HashMap(100000);
      
      new Thread( new QueueWatcher( this ) ).start();
   }
   
   //=================================================================
   
   public boolean isLocked( url2resolve u )
   {
      if ( u.isBad() )
         return false;
      
      String host = u.getHost();
      if (host == null || host.length() <= 0)
         return false;
      
      synchronized ( locks )
      {
         MyInteger x = (MyInteger)locks.get(host); 
         host = null;
         return ( (x != null) && (x.intValue() >= max_locks));
      }
   }
   
   
   //=================================================================
   
   public void setLock( url2resolve u )
   {
      if ( u == null )
         return;
      
      if ( u.isBad() )
         return;
      
      String host = u.getHost();
      if (host == null || host.length() <= 0)
         return;
      
      synchronized ( locks )
      {
         MyInteger x = (MyInteger)locks.get(host);
         
         if ( x == null )
            locks.put( host, new MyInteger( 1 ) );
         else
            x.increase();
         
         x=null;
      }
      host = null;
   }
   
   
   //=================================================================
   
   public void unsetLock( url2resolve u )
   {
      if (u==null)
         return;
      
      String host = u.getHost();
      if (host == null || host.length() <= 0)
         return;
      
      synchronized ( locks )
      {
         MyInteger x = (MyInteger)locks.get( host );
         
         if ( x == null )
         {
            SessionBuffer.log( "Warning: lost IP in locks URL_Queue", 3 );
            SessionBuffer.log( "Host: " + host, 4 );
            SessionBuffer.log(u.getStatus() + " =?= ", 4);
            SessionBuffer.log("URL: " + u.getURL(), 4);
         }
         else
         {
            x.decrease();
            if (x.intValue() <= 0)
               locks.remove(host);
         }
         x=null;
      }
      host = null;
      
      synchronized ( semaphore )
      {
         semaphore.notify();
      }
   }
   
   
   //=================================================================
   //=================================================================
   public void put( BingoDocument bd )
   {
      SessionBuffer.log("URL_Queue.put(bd)", 7);
      put( bd.getUrl2r() );
   }
   
   
   public void put( url2resolve u )
   {
      if (u == null)
         return;
      if (u.isBad())
         return;
      if (u.parentDoc == null)
         return;
      
      Long key = new Long(u.parentDoc.getId());
      boolean seen = false;
      synchronized ( visitedPath )
      {
         if (visitedPath.contains(key))
            seen = true;
         else
            visitedPath.add(key);
      }
      key = null;
      
      if (seen)
         return;
      
      URL url = u.getURL();
      if (url==null)
         return;
      String host = url.getHost();
      url = null;
      if (host == null || host.length() == 0)
         return;
      Integer hostHash = new Integer(host.hashCode());
      host = null;
      MyInteger myInt = (MyInteger)hostCounter.get(hostHash);
      if (myInt == null)
         hostCounter.put(hostHash, new MyInteger(1));
      else
      {
         if (myInt.intValue() < maxPerHost)
         {
           myInt.increase();
         }   
         else
         {
            return;
         }
      }
      myInt = null;
      hostHash = null;
       
      
      if ( u.getOntoClass() == null)
      {
         SessionBuffer.log("Warnung: url2resolve Ontoclass NULL", 3);
         return;
      }
      
      Class_Queue current = null;
      synchronized(queues)
      {
         current = (Class_Queue)queues.get(u.getOntoClass());
         if (current == null)
         {
            current = new Class_Queue(u.getOntoClass(),this);
            queues.put(u.getOntoClass(), current);
            
            if (queuesArray == null)
            {
               queuesArray = new Comparable[1];
            }
            else
            {
               Comparable[] newQueuesArray = new Comparable[queuesArray.length + 1];
               for (int i=0; i<queuesArray.length; i++)
                  newQueuesArray[i] = queuesArray[i];
               queuesArray = newQueuesArray;
               newQueuesArray = null;
            }
            queuesArray[queuesArray.length-1] = current;
         }
      }
      current.addLink(u);
   }
   
   //=================================================================
   
   public void forget( Object u )
   {
      if (u == null || ((url2resolve)u).parentDoc == null)
         return;
      
      Long key = new Long(( (url2resolve) u ).parentDoc.getId());
      synchronized ( visitedPath )
      {
         visitedPath.remove( key );
      }
      key = null;
   }
   
   
   //=================================================================
   
   public void addVisitedPath( long k)
   {
      Long key = new Long(k);
      synchronized (visitedPath)
      {
         visitedPath.add( key );
      }
      key = null;
   }
   
   public void addVisitedPath( Object x )
   {
      url2resolve u = (url2resolve) x;
      
      if (u.isBad())
         return;
      
      Long key = new Long(u.parentDoc.getId());
      u = null;
      
      synchronized ( visitedPath )
      {
         if ( visitedPath.contains( key ) )
         {
            SessionBuffer.log( "Alarm! Key bereits drin!", 3 );
            //System.exit( 1 ); // auskommentiert 3.9.03 kaster
         }
         else
            visitedPath.add(key);
      }
      key = null;
   }
   
   
   //==================================================================
   
   public boolean isVisitedPath( Object x )
   {
      url2resolve u = (url2resolve) x;
      boolean result = false;
      
      if (u.isBad())
         return result;
      
      Long key = new Long(( u.parentDoc ).getId());
      u = null;
      
      synchronized ( visitedPath )
      {
         result = visitedPath.contains( key ) ;
      }
      key = null;
      
      return result;
   }
   
//==================================================================
   
   public void addAlreadyParsed(long md5, long id)
   {
     if (md5 == 0) 
        return;
     
     synchronized ( parsedPath )
     {
       parsedPath.put(new Long(md5), new Long(id));
     }
   }
   
   public boolean isAlreadyParsed( BingoDocument bd )
   {
      boolean result = false;
      
      if (bd == null)
         return result;
      
      long md5 = bd.getMD5();
      
      if (md5 == 0)
         return result;
      
      synchronized ( parsedPath )
      {
         Object key = new Long(md5);
         Object value = parsedPath.get( key ) ;
         if (value != null)
         {
            result = true;
            bd.setParsedBefore( ((Long)value).longValue() );
         }  
         else
         {   
           bd.setParsedBefore( (long)0 ); 
           parsedPath.put(key, new Long(bd.getId()));
         }   
         key = null;
      }
      return result;
   }
   
   public void putSuspended( url2resolve u )
   {
      // erst den ganzen Host als schlecht vormerken (Annahme: IP ist bereits bekannt)
      // Das macht man nur, wenn dieser Host noch nicht eingetragen ist.
      /*
      synchronized ( slowHosts )
      {
         Object found = slowHosts.get( u.IP );
       
         if ( found == null )
         {
            resolvedHost rh = new resolvedHost( u.getTruehost(), null, null, Rcode.SERVFAIL, bad_ttl );
       
            slowHosts.put( u.getIP(), rh );
         }
      }
       */
      
      Class_Queue classqueue = null;
      String label = u.getOntoClass();
      synchronized (queues)
      {
         classqueue = (Class_Queue)queues.get(label);
      }
      label=null;
      
      if (classqueue != null)  // normalerweise, daher kein insert
         classqueue.addSuspended( u );
      
      classqueue=null;
   }
   
   //=================================================================
   
   public boolean hasSlowHost( url2resolve u )
   {
      String host = u.getHost();
      if (host == null || host.length() <= 0)
         return false;
      
      Object found = null;      
      synchronized ( slowHosts )
      {
         found = slowHosts.get( host );
      }
      
      return ( found != null );
   }
   
   //=================================================================
   
   public url2resolve getBest()
   {
      url2resolve u_tmp = null;
      if (queuesArray ==null || queuesArray.length <=0)
         return null;
      
      int counter = 0;
      synchronized (sortBatch)
      {
         sortBatch.increase();
         counter = sortBatch.intValue() % sortBatchSize;
         if (counter == 0)
            sortBatch.setValue(0);
      }
      if (counter == 0)
         Arrays.sort(queuesArray);
      
      for (int i=0; i<queuesArray.length; i++)
      {
         u_tmp = ((Class_Queue)queuesArray[i]).getBest();
         if (u_tmp != null)
            break;
      }
      
      if ( u_tmp != null )
      {
         setLock( u_tmp );
         if (u_tmp.isBad())
            SessionBuffer.log("Warnung: ausgegeben mit status BAD", 3);
      }
      return u_tmp;
   }
   
   //===== Debugging methods ==============================================
   
   public int getQueueSize()  // unpraezise, da ohne synchronisation
   {
      int a = 0;
      if (queuesArray == null || queuesArray.length <= 0)
         return a;
      
      for (int i=0;i<queuesArray.length; i++)
         a+= ((Class_Queue)queuesArray[i]).sortedQueue.length;
      
      return a;
   }
   
   
   public int getNumSuspended()
   {
      int a = 0;
      if (queuesArray == null || queuesArray.length <= 0)
         return a;
      
      for (int i=0;i<queuesArray.length; i++)
         a+= ((Class_Queue)queuesArray[i]).suspended.treeSet.size();
      
      return a;
   }
   
   public int getNumSlowHosts()
   {
      return slowHosts.getSize();
   }
}