package bingo.crawler.frontier;

import java.io.InterruptedIOException;
import java.io.Serializable;
import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.*;
import bingo.util.*;

public class url2resolve implements Serializable, Comparable, timedListElement
{
   static long TIMEOUT = 15 * 1000;
   // Max Wartezeit auf Verbindung
   
   static short BAD = 10;
   static short DNS_ERROR = 10;
   static short UNKNOWN = 10000;
   static short RESOLVED = 0;
   
   static int BUFFER = 1024;
   // Grsse des Buffers im BufferedReader
   
   static int maxLength = 1000;
   // max Length der absoluten URL (hier grosszgig, wird noch vom UrlVerifier kontrolliert)
   
   private float prio = 0;
   private int depth = 0;
   static MyInteger counter;
   public BingoDocument parentDoc;
   public URL url;
   // Initiale URL und die neue mit aufgelster IP
   
   short status;
   HashSet allowedProtocols = null;
   // Zustand: 0 = unbekannt, -1 = bad, +1 = good
   public Long IP;
   boolean ready = false;
   long timer = 0;
   
   int attempts = 0;
   // Anzahl frherer Versuche die Verbindung herzustellen
   
   public int id;
   
   private long ttl;
   // fr Eintrag als TimedListElement
   private long created;
   // fr Eintrag als TimedListElement
   
   public void restoreStatic()
   {
      TIMEOUT = 15 * 1000;
      BAD = 10;
      DNS_ERROR = 10;
      UNKNOWN = 10000;
      RESOLVED = 0;
      BUFFER = 1024;
      maxLength = 1000;
   }
   
   public url2resolve()
   {
      url = null;
      status = UNKNOWN;
      IP = null;
      prio = 0;
      attempts = 0;
      
      synchronized ( this.getClass() )
      {
         if ( counter == null )
         {
            counter = new MyInteger( 0 );
            restoreStatic();
         }
      }
      
      synchronized ( counter )
      {
         counter.increase();
         this.id = counter.intValue();
      }
      
      if ( allowedProtocols == null )
      {
         // genehmigte Protokolle
         allowedProtocols = new HashSet( 1 );
         allowedProtocols.add("http");
         restoreStatic();
      }
   }
   
   public url2resolve( URL u, float p, BingoDocument bd )
   {
      this();
      this.url = u;
      this.parentDoc = bd;
      this.status = UNKNOWN;
      this.prio = p;
      this.attempts = 0;
      this.IP = null;
      
      if ( url.getHost() == null || url.getHost().trim().equals( "" ) )
      {
         this.setStatus(BAD);
         this.destroy();
      }
      
      if ( url.getHost().length() > 255 )
      {
         this.setStatus(BAD);
         this.destroy();
      }
   }
   
   public url2resolve( URL orig, String u, float p, BingoDocument bd )
   {
      this();
      
      parentDoc = bd;
      status = UNKNOWN;
      prio = p;
      attempts = 0;
      IP = null;
      
      try
      {
         // URL parsen, Syntax checken
         
         if ( orig == null )
            url = new URL( u );
         else
            url = new URL( orig, u );
         
         if ( url.getHost() == null || url.getHost().trim().equals( "" ) )
         {
            // Irrefhrende leere Hostbezeichner
            throw new MalformedURLException();
         }
         
         if ( !allowedProtocols.contains( url.getProtocol() ) )
         {
            // Nur erlaubte Protokolle werden untersttzt
            throw new MalformedURLException();
         }
         
         if ( url.getHost().length() > 255 )
         {
            // Irrefhrende endlos lange Hostnamen
            throw new MalformedURLException();
         }
         
         if ( url.toString().length() > maxLength )
         {
            // so lange URLs gibt es nicht
            throw new MalformedURLException();
         }
      }
      catch ( MalformedURLException e )
      {
         url = null;
         status = BAD;
      }
      catch ( Exception e )
      {
         SessionBuffer.log(e);
         status = BAD;
      }
   }
   
   //====================================================================================
   
   public URL getURL()
   {
      return url;
   }
   
   
   public String getHost()
   {
      if ( url == null )
         return null;
      else
         return url.getHost();
   }
   
   
   public String getPath()
   {
      if ( url == null )
         return null;
      else
         return url.getFile();
      // return oldURL.getPath();  // Achtung, GET-Query teil wird abgeschnitten
   }
   
   
   public String getFile()
   {
      if ( url == null )
         return null;
      else
         return url.getFile();
   }
   
   public int getPort()
   {
      if (url == null)
         return 0;
      
      int port = url.getPort();
      
      if ( port <= 0 )
         return 80;
      else
         return port;
   }
   
   
   public void setStatus( short s )
   {
      status = s;
   }
   
   public short getStatus()
   {
      return status;
   }
   
   
   public void setIP( long x )
   {
      IP = new Long(x);
   }
   
   
   public long getIP()
   {
      if (IP == null)
         return (long)0;
      
      return IP.longValue();
   }
   
   
   public boolean isProcessed()
   {
      return ( status != UNKNOWN );
   }
   
   
   public boolean isBad()
   {
      return (status == BAD);
   }
   
   
   public String getProtocol()
   {
      if (url == null)
         return null;
      
      return url.getProtocol();
   }
   
   
   //====================================================================
   
   public HttpURLConnection getConnection() throws InterruptedIOException
   {
      HttpURLConnection connection = null;
      
      ConnectionOpener opener = new ConnectionOpener(this);
      opener.start();
      
      synchronized ( this )
      {
         try
         {
            this.wait( TIMEOUT );
         }
         catch ( InterruptedException e )
         {}
      }
      
      if ( !opener.ready )
      {
         opener.interrupt();
         throw new InterruptedIOException( "Connect problem" );
      }
      
      if ( opener.connection == null )
      {
         throw new InterruptedIOException( "Broken connection" );
      }
      
      connection = opener.connection;
      opener = null;
      return connection;
   }
   
   //====================================================================
   
   //===== Methoden des TimedList - Interfaces ==========================
   
   public long getTTL()
   {
      return ttl;
   }
   
   
   public void setTTL( long t )
   {
      ttl = t;
   }
   
   
   public long getCreated()
   {
      return created;
   }
   
   
   public void setCreated( long c )
   {
      created = c;
   }
   
   
   //====================================================================
   
   public synchronized int compareTo( Object o )
   {
      url2resolve u2r = (url2resolve) o;
      if ( u2r == null )  return  +1;
      if ( u2r.id == this.id ) return 0;
      
      if (Float.isNaN(u2r.getPrio()) && !Float.isNaN(this.getPrio()))
         return +1;
      if (!Float.isNaN(u2r.getPrio()) && Float.isNaN(this.getPrio()))
         return -1;
      if (Float.isNaN(u2r.getPrio()) && Float.isNaN(this.getPrio()))
      {
         if ( this.id > u2r.id ) return +1;
         if ( this.id < u2r.id ) return -1;
      }
      
      if ( this.prio > u2r.getPrio() ) return +1;
      if ( this.prio < u2r.getPrio() ) return -1;
      if ( this.id > u2r.id ) return +1;
      if ( this.id < u2r.id ) return -1;
      
      return 0;
   }
   
   
   public float getPrio()
   {
      return prio;
   }
   
   
   public void setPrio( float p )
   {
      prio = p;
   }
   
   
   public String getOntoClass()
   {
      if (parentDoc == null)
         return null;
      BingoTreeNode node = parentDoc.getNode();
      if (node == null)
         return null;
      return node.label;
   }
   
   //================================================================
   
   class ConnectionOpener extends Thread
   {
      url2resolve parent = null;
      HttpURLConnection connection = null;
      public boolean ready = false;
      
      public ConnectionOpener(url2resolve p)
      {this.parent = p;}
      
      public void run()
      {
         connection = null;
         ready = false;
         
         try
         {
            connection = prepareTrivialConnection();
            ready = true;
         }
         catch ( Exception e )
         {
            if ( connection != null )
               connection.disconnect();
            connection = null;
            ready = false;
            SessionBuffer.log(e);
         }
         
         if (parent != null)
            synchronized ( this.parent )
            {
               this.parent.notifyAll();
            }
      }
      
      // Der kleine checker, ob der Thread nicht unterbrochen wurde
      public void checkAlive() throws Exception
      {
         if ( this.isInterrupted() )
         {
            throw new Exception( "Interrupted" );
         }
      }
      
      // =================================================================
      // Prepare Mozilla-Faked HTTP Connection
      
      public HttpURLConnection prepareTrivialConnection() throws Exception
      {
         HttpURLConnection result = null;
         
         if (url == null || isBad())
            return result;
         
         // SessionBuffer.log("WARNUNG: trivial connection in use", 4);
         
         IP = new Long(url2resolve.resolveIP(url.getHost()));
         checkAlive();
         
         
         
         result = (HttpURLConnection) url.openConnection();
         //sun.net.www.protocol.http.Handler handler = new sun.net.www.protocol.http.Handler();
         //result = new sun.net.www.protocol.http.HttpURLConnection(url,handler);

         checkAlive();
         
         result.setFollowRedirects( true );
         result.setRequestProperty( "accept-language", "de" );
         result.setRequestProperty( "accept", "image/gif, image/x-xbitmap, image/jpeg, image/pjpeg, application/vnd.ms-excel, application/vnd.ms-powerpoint, application/msword, */*" );
         result.setRequestProperty( "user-agent", "Mozilla/4.0 (compatible; MSIE 6.01; Windows NT 5.0)" );
         result.setRequestMethod( "GET" );
         result.setAllowUserInteraction( false );
         result.setDoOutput( false );
         result.setDoInput( true );
         result.setUseCaches( false );
         result.connect();
         checkAlive();
         
         return result;
      }
   }
   
   public static long resolveIP(String hostname) throws Exception
   {
      if (hostname==null || hostname.length() <=0)
         return (long)0;
      
      InetAddress adr = InetAddress.getByName(hostname);
      long r1 = (long)(adr.getAddress()[0] & 0xff ) << 24;
      long r2 = (long)(adr.getAddress()[1] & 0xff ) << 16;
      long r3 = (long)(adr.getAddress()[2] & 0xff ) << 8;
      long r4 = (long)(adr.getAddress()[3] & 0xff ) << 0;
      
      long result = r1+r2+r3+r4;
      
      return result;
   }
   
   public void destroy()
   {
      this.IP = null;
      this.parentDoc = null;
      this.url = null;
   }
   
   public void setDepth(int d)
   {
      this.depth = d;
   }
   
   public int getDepth()
   {
      return depth;
   }
}
