package bingo.crawler.handler;

import java.net.HttpURLConnection;
import java.util.Date;

import bingo.crawler.frontier.url2resolve;
import bingo.util.BingoDocument;
import bingo.util.SessionBuffer;

//=======================================================================================

public class HandlerManager
{
    int responseCode = 0;
    String responseMessage = null;
    String contentType = null;
    int contentLength = -1;
    String contentEncoding = null;
    long lastModified = (long) 0.0;
    long expiration = (long) 0.0;
    
    private SessionBuffer session = SessionBuffer.getInstance();
    
    public CrawlHandler getHandler( url2resolve u, HttpURLConnection conn )
    {
        estimateProperties( conn );
        
        try
        {
        	BingoDocument bd =u.parentDoc;
        if (bd.getStatus() == 'L')
        session.log( responseCode 
                    + " -> " 
                    + bd.getDepth() 
                    + " " + bd.getStatus() 
                    + " " + bd.getUrl2r().getURL().toString()
                    + "[" + bd.getParentID() + "->]"
                    + " Prio: " + bd.getPriority(), 1 );

            
            bd.setResponseCode( responseCode );
            bd.setContentType( contentType );
            bd.setContentLength( contentLength );
            bd.setLastModified( lastModified );
            bd.setContentEncoding( contentEncoding );
            bd.setExpiration( expiration );
            
            responseMessage = null;
            contentType = null;
            contentEncoding = null;
            
            if ( bd.getResponseCode() == HttpURLConnection.HTTP_MOVED_PERM ||
            		bd.getResponseCode() == HttpURLConnection.HTTP_MOVED_TEMP ||
            		bd.getResponseCode() == HttpURLConnection.HTTP_SEE_OTHER ||
            		bd.getResponseCode() == HttpURLConnection.HTTP_GONE )
            {
                return new RedirectHandler();
            }
            
            if ( bd.getResponseCode() == HttpURLConnection.HTTP_INTERNAL_ERROR )
            {
                return new SuspendHandler();
            }
            
            if ( bd.getResponseCode() != HttpURLConnection.HTTP_OK )
            {
                return new ErrorHandler();
            }
/*            
            if ( link.getStatus() == 'G' )
            {
                // return new GoogleHandler();
            }
            
            if ( link.getStatus() == 'D' )
            {
                // return new GoogleDirectoryHandler();
            }
  */          
            if ( bd.getContentType().indexOf( "text/html" ) >= 0 ||
            		bd.getContentType().indexOf( "text/plain" ) >= 0 ||
            		bd.getContentType().indexOf( "text/richtext" ) >= 0 ||
            		bd.getContentType().indexOf( "application/plain" ) >= 0
            )
            {
                return new HtmlHandler();
            }
            
            if ( bd.getContentType().indexOf( "application/pdf" ) >= 0 )
            {
                return new PDFHandler();
            }

            // diese Liste wird fortgesetzt !
            
        }
        catch ( Exception e )
        {
        	session.log(e);
        }
        return new DefaultHandler();
    }
    
    
    public void estimateProperties( HttpURLConnection conn )
    {
        try
        {
            
            responseCode = conn.getResponseCode();
            if ( responseCode == -1 )
            {
                responseCode = HttpURLConnection.HTTP_NOT_IMPLEMENTED;
            }
            
            responseMessage = conn.getResponseMessage();
            if ( responseMessage == null )
            {
                responseMessage = "";
            }
            
            contentType = conn.getContentType();
            if ( contentType == null )
            {
                contentType = "text/html";
            } else
            {
                contentType = contentType.toLowerCase();
            }
            
            contentLength = conn.getContentLength();
            /*
            if ( contentLength == -1 )
            {
            }
            */
            contentEncoding = conn.getContentEncoding();
            if ( contentEncoding == null )
            {
                contentEncoding = "unknown";
            }
            
            lastModified = conn.getLastModified();
            if ( lastModified == 0 )
            {
                lastModified = new Date().getTime();
            }
            
            expiration = conn.getExpiration();
            /*
            if ( expiration == 0 )
            {
            }*/
        }
        catch ( Exception e )
        {
            session.log(e);
        }
    }
}
