package bingo.crawler.handler;

import java.io.InputStream;
import java.net.HttpURLConnection;

import bingo.crawler.URL_Crawler;
import bingo.crawler.frontier.url2resolve;
import bingo.crawler.handler.parser.HtmlParser;
import bingo.crawler.handler.parser.StringCentenizer;
import bingo.svmlight.TFBuilder;
import java.util.*;
import bingo.util.*;

//=======================================================================================

public class HtmlHandler extends CrawlHandler
{
   public static String mime = "text/html";
   
   public void handle( URL_Crawler uc, url2resolve u, HttpURLConnection c )
   {
      
      if ( !init( uc, u, c ) )
      {
         destroy();
         return;
      }
      readContent();
      
      // System.out.println (new String(byteArray, 0, byteArraySize));
      
      processContent();
      
      /*
      System.out.println ("MD5: " + bingoDoc.getMD5());
      System.out.println("Frames > " + bingoDoc.hasFrames());
      if (3<5)
      return;
       */
      
      if ( bingoDoc.isBroken() )
      {
         new ErrorHandler().handle(uc,u,c);
         destroy();
         return;
      }
      
      if ( bingoDoc.hasRedirect() )
      {
         new LinkHandler().handle(uc,u,c);
         bingoDoc.setStatus('R');
         new StoreHandler().handle(uc,u,c);
      }
      else if (bingoDoc.hasFrames())
      {
         new LinkHandler().handle(uc,u,c);
         bingoDoc.setStatus('F');
         new StoreHandler().handle(uc,u,c);
      }
      else
      {
         if (uc.isAlreadyParsed(bingoDoc))    // text kam im anderen Dokument bereits vor
         {
            bingoDoc.setStatus('D');
            bingoDoc.setAnchors(null);
            bingoDoc.setBulk(null);
            bingoDoc.setFeatureVector(null);
            bingoDoc.setLinks(null);
            bingoDoc.setText(null);
            bingoDoc.setTitle(null);
            bingoDoc.setPreview(null);
            new StoreHandler().handle( uc, u, c );
         }
         else if (bingoDoc.isForeign())   // Andere Sprache vermutet, nicht klassifizieren
         {
            session.log("Fremdsprache: " + bingoDoc.getUrl(), 1);
            bingoDoc.setStatus('X');
            bingoDoc.setAnchors(null);
            bingoDoc.setBulk(null);
            bingoDoc.setFeatureVector(null);
            // bingoDoc.setLinks(null);   // Links kann man mitspeichern
            bingoDoc.setText(null);
            bingoDoc.setTitle(null);
            bingoDoc.setPreview(null);
            
            bingoDoc.setNode(session.getRootNode());
            bingoDoc.setSvmRank(new Double(-1.0));
            
            new LinkHandler().handle(uc,u,c);
            new StoreHandler().handle( uc, u, c );
         }
         else
         {
            bingoDoc.setFeatureVector( new TFBuilder().getTFFeatures( bingoDoc.getText() ) );
            
            if ( bingoDoc.getStatus() != 'B' &&  bingoDoc.getStatus() != 'J')
            {
               //Bookmarks nicht klassifizieren
               new ClassifierHandler().handle( uc, u, c );
            }
            
            
            if ( bingoDoc.getNode() != rootNode)
            {
               StringCentenizer.setPreview(bingoDoc);
               if (session.storeStreams)
                  bingoDoc.setBulk(byteArray);
               
               try
               {
                  //new ParentDirectoryHandler().handle( uc, u, c );
               }
               catch ( Exception any )
               {
                  session.log(any);
               }
            }
            
            new StoreHandler().handle( uc, u, c );
            new LinkHandler().handle( uc, u, c );
         }
      }
      
      bingoDoc.content = null;
      bingoDoc.contentCounter = 0;
      destroy();
   }
   
   public void processContent()
   {
      InputStream is = null;
      try
      {
         is = getContentStream();
         if (is == null)
         {
            bingoDoc.setBroken(true);
            return;
         }
         
         HtmlParser htmlParser = new HtmlParser( this, is, url2r);
         htmlParser.run();  // kein Zusatz-Thread
         is.close(); is = null;
         htmlParser.destroy();
         htmlParser = null;
      }
      catch (Exception e)
      {
         try
         {is.close();}
         catch (Exception e1)
         {}
         bingoDoc.setBroken(true);
      }
      finally
      {
         is = null;
      }
   }
   
   public boolean init( URL_Crawler uc, url2resolve u, HttpURLConnection c )
   {
      if (!super.init(uc,u,c))
         return false;
      
      maxsize = session.getMaxsizeFor( mime );
      session.log("### maxsize: " + maxsize + " cL: " + bingoDoc.getContentLength(), 10);
      if ( bingoDoc.getContentLength() > maxsize )
         return false;
      return true;
   }
   
   public void destroy()
   {
      super.destroy();
   }
}