package bingo.crawler.handler;
import java.net.HttpURLConnection;
import java.util.*;
import java.net.URL;
import bingo.util.BingoDocument;
import bingo.crawler.URL_Crawler;
import bingo.crawler.frontier.url2resolve;

//=======================================================================================

public class LinkHandler extends CrawlHandler
{
   Collection ziele;
   
   static final int CLASSIFIER_DEPTH_FIRST = 0;
   static final int CLASSIFIER_BREADTH_FIRST = 1;
   static final int CLASSIFIER = 2;
   static final int DEPTH_FIRST = 3;
   static final int BREADTH_FIRST = 4;
   static final int FIFO = 5;
   
   public void handle(URL_Crawler uc, url2resolve u, HttpURLConnection c)
   {
      if (!init(uc, u, c))
      {
         destroy();
         return;
      }
      
      if (bingoDoc.fromNode == null)
         bingoDoc.fromNode = bingoDoc.getNode();
      
      if (bingoDoc.hasFrames())
         processFrames();
      else if (bingoDoc.hasRedirect())
         processRedirect();
      else
         processPlain();
      
      destroy();
   }
   
   public void handle(URL_Crawler uc, BingoDocument bd)
   {
      if (!init(uc, bd))
         return;
      
      if (bingoDoc.hasFrames())
         processFrames();
      else if (bingoDoc.hasRedirect())
         processRedirect();
      else
         processPlain();
      
      destroy();
   }
   
   //===============================================
   
   public void destroy()
   {
      super.destroy();
      ziele = null;
   }
   
   public boolean init(URL_Crawler uc, url2resolve u, HttpURLConnection c)
   {
      if (!super.init(uc, u, c))
         return false;
      ziele = bingoDoc.getLinks();
      if (ziele == null || ziele.size() <= 0)
         return false;
      return true;
   }
   
   public boolean init(URL_Crawler uc, BingoDocument bd)
   {
      if (!super.init(uc, bd))
         return false;
      ziele = bingoDoc.getLinks();
      if (ziele == null || ziele.size() <= 0)
         return false;
      
      return true;
   }
   
   public boolean canCrawl(URL target)
   {
      if (target == null)
         return false;
      
      if (!session.softFocus)   // focus soft/hard checken
      {
         if (bingoDoc.getNode() == null || bingoDoc.fromNode == null || bingoDoc.getNode().isRoot())
            return false;
         
         if (! bingoDoc.getNode().isNodeDescendant(bingoDoc.fromNode))
         {
            //System.out.println("Rejected FROM: " + bingoDoc.fromNode.label + " TO: " + bingoDoc.getNode().label );
            return false;
         }  
         else
         {
            //System.out.println("Accepted FROM: " + bingoDoc.fromNode.label + " TO: " + bingoDoc.getNode().label );  
         }   
      }
      
      if (session.hostCrawling)  // host crawling checken
      {
         url2resolve u1 = bingoDoc.getUrl2r();
         if (u1 == null)
            return false;
         
         String host1 = u1.getHost();
         String host2 = target.getHost();
         
         if (host1 == null || host2 == null || host1.length() == 0 || host2.length() == 0)	return false;
         
         host1 = host1.toLowerCase().trim();
         host2 = host2.toLowerCase().trim();
         
         if (!host1.equals(host2))
            return false;
         
         host1 = null;
         host2 = null;
      }
      
      return true;
   }
   
   public void processPlain()
   {
      if (bingoDoc.getDepth() >= session.maxCrawlingDepth)  // too deep
         return;
      
      if ( (bingoDoc.getNode() == rootNode || bingoDoc.getNode() == null)  // tunnel ende
      && (bingoDoc.getClassDistance() > session.maxTunnelDistance ))
         return;
      
      ArrayList linksArray = new ArrayList(ziele);
      Random rd = new Random(System.currentTimeMillis());
      int linkCount = 0;
      
      while (linksArray.size() > 0 && linkCount <= session.linksPerDoc)
      {
         int nextPos = rd.nextInt(linksArray.size());
         
         URL tmp = (URL) linksArray.get(nextPos);
         linksArray.remove(nextPos);
         
         if (!canCrawl(tmp))
            continue;
         
         url2resolve tempU = UrlVerifier.verifyURL(tmp);
         tmp = null;
         if (tempU == null)
            continue;
         
         BingoDocument target = new BingoDocument();
         target.setUrl2r(tempU);
         target.setNode(bingoDoc.getNode());
         if (!session.softFocus)
         {
            target.fromNode = bingoDoc.fromNode;
         }
         
         // target.setStatus(bingoDoc.getStatus());
         target.setStatus('L');
         target.setDepth(bingoDoc.getDepth() + 1);
         
         if (bingoDoc.getNode() == null || bingoDoc.getNode() == rootNode)
         {
            target.setPriority((float) (bingoDoc.getPriority() / 2.0));
            target.setClassDistance(bingoDoc.getClassDistance() + 1);
            // Prio reduzieren
            //target.setRootNode(bingoDoc.getRootNode());
         }
         else
         {
            double prio = bingoDoc.getSvmRank().doubleValue();
            target.setPriority(this.getPrio(prio));
            //target.setRootNode(bingoDoc.getClassifiedNode());
            target.setClassDistance(0);
         }
         
         crawler.addLink(target.getUrl2r());
         linkCount++;
      }
      linksArray = null;
      rd =null;
   }
   
   public void processFrames()
   {
      if (bingoDoc.fromFrame)  // Frames in Frames sind nicht erlaubt !
         return;
      
      Iterator it = ziele.iterator();
      int linkCount = 0;
      
      while (it.hasNext() && linkCount <= session.linksPerDoc)
      {
         URL tmp = (URL) it.next();
         
         if (!canCrawl(tmp))
            continue;
         
         url2resolve tempU = UrlVerifier.verifyURL(tmp);
         tmp = null;
         if (tempU == null)
            continue;
         
         BingoDocument target = new BingoDocument();
         target.setUrl2r(tempU);
         
         target.setNode(bingoDoc.getNode());
         if (!session.softFocus)
         {
            target.fromNode = bingoDoc.fromNode;
         }
         target.setPriority((float) (bingoDoc.getPriority()));
         target.setClassDistance(bingoDoc.getClassDistance());
         target.setStatus(bingoDoc.getStatus());
         target.setDepth(bingoDoc.getDepth());
         
         target.fromFrame = true;
         
         crawler.addLink(target.getUrl2r());
         linkCount++;
      }
      it = null;
   }
   
   public void processRedirect()
   {
      Iterator it = ziele.iterator();
      int linkCount = 0;
      
      if (bingoDoc.getHops() >= crawler.maxHops)
         return;
      
      while (it.hasNext() && linkCount < 1)
      {
         linkCount++;
         URL tmp = (URL) it.next();
         
         if (!canCrawl(tmp))
            continue;
         
         url2resolve tempU = UrlVerifier.verifyURL(tmp);
         tmp = null;
         if (tempU == null)
            continue;
         
         BingoDocument target = new BingoDocument();
         target.setUrl2r(tempU);
         
         target.setNode(bingoDoc.getNode());
         if (!session.softFocus)
         {
            target.fromNode = bingoDoc.fromNode;
         }
         
         target.setPriority((float) (bingoDoc.getPriority()));
         target.setHops(bingoDoc.getHops() + 1);
         target.setDepth(bingoDoc.getDepth());
         target.setStatus(bingoDoc.getStatus());
         
         crawler.addLink(target.getUrl2r());
      }
      it = null;
   }
   
   public float getPrio(double svmscore)
   {
      float result = (float)svmscore;
      float depth = (float)bingoDoc.getDepth();
      switch(session.crawlStrategy)
      {
         case CLASSIFIER_DEPTH_FIRST:
            result = result * depth;
            break;
         case CLASSIFIER_BREADTH_FIRST:
            result = result / (depth + (float)1.0);
            break;
         case CLASSIFIER:
            break;
         case DEPTH_FIRST:
            result = depth;
            break;
         case BREADTH_FIRST:
            result = (float)1.0 / (depth + (float)1.0);
            break;
         case FIFO:
         default:
            result = (float)1.0;
      }
      return result;
   }
}
