package bingo.crawler.handler;

import java.net.HttpURLConnection;

import bingo.crawler.URL_Crawler;
import bingo.crawler.frontier.url2resolve;
import bingo.crawler.handler.parser.PDFParser;
import bingo.svmlight.TFBuilder;
import bingo.crawler.handler.parser.StringCentenizer;

//=======================================================================================

public class PDFHandler extends CrawlHandler
{
   public static String mime = "application/pdf";
   
   public void handle( URL_Crawler uc, url2resolve u, HttpURLConnection c )
   {
      if ( !init( uc, u, c ) )
      {
         destroy();
         return;
      }
      
      readContent();
      processContent();
      
      if ( bingoDoc.isBroken() )
      {
         new ErrorHandler().handle(uc,u,c);
         destroy();
         return;
      }
      
      if (uc.isAlreadyParsed(bingoDoc))    // text kam im anderen Dokument bereits vor
      {
         bingoDoc.setStatus('D');
         bingoDoc.setAnchors(null);
         bingoDoc.setBulk(null);
         bingoDoc.setFeatureVector(null);
         bingoDoc.setLinks(null);
         bingoDoc.setText(null);
         bingoDoc.setTitle(null);
         bingoDoc.setPreview(null);
         new StoreHandler().handle( uc, u, c );
      }
      else if (bingoDoc.isForeign())   // Andere Sprache vermutet, nicht klassifizieren
      {
         session.log("Fremdsprache: " + bingoDoc.getUrl(), 1);
         bingoDoc.setStatus('X');
         bingoDoc.setAnchors(null);
         bingoDoc.setBulk(null);
         bingoDoc.setFeatureVector(null);
         // bingoDoc.setLinks(null);   // Links kann man mitspeichern
         bingoDoc.setText(null);
         bingoDoc.setTitle(null);
         bingoDoc.setPreview(null);
         
         bingoDoc.setNode(session.getRootNode());
         bingoDoc.setSvmRank(new Double(-1.0));
         
         new StoreHandler().handle( uc, u, c );
      }
      else
      {
         
         bingoDoc.setFeatureVector( new TFBuilder().getTFFeatures( bingoDoc.getText() ) );
         
         if ( bingoDoc.getStatus() != 'B' && bingoDoc.getStatus() != 'J')
         {
            //Bookmarks nicht klassifizieren
            new ClassifierHandler().handle( uc, u, c );
         }
         
         if ( bingoDoc.getNode() != session.getRootNode() )
         {
            StringCentenizer.setPreview(bingoDoc);
            if (session.storeStreams)
               bingoDoc.setBulk(byteArray);
            
            try
            {
               // new GoogleRelatedQueryHandler().handle (uc,u,c);
               // new GoogleLinkQueryHandler().handle (uc,u,c);
               // new ParentDirectoryHandler().handle( uc, u, c );
            }
            catch ( Exception any )
            {
               session.log(any);
            }
         }
         new StoreHandler().handle( uc, u, c );
      }
      bingoDoc.content = null;
      bingoDoc.contentCounter = 0;
      
      destroy();
   }
   
   public void processContent()
   {
      PDFParser pdfParser = new PDFParser( this, url2r);
      pdfParser.run();
      pdfParser.destroy();
      pdfParser = null;
   }
   
   public boolean init( URL_Crawler uc, url2resolve u, HttpURLConnection c )
   {
      if (!super.init(uc,u,c))
         return false;
      
      maxsize = session.getMaxsizeFor( mime );
      if ( bingoDoc.getContentLength() > maxsize )
         return false;
      
      //    myAssistants = new Vector();
      
      return true;
   }
}