package bingo.crawler.handler;
import bingo.util.BingoDocument;
import bingo.crawler.frontier.url2resolve;
import bingo.util.*;
import java.net.URL;
import java.util.*;

//=======================================================================================
public class UrlVerifier
{
   private SessionBuffer session;
   private url2resolve url2r = null;
   private BingoDocument bd = null;
   private static String[] forbiddenMatches =
   {"google", "\\\\", "///", "\\.", ".\\",
    "{", "}", "]", "["};
    private static HashSet allowedProtocols = null;
    
    public UrlVerifier(url2resolve base)
    {
       if ( allowedProtocols == null )
       {
          synchronized (this.getClass())
          {
             // genehmigte Protokolle
             allowedProtocols = new HashSet(1);
             allowedProtocols.add("http");
          }
       }
       
       url2r = base;
       if (url2r != null)
       {
          bd = url2r.parentDoc;
       }
       session = SessionBuffer.getInstance();
    }
    
    public static URL checkAbsoluteLink(String target)
    {
       if (target == null || target.length() > 300 || target.length() == 0)
          return null;
       
       int pos = target.indexOf("#");        // Schwanz abschneiden
       if (pos >= 0)
       {
          target = target.substring(0, pos);
          target = target.trim();
       }
       
       if (target.length() == 0)
          return null;
       
       URL newURL = null;
       try
       {
          newURL = new URL(target);
       }
       catch (Exception e)
       {
          newURL = null;
       }
       
       if (newURL.getHost() == null || newURL.getHost().equals(""))
          newURL = null;
       
       if (SessionBuffer.allowedMimes.get(guessContentTypeFromName(newURL.getFile())) == null)
          newURL = null;
       
       return newURL;
    }
    
    public URL checkLink(String target)
    {
       if (url2r == null)
          return null;
       
       if (target == null || target.length() > 300 || target.length() == 0)
          return null;
       
       int pos = target.indexOf("#");        // Schwanz abschneiden
       if (pos >= 0)
       {
          target = target.substring(0, pos);
          target = target.trim();
       }
       
       if (target.length() == 0)
          return null;
       
       URL newURL = null;
       try
       {
          newURL = new URL(url2r.getURL(), target);
       }
       catch (Exception e)
       {
          newURL = null;
       }
       
       if (newURL == null)
          return null;
       
       if (newURL.toExternalForm().length() > 300)
          return null;
       
       if (newURL.getHost() == null || newURL.getHost().equals("") || newURL.getProtocol() == null || !allowedProtocols.contains(newURL.getProtocol()))
          return null;
      
       if (session.allowedMimes.get(guessContentTypeFromName(newURL.getFile())) == null)
          return null;
       
       return newURL;
    }
    
    public static url2resolve verifyURL(URL target)
    {
       if (target == null)
          return null;
       
       String check = target.toString();
       if (check == null || check.length() <= 0 || check.length() > 300)
          return null;
       
       for (int i = 0; i < forbiddenMatches.length; i++)
       {
          if (check.indexOf(forbiddenMatches[i]) >= 0)
             return null;
       }
       check = null;
       
       url2resolve tmpurl = new url2resolve(target, (float)0.0, null);
       if (tmpurl.isBad())
          return null;
       
       return tmpurl;
    }
    
    //================================================
    
    public static String guessContentTypeFromName(String test)
    {
       String result = "text/html";
       // Default-Wert, falls alles andere nicht greift
       int pos = -1;
       if (test == null)
          return result;
       
       String who = test.trim().toLowerCase();
       if (who.length() == 0)
          return result;
       
       pos = who.indexOf('?');
       // Forms sind zuerst immer als text/html zu betrachten
       if (pos >= 0)
          return result;
       
       pos = who.lastIndexOf('.');
       if (pos < 0)
          return result;
       
       MimeType mime = (MimeType) SessionBuffer.allMimes.get(who.substring(pos,who.length()).trim().toLowerCase());
       if (mime != null)
          result = mime.getDescription();
       
       who = null;
       return result;
    }
    
    //===========================================================
    
    public void destroy()
    {
       //bd = null;
       url2r = null;
       session = null;
    }
    
    public static BingoDocument string2BingoDocument(String absURL)
    {
       URL url = UrlVerifier.checkAbsoluteLink(absURL);
       if (url == null)
          return null;
       url2resolve url2r = UrlVerifier.verifyURL(url);
       if (url2r == null)
          return null;
       BingoDocument bd = new BingoDocument();
       bd.setUrl2r(url2r);
       
       url2r=null; url=null;
       
       return bd;
    }
}
