package bingo.crawler.handler.parser;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.Reader;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.text.html.HTMLEditorKit.ParserCallback;
import javax.swing.text.html.parser.ParserDelegator;

import bingo.util.BingoDocument;
import bingo.crawler.URL_Crawler;
import bingo.crawler.frontier.url2resolve;
import bingo.crawler.handler.CrawlHandler;
import bingo.crawler.handler.UrlVerifier;
import bingo.util.SessionBuffer;

//=======================================================================================

public class GoogleDirectoryParser implements Runnable
{
   SessionBuffer session;
   PipedInputStream in = null;
   Reader reader = null;
   ParserCallback callback = null;
   url2resolve url2r;
   URL_Crawler crawler;
   GoogleDirectoryItem google;
   
   BingoDocument bd = null;
   BingoDocument tmp = null;

   public CrawlHandler parent;
   
   
   public GoogleDirectoryParser( CrawlHandler p, PipedOutputStream out, url2resolve u )
   {
      parent = p;
      url2r = u;
      session = SessionBuffer.getInstance();
      crawler = session.getCrawler();
      bd = u.parentDoc;
      
      try
      {
         in = new PipedInputStream( out );
         reader = new BufferedReader( new InputStreamReader( in ) );
         callback = new GoogleDirectoryParserCallback( bd );
      }
      catch ( Exception e )
      {
         bd.setBroken(true);
      }
   }
   
   
   public void run()
   {
      if ( bd.isBroken() )
      {
         return;
      }
      
      try
      {
         new ParserDelegator().parse( reader, callback, true );
         reader.close();
         in.close();

         Iterator it = bd.getGoogle().iterator();
         while ( it.hasNext() )
         {
            google = (GoogleDirectoryItem) it.next();
            tmp = UrlVerifier.string2BingoDocument(google.getLink());
            if (tmp == null)
               continue;
            
            if ( google.isDirectory() )
            {
               tmp.setStatus( 'D' );
            } 
            else
            {
               tmp.setStatus( 'L' );
            }
            finish( tmp );
         }
         it = null;
      }
      catch ( Exception e )
      {
         bd.setBroken(true);
      }
      finally
      {
         Vector register = parent.getAssistants();
         
         if ( register != null )
         {
            synchronized ( register )
            {
               register.remove( this );
               register.notifyAll();
            }
         }
         register = null;
         destroy();
      }
   }
   
   
   public void finish( BingoDocument temp )
   {
      temp.setNode( bd.getNode() );
      temp.setPriority( bd.getPriority() );
      //temp.setDepth( bd.getDepth() + 1 );
      temp.setClassDistance(bd.getClassDistance() + 1);
      //temp.setRootNode( bd.getRootNode() );
      //temp.setRootLink( bd.getUrl() );
      
//      if ( tmp.getDepth() <= session.maxCrawlingDepth )
//      {
//         page.addLink( tmp );
//      }
   }
   
   
   public void destroy()
   {
   	  bd = null;
   	  tmp = null;
      url2r = null;
      session = null;
      crawler = null;
      reader = null;
      callback = null;
      in = null;
      google = null;
      reader = null;
      in = null;
      parent = null;
      tmp = null;
   }
}
