package bingo.crawler.handler.parser;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.Reader;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.text.html.HTMLEditorKit.ParserCallback;
import javax.swing.text.html.parser.ParserDelegator;

import bingo.util.BingoDocument;
import bingo.crawler.URL_Crawler;
import bingo.crawler.frontier.url2resolve;
import bingo.crawler.handler.CrawlHandler;
import bingo.crawler.handler.UrlVerifier;
import bingo.util.SessionBuffer;

//=======================================================================================

public class GoogleParser implements Runnable {
    SessionBuffer session;
    PipedInputStream in = null;
    Reader reader = null;
    ParserCallback callback = null;
    url2resolve url2r;
    URL_Crawler crawler;
    UrlVerifier urlVerifier;
    GoogleItem google;
    
    BingoDocument bd = null;
    BingoDocument temp = null;

    public CrawlHandler parent;
    
    
    public GoogleParser( CrawlHandler p, PipedOutputStream out, url2resolve u ) {
        parent = p;
        url2r = u;
        session = SessionBuffer.getInstance();
        crawler = session.getCrawler();
        bd = u.parentDoc;
        
        try {
            in = new PipedInputStream( out );
            reader = new BufferedReader( new InputStreamReader( in ) );
            callback = new GoogleParserCallback( bd );
        }
        catch ( Exception e ) {
            bd.setBroken(true);
        }
    }
    
    
    public void run() {
        if ( bd.isBroken() ) {
            return;
        }
        
        try {
            new ParserDelegator().parse( reader, callback, true );
            reader.close();
            in.close();

            Iterator it = bd.getGoogle().iterator();
            while ( it.hasNext()) {
                google = (GoogleItem) it.next();
                
                if ( google.getTarget() != null )
                {
                    temp = UrlVerifier.string2BingoDocument(google.getTarget());
                    if ( temp != null ) {
                        temp.setStatus( 'L' );
                        finish( temp );
                    }
                }
                
                if ( google.getCategory() != null )
                {
                    temp = UrlVerifier.string2BingoDocument(google.getCategory());
                    if ( temp != null ) 
                    {
                        temp.setStatus( 'D' );
                        finish( temp );
                    }
                }
                
                if ( google.getTranslate() != null )
                {
                    temp = UrlVerifier.string2BingoDocument(google.getTranslate());
                    if ( temp != null ) {
                        temp.setStatus( 'L' );
                        finish( temp );
                    }
                }
            }
            it=null;
        }
        catch ( Exception e ) {
            bd.setBroken(true);
        }
        finally {
            Vector register = parent.getAssistants();
            
            if ( register != null ) {
                synchronized ( register ) {
                    register.remove( this );
                    register.notifyAll();
                }
            }
            register = null;
            destroy();
        }
    }
    
    
    public void finish( BingoDocument tmp ) {
        tmp.setNode( bd.getNode() );
        tmp.setPriority( bd.getPriority() );
//        tmp.setDepth( link.getDepth() + 1 );
        tmp.setClassDistance(bd.getClassDistance() + 1);
        //tmp.setRootNode( bd.getRootNode() );
        //tmp.setRootLink( bd.getUrl() );
//        if ( tmp.getDepth() <= session.maxCrawlingDepth ) {
//            page.addLink( tmp );
//        }
    }
    
    
    public void destroy() {
        bd = null;
        temp = null;
        url2r = null;
        session = null;
        crawler = null;
        reader = null;
        callback = null;
        in = null;
        if ( urlVerifier != null ) {
            urlVerifier.destroy();
        }
        urlVerifier = null;
        google = null;
        parent = null;
        reader = null;
        in = null;
    }
}