package bingo.crawler.handler.parser;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.Reader;
import java.util.Vector;

import javax.swing.text.html.parser.ParserDelegator;

import bingo.crawler.frontier.url2resolve;
import bingo.crawler.handler.CrawlHandler;
import bingo.crawler.handler.ErrorHandler;
import bingo.util.BingoDocument;
import bingo.util.SessionBuffer;

//=======================================================================================

public class HtmlParser implements Runnable
{
  SessionBuffer session;
  InputStream in = null;
  Reader reader = null;
  HtmlParserCallback callback = null;
  url2resolve url2r;
  
  private BingoDocument bd;
  
  public CrawlHandler parent;
  
  
  public HtmlParser( CrawlHandler p, InputStream i, url2resolve u )
  {
    session = SessionBuffer.getInstance();
    parent = p;
    url2r = u; 
    try
    {
      in = i;
      bd = u.parentDoc;
      if (bd == null) System.exit(1);
      reader = new InputStreamReader( in );
      callback = new HtmlParserCallback( bd );
    }
    catch ( Exception e )
    {
      if ( in != null )
      {
        try
        {
          in.close();
        }
        catch ( Exception e1 )
        {}
      }
      
      if ( reader != null )
      {
        try
        {
          reader.close();
        }
        catch ( Exception e1 )
        {}
      }
      
      new ErrorHandler().handle( parent.getCrawler(), url2r, null );
    }
  }
  
  
  public void run()
  {
    try
    {
      if ( !bd.isBroken() )
      {
        new ParserDelegator().parse( reader, callback, true );
        reader.close();
        in.close();
        callback.flush();
      }
    }
    catch ( Exception e )
    {
      if ( parent != null )
      {
        new ErrorHandler().handle( parent.getCrawler(), url2r, null );
      }
    }
    finally
    {
      Vector register = null;
      
      if ( parent != null )
      {
        register = parent.getAssistants();
      }
      if ( register != null )
      {
        synchronized ( register )
        {
          register.remove( this );
          register.notifyAll();
        }
      }
      register = null;
    }
    destroy();
  }
  
  
  public void destroy()
  {
    session = null;
    
    if ( reader != null )
    {
      try
      {
        reader.close();
      }
      catch ( Exception e )
      {}
    }
    
    reader = null;
    
    if ( in != null )
    {
      try
      {
        in.close();
      }
      catch ( Exception e )
      {}
    }
    in = null;
    
    if ( callback != null )
    {
      callback.destroy();
    }
    callback = null;
    
    url2r = null;
    bd = null;
    parent = null;
  }
  
}
