package bingo.crawler.handler.parser;
import java.nio.CharBuffer;
import java.util.*;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.net.URL;
import javax.swing.text.MutableAttributeSet;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTMLEditorKit;
import bingo.crawler.frontier.url2resolve;
import bingo.crawler.handler.UrlVerifier;
import bingo.util.BingoDocument;
import bingo.util.SessionBuffer;
import java.security.*;

public class HtmlParserCallback extends HTMLEditorKit.ParserCallback
{
   
   private static int maxLinksPerDoc = 50;
   
   private int active = 0;   // >=1 ein, <=0 aus
   
   public static Pattern[] patterns =
   {Pattern.compile("<"),
    Pattern.compile(""),
    Pattern.compile(""),
    Pattern.compile(""),
    Pattern.compile(""),
    Pattern.compile(""),
    Pattern.compile(""),
    Pattern.compile(""),
    Pattern.compile("[^\\p{Graph}]"),
    Pattern.compile("\\s++")
   };
   
   public static String[] replaces =
   {" ",
    "ae",
    "ue",
    "oe",
    "Ae",
    "Ue",
    "Oe",
    "ss",
    " ",
    " "};
    //    private CharBuffer preview = null;
    private CharBuffer body = null;
    private CharBuffer title = null;
    String url;
    //  boolean isAnchor = false;
    boolean isTitle = false;
    private BingoDocument bd;
    private url2resolve url2r;
    String redirect = null;
    private SessionBuffer session;
    private UrlVerifier urlVerifier;
    //    private int previewLength = 0;
    private int titleLength = 0;
    private int linkCount = 0;
    
    public HtmlParserCallback(BingoDocument bd)
    {
       session = SessionBuffer.getInstance();
       redirect = null;
       //    sd = new StemmerDriver();
       this.bd = bd;
       this.url2r = bd.getUrl2r(); //
       
       active = 0;
       
       urlVerifier = new UrlVerifier(url2r);
       //       previewLength = session.getPreviewLength();
       titleLength = session.getTitleLength();
       //       preview = CharBuffer.allocate(previewLength);
       title = CharBuffer.allocate(titleLength);
       if (bd.getBulkSize() > 0)
          body = CharBuffer.allocate(bd.getBulkSize());
       linkCount = 0;
    }
    public void handleText(char[] text, int pos)
    {
       if (active <= 0)
          return;
       
       try
       {
          int len = 0;
          if (isTitle && title != null) // noch nicht abgeschlossen
          {
             len = Math.min(title.remaining(), text.length);
             if (len > 0)
                title.put(text, 0, len);
             if (title.remaining() <= 0)
             {
                bd.setTitle(finish(title));
                title = null;
             } else
                title.put(' ');
          }
/*
          if (!isTitle && preview != null) // noch nicht abgeschlossen
          {
             len = Math.min(preview.remaining(), text.length);
             if (len > 0)
                preview.put(text, 0, len);
             if (preview.remaining() <= 0)
             {
                bd.setPreview(finish(preview));
                preview = null;
             } else
                preview.put(' ');
          }
 */       
          if (body != null && body.remaining() > text.length + 1)
          {
             body.put(text);
             body.put(' ');
          }
       } catch (Exception e)
       {
          session.log(e);
       }
    }
    public void handleStartTag(HTML.Tag tag, MutableAttributeSet a, int pos)
    {
       isTitle = false;
       
       if (tag == HTML.Tag.HTML)
       {   
           active = 1;
           return;
       }    
       
       if (tag == HTML.Tag.SCRIPT)
       {   
           active-- ;
           return;
       }    

       if (tag == HTML.Tag.NOFRAMES)
       {   
           active-- ;
           body = null;
           return;
       } 
       
       if (active <= 0)
          return;
       
       try
       {
          if (
          tag == HTML.Tag.P ||
          tag == HTML.Tag.A ||
          tag == HTML.Tag.UL ||
          tag == HTML.Tag.OL ||
          tag == HTML.Tag.TD ||
          tag == HTML.Tag.LI ||
          tag == HTML.Tag.BLOCKQUOTE ||
          tag == HTML.Tag.H1 ||
          tag == HTML.Tag.H2 ||
          tag == HTML.Tag.H3 ||
          tag == HTML.Tag.H4 ||
          tag == HTML.Tag.H5 ||
          tag == HTML.Tag.H6 ||
          tag == HTML.Tag.TITLE ||
          tag == HTML.Tag.BODY ||
          tag == HTML.Tag.HTML )
          {
             if (body != null && body.remaining() > 0)
             {
                body.put('.');
             }
          }
          
          if (tag == HTML.Tag.P && body != null && body.remaining() >= 1)
             body.put((char)7);
          
          if ((
              tag == HTML.Tag.H1 || 
              tag == HTML.Tag.H2 || 
              tag == HTML.Tag.H3 || 
              tag == HTML.Tag.H4 || 
              tag == HTML.Tag.H5 || 
              tag == HTML.Tag.H6 ) && body != null && body.remaining() >= 2)
          {   
             body.put((char)7);
             body.put((char)7);
          } 

          else if (tag == HTML.Tag.A)
          {
             if (a.getAttribute(HTML.Attribute.HREF) != null)
             {
                String ziel = (a.getAttribute(HTML.Attribute.HREF)).toString().trim();
                // session.log("###### ziel vor verify: " + ziel, 10);
                URL toAdd = urlVerifier.checkLink(ziel);
                if (toAdd != null && linkCount < maxLinksPerDoc )
                {
                   if (bd.addLink(toAdd))
                      linkCount ++;
                }
                ziel = null; toAdd = null;
                
                /*
                BingoDocument newDoc = urlVerifier.verifyURL(ziel);
                // session.log("###### ziel nach verify: " + newDoc.getUrl(),10);
                if (newDoc != null)
                {
                   //          isAnchor = true;
                   newDoc.setDepth(bd.getDepth() + 1);
                   // session.log("########### HTMLParserCallback: Link found! set depth = " + newDoc.getDepth(), 7);
                   newDoc.setStatus('L');
                   newDoc.setParentID(bd.getId());
                   session.log("parentID: " + newDoc.getParentID()+ " ID: " + newDoc.getId(), 10);
                   bd.addLink(newDoc);
                   //          actualAnchor = new URLAnchors( newlink.getURL() );
                }
                 */
             }
          }
          else if (tag.equals(HTML.Tag.TITLE))
          {
             isTitle = true;
          }
       } catch (Exception e)
       {
          session.log(e);
       }
    }
    
    public void handleEndTag(HTML.Tag tag, int pos)
    {
       isTitle = false;     // also nur kurze Titles ohne Formatierung zulassen
       
       if (tag == HTML.Tag.HTML || tag == HTML.Tag.BODY)    // ab hier kein Parsen
       {   
           active = -1000000;
           return;
       }    
       else if (tag == HTML.Tag.FRAMESET)  // ab hier kein Parsen
       {   
           active = -1000000;
           body = null;
           return;
       }   
       else if (tag == HTML.Tag.SCRIPT)
       {   
           active ++;
           return;
       }    
       else if (tag == HTML.Tag.NOFRAMES)
       {   
           active ++;
           body = null;
           return;
       }
       
       if (active <= 0)
          return;
       
       try
       {
          
          if (
          tag == HTML.Tag.P ||
          tag == HTML.Tag.A ||
          tag == HTML.Tag.UL ||
          tag == HTML.Tag.OL ||
          tag == HTML.Tag.TD ||
          tag == HTML.Tag.LI ||
          tag == HTML.Tag.BLOCKQUOTE ||
          tag == HTML.Tag.H1 ||
          tag == HTML.Tag.H2 ||
          tag == HTML.Tag.H3 ||
          tag == HTML.Tag.H4 ||
          tag == HTML.Tag.H5 ||
          tag == HTML.Tag.H6 ||
          tag == HTML.Tag.TITLE ||
          tag == HTML.Tag.BODY ||
          tag == HTML.Tag.HTML )
          {
             if (body != null && body.remaining() > 0)
             {
                body.put('.');
             }
          }
          /*
           * else if ( tag == HTML.Tag.A ) { // isAnchor = false; // if (
           * actualAnchor != null ) // { // page.addAnchors( actualAnchor ); //
           * actualAnchor = null; // }
           */

       } catch (Exception e)
       {
          session.log(e);
       }
    }
    public void handleSimpleTag(HTML.Tag tag, MutableAttributeSet a, int pos)
    {
       isTitle = false;
       
       if (active <= 0)
          return;
       
       try
       {
          if (
          tag == HTML.Tag.BR ||
          tag == HTML.Tag.HR
          )
          {
             if (body != null && body.remaining() > 0)
             {
                body.put('.');
             }
          }
          
          if (tag == HTML.Tag.FRAME)
          {
             body = null;
             bd.setHasFrames(true); // Frames als Bestandteile einer Seite betrachten
             Object attr = a.getAttribute(HTML.Attribute.SRC);
             if (attr == null)
                return;
             String check = attr.toString().trim();
             attr = null;
             URL toAdd = urlVerifier.checkLink(check);
             if (toAdd != null && linkCount < maxLinksPerDoc )
             {
                if (bd.addLink(toAdd))
                   linkCount ++;
             }
             check = null; toAdd = null;
             
             /*
             BingoDocument newDoc = urlVerifier.verifyURL(check);
             check = null;
             if (newDoc != null)
             {
                newDoc.setDepth( bd.getDepth() );
                // anders als bei den "normalen" Links
                newDoc.setStatus( bd.getStatus() );
                newDoc.setPriority(bd.getPriority());
                bd.setHasFrames(true);
                bd.addLink( newDoc );
             }
              */
          }
          else if (tag == HTML.Tag.META)
          {
             if (a.getAttribute(HTML.Attribute.HTTPEQUIV) != null)
             {
                String check = a.getAttribute(HTML.Attribute.HTTPEQUIV).toString().trim();
                if (check.indexOf("refresh") >= 0)
                {
                   String content = a.getAttribute(HTML.Attribute.CONTENT).toString();
                   if (content != null)
                   {
                      content = content.toLowerCase();
                      int point = content.indexOf("url=");
                      if (point >= 0)
                      {
                         String redirect = content.substring(point + 4,
                         content.length());
                         redirect = redirect.trim();
                         point = redirect.indexOf(" ");
                         if (point >= 0)
                            redirect = redirect.substring(0, point);
                         
                         URL toAdd = urlVerifier.checkLink(redirect);
                         if (toAdd != null)
                            bd.addLink(toAdd);
                         bd.setRedirect(true);
                         
                         redirect = null; toAdd = null;
                         // Umleitung auf diese Seite !
                         active = -1000000;
                         body = null;
                      }
                   }
                   content = null;
                }
                check = null;
             }
          }
          else if (tag == HTML.Tag.INPUT && bd.hasForms())
          {
             if (a.toString().toLowerCase().indexOf("type=text") > -1)
             {
                bd.setHasForms(true);
             }
          }
       } catch (Exception e)
       {
          session.log(e);
       }
    }
    
    public void handleComment(char[] comment, int pos)
    {
    }
    
    public void handleError(String errorMsg, int pos)
    {
       //System.err.println( "[Error] " + errorMsg );
    }
    public void flush()
    {
       /*
       System.out.println ("BASE: " + bd.getUrl2r().getURL());
       if (bd.getLinks() == null)
       {
         System.out.println ("-> none");
       }   
       else
       {
         Iterator it = bd.getLinks().iterator();
         while (it.hasNext())
           System.out.println ("Link -> " + (URL)it.next())   ;
       }
       */
       
       if (title != null)
       {
          bd.setTitle(finish(title));
          title = null;
       }
       
/*
       if (preview != null)
       {
          bd.setPreview(finish(preview));
          preview = null;
       }
 */
       
       if (body != null && body.array().length > 0)
       {
          parseBody(bd);
          
          bd.content = body.array();
          bd.contentCounter = body.position();
          
          //System.out.println("MD5 auf:");
          //System.out.println(new String(bd.content, 0, bd.contentCounter));
          
          try
          {
             MessageDigest md = MessageDigest.getInstance("MD5");
             byte[] hash = md.digest(jodd.util.CharUtil.toByteArray(body.array()));
             if (hash != null && hash.length >=8)
             {
                long signature = 0;
                for (int i=0; i<8; i++)
                {
                   long tmp = (long)hash[i];
                   tmp = (tmp & 0xff) << (i*8);
                   signature += tmp;
                }
                bd.setMD5(signature);
                
                //for (int i=0; i<hash.length; i++)
                //System.out.print("[" + hash[i] + "]") ;
                //System.out.println();
             }
             md=null;
             hash = null;
          }
          catch (Exception e)
          {
             e.printStackTrace();
          }
          
          body = null;
       }
       destroy();
    }
    
    public String finish(CharBuffer buffer)
    {
       if (buffer == null || bd == null)
          return null;
       String tmp = new String(buffer.array(), 0, buffer.position());
       return finish(tmp);
    }
    
    public static String finish(String tmp)
    {
       for (int i = 0; i < patterns.length; i++)
       {
          Matcher m = patterns[i].matcher(tmp);
          tmp = m.replaceAll(replaces[i]);
          m = null;
       }
       return tmp.trim();
    }
    
    public void destroy()
    {
       url = null;
       bd = null;
       url2r = null;
       redirect = null;
       session = null;
       if (urlVerifier != null)
          urlVerifier.destroy();
       urlVerifier = null;
       body = null;
       //       preview = null;
       title = null;
    }
    
    public void parseBody(BingoDocument bd)
    {
       StemmerDriver sd = new StemmerDriver();
       LinkedList result = sd.stem(body);
       bd.setText(result);
       bd.setForeign(sd.isForeign);
       sd = null;
    }
}
