package bingo.crawler.handler.parser;

import java.lang.reflect.Method;
import java.nio.CharBuffer;
import java.util.LinkedList;
import java.util.Collection;
import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import net.sf.snowball.SnowballProgram;
import bingo.util.SessionBuffer;

public class StemmerDriver
{
  private SessionBuffer session;
  private SnowballProgram stemmer;
  private Method stemMethod;
  private Object[] emptyArgs;
  private int repeat = 1;
  
  public boolean isForeign = false;
  public static int minStopwords = 3;
  
  private static int maxTermsPerDoc = 5000;
  
  public static Pattern[] patterns =
  {
    Pattern.compile(""),
    Pattern.compile(""),
    Pattern.compile(""),
    Pattern.compile("")
//    Pattern.compile(""),
//    Pattern.compile(""),
//    Pattern.compile("")
    
//    Pattern.compile("[^a-z]")
//    Pattern.compile("\\s"),
  };
  
  public static String[] replacements =
  {
    "ae",
    "ue",
    "oe",
    "ss"
//    "AE",
//    "UE",
//    "OE"
//    ""
//    " "
  };
  
  public StemmerDriver()
  {
    session = SessionBuffer.getInstance();
    isForeign = false;
    try
    {
      stemmer = (SnowballProgram) ( session.stemClass.newInstance() );
      stemMethod = session.stemClass.getMethod( "stem", new Class[0] );
      emptyArgs = new Object[0];
      repeat = session.repeatStemmer;
    }
    catch ( Exception e )
    {
      SessionBuffer.log(e);
    }
  }
   
  
  public LinkedList stem(CharBuffer buffer)
  {
    return stem(buffer.array(), 0, buffer.position());
  }
  
  
  public LinkedList stem( char[] data )
  {
    return stem(data, 0, data.length);
  }
  
  public static LinkedList parse( char[] data )
  {
    return parse(data, data.length);
  }
  
  public static LinkedList parse (char[] array, int pos)
  {
    return parse (array, 0, pos);
  }
  
  public static LinkedList parse (char[] array, int start, int end)
  {
    LinkedList tokens = new LinkedList();
    int count = 0;
    int a=start, b=0;
    while (a < end)
    {  
       if  ((array[a] >= 65 && array[a] <= 90 ) ||
              (array[a] >= 97 && array[a] <= 122 ) ||
              (array[a] == 196) ||
              (array[a] == 214) ||
              (array[a] == 220) ||
              (array[a] == 228) ||
              (array[a] == 246) ||
              (array[a] == 252) ||
              (array[a] == 223))
        {
        b = a + 1; 
        while (b < end &&
              (
              (array[b] >= 65 && array[b] <= 90 ) ||
              (array[b] >= 97 && array[b] <= 122 ) ||
              (array[b] == 196) ||
              (array[b] == 214) ||
              (array[b] == 220) ||
              (array[b] == 228) ||
              (array[b] == 246) ||
              (array[b] == 252) ||
              (array[b] == 223)
              ))
        b++;
        
        int len = b-a;
        if (len > 1)
        {  
          String token = new String (array, a, len).toLowerCase();
          for (int i=0;i<patterns.length;i++)
          {
            Matcher m = patterns[i].matcher(token);
            token = m.replaceAll(replacements[i]);
            m = null;
          }  
          tokens.add(token);
          count ++;
          token = null;
        }
        a = b+1;
       }
       else
         a++; 
       
       if (count > maxTermsPerDoc)
          break;
    }  
    return tokens;
  }
  
  public LinkedList stem (char[] array, int start, int end)
  {
    Collection tokens = parse (array, start, end);
    LinkedList result = stem (tokens);
    tokens = null;
    return result;
  }

  public LinkedList stem( Collection data )
  {
    LinkedList resultWords = new LinkedList();

    int pos = 0;
    int stopwordCounter = 0;
    Iterator it = data.iterator();
    while ( it.hasNext() )
    {
      String token = (String) it.next();
      
      if (session.isForeignStopword(token))
      {
        stopwordCounter -- ; 
        continue;
      }  
      else if (session.isStopword(token))
      {  
        stopwordCounter ++ ;
        continue;
      }  
      
      TextItem item = new TextItem( token );
      item.isStopword = false;
      item.position = ++pos;
      token = null;
 
/*
      if ( item.stem.length() <= 1 || session.isStopword( item.term ) ) {
        item.isStopword = true;
      } else {
        item.isStopword = false;
      }
 */      
      
      if ( session.bSTEM )
      {
        stemmer.setCurrent( item.term );
        
        for ( int i = repeat; i != 0; i-- )
        {
          try
          {
            stemMethod.invoke( stemmer, emptyArgs );
          }
          catch ( Exception e )
          {
            SessionBuffer.log(e);
          }
        }
        
        item.stem = stemmer.getCurrent();
      } else
      {
        item.stem = item.term;
      }
      
      if ( item.stem.length() > 1 )
        resultWords.add( item );
      
      item = null;
    }
    it = null;
    
    if (stopwordCounter <= minStopwords)
       isForeign = true;
     
    return resultWords;
  }
}
