package bingo.crawler.handler.parser;

import java.util.*;
import bingo.util.SessionBuffer;
import bingo.util.BingoTreeNode;
import bingo.util.BingoDocument;
import bingo.svmlight.TFBuilder;
import bingo.svmlight.Feature;

/**
 * Creates centence-based annotation of the given document
 */
public abstract class StringCentenizer
{
   public static ArrayList createCentences(char[] data)
   {
      if (data == null || data.length <= 0)
         return null;
      
      return createCentences(data, data.length);
   }
   
   public static ArrayList createCentences(char[] array, int pos)
   {
      if (array == null || array.length <= 0 || pos >= array.length || pos <= 0)
         return null;
      
      ArrayList centences = new ArrayList(100);
      int count = 0;
      int a=0, b=0;
      while (a < pos)
      {
         b = a;
         int bells = 0;
         
         while (array[b] != 33 && array[b] != 46 && array[b] != 63 && b < pos-1)
         {
            if (array[b] == 7)  // Bell
               bells ++;
            b++;
         }   
         
         if (a < (b-10))
         {
            Centence centence = new Centence(a, b);
            centence.pos = ++count;
            centence.bells = bells;
            centences.add(centence);
            // System.out.println ("Centence: [" + a + "," + b + "]");
         }
         a = b+1; 
      }
      return centences;
   }
   
   public static void parseCentences(char[] array, ArrayList centences)
   {
      if (centences == null || centences.size() <= 0 || array == null || array.length <= 0)
         return;
      
      Iterator it = centences.iterator();
      while (it.hasNext())
      {
         Centence centence = (Centence)it.next();
         if (centence == null || centence.start < 0 || centence.end >= array.length || centence.start >= centence.end)
            continue;
         centence.terms = new StemmerDriver().stem(array, centence.start, centence.end);
       //  System.out.println ("Sentence: " + new String(array, centence.start, centence.end - centence.start + 1));
      }
      it = null;
   }
   
   public static String annotate(char[] array, Collection centences, BingoTreeNode node)
   {
      if (node == null || centences == null || centences.size() <= 0 || array == null || array.length <= 0)
         return null;
      
      Iterator it = centences.iterator();
      while (it.hasNext())
      {
         Centence centence = (Centence)it.next();
         centence.weight = (double)0.0;
         Collection features = TFBuilder.getTFFeatures( centence.terms );
         Object[] mapping = node.documentMapping(features);
         if (mapping == null || mapping.length <= 0)
         {
            mapping = null;
            features = null;
            centence.terms = null;
            centence = null;
            continue;
         }
         
         for (int i=0; i<mapping.length; i++)
         {
            double tmp = node.getLocalFeatureWeight ( ((Feature)mapping[i]).getTerm() );
            if (tmp > 0)
               centence.weight += tmp;
         }
         
         // centence.weight = centence.weight / (double)(features.size());
         
         if (centence.bells > 0)
            centence.weight = centence.weight * (double)(centence.bells);
         
         mapping = null;
         features = null;
         centence.terms = null;
         centence = null;
      }
      
      Object[] centArray = centences.toArray();
      Arrays.sort(centArray);
      
      //System.out.println("Centences:");
      //for (int i=0; i<centArray.length; i++)
        //System.out.println ("[" + ((Centence)centArray[i]).weight + "] ->" + new String(array, ((Centence)centArray[i]).start, ((Centence)centArray[i]).end - ((Centence)centArray[i]).start + 1))   ;
      
      StringBuffer buffer = new StringBuffer (500);
      int added = 0;
      for(int i=0; added <= 400 && i < centArray.length && ((Centence)centArray[i]).weight > 0; i++)
      {
          if ( ((Centence)centArray[i]).end - ((Centence)centArray[i]).start < (400 - added) )
          {
            buffer.append(array, ((Centence)centArray[i]).start, ((Centence)centArray[i]).end - ((Centence)centArray[i]).start + 1);
            added += ((Centence)centArray[i]).end - ((Centence)centArray[i]).start + 1;
          }
          else
          {   
            buffer.append(array, ((Centence)centArray[i]).start, (400-added));
            added = 500;
          }  
          buffer.append(' ');
      }   
      centArray = null;
      return buffer.toString();
   }
   
   static String nopreview = "NO Preview";
   public static void setPreview (BingoDocument doc) 
   {
     //System.out.println ("bulk -> " + bulk.length + " , doc:  " + doc); 
     if (doc == null || doc.contentCounter <= 0 || doc.content == null || doc.content.length <= 0)
     {   
        doc.setPreview(nopreview);
        return;
     }
     
     char[] bulk = doc.content;
     int position = doc.contentCounter;
     
     ArrayList centences = createCentences(bulk, position);
     parseCentences(bulk, centences);
     String annotation = annotate (bulk, centences, doc.getNode());
     if (annotation == null || annotation.length() <= 0)
        doc.setPreview(nopreview);
     else
        doc.setPreview(HtmlParserCallback.finish(annotation));
     
     annotation = null;
     centences = null;
     bulk = null;
     
     return;
   }
}
