package bingo.db;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.sql.ResultSet;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Vector;
import javax.swing.JButton;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.table.TableColumn;
import oracle.jdbc.driver.OracleStatement;
import bingo.base.*;
import bingo.crawler.CrawlerFrame;
import bingo.crawler.handler.UrlVerifier;
import bingo.language.LanguageConstants;
import bingo.util.*;
import java.text.*;
/**
 * Interface fr die Datenbankanbindung. Anzeige, Lschen und Auswahl von
 * DB-Inhalten. bernahme von Eintgen in die Trainingsdaten bzw. als Root-URLs
 * fr den Crawler.
 *
 * @author martin @created 26. Juni 2002
 */
public class DBFrame extends JInternalFrame implements ActionListener, Runnable
{
   private static LanguageConstants lc = LanguageConstants.getInstance();
   private SessionBuffer session = SessionBuffer.getInstance();
   private DBInterface dbinterface;
   private JMenuBar menuBar;
   private JPanel panel1 = new JPanel();
   private JPanel panel2 = new JPanel();
   private JLabel selLabel = new JLabel("0 " + lc.get("dbMark"));
   private JLabel whereLabel = new JLabel(lc.get("dbWhere") + ":");
   private JTextField queryField = new JTextField("url like '%'", 40);
   private JButton bCrawl = new JButton(lc.get("crawlerButton"));
   private JButton bTrain = new JButton(lc.get("trainingButton"));
   private JButton bView = new JButton(lc.get("sourceButton"));
   private JButton bBrowser = new JButton(lc.get("browserButton"));
   private JButton bShowFeat = new JButton(lc.get("featureButton"));
   private JButton bShowLinks = new JButton( lc.get( "linkButton" ) );
   //   private JButton bFeatures = new JButton(lc.get("computeButton"));
   //   private JButton bClassify = new JButton(lc.get("classifyButton"));
   private JButton bSelect = new JButton(lc.get("dbSelect"));
   private JButton bStop = new JButton(lc.get("stop"));
   private JMenuItem miCrawl = new JMenuItem(lc.get("crawlerButton"));
   private final String[] columnNames =
   {"ID", "URL", "CLASS", "AUTHORITY", "HUB",
    "PAGERANK", "SVM", "STATUS", "HOST", "MIME", "FEATURES"};
    private static String sTable;
    private CrawlerFrame crawlerFrame;
    public DBTableModel dbTableModel;
    private TableSorter sorter;
    private static boolean singleStart = false;
    private final JTable table;
    private static OracleStatement generalStmt = SessionBuffer.getStatement();
    private static OracleStatement innerStmt = SessionBuffer.getStatement();
    boolean stopped = false;
    Font f = new Font("Serif", Font.PLAIN, 14);
    Font ff = new Font("Serif", Font.ITALIC, 18);
    
    
    protected DBFrame()
    {
       super(lc.get("bDDatabase"), true,
       //resizable
       true,
       //closable
       true,
       //maximizable
       true);
       //iconifiable
       setDefaultCloseOperation(HIDE_ON_CLOSE);
       sTable = session.sTable;
       // Aufbau der GUI
       this.setTitle(lc.get("dbTitle"));
       if (!session.dbConnect)
       {
          session.log("Invalid DB Connection", 1);
          this.setTitle(lc.get("dbTitle") + " - " + lc.get("dbDisconnected"));
       }
       else
       {
          dbinterface = DBInterface.getInstance();
          if (dbinterface == null || !dbinterface.isValid())
          {
             session.log("Invalid DB Connection", 1);
             this.setTitle(lc.get("dbTitle") + " - " + lc.get("dbDisconnected"));
          }
       }
       Container content = this.getContentPane();
       content.setLayout(new BorderLayout());
       dbTableModel = new DBTableModel();
       sorter = new TableSorter(dbTableModel);
       dbTableModel.setColumnNames(columnNames);
       table = new JTable(sorter);
       sorter.addMouseListenerToHeaderInTable(table);
       MouseAdapter listMouseListener = new MouseAdapter()
       {
          public void mouseReleased(MouseEvent e)
          {
             selLabel.setText(table.getSelectedRows().length + " "
             + lc.get("dbMark"));
          }
       };
       table.addMouseListener(listMouseListener);
       TableColumn column = null;
       table.getColumnModel().getColumn(0).setPreferredWidth(200);
       for (int i = 0; i < 2; i++)
       {
          column = table.getColumnModel().getColumn(i);
          column.setPreferredWidth(100);
       }
       table.setBackground(Color.gray);
       table.setForeground(Color.white);
       table.setPreferredScrollableViewportSize(new Dimension(900, 600));
       queryField.setBackground(Color.gray);
       queryField.setForeground(Color.white);
       panel1.setLayout(new BorderLayout());
       panel2.setLayout(new FlowLayout());
       panel2.add(bCrawl);
       panel2.add(bTrain);
       panel2.add(bView);
       panel2.add(bBrowser);
       panel2.add(bShowFeat);
       panel2.add(bShowLinks);
       //       panel2.add(bFeatures);
       //       panel2.add(bClassify);
       panel2.add(selLabel);
       JPanel queryPanel = new JPanel();
       queryPanel.add(whereLabel);
       queryPanel.add(queryField);
       queryPanel.add(bSelect);
       queryPanel.add(bStop);
       JScrollPane scrollpane = new JScrollPane(table);
       content.add(queryPanel, BorderLayout.NORTH);
       content.add(scrollpane, BorderLayout.CENTER);
       content.add(panel2, BorderLayout.SOUTH);
       // MenLeiste
       menuBar = new JMenuBar();
       setJMenuBar(menuBar);
       // Men Datenbank
       JMenu mFile = new JMenu(lc.get("bDDatabase"));
       menuBar.add(mFile);
       JMenuItem miFeatures = addMenuItem(mFile, lc.get("computeButton"));
       mFile.add(miCrawl);
       mFile.addSeparator();
       JMenuItem miClose = addMenuItem(mFile, lc.get("close"));
       // Bearbeiten
       JMenu mEdit = new JMenu(lc.get("edit"));
       menuBar.add(mEdit);
       JMenuItem miCut = addMenuItem(mEdit, lc.get("cut"));
       JMenuItem miCopy = addMenuItem(mEdit, lc.get("copy"));
       JMenuItem miPaste = addMenuItem(mEdit, lc.get("paste"));
       mEdit.addSeparator();
       JMenuItem miSelectAll = addMenuItem(mEdit, lc.get("selectAll"));
       mEdit.addSeparator();
       JMenuItem miClear = addMenuItem(mEdit, lc.get("clearSelection"));
       JMenuItem miClearAll = addMenuItem(mEdit, lc.get("deleteTable"));
       JMenuItem miRemoveFalse = addMenuItem(mEdit, lc.get("deleteFalse"));
       mEdit.addSeparator();
       JMenuItem miSearch = addMenuItem(mEdit, lc.get("search"));
       // ActionListener fr alle Mens registrieren
       bCrawl.addActionListener(this);
       bTrain.addActionListener(this);
       bView.addActionListener(this);
       bBrowser.addActionListener(this);
       bShowFeat.addActionListener(this);
       bShowLinks.addActionListener(this);
       //       bFeatures.addActionListener(this);
       //       bClassify.addActionListener(this);
       bSelect.addActionListener(this);
       bStop.addActionListener(this);
       miCrawl.addActionListener(this);
       miClose.addActionListener(this);
       miClear.addActionListener(this);
       miClearAll.addActionListener(this);
       miSelectAll.addActionListener(this);
       miFeatures.addActionListener(this);
       miSearch.addActionListener(this);
       miCut.addActionListener(this);
       miCopy.addActionListener(this);
       miPaste.addActionListener(this);
       miRemoveFalse.addActionListener(this);
       KeyAdapter keyAdapterQuery = new KeyAdapter()
       {
          public void keyPressed(KeyEvent e)
          {
             if (e.getKeyCode() == KeyEvent.VK_ENTER)
             {
                initializeSelection();
             }
          }
       };
       queryField.addKeyListener(keyAdapterQuery);
       KeyAdapter keyAdapterTable = new KeyAdapter()
       {
          public void keyPressed(KeyEvent e)
          {
             selLabel.setText(table.getSelectedRows().length + " "
             + lc.get("dbMark"));
          }
          public void keyReleased(KeyEvent e)
          {
             selLabel.setText(table.getSelectedRows().length + " "
             + lc.get("dbMark"));
          }
       };
       table.addKeyListener(keyAdapterTable);
       this.setSize(session.defaultFrameWidth, session.defaultFrameHeight);
       this.setVisible(true);
    }
    public void actionPerformed(ActionEvent e)
    {
       String cmd = e.getActionCommand();
       stopped = true;
       try
       {
          if (cmd == (lc.get("close")))
          {
             if (singleStart)
             {
                session.closeDB();
                System.exit(0);
             }
             else
             {
                dbinterface.close();
                dbinterface = null;
             }
             setVisible(false);
          }
          else if (cmd == (lc.get("crawlerButton")))
          {
             initializeCrawler();
          }
          else if (cmd == (lc.get("trainingButton")))
          {
             initializeTraining();
          }
          else if (cmd == (lc.get("dbSelect")))
          {
             table.clearSelection();
             initializeSelection();
          } else if (cmd == (lc.get("clearSelection")))
          {
             removeFromTable(table.getSelectedRows());
             table.clearSelection();
             initializeSelection();
          } else if (cmd == (lc.get("deleteTable")))
          {
             table.selectAll();
             removeAllFromTable();
             table.clearSelection();
             initializeSelection();
          } else if (cmd == (lc.get("selectAll")))
          {
             table.selectAll();
             selLabel.setText(table.getSelectedRows().length + " "
             + lc.get("dbMark"));
          }
          else if (cmd == (lc.get("sourceButton")))
          {
             BingoDocument bd = getBingoDoc(table.getSelectedRow());
             if (bd != null)
                new ParserFrame(getBingoDoc(table.getSelectedRow()));
          }
          
          else if ( cmd == ( lc.get( "browserButton" ) ) )
          {
             BingoDocument selectedURL = getBingoDoc(table.getSelectedRow());
             if (selectedURL != null)
             {
                String target = null;
                if (selectedURL.getUrl2r() != null && selectedURL.getUrl2r().getURL() != null)
                {
                   target = selectedURL.getUrl2r().getURL().toString();
                }
                else
                {
                   target = "http://" + selectedURL.getUrl();
                }
                
                if (target != null)
                {
                   try
                   {
                      bingo.util.BrowserLauncher.openURL(target);
                   }
                   catch (Exception ble)
                   {  }
                }
                target = null;
             }
          }
          
          else if (cmd == (lc.get("computeButton")))
          {
             Vector rootVector = getBingoDocs();
             removeFromTable(table.getSelectedRows());
             FeatureDialog fd = new FeatureDialog(
             BINGODesktop.getInstance(), rootVector);
             fd.setVisible(true);
             table.clearSelection();
             initializeSelection();
          }
          else if (cmd == (lc.get("featureButton")))
          {
             BingoDocument bd = getBingoDoc(table.getSelectedRow());
             if (bd != null)
                new FeatureViewer(this, bd).setVisible(true);
          }
          else if (cmd == ( lc.get( "linkButton" ) ) )
          {
             BingoDocument bd = getBingoDoc(table.getSelectedRow());
             if (bd != null)
                new LinksViewer( this, bd ).setVisible( true );
          }
          else if (cmd == (lc.get("cut")))
          {
             Vector v = getBingoDocs();
             removeFromTable(table.getSelectedRows());
             session.clearURLBuffer();
             for (Enumeration enum = v.elements(); enum.hasMoreElements();)
             {
                BingoDocument temp = (BingoDocument) enum.nextElement();
                session.addURLtoBuffer(temp);
             }
             table.clearSelection();
             initializeSelection();
          } else if (cmd == (lc.get("copy")))
          {
             Vector v = getBingoDocs();
             session.clearURLBuffer();
             for (Enumeration enum = v.elements(); enum.hasMoreElements();)
             {
                BingoDocument temp = (BingoDocument) enum.nextElement();
                session.addURLtoBuffer(temp);
             }
          } else if (cmd == (lc.get("paste")))
          {
             Vector v = session.getURLBuffer();
             FeatureDialog fd = new FeatureDialog(
             BINGODesktop.getInstance(), v);
             fd.setVisible(true);
             table.clearSelection();
             initializeSelection();
          } else if (cmd == (lc.get("deleteFalse")))
          {
             table.clearSelection();
             for (int i = 0; i < dbTableModel.getRowCount(); i++)
             {
                if (dbTableModel.getValueAt(i, 9).toString().equals("false"))
                {
                   addSelection(i);
                }
             }
             removeFromTable(table.getSelectedRows());
             table.clearSelection();
             initializeSelection();
          } else if (cmd == (lc.get("search")))
          {
             new DBSearchDialog().setVisible(true);
          } else if (cmd == (lc.get("stop")))
          {
             stopped = true;
          }
       } catch (Exception any)
       {
          session.log(any);
       }
    }
    private JMenuItem addMenuItem(JMenu target, String text)
    {
       JMenuItem mi = new JMenuItem(text);
       target.add(mi);
       return mi;
    }
    public void addSelection(int i)
    {
       try
       {
          table.addRowSelectionInterval(i, i);
       } catch (Exception any)
       {
       }
    }
    public void initializeSelection()
    {
       stopped = true;
       new Thread(this).start();
    }
    public void run()
    {
       if (!session.dbConnect)
       {
          return;
       }
       selLabel.setText("0 " + lc.get("dbMark"));
       table.clearSelection();
       dbTableModel.removeAllRows();
       int i = 0;
       try
       {
          String query = "select url, id, class, authority, hub, pagerank, svm, status, host, mime from documents where "
          + queryField.getText() + " ";
          ResultSet generalSet = generalStmt.executeQuery(query);
          DecimalFormat formatter = new DecimalFormat("#0.#####");
          stopped = false;
          while (generalSet.next() && !stopped)
          {
             //                DateFormat df = DateFormat.getDateTimeInstance(
             // DateFormat.SHORT, DateFormat.SHORT );
             String url = generalSet.getString("URL");
             String id = generalSet.getString("ID");
             String sClass = generalSet.getString("CLASS");
             String host = generalSet.getString("HOST");
             String mime = generalSet.getString("MIME");
             String status = generalSet.getString("STATUS");
             String svm = formatter.format(generalSet.getDouble("SVM"));
             String auth = formatter.format(generalSet.getDouble("AUTHORITY"));
             String hub = formatter.format(generalSet.getDouble("HUB"));
             String pagerank = formatter.format(generalSet.getDouble("PAGERANK"));
             ResultSet innerSet = innerStmt
             .executeQuery("select id from DocumentFeatures where id = "
             + String.valueOf(id));
             if (innerSet.next())
             {
                Object[] sRow =
                {id, url, sClass, auth, hub, pagerank, svm,
                 status, host, mime, "true", id};
                 dbTableModel.addRow(sRow);
             } else
             {
                Object[] sRow =
                {id, url, sClass, auth, hub, pagerank, svm,
                 status, host, mime, "false", id};
                 dbTableModel.addRow(sRow);
             }
             i++;
             innerSet.close();
          }
          generalSet.close();
          formatter = null;
       } catch (Exception any)
       {
          session.log(any);
       }
       dbTableModel.fireTableDataChanged();
       this.setTitle(lc.get("dbTitle") + " - " + lc.get("dbConnected") + " '"
       + sTable.toUpperCase() + "', " + i + " " + lc.get("dbEntries"));
    }
    
    private void removeFromTable(int[] iRows)
    {
       if (!session.dbConnect)
       {
          return;
       }
       String query = "delete from bingodocuments where url in ( ";
       try
       {
          for (int i = 0; i < iRows.length; i++)
          {
             String url = (String) table.getValueAt(iRows[i], 0);
             query += "'" + url + "',";
          }
          query = query.substring(0, query.length() - 1);
          query += " )";
          generalStmt.executeQuery(query);
          generalStmt.executeQuery("commit");
       } catch (Exception any)
       {
          session.log(any);
       }
    }
    private void removeAllFromTable()
    {
       if (!session.dbConnect)
       {
          return;
       }
       try
       {
          generalStmt.executeQuery("delete from " + sTable + "");
          generalStmt.executeQuery("commit");
       } catch (Exception any)
       {
          session.log(any);
       }
    }
    public static void close() throws Exception
    {
       if (innerStmt != null)
       {
          innerStmt.close();
       }
       if (generalStmt != null)
       {
          generalStmt.close();
       }
    }
    
    private void initializeCrawler()
    {
       if (!session.dbConnect)
       {
          return;
       }
       crawlerFrame = CrawlerFrame.getInstance();
       crawlerFrame.setVisible(true);
       Vector v = getBingoDocs();
       for (Enumeration enum = v.elements(); enum.hasMoreElements();)
       {
          crawlerFrame.addLink((BingoDocument) enum.nextElement());
       }
    }
    
    private void initializeTraining()
    {
       if (session.getRootNode() == null)
          return;
       
       DBInterface dbi = null;
       if (session.dbConnect)
          dbi = DBInterface.getInstance();
       
       Vector v = getBingoDocs();
       for (Enumeration enum = v.elements(); enum.hasMoreElements();)
       {
          BingoDocument bd = (BingoDocument) enum.nextElement();
          if (bd == null)
             continue;
          
          BingoTreeNode hisNode = bd.getNode();
          if (hisNode != null && hisNode != session.getRootNode())
          {
             hisNode.addDocument(bd);
             bd.setStatus('M');
             
             if (dbi != null)
                dbi.updateStatus(bd, 'M');
          }
          bd=null; hisNode=null;
       }
       v=null; dbi=null;
    }
    
    
    /**
     * Teilt dem Object mit, dass es im Einzelstart-Modus (ohne
     * Klassifikations-Modul) laufen soll.
     *
     * @param b
     *            Einzelstart ja oder nein
     */
    public void setSingleStart(boolean b)
    {
       this.singleStart = b;
    }
    private Vector getBingoDocs()
    {
       Vector vURLs = new Vector(10);
       int[] rows = table.getSelectedRows();
       for (int i = 0; i < rows.length; i++)
       {
          BingoDocument temp = getBingoDoc(rows[i]);
          if (temp!=null)
             vURLs.addElement(temp);
       }
       return vURLs;
    }
    // erzeugt neues BingoDocument
    private BingoDocument getBingoDoc(int position)
    {
       if (!session.dbConnect)
          return null;
       
       DBInterface dbInterface = DBInterface.getInstance();
       if (dbInterface == null || !dbInterface.isValid())
          return null;
       
       //     String stringID = (String) dbTableModel.getRowAt(position)[0];
       String stringID = (String) table.getValueAt( position, 0 );
       if (stringID == null)
          return null;
       
       long id = 0;
       try
       {
          id = Long.parseLong(stringID);
       }
       catch (NumberFormatException ne)
       {
          session.log("Konvertierung String-Int fuer ID: " + stringID + " fehlgeschlagen", 10);
          return null;
       }
       
       BingoDocument bd = dbInterface.pullDocument(id);
       String stringClass = (String)dbTableModel.getRowAt(position)[2];
       if (bd != null && stringClass != null)
          bd.setNode( session.getRootNode().getNode(stringClass) );
       if (bd.getNode() == null)
          bd.setNode(session.getRootNode());
       
       return bd;
    }
    
    //    /**
    //     * Interfacemethode zu BaseTreeCaller: ndert die Kategorie eines
    //     * Tabelleneintrags
    //     *
    //     * @param s Die Kategorie als String
    //     * @param n The new path value
    //     * @param rank The new path value
    //     */
    //
    //    public void setPath( String s, DefaultMutableTreeNode n, Double rank ) {
    //        try {
    //            table.setValueAt( s, table.getSelectedRow(), 1 );
    //            table.setValueAt( String.valueOf( rank ), table.getSelectedRow(), 5 );
    //            sorter.fireTableCellUpdated( table.getSelectedRow(), 1 );
    //            sorter.fireTableCellUpdated( table.getSelectedRow(), 5 );
    //            if ( SessionBuffer.dbConnect ) {
    //                Link link = new Link((String) table.getValueAt( table.getSelectedRow(),
    // 0 ) );
    //                session.dbinterface.updateSVMResult(link.id, s, rank );
    //                link = null;
    //            }
    //        }
    //        catch ( Exception any ) {
    //            session.log(any);
    //        }
    //    }
    // Singleton
    private static DBFrame dbFrame = null;
    /**
     * Die Methode dient dem Aufruf der (einzigen) Instanz der Klasse.
     *
     * @return The instance value
     */
    public static DBFrame getInstance()
    {
       if (dbFrame == null)
       {
          dbFrame = new DBFrame();
          BINGODesktop.getInstance().add(dbFrame);
          dbFrame.initializeSelection();
       }
       return dbFrame;
    }
    public void setSelectedIndex(int i)
    {
       try
       {
          table.setRowSelectionInterval(i, i);
       } catch (Exception any)
       {
       }
    }
}
