package bingo.db;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.sql.Connection;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.net.URL;

import bingo.crawler.frontier.url2resolve;
import bingo.crawler.handler.UrlVerifier;
import bingo.crawler.handler.parser.TextItem;
import bingo.svmlight.Feature;
import bingo.util.BingoDocument;
import bingo.util.BingoTreeNode;
import bingo.util.SessionBuffer;
public abstract class DBInterface
{
   protected boolean timedOut = false;
   protected static SessionBuffer session;
   public Connection conn = null;
   protected Statement stmt = null;
   protected PreparedStatement pstmt, pstmt1;
   protected ResultSet rset;
   protected String query;
   protected static int chunkSize;
   protected static int batchSize;
   protected static int counter = 0;
   protected boolean valid;
   /**
    * Creates a new instance of DBInterface
    */
   static String oraString = "oracle";
   public static DBInterface getInstance()
   {
      session = SessionBuffer.getInstance();
      batchSize = session.getBatchSize();
      chunkSize = session.getChunkSize();
      DBInterface result;
      if (session.dbType.equals(oraString))
         result = new OracleInterface();
      else
         result = new MySQLInterface();
      result.initConnection();
      return result;
   }
   
   static String pullTrainQuery = "SELECT ID FROM BingoDocuments WHERE Class_ID = ? AND Status IN ('B', 'M')";
   public synchronized ArrayList pullTrainingDocuments(BingoTreeNode klasse)
   {
      ArrayList result = new ArrayList(100);
      try
      {
         pstmt = conn.prepareStatement(pullTrainQuery);
         pstmt.setInt(1, klasse.ID);
         ResultSet rset = pstmt.executeQuery();
         while (rset.next())
         {
            BingoDocument tmp = pullDocument(rset.getLong(1));
            if (tmp!=null && tmp.getFeatureVector()!=null &&  tmp.getFeatureVector().size() > 0)
            {
               tmp.setNode(klasse);
               tmp.setStatus('B');
               result.add(tmp);
            }
         }
         rset.close();rset=null;
         pstmt.close();pstmt=null;
         
      }
      catch (Exception e)
      {
         session.log(e);
         return result;
      }
      return result;
   }
   
   static String pullJunkQuery = "SELECT ID FROM BingoDocuments WHERE Class_ID = ? AND Status = 'J'";
   public synchronized ArrayList pullJunkDocuments(BingoTreeNode klasse)
   {
      ArrayList result = new ArrayList(100);
      try
      {
         pstmt = conn.prepareStatement(pullJunkQuery);
         pstmt.setInt(1, klasse.ID);
         ResultSet rset = pstmt.executeQuery();
         while (rset.next())
         {
            BingoDocument tmp = pullDocument(rset.getLong(1));
            if (tmp!=null && tmp.getFeatureVector()!=null &&  tmp.getFeatureVector().size() > 0)
            {
               tmp.setNode(klasse);
               tmp.setStatus('J');
               result.add(tmp);
            }
         }
         rset.close();rset=null;
         pstmt.close();pstmt=null;
         
      }
      catch (Exception e)
      {
         session.log(e);
         return result;
      }
      return result;
   }
   
   static String urlbyidquery = "SELECT URL FROM BingoDocuments WHERE ID = ?";
   public BingoDocument pullDocument(long docID)
   {
      BingoDocument doc = null;
      try
      {
         pstmt = conn.prepareStatement(urlbyidquery);
         pstmt.setLong(1, docID);
         ResultSet rset = pstmt.executeQuery();
         if (rset.next())
            doc = UrlVerifier.string2BingoDocument(rset.getString(1));
         
         rset.close(); rset=null;
         pstmt.close(); pstmt=null;
         
         if (doc == null)
            return doc;
         
         doc.id = docID;
         fillDocument(doc);
      }
      catch (Exception e)
      {
         session.log(e);
         return doc;
      }
      
      return doc;
   }
   
   public void fillDocument(BingoDocument doc) throws Exception
   {
      pullLink(doc);
      doc.setLinks(pullLinks(doc.getId()));
      doc.setText(pullTerms(doc.getId()));
      doc.setFeatureVector(pullFeatures(doc.getId()));
   }
   
   // ausgelagerte Methoden
   //---------------------------
   public synchronized ArrayList pullFeatures(long ID)
   {
      ArrayList features = new ArrayList();
      try
      {
         pstmt = conn.prepareStatement("SELECT Term, TF FROM DocumentFeatures WHERE ID = ?");
         pstmt.setLong(1, ID);
         rset = pstmt.executeQuery();
         while (rset.next())
            features.add(new Feature(rset.getString("TERM"), rset.getInt("TF")));
         rset.close();
         pstmt.close();
         
         session.log("Loading features for docID: " + ID, 7);
                        /*			if (features == null || features.size() == 0) {
                                features = new TFBuilder().getTFFeatures(pullTerms(ID));
                        }*/
      } catch (Exception e)
      {
         features = null;
         session.log(e);
      }
      return features;
   }
   
   public synchronized long[] getVisitedLinks()
   {
      try
      {
         rset = stmt.executeQuery("select count(*) from bingodocuments");
         rset.next();
         int numLinks = rset.getInt(1);
         rset.close();
         rset = null;
         if (numLinks <= 0)
            return null;
         rset = stmt.executeQuery("select ID from bingodocuments");
         long[] result = new long[numLinks];
         int i = 0;
         while (rset.next())
         {
            result[i] = rset.getLong(1);
            i++;
         }
         rset.close();
         rset = null;
         return result;
      } catch (Exception e)
      {
         session.log(e);
      }
      return null;
   }
   
   public synchronized long[][] getParsedLinks()
   {
      try
      {
         rset = stmt.executeQuery("select count(*) from bingodocuments where md5 <> 0");
         rset.next();
         int numLinks = rset.getInt(1);
         rset.close();
         rset = null;
         if (numLinks <= 0)
            return null;
         rset = stmt.executeQuery("select md5, id from bingodocuments where md5 <> 0");
         long[][] result = new long[numLinks][2];
         int i = 0;
         while (rset.next() && i<result.length)
         {
            result[i][0] = rset.getLong(1);
            result[i][1] = rset.getLong(2);
            i++;
         }
         rset.close();
         rset = null;
         return result;
      } catch (Exception e)
      {
         session.log(e);
      }
      return null;
   }
   
   
   static String pullLinksQuery1 = "SELECT Target From Links WHERE ID = ?";
   public synchronized LinkedList pullLinks(long ID)
   {
      LinkedList links = new LinkedList();
      try
      {
         PreparedStatement pstmt = conn.prepareStatement(pullLinksQuery1);
         pstmt.setLong(1,ID);
         ResultSet rset = pstmt.executeQuery();
         while (rset.next())
         {
            URL tmp = UrlVerifier.checkAbsoluteLink(rset.getString(1));
            if (tmp != null)
              links.add(tmp);
         }
         rset.close(); rset=null;
         pstmt.close(); pstmt=null;
      }
      catch (Exception e)
      {
         links = null;
         session.log(e);
      }
      return links;
   }
   
   public synchronized LinkedList pullTerms(long ID)
   {
      LinkedList terms = new LinkedList();
      try
      {
         pstmt = conn.prepareStatement("SELECT Term, Pos FROM Terms WHERE ID = ? ORDER BY Pos Asc");
         pstmt.setLong(1, ID);
         rset = pstmt.executeQuery();
         while (rset.next())
         {
            TextItem ti = new TextItem(rset.getString("Term"), rset.getInt("Pos"));
            ti.stem = ti.term;
            terms.add(ti);
            ti=null;
         }
         rset.close();rset=null;
         pstmt.close();
      } 
      catch (Exception e)
      {
         session.log(e);
      }
      //System.out.println("+++++++++++++++++++++++++++++++++++++++++++ " + terms);
      return terms;
   }
   
   public abstract void pushModel(int iteration) throws Exception;
   public abstract void pullModel(int id) throws Exception;
   public synchronized void pullModel(long modelDate) throws Exception
   {
   }
   //---------------------------
   public void initConnection()
   {
      conn = null;
      stmt = null;
      valid = false;
      try
      {
         if (session.dbConnect)
         {
            conn = session.getDBConnection();
         }
      } catch (Exception e)
      {
         session.log(e);
      }
      if (conn != null)
      {
         try
         {
            stmt = conn.createStatement();
         } catch (Exception e)
         {
            session.log(e);
         }
      }
      if (conn != null && stmt != null)
      {
         valid = true;
      }
      if (!valid)
      {
         SessionBuffer.log("Warning: DB-Connection invalid", 2);
      }
   }
   public void reset()
   {
      close();
      initConnection();
   }
   public void rollback()
   {
      try
      {
         if (conn != null)
         {
            conn.rollback();
         }
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   public void commit()
   {
      try
      {
         if (conn != null)
            conn.commit();
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   public Statement getStatement()
   {
      Statement st = null;
      if (valid && conn != null)
      {
         try
         {
            st = conn.createStatement();
         } catch (Exception e)
         {
            session.log(e);
         }
      }
      return st;
   }
   /**
    * Schreibt die hierarchische Kategoriezuordnung einer URL in die
    * Datenbank.
    *
    * @param sURL
    *            Die URL des Dokuments.
    * @param className
    *            String mit der Kategorie im Baum. Trenner muss '/' sein.
    * @param rank
    *            Description of the Parameter
    */
   public synchronized void updateSVMResult(long id, String className, Double rank)
   {
      try
      {
         pstmt = conn
         .prepareStatement("UPDATE bingodocuments SET Class_ID = ?, SVMScore = ? WHERE ID = ? ");
         pstmt.setInt(1, BingoTreeNode.getID(className));
         pstmt.setDouble(2, rank.doubleValue());
         pstmt.setLong(3, id);
         pstmt.executeUpdate();
         pstmt.close();
         pstmt = null;
         stmt.executeQuery("COMMIT");
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   /**
    * Schreibt einen Status-Character zu einer URL in die Datenbank. Der
    * Status gibt die Art der Klassifikation eines Dokuments an.
    *
    * @param sURL
    *            Die URL des Dokuments.
    * @param c
    *            Der Status.
    */
   public synchronized void updateStatus(BingoDocument bd, char c)
   {
      if (bd == null)
         return;
      updateStatus(bd.getId(), c);
   }
   private synchronized void updateStatus(long id, char c)
   {
      try
      {
         pstmt = conn.prepareStatement("UPDATE bingodocuments SET Status = ? WHERE ID = ?");
         char[] ch =
         {c};
         pstmt.setString(1, String.valueOf(ch));
         pstmt.setLong(2, id);
         pstmt.executeUpdate();
         pstmt.close();
         stmt.executeQuery("COMMIT");
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   /**
    * Generiert einen Datenbankeintrag aus einem Link (erhalten per Crawling).
    * Der Content wird ber die bereits heruntergeladene Page des Links
    * eingelesen und gespeichert.
    *
    * @param link
    *            Der Link.
    * @return URL neu eingetragen.
    */
   public synchronized boolean pushLink(BingoDocument bd)
   {
      if (bd == null || bd.isBroken())
         return false;
      
      long ID = bd.getId();
      boolean up2date = false;
      url2resolve u = bd.getUrl2r();
      if (u == null || u.getURL() == null)
         return false;
      char status = bd.getStatus();
      try
      {
         up2date = true;
         BingoTreeNode node = bd.getNode();
         int Class_ID = 0;
         if (node != null)
            Class_ID = node.ID;
         node = null;
         char[] status1 =
         {status};
         
         pstmt = conn.prepareStatement("INSERT INTO BingoDocuments (ID, URL, Hostname, IP, Port, Filename, Lastmod, mime, ResponseCode, ContentEncoding, Expiration, Status, Depth, Class_ID, SvmScore, HashCode, md5) VALUES (?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?)");
         pstmt.setLong(1, ID);
         pstmt.setString(2, u.getURL().toString());
         // NICHT sURL wegen gross/kleinschreibung
         pstmt.setString(3, u.getHost());
         pstmt.setLong(4, u.getIP());
         pstmt.setInt(5, u.getPort());
         pstmt.setString(6, u.getFile());
         pstmt.setDate(7, new Date(bd.getLastModified()));
         pstmt.setString(8, bd.getContentType());
         pstmt.setInt(9, bd.getResponseCode());
         pstmt.setString(10, bd.getContentEncoding());
         pstmt.setLong(11, bd.getExpiration());
         pstmt.setString(12, String.valueOf(status1));
         pstmt.setInt(13, bd.getDepth());
         pstmt.setInt(14, Class_ID);
         pstmt.setDouble(15, bd.getSvmRank().doubleValue());
         
         // Signatur: Host-IP + Groesse des Dokuments in Bytes
         long signatur = ((u.getIP()) << 32) + (long)bd.getBulkSize();
         // session.log ("Signatur: " + signatur, 1);
         pstmt.setLong(16, signatur);
         
         pstmt.setLong(17, bd.getMD5());
         
         pstmt.executeUpdate();
         pstmt.close();
         /*
         if (bd.hasForms())
         {
            pstmt = conn
            .prepareStatement("INSERT INTO hasForms (ID) VALUES (?)");
            pstmt.setInt(1, ID);
            pstmt.executeUpdate();
            pstmt.close();
         }
          */
      }
      catch (Exception e)
      {
         bd.setBroken(true);
         try
         {
            // conn.rollback();
            pstmt.close();
            pstmt = null;
         }
         catch (Exception e1)
         {
         }
         session.log(e);
      }
      bd = null;
      u = null;
      return up2date;
   }
   
   static String insertDuplicateString = "INSERT INTO Duplicates (id, link_id, MD5) VALUES (?,?,?)";
   public synchronized boolean pushDuplicate(BingoDocument bd)
   {
      if (bd == null || bd.isBroken())
         return false;
      
      boolean up2date = true;
      try
      {
         pstmt = conn.prepareStatement(insertDuplicateString);
         pstmt.setLong(1, bd.getId());
         pstmt.setLong(2, bd.getParsedBefore());
         pstmt.setLong(3, bd.getMD5());
         pstmt.executeUpdate();
         pstmt.close(); pstmt=null;
         
         System.out.println(bd.getParsedBefore() + " D -> " + bd.getUrl2r().getURL());
      }
      catch (Exception e)
      {
         up2date = false;
         
         if (bd != null)
         {
            bd.setBroken(true);
         }
         
         try
         {
            pstmt.close();
            pstmt = null;
         }
         catch (Exception e1)
         { }
      }
      return up2date;
   }
   
   public synchronized boolean pushTrainingDocument(BingoDocument bd)
   {
      long ID = bd.getId();
      boolean up2date = false;
      char status = bd.getStatus();
      //session.log("id: " + ID + "; status: " + status, 5);
      try
      {
         up2date = true;
         // Eintrag in bingodocuments machen bzw. korrigieren
         BingoTreeNode node = bd.getNode();
         int Class_ID = 0;
         if (node != null)
            Class_ID = node.ID;
         node = null;
         char[] status1 =
         {status};
         pstmt = conn.prepareStatement("INSERT INTO bingodocuments "
         + "(ID, URL, Status, Depth, Class_ID, SvmScore) "
         + "VALUES (?,?,?,?,?,?)");
         pstmt.setLong(1, ID);
         pstmt.setString(2, bd.getUrl());
         pstmt.setString(3, String.valueOf(status1));
         pstmt.setInt(4, 1);
         pstmt.setInt(5, Class_ID);
         pstmt.setDouble(6, bd.getSvmRank().doubleValue());
         pstmt.executeUpdate();
         pstmt.close();
      } catch (Exception e)
      {
         String emsg = e.getMessage();
         if (!emsg.startsWith("ORA-00001"))
            session.log(e);
         try
         {
            conn.rollback();
            pstmt.close();
         } catch (Exception e1)
         {
         }
      }
      return up2date;
   }
   /**
    * Schreibt die geparsten Terme zusammen mit ihrer relativen Position im
    * Dokument als Nested-Table in die Datenbank. Berechnet optional auch die
    * Features mit TF-Werten fr dieses Dokument.
    *
    * @param buildFeatures
    *            Berechne Feature-Tabelle.
    * @param link
    *            Description of Parameter
    */
   public synchronized void pushNested(BingoDocument bd)
   {
      if (bd == null)
         return;
      if (bd.getStatus() == 'B')
      {
         pushTerms(bd.getId(), bd.getText());
      }
      pushLinks(bd.getId(), bd.getLinks());
      pushFeatures(bd.getId(), bd.getFeatureVector());
      pushPreviews(bd.getId(), bd.getTitle(), bd.getPreview(), bd.getBulk(),
      bd.getBulkSize());
   }
   //	public synchronized void pushNested(BaseURL link) {
   //		if (link == null)
   //			return;
   //		pushTerms(link.id, link.getText());
   //		pushLinks(link.id, link.getLinks());
   //		pushFeatures(link.id, link.getFeatures());
   //		pushPreviews(
   //			link.id,
   //			link.getTitle(),
   //			link.getPreview(),
   //			link.bulk,
   //			link.bulkSize);
   //	}
   //===============================================================================
   static String pushFeaturesQuery = "INSERT INTO DocumentFeatures (ID, Term, tf, rtf, tf_idf) VALUES (?,?,?,?,?)";
   public synchronized void pushFeatures(long ID, Collection features)
   {
      if (features == null || features.size() == 0)
         return;
      // Erster Scan zur Berechnung von RTF
      double count = 0.0;
      Iterator it = features.iterator();
      while (it.hasNext())
      {
         count += ((Feature) it.next()).getTF();
      }
      it = null;
      try
      {
         int k = 0;
         pstmt = conn.prepareStatement(pushFeaturesQuery);
         it = features.iterator();
         while (it.hasNext())
         {
            Feature feature = (Feature) it.next();
            pstmt.setLong(1, ID);
            pstmt.setString(2, feature.getTerm().toString());
            pstmt.setDouble(3, feature.getTF());
            if (count > 0)
            {
               pstmt.setDouble(4, feature.getTF() / count);
            } else
            {
               pstmt.setDouble(4, feature.getTF());
            }
            //         ACHTUNG!!! Unbedingt beachten !!!
            //         TFIDF wird on demand dynamisch berechnet!
            //         Die Endberechnung fuer die Datenbank soll NACH dem Crawl
            // problemspezifisch erfolgen !!!!
            pstmt.setDouble(5, 0.0);
            pstmt.addBatch();
            if (++k % batchSize == 0)
               pstmt.executeBatch();
         }
         it = null;
         pstmt.executeBatch();
         pstmt.close();
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   /**
    * Schreibt ein Feld von Links als Nested-Table zu einer URL.
    *
    * @param link
    *            Description of the Parameter
    */
   static String pushLinksQuery = "INSERT INTO Links (Target, Hostname, Link_ID, ID) VALUES (?,?,?,?)";
   public synchronized void pushLinks(long ID, Collection links)
   {
      if (links == null || links.size() <= 0)
         return;
      try
      {
         int k = 0;
         pstmt = conn.prepareStatement(pushLinksQuery);
         Iterator it = links.iterator();
         while (it.hasNext())
         {
            URL tmp = (URL)it.next();
            String tmpString = tmp.toString();
            pstmt.setString(1, tmpString);
            pstmt.setString(2, tmp.getHost());
            pstmt.setLong(3, BingoDocument.getID(tmpString));
            pstmt.setLong(4, ID);
            pstmt.addBatch();
            if (++k % batchSize == 0)
               pstmt.executeBatch();
            
            tmpString = null;
            tmp = null;
         }
         it = null;
         pstmt.executeBatch();
         pstmt.close();
         pstmt = null;
         //conn.commit();
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   /**
    * Aktualisiert die Stopwort-Liste in der Datenbank
    */
   static String deleteStopwordsQuery = "delete from stopwords";
   static String pushStopwordsQuery = "INSERT INTO Stopwords (Term) VALUES (?)";
   public synchronized void pushStopwords()
   {
      int k;
      try
      {
         String next = null;
         Statement stmt = conn.createStatement();
         stmt.executeUpdate(deleteStopwordsQuery);
         stmt.close();
         stmt = null;
         conn.commit();
         pstmt = conn.prepareStatement(pushStopwordsQuery);
         k = 0;
         Iterator iter = session.getStopwords();
         while (iter.hasNext())
         {
            next = (String) iter.next();
            pstmt.setString(1, next);
            pstmt.addBatch();
            if (++k % batchSize == 0)
               pstmt.executeBatch();
         }
         iter = null;
         pstmt.executeBatch();
         pstmt.close();
         pstmt = null;
         conn.commit();
      } catch (Exception e)
      {
         session.log(e);
      }
   }
   /**
    * Aktualisiert die Stopwort-Liste in der Datenbank
    *
    * @return Description of the Return Value
    */
   public HashSet pullStopwords()
   {
      HashSet stopwords = new HashSet();
      try
      {
         Statement stmt = conn.createStatement(ResultSet.TYPE_FORWARD_ONLY,
         ResultSet.CONCUR_READ_ONLY);
         ResultSet rset = stmt.executeQuery("SELECT Term FROM Stopwords");
         while (rset.next())
         {
            stopwords.add(rset.getString(1));
         }
         rset.close();
         stmt.close();
      } catch (Exception e)
      {
         session.log(e);
      }
      return stopwords;
   }
   public void pushClasses(BingoTreeNode rootNode)
   {
      try
      {
         stmt.executeUpdate("delete from classes");
         pstmt = conn
         .prepareStatement("INSERT INTO Classes (Class_ID, Name) VALUES (?,?)");
         Iterator it = rootNode.allClasses.values().iterator();
         while (it.hasNext())
         {
            BingoTreeNode node = (BingoTreeNode) it.next();
            pstmt.setInt(1, node.ID);
            pstmt.setString(2, node.label);
            pstmt.addBatch();
         }
         pstmt.executeBatch();
         pstmt.close();
         stmt.executeUpdate("commit");
      } catch (Exception any)
      {
         session.log(any);
      }
   }
   /**
    * Liest die SVM-Modellhierarchie mit neuestem Erstellungsdatum aus der
    * Datenbank und speichert sie im Objekt session.
    *
    * @see session
    */
   public abstract void pullLatestModel();
   public synchronized void close()
   {
      try
      {
         if (rset != null)
         {
            rset.close();
         }
      } catch (Exception e)
      {
      }
      try
      {
         if (stmt != null)
         {
            stmt.close();
         }
      } catch (Exception e)
      {
      }
      try
      {
         if (pstmt != null)
         {
            pstmt.close();
         }
      } catch (Exception e)
      {
      }
      try
      {
         if (conn != null)
         {
            conn.close();
         }
      } catch (Exception e)
      {
      }
      rset = null;
      pstmt = null;
      stmt = null;
      conn = null;
      valid = false;
   }
   public boolean isValid()
   {
      return valid;
   }
   /**
    * Schreibt die geparsten Terme zusammen mit ihrer relativen Position im
    * Dokument als Nested-Table in die Datenbank. Berechnet optional auch die
    * Features mit TF-Werten fr dieses Dokument.
    *
    * @param link
    *            Description of the Parameter
    */
   protected synchronized void pushTerms(long ID, Collection terms)
   {
      if (terms == null || terms.size() == 0)
         return;
      try
      {
         int k = 0;
         pstmt = conn.prepareStatement("INSERT INTO Terms (Term, Stem, Pos, ID) VALUES (?, ?, ?, ?)");
         Iterator it = terms.iterator();
         while (it.hasNext())
         {
            TextItem item = (TextItem) it.next();
            if (!session.saveStopwordTerms && item.isStopword)
               continue;
            pstmt.setString(1, item.term);
            pstmt.setString(2, item.stem);
            pstmt.setInt(3, item.position);
            pstmt.setLong(4, ID);
            pstmt.addBatch();
            if (++k % batchSize == 0)
            {
               pstmt.executeBatch();
            }
         }
         it = null;
         pstmt.executeBatch();
         pstmt.close();
      } catch (Exception e)
      {
         session.log(e);
         try
         {
            stmt.executeQuery("ROLLBACK WORK");
            pstmt.close();
         } catch (Exception e1)
         {
         }
      }
   }
   /**
    * Hilfsmethode zur Berechnung der Datenbank-ID aus dem URL-String In der
    * aktuellen Version : HashCode
    *
    * @param from
    *            Description of the Parameter
    * @return The iD value
    */
   public ArrayList getStartLinks()
   {
      ArrayList result = null;
      try
      {
         Statement stmt = getStatement();
         ResultSet rset = stmt.executeQuery("SELECT * FROM start_urls");
         result = new ArrayList();
         while (rset.next())
            result.add(rset.getString(1));
         rset.close();
         rset = null;
         stmt.close();
         stmt = null;
      } catch (Exception e)
      {
         session.log(e);
         result = null;
      }
      return result;
   }
   
   public abstract void pushPreviews(long ID, String title, String preview, byte[] content, int contentLength);
   
   public void verifyConnection()
   {
      boolean isBad = false;
      try
      {
         if (conn == null || conn.isClosed() == true)
            isBad = true;
      } catch (Exception e)
      {
         isBad = true;
      }
      if (isBad)
         reset();
      if (valid)
      {
         try
         {
            conn.clearWarnings();
         } catch (Exception e)
         {
            close();
         }
      }
   }
   //==============================================================
   public void createIndex()
   {
      String userDir = System.getProperty("user.dir");
      String s = System.getProperty("file.separator");
      try
      {
         processFile(userDir + s + "schema" + s + "index_create.sql");
      } catch (Exception e)
      {
         session.log(e);
      }
      userDir = null;
      s = null;
   }
   public void dropIndex()
   {
      String userDir = System.getProperty("user.dir");
      String s = System.getProperty("file.separator");
      try
      {
         processFile(userDir + s + "schema" + s + "index_drop.sql");
      } catch (Exception e)
      {
         session.log(e);
      }
      userDir = null;
      s = null;
   }
   public void processFile(String myFile) throws Exception
   {
      BufferedReader reader = new BufferedReader(new FileReader(new File(
      myFile)));
      boolean finished = false;
      StringBuffer command = new StringBuffer();
      while (!finished)
      {
         String line = reader.readLine();
         if (line == null)
         {
            if (command.length() > 0)
            {
               String befehl = command.toString().replaceAll(";", "");
               try
               {
                  stmt.executeQuery(befehl);
               } catch (Exception e)
               {
                  session.log(e);
               }
               session.log(befehl, 4);
               befehl = null;
               command = null;
            }
            finished = true;
         } else if (line.trim().length() == 0 || line.startsWith("--")) // Leerzeile
            // oder
            // Kommentar
         {
            if (command.length() > 0)
            {
               String befehl = command.toString().replaceAll(";", "");
               try
               {
                  stmt.executeQuery(befehl);
               } catch (Exception e)
               {
                  session.log(e);
               }
               session.log(befehl, 4);
               befehl = null;
               command = new StringBuffer();
            }
         } else
         {
            line = line.replaceAll("\\$user\\$", session.defaultUser);
            line = line.replaceAll("\\$password\\$",
            session.defaultUserPasswd);
            command.append(line);
            command.append(" ");
         }
      }
   }
   /**
    * Holt die Inhalte zu einem BingoDocument aus der Datenbank
    */
   public synchronized void pullLink(BingoDocument bd) throws Exception
   {
      if (bd == null)
         return;
      long ID = bd.getId();
      url2resolve u = bd.getUrl2r();
      if (u == null || u.getURL() == null)
         return;
      Statement stmt = conn.createStatement();
      ResultSet rset = stmt.executeQuery("SELECT * FROM BingoDocuments WHERE ID=" + ID);
      if (!rset.next())
         return;
      u.setIP(rset.getLong("IP"));
      bd.setLastModified(rset.getDate("LASTMOD").getTime());
      bd.setContentType(rset.getString("MIME"));
      bd.setResponseCode(rset.getInt("RESPONSECODE"));
      bd.setContentEncoding(rset.getString("CONTENTENCODING"));
      bd.setExpiration(rset.getLong("EXPIRATION"));
      bd.setStatus(rset.getString("STATUS").toCharArray()[0]);
      bd.setDepth(rset.getInt("DEPTH"));
      int classID = rset.getInt("CLASS_ID");
      bd.setSvmRank(new Double(rset.getDouble("SVMSCORE")));
      rset.close();
      rset = null;
      rset = stmt.executeQuery("SELECT Name FROM Classes WHERE CLASS_ID = " + classID);
      if (rset.next())
      {
         String classname = rset.getString(1);
         BingoTreeNode root = session.getRootNode();
         if (root == null)
            bd.setNode(null);
         else
         {
            bd.setNode(root.getNode(classname));
         }
      }
      rset.close();
      rset = null;
      rset = stmt.executeQuery("SELECT Title, Preview From ArchiveFiles WHERE ID=" + ID);
      if (rset.next())
      {
         bd.setTitle(rset.getString(1));
         bd.setPreview(rset.getString(2));
      }
      rset.close();
      rset = null;
      stmt.close();
      stmt = null;
      return;
   }
   /** Extrahieren der Preview-, Source- und Title-Informationen aus der DB */
   public abstract BingoDocument pullArchiveFile(long id);
   
   public HashSet getIndicators(int nodeID)
   {
      HashSet result = new HashSet(500);
      try
      {
         PreparedStatement pstmt = conn.prepareStatement("SELECT Name FROM Classes WHERE Class_ID = ?");
         pstmt.setInt(1, nodeID);
         ResultSet rset = pstmt.executeQuery();
         String className = null;
         if (rset.next())
         {
            className = rset.getString(1) + '%';
         }
         rset.close(); rset=null;
         pstmt.close(); pstmt=null;
         
         session.log("Read Indicators for " + className, 1);
         
         if (className == null)
            return result;
         
         pstmt = conn.prepareStatement("SELECT DISTINCT Term FROM ClassFeatures WHERE Class_ID IN (SELECT Class_ID FROM Classes WHERE Name LIKE ?) AND Evaluation > 0");
         pstmt.setString(1, className);
         rset = pstmt.executeQuery();
         while (rset.next())
         {
           result.add(rset.getString(1));
         }
         rset.close(); rset = null;
         pstmt.close(); pstmt = null;
      }
      catch (Exception e)
      {
         session.log(e);
         try
         {rset.close(); rset=null;}
         catch (Exception e1)
         { }
         try
         {pstmt.close(); pstmt=null;}
         catch (Exception e2)
         { }
      }
      session.log("Found Indicators " + result.size(), 1);
      return result;
   }
}
