/*
 * OracleInterface.java
 * 
 * Created on 6. Juni 2003, 10:56
 */
package bingo.db;

import java.io.BufferedInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Blob;

import bingo.util.BingoDocument;
import bingo.util.BingoTreeNode;

/**
 * @author kaster @created 19. August 2003
 */
public class MySQLInterface extends DBInterface {
	/**
	 * Schreibt die aktuelle SVM-Modellhierarchie des Objekts session als
	 * serialisierten BLOB in die Datenbank.
	 * 
	 * @param iteration
	 *            Description of Parameter
	 * @exception Exception
	 *                Description of the Exception
	 * @see session
	 */
	public synchronized void pushModel(int iteration) throws Exception {
		BingoTreeNode rootNode = session.getRootNode();
		if (rootNode == null)
			return;

		int mType = session.getModellingType();

		try {
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			ObjectOutputStream oos = new ObjectOutputStream(baos);

			oos.writeObject(rootNode);
			oos.flush();
			baos.flush();

			long timestamp = System.currentTimeMillis();

			pstmt =
				conn.prepareStatement(
					"INSERT INTO Classifier (Timestamp, Type, Step, Classifier) VALUES ( ?, ?, ?, ?)");
			pstmt.setLong(1, timestamp);
			pstmt.setInt(2, mType);
			pstmt.setInt(3, iteration);
			pstmt.setBytes(4, baos.toByteArray());
			pstmt.executeUpdate();
			pstmt.close();
		} catch (Exception e) {
			session.log(e);
		}
	}

	/**
	 * Liest die SVM-Modellhierarchie mit angegebenem Erstellungsdatum aus der
	 * Datenbank und speichert sie im Objekt session.
	 * 
	 * @param modelDate
	 *            Description of the Parameter
	 * @exception Exception
	 *                Description of the Exception
	 * @see session
	 */
	public synchronized void pullModel(int id) throws Exception {
		//pstmt = conn.prepareStatement( "SELECT Classifier FROM Classifier
		// WHERE Timestamp=to_date( ?, 'YYYY.MM.DD HH24:MI:SS' ) for UPDATE "
		// );
		pstmt =
			conn.prepareStatement(
				"SELECT Classifier, Type FROM Classifier WHERE id = ?");

		pstmt.setInt(1, id);
		//pstmt.setDate(1, myDate);
		rset = pstmt.executeQuery();

		if (!rset.next()) {
			pstmt.close();
			return;
		}

		Blob blob = rset.getBlob(1);
		BufferedInputStream buf =
			new BufferedInputStream(blob.getBinaryStream(), 1024 * 1024);
		ObjectInputStream ois = new ObjectInputStream(buf);
		BingoTreeNode newnode = (BingoTreeNode) ois.readObject();

		if (newnode != null) {
			newnode.restoreSerialization();
			session.setRootNode(newnode);
			session.setModellingType(rset.getInt(2));
		}
		rset.close();
		pstmt.close();
	}

	////////////////////////////////////////////////////////////////////
	// nicht fr MySQL implementiert
	///////////////////////////////////////////////////////////////////

	public synchronized void pushBody(BingoDocument bd) {
		// DB-spezifisch
	}

	public synchronized void pushPreviews(
		long ID,
		String title,
		String preview,
		byte[] content,
		int contentLength) {
		// DB-spezifisch
	}

	public synchronized void pullLatestModel() {
		// DB-spezifisch
	}

	public BingoDocument pullArchiveFile(long id) {
		// DB-spezifisch
		return null;
	}

}
