/*
 * OracleInterface.java
 */
package bingo.db;

import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.sql.PreparedStatement;

import oracle.jdbc.OracleResultSet;
import oracle.sql.BLOB;
import bingo.util.BingoDocument;
import bingo.util.BingoTreeNode;
import bingo.util.SessionBuffer;

/**
 * @author     kaster
 * @created    19. August 2003
 */
public class OracleInterface extends DBInterface {
    
    /**
     * Creates a new instance of OracleInterface
     */
    public OracleInterface() {
        super();
        session = SessionBuffer.getInstance();
    }
    
    /**
     *  Schreibt die aktuelle SVM-Modellhierarchie des Objekts session als
     *  serialisierten BLOB in die Datenbank.
     *
     * @param  iteration      Description of Parameter
     * @exception  Exception  Description of the Exception
     * @see                   session
     */
    
    public synchronized void pushModel( int iteration ) throws Exception {
        BingoTreeNode rootNode = session.getRootNode();
        
        if (rootNode == null )
            return;
        session.log("OracleInterface: prepareSerialization", 7);
        rootNode.prepareSerialization();
        session.log("OracleInterface: Serialization prepared!", 7);
        int mType = session.getModellingType();
        
        try {
            int id = 1;
            rset = stmt.executeQuery("SELECT MAX(ID) FROM CLassifier");
            if (rset.next())
                id = rset.getInt(1) + 1;
            rset.close(); rset = null;
            
            pstmt = conn.prepareStatement( "INSERT INTO Classifier (ID, Timestamp, Type, Step) VALUES (?, sysdate, ?, ? )" );
            pstmt.setInt( 1, id );
            pstmt.setInt( 2, mType );
            pstmt.setInt( 3, iteration );
            
            pstmt.executeUpdate();
            pstmt.close(); pstmt=null;
            
            pstmt = conn.prepareStatement( "SELECT Classifier FROM Classifier WHERE ID = ? for UPDATE " );
            
            pstmt.setInt( 1, id );
            rset = pstmt.executeQuery();
            rset.next();     // soll existieren!!
            oracle.sql.BLOB blob = ( (OracleResultSet) rset ).getBLOB( "CLASSIFIER" );
            OutputStream blobOut = blob.getBinaryOutputStream();
            ObjectOutputStream oos = new ObjectOutputStream( blobOut );
            session.log("OracleInterface: writeObject(rootNode)", 7);
            oos.writeObject( rootNode );
            session.log("OracleInterface: writing done!", 7);
            oos.flush();
            blobOut.flush();
            blobOut.close();blobOut=null;oos=null;
            rset.close(); rset=null;
            pstmt.close(); pstmt=null;
            conn.commit();
        }
        catch ( Exception e ) {
                        session.log(e);
        }
        
        rootNode.restoreSerialization();
    }
    
    
    /**
     *  Liest die SVM-Modellhierarchie mit angegebenem Erstellungsdatum aus der
     *  Datenbank und speichert sie im Objekt session.
     *
     * @param  modelDate      Description of the Parameter
     * @exception  Exception  Description of the Exception
     * @see                   session
     */
    public synchronized void pullModel( int id ) throws Exception {
        //pstmt = conn.prepareStatement( "SELECT Classifier FROM Classifier WHERE Timestamp=to_date( ?, 'YYYY.MM.DD HH24:MI:SS' )" );
        pstmt = conn.prepareStatement( "SELECT Classifier, Type FROM Classifier WHERE id=? " );
        pstmt.setInt( 1, id );
        rset = pstmt.executeQuery();
        
        if ( !rset.next() ) {
            rset.close();
            pstmt.close();
            return;
        }
        
        oracle.sql.BLOB blob = ( (OracleResultSet) rset ).getBLOB( "CLASSIFIER" );
        InputStream blobIn = blob.getBinaryStream();
        ObjectInputStream ois = new ObjectInputStream( blobIn );
        BingoTreeNode newRootNode = (BingoTreeNode) ois.readObject();
        
        if (newRootNode != null) {
            newRootNode.restoreSerialization();
            session.setRootNode( newRootNode );
            session.setModellingType(rset.getInt(2));
        }
        
        ois.close();
        blobIn.close();
        rset.close();
        pstmt.close();
    }
    
    static String noTitle = "NO Title";
    static String noPreview = "NO Preview";
    static String pushPreviewsQuery = "INSERT INTO ArchiveFiles (ID, Title, Preview) VALUES (?,?,?)";
    static String pushBlobQuery = "SELECT Source FROM ArchiveFiles WHERE ID=? FOR UPDATE";
    public synchronized void pushPreviews(long ID, String title, String preview, byte[] content, int contentLength) {
        if (title == null || title.length() <= 0)
            title = noTitle;
        
        if (preview == null || preview.length() <= 0)
            preview = noPreview;
        
        try {
            PreparedStatement pstmt = conn.prepareStatement(pushPreviewsQuery);
            pstmt.setLong(1, ID);
            pstmt.setString(2, title);
            pstmt.setString(3, preview);
            pstmt.execute();
            pstmt.close(); pstmt = null;
            conn.commit();
            
            if (session.storeStreams) {
                if (content != null && content.length>0 && contentLength<=content.length && contentLength>0) {
                    pstmt = conn.prepareStatement(pushBlobQuery);
                    pstmt.setLong(1, ID);
                    OracleResultSet rset = (OracleResultSet)pstmt.executeQuery();
                    rset.next();
                    BLOB blob = rset.getBLOB( 1 );
                    OutputStream out = blob.getBinaryOutputStream(0);
                    out.write(content,0,contentLength);
                    out.flush();
                    out.close();
                    out = null;
                    rset.close(); rset = null;
                    conn.commit();
                    pstmt.close(); pstmt = null;
                }
            }
        }
        catch (Exception e) {
            String message = e.getMessage();
            if (!message.startsWith("ORA-00001")) session.log(e);
            //session.log(e);
        }
    }
    
    public synchronized void pullLatestModel() {
        try {
            rset = stmt.executeQuery( "SELECT Classifier FROM Classifier ORDER BY ID DESC" );
            if ( !rset.next() )
                return;
            
            oracle.sql.BLOB blob = ( (OracleResultSet) rset ).getBLOB( "CLASSIFIER" );
            InputStream blobIn = blob.getBinaryStream();
            ObjectInputStream ois = new ObjectInputStream( blobIn );
            BingoTreeNode node = (BingoTreeNode) ois.readObject();
            if ( node != null ) {
                node.restoreSerialization();
                session.setRootNode(node);
            }
            ois.close();
            blobIn.close();
            rset.close();
            
        }
        catch ( Exception any ) {
            session.log(any);
        }
    }
    
    public synchronized BingoDocument pullArchiveFile(long id) {
        try {
            BingoDocument ret = null;
            pstmt = conn.prepareStatement("SELECT title, preview, source FROM Archivefiles WHERE ID=?");
            pstmt.setLong(1, id);
            
            rset = pstmt.executeQuery();
            
            if ( !rset.next() )
                return ret;
            
            ret = new BingoDocument();
            
            ret.setTitle(rset.getString( "title" ));
            ret.setPreview(rset.getString("preview"));
            
            oracle.sql.BLOB blob = ((OracleResultSet)rset).getBLOB("source");

            byte[] source = blob.getBytes();
            
            if ( source != null && source.length > 0) {
                ret.setBulk(source);
            }
            rset.close();
            
            return ret;
        }
        catch ( Exception any ) {
            session.log(any);
            return null;
        }
    }
    
}
