package bingo.db;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.text.SimpleDateFormat;
import java.util.Locale;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;

import bingo.language.LanguageConstants;
import bingo.svmlight.NodeModel;
import bingo.util.SessionBuffer;
import bingo.util.URLClassTableModel;

/**
 * Dialog zum Einlesen eines hierarchischen SVM-Modells in das Objekt
 * SessionBuffer fr systemweiten Zugriff.
 * 
 * @author T @created 18. Juli 2001
 */
public class ReadModelDialog extends JDialog implements ActionListener {

	private static LanguageConstants lc = LanguageConstants.getInstance();
	private SessionBuffer session = SessionBuffer.getInstance();
	private JButton ok = new JButton(lc.get("load"));
	private JButton remove = new JButton(lc.get("delButton"));
	private JButton cancel = new JButton(lc.get("cancel"));

	private URLClassTableModel tableModel;
	private final JTable table;

	/**
	 * Konstruktor fr das ReadModelDialog Objekt.
	 * 
	 * @param parentFrame
	 *            Der Vater-Frame.
	 */
	public ReadModelDialog(JFrame parentFrame) {
		super(parentFrame, lc.get("rmTitle"), true);
		setSize(400, 280);

		// Aufbau der GUI
		Container content = this.getContentPane();

		String[] columnNames =
			{ "ID", lc.get("rmCol_1"), lc.get("rmCol_2"), lc.get("rmCol_3")};

		tableModel = new URLClassTableModel() {
			public boolean isCellEditable(int x, int y) {
				return false;
			}
		};

		tableModel.setColumnNames(columnNames);
		table = new JTable(tableModel);
		table.setPreferredScrollableViewportSize(new Dimension(400, 240));
		table.getColumnModel().getColumn(2).setPreferredWidth(30);

		table.setBackground(Color.gray);
		table.setForeground(Color.white);

		JScrollPane scrollpane = new JScrollPane(table);
		JPanel panel = new JPanel();

		panel.setLayout(new FlowLayout());
		panel.add(ok);
		panel.add(remove);
		panel.add(cancel);

		content.add(scrollpane, BorderLayout.CENTER);
		content.add(panel, BorderLayout.SOUTH);

		// ActionListener fr alle Mens registrieren
		ok.addActionListener(this);
		remove.addActionListener(this);
		cancel.addActionListener(this);

		setLocationRelativeTo(parentFrame);

		if (session.dbConnect) {
			readTable();
		}
	}

	/**
	 * berwacht den Button.
	 * 
	 * @param e
	 *            Das Event.
	 */
	public void actionPerformed(ActionEvent e) {
		String cmd = e.getActionCommand();

		try {
			if (cmd == (lc.get("load"))) {
				DBInterface dbinterface = DBInterface.getInstance();
				dbinterface.pullModel(
					Integer.parseInt(
						(String) table.getValueAt(table.getSelectedRow(), 0)));
				dbinterface.close();
				dbinterface = null;

				super.dispose();
			}
			if (cmd == (lc.get("delButton"))) {
				removeModels();
			} else if (cmd == (lc.get("cancel"))) {
				super.dispose();
			}
		} catch (Exception any) {
			session.log(any);
		}
	}

	private void readTable() {
		tableModel.removeAllRows();
		try {
			Connection conn = session.getDBConnection();
			Object created = null;
			//Statement stmt = session.getStatement();
			Statement stmt = conn.createStatement();
			ResultSet rset =
				stmt.executeQuery(
					"select id, timestamp, step, type from Classifier order by id desc");

			while (rset.next()) {
				java.sql.Date timestamp = rset.getDate("timestamp");

				created =
					new SimpleDateFormat(
						"yyyy-MM-dd HH:mm:ss",
						Locale.ENGLISH).format(
						timestamp);

				//created = created.substring( 0, created.lastIndexOf( "." )
				// );
				Object[] row =
					{
						String.valueOf(rset.getInt(1)),
						created,
						NodeModel.types[rset.getInt("type")],
						rset.getString("step")};

				tableModel.addRow(row);
			}
			rset.close();
			stmt.close();
			if (tableModel.getRowCount() > 0) {
				table.addRowSelectionInterval(0, 0);
			}
			conn.close();
		} catch (Exception any) {
			session.log(any);
		}

	}

	private void removeModels() {
		PreparedStatement pstmt;
		try {
			Connection conn = session.getDBConnection();
			int[] rows = table.getSelectedRows();

			for (int i = 0; i < rows.length; i++) {
				if (session.dbType.equals("oracle")) {
					pstmt =
						conn.prepareStatement(
							"DELETE FROM CLASSIFIER WHERE id=?");
					String temp = (String) table.getValueAt(rows[i], 0);
					pstmt.setString(1, temp);
				} else {
					/// FIXME!!! An mySQL angleichen
					pstmt =
						conn.prepareStatement(
							"DELETE FROM CLASSIFIER WHERE timestamp=to_date( ?, 'YYYY.MM.DD HH24:MI:SS' ) ");
					pstmt.setLong(
						1,
						((Long) table.getValueAt(rows[i], 0)).longValue());
				}
				pstmt.execute();
				pstmt.close();
			}

			Statement stmt = conn.createStatement();

			stmt.executeQuery("commit");
			stmt.close();
			conn.close();
		} catch (Exception any) {
			session.log(any);
		}
		readTable();
	}

}
