package bingo.google;

import java.io.InputStreamReader;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;

import javax.swing.text.MutableAttributeSet;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.text.html.parser.ParserDelegator;

import bingo.util.SessionBuffer;

/**
 *  Implementierung einer automatischen Google-Related (=hnliche) Anfrage zu
 *  einer vorgegebenen Ziel-URL. Antwort ist ein HTML-Dokument, das geparst wird
 *  und dessen Related-Links extrahiert als Ergebnis zurckgegeben werden
 *
 * @author     T & Sergej Sizov.
 * @created    15. Juni 2001
 */
public class google_similar implements Runnable {

  private Exception exception;
  private String ziel, zahl;
  private boolean istFertig;
  //private URL u;
  private Vector results;


  /**
   *  Konstruktor fr das google_related Objekt
   *
   * @param  adresse  Die Ziel-URL der Related-Anfrage.
   * @param  anzahl   Die Anzahl der Ergebniss.
   */
  public google_similar( String adresse, String anzahl ) {
    istFertig = false;
    zahl = anzahl;
    ziel = "http://www.google.de/search?num=" + zahl + "&hl=en&lr=&safe=off&filter=0&q=related%3A" + adresse;
    results = new Vector();
  }


  /**
   *  Startet den Thread.
   */
  public void run() {
    try {
      //---------------------------------------------------------
      // Parser vorbereiten
      //---------------------------------------------------------
      HTMLEditorKit.ParserCallback callback =
        new HTMLEditorKit.ParserCallback() {
          boolean flag_active;
          boolean flag_category;
          boolean flag_new;
          String act_url;
          String[] ress;


          public void handleText( char[] data, int pos ) {
            if ( flag_active ) {
              String zeile = String.valueOf( data );

              if ( zeile.toLowerCase().indexOf( "translate" ) >= 0 ) {
                // Nicht das Ziel, weil frames - sondern direkt zum translator
                ress[2] = "http://translate.google.com/translate_c?hl=en&sl=de&u=" + ress[0];
              } else if ( zeile.toLowerCase().indexOf( "cached" ) >= 0 ) {
                // Google gibt hier relativen Pfad her!!
                ress[1] = "http://www.google.de" + act_url;
              } else if ( zeile.toLowerCase().indexOf( "category" ) >= 0 ) {
                flag_category = true;
              } else if ( zeile.toLowerCase().indexOf( "similar" ) >= 0 ) {
                flag_new = true;
                results.addElement( ress );
              }
            }
          }


          public void handleComment( char[] data, int pos ) { }


          public void handleStartTag( HTML.Tag t, MutableAttributeSet a, int pos ) {
            try {
              //Parser init..
              if ( t == HTML.Tag.HTML ) {
                flag_active = false;
                flag_category = false;
                flag_new = true;
                act_url = "none";
              }

              //parser aus..
              if ( a.getAttribute( HTML.Attribute.CLASS ) != null ) {
                if ( ( t == HTML.Tag.DIV ) && ( a.getAttribute( HTML.Attribute.CLASS ).equals( "nav" ) ) ) {
                  flag_active = false;
                  istFertig = true;
                }
              }

              if ( t == HTML.Tag.TABLE ) {
                flag_active = false;
              }

              //URL bearbeiten
              if ( flag_active && t == HTML.Tag.A ) {
                act_url = parseHREF( a.getAttribute( HTML.Attribute.HREF ).toString() );
                if ( flag_new ) {
                  ress = new String[4];
                  ress[0] = act_url;
                  flag_new = false;
                } else if ( flag_category ) {
                  ress[3] = act_url;
                  flag_category = false;
                }
              }
            }
            catch ( NullPointerException npE ) {
              SessionBuffer.log(npE);
            }
          }


          public void handleEndTag( HTML.Tag t, int pos ) {
            try {
              //Parser ein..
              if ( t == HTML.Tag.TABLE ) {
                flag_active = true;
              }
              if ( t == HTML.Tag.HTML ) {
                flag_active = false;
                istFertig = true;
              }
            }
            catch ( NullPointerException npE ) {
              SessionBuffer.log(npE);
            }
          }


          public void handleSimpleTag( HTML.Tag t, MutableAttributeSet a, int pos ) { }

        };

      exception = null;

      URL u = new URL( ziel );

      HttpURLConnection uc = (HttpURLConnection) u.openConnection();

      uc.setRequestProperty( "user-agent", "Mozilla/4.7 [de] (WinNT; I)" );
      uc.setRequestMethod( "GET" );
      //uc.setFollowRedirects(false);
      uc.setDoOutput( true );
      uc.setDoInput( true );
      uc.setAllowUserInteraction( false );

      if ( ( uc.getResponseCode() ) == 200 ) {
        InputStreamReader rd = new InputStreamReader( uc.getInputStream() );

        new ParserDelegator().parse( rd, callback, true );
        rd.close();
      }
    }
    catch ( Exception e ) {
      exception = e;
    }
  }


  /**
   *  Liefert den Status der Anfrage.
   *
   * @return    Ist Fertig?
   */
  public boolean getStatus() {
    return istFertig;
  }


  /**
   *  Liefert einen Vector mit den Ergebnis-URLs der Anfrage.
   *
   * @return    Der Vektor.
   */
  public Vector getResults() {
    return results;
  }


  /**
   *  Liefert eine Exception falls etwas schief gegangen ist.
   *
   * @return    Die Exception oder null.
   */
  public Exception getException() {
    return exception;
  }


  private String parseHREF( String s ) {
    int a;
    int b;

    a = s.indexOf( "http://" );
    b = s.lastIndexOf( "&e=" );
    if ( a < 0 ) {
      return null;
    }
    if ( b < 0 ) {
      b = s.length();
    }
    return s.substring( a, b );
  }


  public static void main( String[] argv ) {
    google_similar gs = new google_similar( "http://www.cs.uni-sb.de", "5" );

    gs.run();
    for ( Enumeration e = gs.getResults().elements(); e.hasMoreElements();  ) {
      String[] ress = (String[]) e.nextElement();

      System.out.println( ress[0] );
    }
    if ( gs.getException() != null ) {
      SessionBuffer.log(gs.getException());
    }
  }
}

