package bingo.svmlight;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JProgressBar;
import javax.swing.JSplitPane;
import javax.swing.tree.DefaultMutableTreeNode;

import bingo.util.BingoDocument;
import bingo.crawler.Classifier;
import bingo.util.BINGODesktop;
import bingo.util.NoModelDialog;
import bingo.util.SessionBuffer;

/**
 *  Klasse zur Berechnung der Kategrie von Datenbankeintrgen mit SVMs in einem
 *  eigenen Thread.
 *
 * @author     T
 * @created    15. Juni 2001
 */

public class ClassifierDialog extends JDialog implements SVMTreeListener, ActionListener, Runnable {

  private Classifier classifier;
  private SessionBuffer session = SessionBuffer.getInstance();
  private JButton cancel = new JButton( "Abbruch" );
  private JProgressBar pBar1 = new JProgressBar();
  private JProgressBar pBar2 = new JProgressBar();
  private boolean bBreak = false;
  private int[] rows;
  private Vector vURLs;


  /**
   *  Konstruktor fr das ClassifierDialog Objekt
   *
   * @param  vURLs        Die BaseURLs
   * @param  rows         Die markierten Eintrge.
   * @param  caller       Description of the Parameter
   */
  public ClassifierDialog( Vector vURLs, int[] rows ) {
    super( BINGODesktop.getInstance(), "Klassifiziere Dokument", true );
    this.setSize( 400, 115 );
    this.vURLs = vURLs;
    this.rows = rows;

    Container content = this.getContentPane();

    content.setLayout( new BorderLayout() );

    pBar1.setString( " URL " );
    pBar2.setString( " TOTAL " );
    pBar1.setMaximum( 100 );
    pBar1.setMinimum( 0 );
    pBar2.setMinimum( 0 );
    pBar1.setStringPainted( true );
    pBar2.setStringPainted( true );
    pBar1.setValue( 100 );

    JSplitPane splitPane = new JSplitPane( JSplitPane.VERTICAL_SPLIT );

    splitPane.setDividerSize( 5 );
    splitPane.setContinuousLayout( true );
    splitPane.setTopComponent( pBar1 );
    splitPane.setBottomComponent( pBar2 );
    splitPane.setDividerLocation( 30 );

    content.add( splitPane, BorderLayout.CENTER );
    content.add( cancel, BorderLayout.SOUTH );

    cancel.addActionListener( this );

    setLocationRelativeTo( BINGODesktop.getInstance() );
  }


  /**
   *  berwacht den Button.
   *
   * @param  e  Das Event.
   */
  public void actionPerformed( ActionEvent e ) {
    String cmd = e.getActionCommand();

    if ( cmd.equals( "Abbruch" ) ) {
      bBreak = true;
      super.dispose();
    }
  }


  /**
   *  Thread zur Klassifikation.
   */
  public void run() {
    session.setShowParserDialog( true );
    session.setParserBreak( false );
    pBar2.setMaximum( vURLs.size() );

    for ( int i = 0; i < rows.length; i++ ) {
      BingoDocument bd = (BingoDocument) vURLs.elementAt( i );

      if ( bBreak ) {
        break;
      }

      pBar1.setString( bd.getUrl() );

      try {
        classifier.classify( bd );

        session.log( "\n" + bd.getUrl(), 3 );
        session.log( "--> " + bd.getClassString(), 3 );
        Thread.sleep( 60 );

      }
      catch ( Exception e ) {
        handleException( e );
      }

      pBar2.setValue( i );

      int w = ( new Double( pBar2.getPercentComplete() * 100 ) ).intValue();

      pBar2.setString( w + " % TOTAL " );
      if ( session.bParserBreak ) {
        break;
      }
    }
    super.dispose();
  }


  /**
   *  Interfacemethode zu SVMTreeListener. Dokument wurde Kategorieknoten
   *  zugeordnet.
   *
   * @param  d        Das Klassifikationsergebnis (negativ).
   * @param  n        Der Knoten.
   * @param  comment  Ein Kommentar.
   */
  public void classified( double d, DefaultMutableTreeNode n, String comment ) { }


  /**
   *  Interfacemethode zu SVMTreeListener. Dokument wurde Kategorieknoten nicht
   *  zugeordnet.
   *
   * @param  d        Das Klassifikationsergebnis (negativ).
   * @param  n        Der Knoten.
   * @param  comment  Ein Kommentar.
   */
  public void notClassified( double d, DefaultMutableTreeNode n, String comment ) { }


  /**
   *  Interfacemethode zu SVMTreeListener. Knotenwechsel in der Baumhierarchie.
   *
   * @param  n  Der neue Knoten.
   */
  public void nodeChanged( DefaultMutableTreeNode n ) { }


  /**
   *  Interfacemethode zu SVMTreeListener. Erfolgreiche Klassifikation.
   *
   * @param  svmRanks  Description of the Parameter
   * @param  n         Description of the Parameter
   * @param  comment   Description of the Parameter
   */
  public void finished( Vector svmRanks, DefaultMutableTreeNode n, String comment ) { }


  /**
   *  Interfacemethode zu SVMTreeListener. Behandelt eine Exception, die beim
   *  Klassifizieren geworfen wurde.
   *
   * @param  e  Description of the Parameter
   */
  public void handleException( Exception e ) {
    if ( e.getMessage().indexOf( "Sparse" ) < 0 && session.bShowParserDialog ) {
      bBreak = true;
      super.dispose();
      new NoModelDialog( BINGODesktop.getInstance(), "Exception des SVM-Klassifikators:", e.toString().trim(), null ).setVisible( true );
    }
  }
}

