package bingo.svmlight;

import java.io.Serializable;

import javax.swing.tree.TreeNode;

import org.biojava.stats.svm.SVMClassifierModel;

/**
 *  Datenstruktur fr Paare von Baumknoten und zugehrigem SVM-Modell zusammen mit
 *  einem Modellierungstyp abhngig von der Featuregewichtung nach TF oder TF_IDF.
 *
 * @author     T
 * @created    15. Juni 2001
 * @see        SVMClassifierModel
 */

public class NodeModel implements Serializable {

  public final static int IG = 0;
  public final static int MI = 1;

  /**
   *  TF-Gewichtung (zum Erstellen der SparseVectoren bzw.
   *  der Reprsentation von Features im SVM-Modell).
   */
  public final static int L1TF = 0;

  /**
   *  TF*IDF - Gewichtung.
   */
  public final static int L2TF = 1;

  /**
   *  genormte TF-IDF-Gewichtung.
   */
  public final static int W_TF_IDF = 2;

  /**
   * Meine Vectoren
   *
   * @author    Siers
   */
  public final static int MY_VEC = 3;

  /**
   *  Die Bezeichner fr die einzelnen Gewichtungsmethoden.
   */
  public static String[] types = {"L1(TF)=RTF" , "L2(TF)"};

  private TreeNode node;
  private SVMClassifierModel model;
  private int modellingType;
  private int iteration;


  /**
   *  Konstruiert eine neues Objekt.
   *
   * @param  node     Der Kategorieknoten.
   * @param  model    Der SVM-Klassifikationsmodell.
   * @param  type     Die Gewichtungmethode.
   * @param  itDepth  Die Iterationstiefe.
   */
  public NodeModel( TreeNode node, SVMClassifierModel model, int type, int itDepth ) {
    this.node = node;
    this.model = model;
    this.modellingType = type;
    this.iteration = itDepth;
  }


  /**
   * @return    Die Gewichtungsmethode.
   */
  public int getModellingType() {
    return modellingType;
  }


  /**
   * @return    Die Iterationstiefe, in der dieses Modell berechnet wurde.
   */
  public int getIteration() {
    return iteration;
  }


  /**
   * @return    Der Kategorieknoten.
   */
  public TreeNode getNode() {
    return node;
  }


  /**
   * @return    Das SVM-Klassifikationsmodell.
   */
  public SVMClassifierModel getModel() {
    return model;
  }


  public String toString() {
    String nodeModelName = "";

    nodeModelName += "Knoten : " + node.toString() + "\n";
    nodeModelName += "Modell : " + model.toString() + "\n";
    return nodeModelName;
  }

}
