package bingo.svmlight;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.tree.DefaultMutableTreeNode;

import bingo.language.LanguageConstants;
import bingo.util.BINGODesktop;
import bingo.util.BingoTreeNode;
import bingo.util.NoModelDialog;
import bingo.util.SessionBuffer;

/**
 *  Diese Klasse implementiert das Interface SVMModelListener und stellt eine
 *  graphische Ausgabe fr das Modellierungsmodul dar.
 *
 * @author     T
 * @created    15. Juni 2001
 * @see        SVMModelBuilder
 */
public class SVMModelFrame extends JDialog implements SVMModelListener, ActionListener, Runnable {
    
    private static LanguageConstants lc = LanguageConstants.getInstance();
    private SessionBuffer session = SessionBuffer.getInstance();
    //private Object[] typeDomain = {NodeModel.types[NodeModel.L1TF]};
    
    private JTextArea output = new JTextArea( "" );
    private JScrollPane scrollPane = new JScrollPane( output );
    private JProgressBar pBar1 = new JProgressBar();
    private JComboBox cb = new JComboBox( NodeModel.types );
    private JButton starter = new JButton( lc.get( "start" ) );
    private JButton ok = new JButton( lc.get( "ok" ) );
    private JButton close = new JButton( lc.get( "close" ) );
    private JFrame parentFrame;
    
    private ModelBuilder modelBuilder = null;
    
    private int max = 0;
    
    /**
     *  Konstruktor fr das SVMModelFrame Objekt
     *
     * @param  parentFrame  Description of the Parameter
     */
    public SVMModelFrame( JFrame parentFrame ) {
        super( parentFrame, lc.get( "svmTitle" ), true );
        this.parentFrame = parentFrame;
        this.setSize( 550, 400 );
        
        Container content = this.getContentPane();
        
        content.setLayout( new BorderLayout() );
        
        output.setBackground( Color.gray );
        output.setForeground( Color.white );
        output.setEditable( false );
        
        ok.setEnabled( false );
        
        cb.setBackground( Color.white );
        cb.setSelectedIndex( NodeModel.L1TF );
        
        pBar1.setMinimum( 0 );
        pBar1.setStringPainted( true );
        
        JPanel panel1 = new JPanel();
        
        panel1.setLayout( new BorderLayout() );
        panel1.add( cb, BorderLayout.WEST );
        panel1.add( pBar1, BorderLayout.CENTER );
        
        JPanel panel2 = new JPanel();
        
        panel2.add( starter );
        panel2.add( ok );
        panel2.add( close );
        
        content.add( panel1, BorderLayout.NORTH );
        content.add( scrollPane, BorderLayout.CENTER );
        content.add( panel2, BorderLayout.SOUTH );
        
        starter.addActionListener( this );
        ok.addActionListener( this );
        close.addActionListener( this );
        
        setLocationRelativeTo( BINGODesktop.getInstance() );
        
        BingoTreeNode rootNode = session.getRootNode();
        if (rootNode == null)
            return;
        
        HashMap allClasses = rootNode.allClasses;
        if (allClasses == null)
            return;
        
        max = allClasses.size();
    }
    
    
    /**
     *  berwacht die Buttons.
     *
     * @param  e  Das Event.
     */
    public void actionPerformed( ActionEvent e ) {
        String cmd = e.getActionCommand();
        
        if ( cmd == ( lc.get( "start" ) ) ) {
            new Thread( this ).start();
            starter.setEnabled( false );
        } else if ( cmd == ( lc.get( "ok" ) ) ) {
            super.dispose();
        }
        else if ( cmd == ( lc.get( "close" ) ) ) {
            
            super.dispose();
        }
    }
    
    
    /**
     *  Startet den Thread.
     */
    public void run() {
        BingoTreeNode rootNode = session.getRootNode();
        if (rootNode == null)
            return;
        
        HashMap allClasses = rootNode.allClasses;
        if (allClasses == null)
            return;
        
        pBar1.setMinimum( 0 );
        pBar1.setMaximum( max );
        
        modelBuilder = new SVMModelBuilder( );
        modelBuilder.registerListener( this );
        modelBuilder.buildAllModels();
        
        synchronized (this)
        {
          this.notifyAll();
        }
    }
    
    
    /**
     *  Fortschritt.
     *
     * @param  value  Neuer Wert des JProgressBars.
     * @param  node   Aktueller Kategorieknoten.
     */
    public void barChanged( int value, DefaultMutableTreeNode node ) {
        pBar1.setValue( value );
        pBar1.setString( " " + lc.get( "svmNode" ) + ": " + node );
    }
    
    /**
     *  Interfacemethode zu SVMModelListener. Traingsdaten werden als "positives"
     *  Beispiel dem Modell zugeordnet.
     *
     * @param  sample   Beschreibung.
     * @param  comment  Ein Kommentar.
     */
    public void positiveSample( String sample, String comment ) {
        output.append( "      +1 - " + sample + " [" + comment + "]\n" );
    }
    
    
    /**
     *  Interfacemethode zu SVMModelListener. Traingsdaten werden als "negatives"
     *  Beispiel dem Modell zugeordnet.
     *
     * @param  sample   Beschreibung.
     * @param  comment  Ein Kommentar.
     */
    public void negativeSample( String sample, String comment ) {
        output.append( "      -1 - " + sample + " [" + comment + "]\n" );
    }
    
    
    /**
     *  Interfacemethode zu SVMModelListener. Es wird ein neuer Knoten innerhalb
     *  der Hierarchie betrachtet.
     *
     * @param  n  Der neue Kategorieknoten.
     */
    public void nodeChanged( DefaultMutableTreeNode n ) {
        output.append( lc.get( "svmLookNode" ) + ": <" + n.toString() + "> \n" );
    }
    
    
    /**
     *  Interfacemethode zu SVMModelListener. Modellierung abgeschlossen.
     *
     * @param  comment  Ein Kommentar.
     */
    public void finished( String comment ) {
        close.setEnabled( false );
        ok.setEnabled( true );
        output.append( lc.get( "svmFinish" ) + "\n" );
    }
    
    
    public void warning( String comment ) {
        close.setEnabled( false );
        ok.setEnabled( true );
        output.append( " - Warning: " + comment + "\n" );
    }
    
    public void message( String comment ) {
        output.append( " - " + comment + "\n" );
    }
    
    
    /**
     *  Interfacemethode zu SVMModelListener. Behandelt eine Exception.
     *
     * @param  e  Die Exception.
     */
    public void handleException( Exception e ) {
        this.setVisible( false );
        session.log(e);
        new NoModelDialog( parentFrame, lc.get( "svmModelExc" ) + "!", e.toString(), null ).setVisible( true );
        super.dispose();
    }
}

