/*
 *  BINGODesktop.java
 *
 *  Created on 12. Juli 2002, 11:09
 */
package bingo.util;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.FileDialog;
import java.awt.FlowLayout;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.ArrayList;
import java.util.Iterator;

import javax.swing.JButton;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JDesktopPane;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;

import stored.berechne_df;
import stored.linkanal.WebRanker;
import stored.mi.MiManager;
import bingo.base.BaseFrame;
import bingo.util.BingoDocument;
import bingo.crawler.ControlDialog;
import bingo.crawler.CrawlerFrame;
import bingo.db.DBFrame;
import bingo.db.DBInterface;
import bingo.db.ReadModelDialog;
import bingo.language.LanguageConstants;
import bingo.main.MiniClient;
import bingo.svmlight.SVMModelFrame;

/**
 * @author     martin
 * @created    14. August 2002
 */
public class BINGODesktop extends JFrame implements ActionListener, ItemListener {
    
    public static KeyAdapter keyAdapter =
    new KeyAdapter() {
        public void keyPressed( KeyEvent e ) {
            if ( e.getKeyCode() == KeyEvent.VK_F1 ) {
                common.modelFromBookmarks();
            } else if ( e.getKeyCode() == KeyEvent.VK_F2 ) {
                common.modelCrawlFromDB();
            } else if ( e.getKeyCode() == KeyEvent.VK_F3 ) {
                common.modelCrawlFromDB();
            } else if ( e.getKeyCode() == KeyEvent.VK_F4 ) {
            	new ReadModelDialog( BINGODesktop.getInstance() ).setVisible( true );
            }
        }
    };
    
    private static LanguageConstants lc = LanguageConstants.getInstance();
    private static CommonTasks common = CommonTasks.getInstance();
    
    private SessionBuffer session;
    private JDesktopPane desktopPane;
    
    private JMenuBar menuBar = new JMenuBar();
    
    private JCheckBoxMenuItem miStem = new JCheckBoxMenuItem( lc.get( "bDStemmer" ) );
    private JCheckBoxMenuItem miCut = new JCheckBoxMenuItem( lc.get( "bDStopwords" ) );
    private JCheckBoxMenuItem miSelectionType = new JCheckBoxMenuItem( lc.get( "bDGlobalClassifierOption" ) );
    private JCheckBoxMenuItem miModellingType = new JCheckBoxMenuItem( lc.get( "bDGlobalModelOption" ) );
    public JCheckBoxMenuItem miStored = new JCheckBoxMenuItem( lc.get( "bDStoredProcOption" ), false );
    
    private JMenuItem miReadModelDB;
    private JMenuItem miCleanModel;
    private JMenuItem miWriteModelDB;
    private JMenuItem miReadIDFDB;
    
    private JMenuItem miHITS;
    private JMenuItem miDF;
    private JMenuItem miMI;
    
    private JButton bBatchBM = new JButton( lc.get( "bDBatchBMButton" ) );
    private JButton bBatchDB = new JButton( lc.get( "bDBatchDBButton" ) );
    private JButton bCrawlDB = new JButton( lc.get( "bDCrawlDBButton" ) );
    private JButton bModelFromDB = new JButton ( lc.get( "bDModelFromDBButton" ) );
    
    static int openFrameCount = 0;
    final static int xOffset = 50, yOffset = 50;
    
    protected BINGODesktop() {
        
        session = SessionBuffer.getInstance();
        
        addWindowListener( new CloseWindow( this ) );
        
        setTitle( lc.get( "bDTitle" ) );
        
        // MenLeiste
        setJMenuBar( menuBar );
        
        // Men Datei
        JMenu mFile = new JMenu( lc.get( "bDBINGO" ) );
        menuBar.add( mFile );
        JMenuItem miBatchBookmarks = addMenuItem( mFile, lc.get( "bDBatchBookmarks" ) );
        JMenuItem miBatchDB = addMenuItem( mFile, lc.get( "bDBatchDB" ) );
        JMenuItem miCrawlDB = addMenuItem( mFile, lc.get( "bDCrawlDB" ) );
        mFile.addSeparator();        
        JMenuItem miRetrainDB = addMenuItem( mFile, lc.get( "bDBatchRetrain" ) );
        mFile.addSeparator();
        JMenuItem miExit = addMenuItem( mFile, lc.get( "bDClose" ) );
        
        // Men Base
        JMenu mBase = new JMenu( lc.get( "bDTrainingBase" ) );
        
        menuBar.add( mBase );
        
        JMenuItem miBase = addMenuItem( mBase, lc.get( "bDEditTrainingBase" ) );
        
        // Men Features
        JMenu mFeatures = new JMenu( lc.get( "bDFeatureSpace" ) );
        
        menuBar.add( mFeatures );
        miDF = addMenuItem( mFeatures, lc.get( "bDComputeDF" ) );
        miMI = addMenuItem( mFeatures, lc.get( "bDComputeMI" ) );
        
        // Men SVM
        JMenu mSVM = new JMenu( lc.get( "bDSVM" ) );
        menuBar.add( mSVM );
        JMenuItem miModel = addMenuItem( mSVM, lc.get( "bDSVMModel" ) );
        
        // Men BingoTreeNode
        JMenu mTree = new JMenu( lc.get( "bDTree" ) );
        menuBar.add( mTree );
        
        JMenu readModel = new JMenu( lc.get( "bDTreeRead" ) );
        mTree.add( readModel );
        
        JMenuItem miReadModelFile = addMenuItem( readModel, lc.get( "bDTreeReadFile" ) );
        miReadModelDB = addMenuItem( readModel, lc.get( "bDTreeReadDB" ) );
        
        JMenu writeModel = new JMenu( lc.get( "bDTreeWrite" ) );
        mTree.add( writeModel );
        
        JMenuItem miWriteModelFile = addMenuItem( writeModel, lc.get( "bDTreeWriteFile" ) );
        miWriteModelDB = addMenuItem( writeModel, lc.get( "bDTreeWriteDB" ) );
        
        miCleanModel = addMenuItem( mTree, lc.get( "bDTreeClean" ) );
        JMenuItem miWriteElem = addMenuItem( mTree, lc.get( "bDElemWrite" ) );
        
        // Men Crawler
        JMenu mCrawler = new JMenu( lc.get( "bDCrawler" ) );
        
        menuBar.add( mCrawler );
        
        JMenuItem miCrawler = addMenuItem( mCrawler, lc.get( "bDCrawlerInterface" ) );
        //JMenuItem miRoots = addMenuItem( mCrawler, lc.get( "bDCrawlerRootURLs" ) );
        
        // Men HITS
        JMenu mLinkAnal = new JMenu( lc.get( "bDLinkanalysis" ) );
        
        menuBar.add( mLinkAnal );
        JMenuItem miHITS = addMenuItem( mLinkAnal, lc.get( "bDComputeHITS" ) );
        JMenuItem miPageRank = addMenuItem( mLinkAnal, lc.get( "bDComputePageRank" ) );
        JMenuItem miPageRankZD = addMenuItem( mLinkAnal, lc.get( "bDComputePageRankZD" ) );
        
        // Men DB
        JMenu mDB = new JMenu( lc.get( "bDDatabase" ) );
        
        menuBar.add( mDB );
        
        JMenuItem miDB = addMenuItem( mDB, lc.get( "bDDBInterface" ) );
        
        mDB.addSeparator();
        JMenuItem miClient = addMenuItem( mDB, lc.get( "bDMiniClient" ) );
        mDB.addSeparator();
        JMenuItem miIndexCreate = addMenuItem( mDB, lc.get( "bDIndexCreate" ) );
        JMenuItem miIndexDrop = addMenuItem( mDB, lc.get( "bDIndexDrop" ) );
        
        // Einstellungen
        JMenu mSettings = new JMenu( lc.get( "bDGlobal" ) );
        
        menuBar.add( mSettings );
        
        JMenuItem miSettings = addMenuItem( mSettings, lc.get( "bDCrawlerOptions" ) );
        JMenu mStemmer = new JMenu( lc.get( "bDParser" ) );
        
        miStem.setState( session.bSTEM );
        miCut.setState( session.bCUT );
        mStemmer.add( miStem );
        mStemmer.add( miCut );
        mSettings.add( mStemmer );
        
        // ActionListener fr alle Mens registrieren
        miExit.addActionListener( this );
        miBatchBookmarks.addActionListener( this );
        miBatchDB.addActionListener( this );
        miCrawlDB.addActionListener( this );
        miRetrainDB.addActionListener( this );
        miBase.addActionListener( this );
        miDF.addActionListener( this );
        miMI.addActionListener( this );
        miModel.addActionListener( this );
        miReadModelFile.addActionListener( this );
        miReadModelDB.addActionListener( this );
        miWriteModelFile.addActionListener( this );
        miWriteModelDB.addActionListener( this );
        miCleanModel.addActionListener( this );
        miWriteElem.addActionListener( this );
        miHITS.addActionListener( this );
        miPageRank.addActionListener( this );
        miPageRankZD.addActionListener( this );
        miDB.addActionListener( this );
        miIndexCreate.addActionListener( this );
        miIndexDrop.addActionListener( this );
        miClient.addActionListener( this );
        miCrawler.addActionListener( this );
        //miRoots.addActionListener( this );
        miSettings.addActionListener( this );
        bBatchBM.addActionListener( this );
        bBatchDB.addActionListener( this );
        bCrawlDB.addActionListener( this );
        miStem.addItemListener( this );
        miCut.addItemListener( this );
        miModellingType.addItemListener( this );
        miSelectionType.addItemListener( this );
        miStored.addItemListener( this );
        
        // KeyListener registrieren
        bBatchBM.addKeyListener( keyAdapter );
        bBatchDB.addKeyListener( keyAdapter );
        bCrawlDB.addKeyListener( keyAdapter );
        bModelFromDB.addKeyListener ( keyAdapter );
        
        desktopPane = new JDesktopPane();
        
        JPanel jumpTo = new JPanel();
        
        jumpTo.setLayout( new FlowLayout() );
        jumpTo.add( new JLabel( lc.get( "bDTasks" ) + ":" ) );
        jumpTo.add(bBatchBM);
        jumpTo.add(bBatchDB);
        jumpTo.add(bCrawlDB);
        jumpTo.add(bModelFromDB);
        
        Container container = new Container();
        
        container.setLayout( new BorderLayout() );
        
        container.add( desktopPane, BorderLayout.CENTER );
        container.add( jumpTo, BorderLayout.SOUTH );
        
        setContentPane( container );
        
        desktopPane.setDragMode( JDesktopPane.OUTLINE_DRAG_MODE );
        
        setVisible( true );
        setExtendedState( MAXIMIZED_BOTH );
        setDefaultCloseOperation( EXIT_ON_CLOSE );
        
        itemStateChanged( null );
        
    }
    
    public void add( JInternalFrame frame ) {
        openFrameCount++;
        
        frame.setVisible( true );
        //necessary as of kestrel
        desktopPane.add( frame );
        
        frame.setLocation( xOffset * openFrameCount, yOffset * openFrameCount );
        
        try {
            frame.setSelected( true );
        }
        catch ( java.beans.PropertyVetoException e )
        {}
        
    }
    
    private JMenuItem addMenuItem( JMenu target, String text ) {
        JMenuItem mi = new JMenuItem( text );
        
        target.add( mi );
        return mi;
    }
    
    
    public void actionPerformed( java.awt.event.ActionEvent e ) {
        String cmd = e.getActionCommand();
        
        try {
            if ( cmd == ( lc.get( "bDClose" ) ) ) {
                System.exit( 0 );
            }
            else if ( cmd == ( lc.get( "bDBatchBookmarks" ) ) ||  cmd == ( lc.get( "bDBatchBMButton" ) )) {
                common.modelFromBookmarks();
            }
            else if ( cmd == ( lc.get( "bDBatchDB" ) ) || cmd == ( lc.get("bDBatchDBButton") ) ) {
                common.modelFromDB();
            }
            else if ( cmd == ( lc.get( "bDCrawlDB" ) ) || cmd == ( lc.get( "bDCrawlDBButton" ) ) ) {
                common.modelCrawlFromDB();
            }
            else if ( cmd == ( lc.get( "bDBatchRetrain" ) ) ) {
                common.modelFromRetrain();
            }
            else if ( cmd == lc.get("bDEditTrainingBase")) {
                BaseFrame bf = BaseFrame.getInstance();
                
                bf.setIcon( false );
                bf.setSelected( true );
                bf.setVisible( true );
            } else if ( cmd == ( lc.get( "bDComputeDF" ) ) ) {
                NumberInDialog nid = new NumberInDialog( lc.get( "bDDFTitle" ), session.numDF );
                
                nid.setVisible( true );
                if ( !nid.canceled ) {
                    new berechne_df( nid.getNumber() ).go();
                }
                session.setFeaturesPerClass( nid.getNumber() );
            }
            else if ( cmd == ( lc.get( "bDComputeMI" ) ) ) {
                NumberInDialog nid = new NumberInDialog( lc.get( "bDMITitle" ), session.numMI );
                
                nid.setVisible( true );
                session.setFeaturesPerClass( nid.getNumber() );
                new MiManager( nid.getNumber() ).go();
                
                try
                {new BingoFeatureReader().importClassFeatures();}
                catch (Exception e1)
                {session.log(e1);}
            }
            else if ( cmd == ( lc.get( "bDComputeHITS" ) ) ) {
                NumberInDialog nid = new NumberInDialog( lc.get( "bDHITSTitle" ), session.numHITS );
                nid.setVisible( true );
                if (nid.canceled)
                    return;
                int numIterations =  nid.getNumber() ;
                WebRanker webranker = new WebRanker();
                webranker.createGraph();
                webranker.calculateHITS(numIterations,10);
                webranker = null;
                session.log("Link Analysis done.", 3);
                //   1.Parameter Anzahl Iterationen, 2.Anzahl Threads
            }
            else if (cmd == ( lc.get( "bDComputePageRank" ) )) {
                NumberInDialog nid = new NumberInDialog( lc.get( "bDHITSTitle" ), session.numHITS );
                nid.setVisible( true );
                if (nid.canceled)
                    return;
                int numIterations =  nid.getNumber() ;
                WebRanker webranker = new WebRanker();
                webranker.createGraph();
                webranker.calculatePageRank(numIterations,10);
                webranker = null;
                session.log("Link Analysis done.", 3);
                //   1.Parameter Anzahl Iterationen, 2.Anzahl Threads
            }
            
            else if (cmd == ( lc.get( "bDComputePageRankZD" ) )) {
                NumberInDialog nid = new NumberInDialog( lc.get( "bDHITSTitle" ), session.numHITS );
                nid.setVisible( true );
                if (nid.canceled)
                    return;
                int numIterations =  nid.getNumber() ;
                //            new HitsManager( nid.getNumber() ).go();
                WebRanker webranker = new WebRanker();
                webranker.createGraph_zd();
                webranker.calculatePageRank_zd(numIterations,10);
                webranker = null;
                session.log("Link Analysis done.", 3);
                //   1.Parameter Anzahl Iterationen, 2.Anzahl Threads
            }
            else if ( cmd == ( lc.get( "bDTreeReadFile" ) ) ) {
                FileDialog fileDialog = new FileDialog( BINGODesktop.getInstance(), lc.get( "bDSVMReadTitle" ), FileDialog.LOAD );
                
                fileDialog.setVisible( true );
                if ( fileDialog.getFile() == null )
                    return;
                
                String sFileName = fileDialog.getDirectory() + fileDialog.getFile();
                BingoTreeNode newRootNode = (BingoTreeNode) FileInterface.readZippedObject( sFileName );
                
                if (newRootNode != null) {
                    newRootNode.restoreSerialization();
                    session.setRootNode(newRootNode);
                }
            }
            else if ( cmd == ( lc.get( "bDTreeReadDB" ) ) ) {
                new ReadModelDialog( BINGODesktop.getInstance() ).setVisible( true );
            }
            else if ( cmd == ( lc.get( "bDTreeWriteFile" ) ) ) {
                FileDialog fileDialog = new FileDialog( BINGODesktop.getInstance(), lc.get( "bDSVMWriteTitle" ), FileDialog.SAVE );
                
                fileDialog.setVisible( true );
                if ( fileDialog.getFile() == null )
                    return;
                String sFileName = fileDialog.getDirectory() + fileDialog.getFile();
                
                BingoTreeNode tmp = session.getRootNode();
                if (tmp!=null) {
                    tmp.prepareSerialization();
                    FileInterface.writeZippedObject( tmp, sFileName );
                    tmp.restoreSerialization();
                }
            }
            else if ( cmd == ( lc.get( "bDTreeWriteDB" ) ) ) {
                if (!session.dbConnect)
                    return;
                DBInterface dbi = DBInterface.getInstance();
                if (dbi==null || !dbi.isValid())
                    return;
                dbi.pushModel( 0 );
                
                dbi.close(); dbi=null;
            }
            else if ( cmd == ( lc.get( "bDSVMModel" ) ) ) {
                new SVMModelFrame( BINGODesktop.getInstance() ).setVisible( true );
            }
            else if ( cmd.startsWith( lc.get( "bDDatabase" ) ) ) {
                DBFrame df = DBFrame.getInstance();
                
                df.setIcon( false );
                df.setSelected( true );
                df.setVisible( true );
            }
            else if ( cmd == ( lc.get( "bDTreeClean" ) ) ) {
                session.getRootNode().removeTrainingData();
            }
            else if ( cmd == ( lc.get( "bDElemWrite" ) ) )  // elementweise speichern
            {
                if (!session.dbConnect)
                    return;
                DBInterface dbinterface = DBInterface.getInstance();
                
                dbinterface.pushClasses(session.getRootNode());
                
                ArrayList allDocs = session.getRootNode().getPathDocuments();
                Iterator it = allDocs.iterator();
                while (it.hasNext()) {
                    BingoDocument bd = (BingoDocument)it.next();
                    dbinterface.pushLink(bd);
                    dbinterface.pushNested(bd);
                    bd=null;
                }
                allDocs = null; it = null;
                
                dbinterface.commit();
                dbinterface.close();
                dbinterface = null;
            }
            else if ( cmd == ( lc.get( "bDMiniClient" ) ) ) {
                MiniClient mc = MiniClient.getInstance();
                
                mc.setVisible( true );
            }
            else if ( cmd.startsWith( lc.get( "bDIndexCreate" ) ) ) {
                DBInterface dbinterface = DBInterface.getInstance();
                dbinterface.createIndex();
                dbinterface.close();
                dbinterface = null;
            }
            else if ( cmd.startsWith( lc.get( "bDIndexDrop" ) ) ) {
                DBInterface dbinterface = DBInterface.getInstance();
                dbinterface.dropIndex();
                dbinterface.close();
                dbinterface = null;
            }
            else if ( cmd.startsWith( lc.get( "bDCrawler" ) ) ) {
                CrawlerFrame cf = CrawlerFrame.getInstance();
                
                cf.setIcon( false );
                cf.setSelected( true );
                cf.setVisible( true );
                cf.setVisible( true );
            }
            else if ( cmd == ( lc.get( "bDCrawlerOptions" ) ) ) {
                ControlDialog cd = ControlDialog.getInstance();
                
                cd.setVisible( true );
            }
        }        catch ( Exception any ) {
            session.log(any);
        }
    }
    
    
    public void itemStateChanged( ItemEvent e ) {
        session.setCUT( miCut.getState() );
        session.setSTEM( miStem.getState() );
        session.setUseUnionSelect( miSelectionType.getState() );
        session.setObeyLevel( miModellingType.getState() );
    }
    
    
    private static BINGODesktop desktop = null;
    
    
    public static BINGODesktop getInstance() {
        if ( desktop == null ) {
            desktop = new BINGODesktop();
        }
        return desktop;
    }
    
    
    public static boolean isNull() {
        return desktop == null;
    }

    public void disableDBMenues() {
        if ( !session.dbConnect ) {
            miReadModelDB.setEnabled( false );
            miWriteModelDB.setEnabled( false );
            miReadIDFDB.setEnabled( false );
            miSelectionType.setEnabled( false );
            miHITS.setEnabled( false );
            miDF.setEnabled( false );
            miMI.setEnabled( false );
            miStored.setEnabled( false );
        }
    }
}
