package bingo.util;

import java.sql.ResultSet;
import java.sql.Statement;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map.*;
import java.util.Iterator;
import java.util.Map.Entry;

import bingo.db.DBInterface;

/**
 *  Klasse zum Einlesen aus der Feature-Tabellle in einem eigenen
 *  Thread. Ausserdem wird Term-Mapping (wichtig fuer SVM Light)
 *  neu angelegt.
 */
public class BingoFeatureReader
{
    private SessionBuffer session = null;
    private BingoTreeNode rootNode = null;
    
    private Statement readStmt = null;
    private ResultSet readSet = null;
    private int featuresPerClass = 500;
    
    private DBInterface dbinterface;
    
    private HashMap featureSpace = null;
    private String query = null;
    
    /**
     *  Konstruktor fr das IDFReader Objekt
     *
     * @param  parentFrame  Der Start-Frame.
     * @param  modal
     */
    public BingoFeatureReader()
    {
        session = SessionBuffer.getInstance();
        rootNode = session.getRootNode();
        featuresPerClass = session.getFeaturesPerClass();
        init();
    }
    
    /**
     *  Thread zum Einlesen
     */
    public void init()
    {
        if (rootNode == null)
          return;
        
        try
        {
            if (!session.dbConnect)
                return;
            
            dbinterface = DBInterface.getInstance();
            readStmt = dbinterface.getStatement();
        }
        
        catch (Exception e)
        {session.log(e);}
    }
    
    public void importClassFeatures() throws Exception
    {
        if (dbinterface == null || rootNode == null)
            return;
        
        HashMap allClasses = readAllClasses();
        
        HashSet allTerms = new HashSet (allClasses.size() * featuresPerClass);
        featureSpace = new HashMap (allClasses.size());

// Einlesen der relevanten features        
        Iterator it = allClasses.entrySet().iterator();
        while (it.hasNext())
        {
            Entry entry = (Entry) it.next();
            String className = (String) entry.getKey();
            Integer classID = (Integer) entry.getValue();
            
            if ( !className.startsWith( "ROOT/" ) )
                continue;
            
            query = "select term, MI from ClassFeatures where class_ID = " + classID.toString() + " ORDER BY MI DESC";
            readSet = readStmt.executeQuery( query );
            query = null;
            
            int count = 0;
            HashMap terms = new HashMap(featuresPerClass);

            while ( readSet.next() && count < featuresPerClass )
            {
                String term = readSet.getString(1);
                Double weight = new Double(readSet.getDouble(2));
                terms.put(term, weight);
                count++;
            }
            readSet.close(); readSet = null;
            featureSpace.put(className, terms);
            allTerms.addAll(terms.keySet());
            className = null; classID = null; entry=null;
        }
        it = null;
        readStmt.close(); readStmt = null;
        
// Create String-to-Integer Mapping         
        HashMap mapping = new HashMap (allTerms.size());
        int count = 1;
        it = allTerms.iterator();
        while (it.hasNext())
        {
          mapping.put (it.next(), new Integer(count));
          count++;
        }    
        it=null; allTerms = null;
        rootNode.termMapping = mapping;
        
// Remap features von String to Integer
        
        HashMap features = new HashMap (featureSpace.size());
        it = featureSpace.entrySet().iterator();
        while (it.hasNext())
        {
          Entry entry = (Entry)it.next();  
          String klasse = (String)entry.getKey();
          HashMap terms = (HashMap)entry.getValue();
          HashMap newterms = new HashMap(terms.size());
          Iterator it1 = terms.entrySet().iterator();
          while (it1.hasNext())
          {
            Entry myterm = (Entry) it1.next();
            String term = (String)myterm.getKey();  
            newterms.put(mapping.get(term), myterm.getValue());
            term = null;
            myterm=null;
          }
          features.put(klasse,newterms);
          it1 = null; terms = null; klasse = null; entry = null; newterms = null;
        }    
        it = null;
        featureSpace = null;
        
     // jetzt lokale feature-spaces den Knoten zuordnen

        it = features.entrySet().iterator();
        while (it.hasNext())
        {
          Entry entry = (Entry)it.next();  
          String klasse = (String)entry.getKey();
          HashMap terms = (HashMap)entry.getValue();
          BingoTreeNode node = rootNode.getNode(klasse);
          node.localFeatures = terms;
           
          entry=null;klasse=null;terms=null;node=null;
        }
        it = null;
        features = null;
        
     // jetzt globale feature-spaces (Vereinigung der Spaces der Kinder) berechnen

        rootNode.buildGlobalFeatures();
        
        session.log("Import Class Features: done", 3);
    }
    
    public HashMap readAllClasses() throws Exception
    {
       HashMap allClasses = new HashMap();
       readSet = readStmt.executeQuery("SELECT Name, Class_ID FROM Classes"); 
       while (readSet.next())
       {
         String key = readSet.getString(1);
         Integer value = new Integer (readSet.getInt(2));
         allClasses.put(key, value);
       }    
       readSet.close(); readSet = null; 
       return allClasses;   
    }
}
