package bingo.util;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.Collection;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeNode;

import org.biojava.stats.svm.SVMClassifierModel;
import org.biojava.stats.svm.SparseVector;

import bingo.util.BingoDocument;
import bingo.svmlight.Feature;
import bingo.db.*;

public class BingoTreeNode extends DefaultMutableTreeNode
{
   public String label;
   public int ID;
   public ArrayList documents;
   public ArrayList junk;
   public static HashMap allClasses;
   
   public double avgSVMScore;
   public double minSVMTreshold;
   
   public static HashMap termMapping;
   public HashMap localFeatures;
   public HashSet globalFeatures;
   public SVMClassifierModel classifier;
   
   private HashMap serializedTermMapping = null;
   private HashSet indicators = null;
   //private String serializedUnknownString=null;
   //private BingoTreeNode serializedUnknownNode;
   
   public double getLocalFeatureWeight (Comparable feature)
   {
     if (feature == null || localFeatures == null || localFeatures.size() <= 0) 
        return (double)0.0;
     
     Object weight = localFeatures.get(feature);
     if (weight == null)
        return (double)0.0;
     else
        return ((Double)weight).doubleValue();
   } 
   
   public void prepareSerialization()
   {
      // alles ausser allclasses in lokale instanz uebertragen
      if (termMapping != null)
         serializedTermMapping = new HashMap(termMapping);
      ArrayList docs = getPathDocuments();
      Iterator iter = docs.iterator();
      while(iter.hasNext())
      {
         BingoDocument b = (BingoDocument)iter.next();
         // Collection myLinks = b.getLinks();
         // SessionBuffer.log("--------> " + myLinks.size(), 7);
         b.setUrl2r(null);
      }
   }
   
   public void restoreSerialization()
   {
      termMapping = serializedTermMapping;
      serializedTermMapping = null;
      this.refreshLabel(); // rebuilds allclasses
      
      // Mitgespeicherte Unterklassen re-initialisieren
      //Link link = new Link(null); // und url2resolve auch drin
      //Page page = new Page(link);
      //link = null;
      //page = null;
   }
   
   public BingoTreeNode(Object l)
   {
      super(l);
      documents = new ArrayList();
      junk = new ArrayList();
      refreshLabel();
   }
   
   public void setUserObject(Object obj)
   {
      super.setUserObject(obj);
      refreshLabel();
   }
   
   public void add(BingoTreeNode newChild)
   {
      super.add(newChild);
      newChild.refreshLabel();
   }
   
   public void remove(BingoTreeNode newChild)
   {
      super.remove(newChild);
      newChild.refreshLabel();
   }
   
   public void refreshLabel()
   {
      String oldLabel = label;
      
      TreeNode p = this.getParent();
      StringBuffer temp = new StringBuffer();
      if (p != null)
         temp.append(((BingoTreeNode) p).label);
      temp.append((String) this.getUserObject());
      temp.append('/');
      this.label = temp.toString();
      temp = null;
      p = null;
      
      synchronized (this.getClass())
      {
         if (allClasses == null)
            allClasses = new HashMap();
         if (oldLabel != null)
            allClasses.remove(oldLabel);
         oldLabel = null;
         allClasses.put(label, this);
      }
      
      Enumeration enum = this.children();
      while (enum.hasMoreElements())
         ((BingoTreeNode) enum.nextElement()).refreshLabel();
      enum = null;
      
      this.ID = getID(this);
   }
   
   public void addDocument(Object document)
   {
      documents.add(document);
   }
   
   public void removeDocument(Object document)
   {
      documents.remove(document);
   }
   
   public void addJunk(Object d)
   {
      junk.add(d);
   }
   
   public void removeJunk(Object d)
   {
      junk.remove(d);
   }
   
   public ArrayList getJunk()
   {
      return junk;
   }
   
   public ArrayList getNodeDocuments()
   {
      return documents;
   }
   
   public ArrayList getPathDocuments()
   {
      ArrayList result = new ArrayList();
      Enumeration enum = this.children();
      while (enum.hasMoreElements())
         result.addAll(
         ((BingoTreeNode) enum.nextElement()).getPathDocuments());
      enum = null;
      result.addAll(documents);
      result.addAll(junk);
      return result;
   }
   
   public ArrayList getOppositeDocuments()
   {
      BingoTreeNode p = (BingoTreeNode) this.getParent();
      if (p == null)
         return null;
      ArrayList result = new ArrayList();
      Enumeration enum = ((BingoTreeNode) parent).children();
      while (enum.hasMoreElements())
      {
         BingoTreeNode temp = (BingoTreeNode) enum.nextElement();
         if (temp != this)
            result.addAll(temp.getNodeDocuments());
      }
      enum = null;
      if (p.junk != null)
         result.addAll(p.junk);
      p = null;
      return result;
   }
   
   public ArrayList getOppositePathDocuments()
   {
      BingoTreeNode p = (BingoTreeNode) this.getParent();
      if (p == null)
         return null;
      ArrayList result = new ArrayList();
      Enumeration enum = ((BingoTreeNode) parent).children();
      while (enum.hasMoreElements())
      {
         BingoTreeNode temp = (BingoTreeNode) enum.nextElement();
         if (temp != this)
            result.addAll(temp.getPathDocuments());
      }
      enum = null;
      if (p.junk != null)
         result.addAll(p.junk);
      p = null;
      return result;
   }
   
   public static BingoTreeNode getNode(String path)
   {
      if (allClasses != null)
         return (BingoTreeNode) allClasses.get(path);
      else
         return null;
   }
   
   public static int getID(String who)
   {
      if (who != null)
         return who.hashCode();
      else
         return -1;
   }
   
   public static int getID(BingoTreeNode node)
   {
      if (node != null)
         return node.label.hashCode();
      else
         return -1;
   }
   
   //================================================================
   // Nachvolgende Methoden sind fuer Verwaltung der Feature-Vektoren
   //================================================================
   
   public HashMap getLocalFeatures()
   {
      return localFeatures;
   }
   
   public HashSet getGlobalFeatures()
   {
      return globalFeatures;
   }
   
   /**
    * Generiert ein Feature-Mapping aus String-Features des Dokuments
    */
   public static Object[] documentMapping(Collection featureVector)
   {
      if (featureVector == null) return null;
      LinkedList result = new LinkedList();
      Iterator it = featureVector.iterator();
      while (it.hasNext())
      {
         Feature feature = (Feature) it.next();
         String term = (String) feature.getTerm();
         Integer id = getTermID(term);
         if (id != null)
            result.add(new Feature(id, feature.getWeight()));
         
         feature = null;
         term = null;
         id = null;
      }
      it = null;
      
      Object[] antwort = result.toArray();
      result = null;
      return antwort;
   }
   
   public static Integer getTermID(String term)
   {
      if (termMapping == null)
         return null;
      return (Integer) termMapping.get(term);
   }
   
   /**
    * Generiert aus vorher erzeugtem allgemeinen Mapping den
    * klassenspezifischen, mit angepassten RTF-Werten
    */
   public Object[] classMapping(Object[] featureVector)
   {
      if (globalFeatures == null)
         return null;
      if (featureVector == null)
         return null;
      
      ArrayList usedFeatures = new ArrayList(featureVector.length);
      for (int i = 0; i < featureVector.length; i++)
      {
         Feature feature = (Feature) featureVector[i];
         Integer term = (Integer) feature.getTerm();
         if (globalFeatures.contains(term))
            usedFeatures.add(new Feature(term, feature.getWeight()));
         term = null;
         feature = null;
      }
      
      if (usedFeatures.size() == 0)
         return null;
      
      // L1 Normalize feature vector with regard to used features
      double count = 0.0;
      Iterator it = usedFeatures.iterator();
      while (it.hasNext())
         count += ((Feature) it.next()).getWeight();
      it = null;
      
      if (count <= 0)
         return null;
      
      it = usedFeatures.iterator();
      Feature[] result = new Feature[usedFeatures.size()];
      int pos = 0;
      while (it.hasNext())
      {
         Feature tmp = (Feature) it.next();
         tmp.setTF(tmp.getTF() / count);
         result[pos] = tmp;
         pos++;
         tmp = null;
      }
      it = null;
      usedFeatures = null;
      
      return result;
   }
   
   /**
    * Generiert einen Trainings-Beispiel fuer SVM Light
    */
   public static String getLightString(Object[] featureVector)
   {
      Arrays.sort(featureVector);
      StringBuffer tmp = new StringBuffer();
      for (int i = 0; i < featureVector.length; i++)
      {
         tmp.append(((Feature) featureVector[i]).getTerm().toString());
         tmp.append(":");
         tmp.append(((Feature) featureVector[i]).getWeight());
         tmp.append(" ");
      }
      String result = tmp.toString().trim();
      tmp = null;
      return result;
   }
   
   public void buildGlobalFeatures()
   {
      if (this.isLeaf()) // keine Kinder -> keine globalen Features
         return;
      
      globalFeatures = new HashSet();
      Enumeration enum = this.children();
      while (enum.hasMoreElements())
      {
         BingoTreeNode child = (BingoTreeNode) enum.nextElement();
         child.buildGlobalFeatures();
         if (child.localFeatures != null)
            globalFeatures.addAll(child.localFeatures.keySet());
         // nur features der direkten Nachfolger betrachtet
         child = null;
      }
      enum = null;
   }
   
   // Klassifikation mi Hilfe des knotenspezifischen SVM Classifiers
   
   public double validate(BingoDocument bd)
   {
      // BaseURL Objekte werden nur implements gegebenen Knoten klassifiziert
      // (Classifier des Knotens, Features des Parents)
      
      if (this.isRoot())
         return -1.0;
      if (this.classifier == null)
         return -1.0;
      
      if (bd.getMappedFeatures() == null)
         bd.setMappedFeatures(documentMapping(bd.getFeatureVector()));
      Object[] featuresForClass =
      ((BingoTreeNode) this.getParent()).classMapping(bd.getMappedFeatures());
      SparseVector sparseVector = buildSparseVector(featuresForClass);
      double result = -1.0;
      
      if (sparseVector != null && sparseVector.size() > 0)
      {
         synchronized (classifier)
         {
            result = classifier.classify(sparseVector);
         }
      }
      
      sparseVector = null;
      featuresForClass = null;
      bd.setMappedFeatures(null);
      return result;
   }
   
   public void classify(BingoDocument bd)
   {
      // Achtung: die Klassifikation erfolgt nur zwischen den Kindern des
      // gegebenen Knotens.
      
      if (bd == null)
         return;
      if (this.isLeaf())
         // in Leaf-Knoten wird Klassifikation nicht fortgesetzt
      {
         SessionBuffer.log("ALARM: Leaf-Node darf nicht klassifizieren", 3);
         bd.setSvmRank(new Double(-1.0));
         bd.setNode((BingoTreeNode)this.getRoot());
         bd = null;
         return;
      }
      
      if (this.isRoot()) // Im Root wird die Klassifikation initialisiert
      {
         bd.setSvmRank(new Double(-1.0));
         bd.setNode(this);
      }
      
      if (bd.getMappedFeatures() == null)
         bd.setMappedFeatures(documentMapping(bd.getFeatureVector()));
      Object[] featuresForClass = this.classMapping(bd.getMappedFeatures());
      SparseVector sparseVector = buildSparseVector(featuresForClass);
      double bestScore = -1;
      BingoTreeNode bestNode = null;
      
      if (sparseVector != null
      && sparseVector.size() >= 3) // min Anzahl Features,
         // Tuning-Parameter
      {
         Enumeration enum = this.children();
         while (enum.hasMoreElements())
         {
            BingoTreeNode child = (BingoTreeNode) enum.nextElement();
            if (child.classifier == null)
               continue;
            
            boolean indicatorsOK = true;
            if (SessionBuffer.getUseIndicators())
            {
              indicatorsOK = child.checkIndicators (bd.getFeatureVector());
            }   
            if (!indicatorsOK)
               continue;
            
            double result = -1.0;
            synchronized (child.classifier)
            {
               result = child.classifier.classify(sparseVector);
            }
            
            if (result > 0)
               result =
               (result + child.minSVMTreshold) / child.avgSVMScore;
            
            if (result > bestScore)
            {
               bestScore = result;
               bestNode = child;
            }
         }
         enum = null;
      }
      
      if (bestScore > 0)
      {
         bd.setNode(bestNode);
         bd.setSvmRank(new Double(bestScore));
         
         if (bestNode.isLeaf())
            bd.setMappedFeatures(null);
         else
            bestNode.classify(bd);
      } else
         bd.setMappedFeatures(null);
      
      bestNode = null;
      bd = null;
   }
   
   public boolean checkIndicators (Collection featureVector)
   {
     if (featureVector == null || featureVector.size() == 0)
        return false;      
     if (this.indicators == null || this.indicators.size() == 0)
        return true;
     Iterator it = featureVector.iterator();
     int count = 0;
     while (it.hasNext() && count < SessionBuffer.getNumIndicators())
     {
       Feature feature = (Feature)it.next();
       String term = (String)feature.term;
       if (this.indicators.contains(term))
          count++;
     }   
     return (count >= SessionBuffer.getNumIndicators());
   }
   
   public SparseVector buildSparseVector(Object[] features)
   {
      if (features == null)
         return null;
      SparseVector sparseVector = new SparseVector(features.length);
      for (int i = 0; i < features.length; i++)
      {
         Feature f = (Feature) features[i];
         int pos = ((Integer) f.getTerm()).intValue();
         double weight = f.getWeight();
         sparseVector.put(pos, weight);
         f = null;
      }
      return sparseVector;
   }
   
   public void removeTrainingData()
   {
      this.localFeatures = null;
      this.documents = new ArrayList(0);
      this.junk = new ArrayList(0);
      Enumeration enum = this.children();
      while (enum.hasMoreElements())
         ((BingoTreeNode) enum.nextElement()).removeTrainingData();
      enum = null;
      
      if (this.isRoot()) // Am Ende Garbage Collector ausfuehren
         System.gc();
   }
   
   public void calcAvgSVMScore()
   {
      avgSVMScore = 1.0;
      
      if (documents != null)
      {
         Iterator it = documents.iterator();
         double count = 0.0;
         while (it.hasNext())
         {
            BingoDocument tmp = (BingoDocument) it.next();
            if (tmp.getSvmRank().doubleValue() > 0)
            {
               avgSVMScore += tmp.getSvmRank().doubleValue();
               count += 1.0;
            }
            tmp = null;
         }
         it = null;
         if (count > 0.0)
            avgSVMScore = avgSVMScore / count;
         
      }
      minSVMTreshold = avgSVMScore * 0.05;
      SessionBuffer.log("Treshold " + label + " : " + minSVMTreshold, 3);
      if (classifier != null)
         classifier.setThreshold(minSVMTreshold);
      
      Enumeration enum = this.children();
      while (enum.hasMoreElements())
      {
         BingoTreeNode node = (BingoTreeNode) enum.nextElement();
         node.calcAvgSVMScore();
         node = null;
      }
      enum = null;
   }
   
   public void loadIndicators()
   {
      if (!SessionBuffer.dbConnect)
         return;
      DBInterface dbi = DBInterface.getInstance();
      indicators = dbi.getIndicators(this.ID);
      dbi = null;

      Enumeration enum = this.children();
      while (enum.hasMoreElements())
      {
         BingoTreeNode node = (BingoTreeNode) enum.nextElement();
         node.loadIndicators();
         node = null;
      }
      enum = null;
      return;
   }
}