package bingo.util;

import java.io.BufferedReader;
import java.io.FileReader;
import java.net.URL;

import javax.swing.text.MutableAttributeSet;
import javax.swing.text.html.HTML;
import javax.swing.text.html.HTMLEditorKit;
import javax.swing.text.html.parser.ParserDelegator;

import bingo.crawler.handler.UrlVerifier;
import bingo.crawler.frontier.url2resolve;
/**
 *  Die Klasse dient dem Generieren von Trainingsdaten durch Parsen von
 *  Netscape-Bookmarkdateien durch Verwendung der Klasse
 *  HTMLEditorKit.ParserCallback. Die spezielle Struktur dieser Dokumente
 *  erlaubt das Extrahieren der darin enhaltenen Links und Erzeugung einer
 *  Baumhierarchie.
 */
public class BookmarkParser
{
   private BingoTreeNode rootNode,currentNode;
   private String sURL;
   public static String rootString = "ROOT";
   public boolean isOthers = false;
   
   /**
    *  Initialisiert diesen BookmarkParser mit der URL der zu parsenden
    *  Bookmarkdatei.
    *
    * @param  sURL  Description of Parameter
    */
   public BookmarkParser( String sURL )
   {
      this.sURL = sURL;
      isOthers = false;
   }
   
   
   /**
    *  Die Bookmarkdatei wird geparst und ein Hierarchiebaum erzeugt. Die
    *  enthaltenen Links werden mit entsprechendem Kategorieeintrag versehen als
    *  Vector von Trainingsdaten zurckgegeben.
    *
    * @return    Root-Knoten der Themenhierarchie
    */
   public BingoTreeNode parseBookmarks()
   {
      HTMLEditorKit.ParserCallback callback =
      new HTMLEditorKit.ParserCallback()
      {
         
         boolean bStart = false;
         
         /**
          *  Description of the Field
          */
         public HTML.Tag lastTag = HTML.Tag.BR;  // Dummy-Platzhalter
         
         
         /**
          *  Description of the Method
          *
          * @param  data  Description of Parameter
          * @param  pos   Description of Parameter
          */
         public void handleText( char[] data, int pos )
         {
            if ( !bStart )
               return;
            if ( lastTag == HTML.Tag.H3 )
            {
               String label = String.valueOf( data );
               if (label.toLowerCase().trim().startsWith("others"))  // ein String fuer alle others classen
               {
                  if (currentNode != null)
                     isOthers = true;
               }
               else
               {
                  isOthers = false;
                  BingoTreeNode temp = new BingoTreeNode( label );
                  
                  if (rootNode == null)
                     rootNode = temp;
                  if (currentNode != null)
                     currentNode.add(temp);
                  currentNode = temp;
               }
            }
            else if ( lastTag == HTML.Tag.A )
            {
               BingoDocument bd = UrlVerifier.string2BingoDocument(sURL);
            
               if (bd == null)
                  return; // da gabs dann wohl nen Fehler
               
               bd.setNode(currentNode);
               
               if (isOthers)
               {
                  bd.setStatus('J');
                  currentNode.addJunk( bd );
               }
               else
               {
                  bd.setStatus('B');
                  currentNode.addDocument( bd );
               }
               bd = null;
            }
         }
         
         public void handleStartTag( HTML.Tag t, MutableAttributeSet a, int pos )
         {
            lastTag = t;
            if ( t == HTML.Tag.TITLE )
            {
               BingoTreeNode temp = new BingoTreeNode( rootString );
               rootNode = temp;
               currentNode = temp;
            }
            else if ( t == HTML.Tag.A)
               sURL = a.getAttribute(HTML.Attribute.HREF).toString();
         }
         
         
         public void handleEndTag( HTML.Tag t, int pos )
         {
            if ( t == HTML.Tag.DL )   // Knoten abgeschlossen
            {
               if (isOthers)
                  isOthers = false;
               else if (currentNode != null)
                  currentNode = (BingoTreeNode)currentNode.getParent();
            }
            else if ( t == HTML.Tag.TITLE )
               bStart = true;
         }
         
         
         public void handleSimpleTag( HTML.Tag t, MutableAttributeSet a, int pos )
         { }
      };
      
      try
      {
         BufferedReader f = new BufferedReader(new FileReader( sURL ));
         new ParserDelegator().parse( f, callback, true );
         f.close(); f=null;
         
      }
      catch ( Exception e )
      {
         SessionBuffer.log(e);
      }
      
      currentNode = null; sURL = null;
      
      return rootNode;
   }
}

