package bingo.util;
import java.awt.FileDialog;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Vector;
import java.util.HashMap;
import java.util.Map.*;
import java.net.URL;

import stored.berechne_df;
import stored.mi.MiManager;
import bingo.base.FeatureDialog;
import bingo.crawler.BINGOCrawler;
import bingo.crawler.handler.UrlVerifier;
import bingo.db.DBInterface;
import bingo.crawler.frontier.url2resolve;
import bingo.language.LanguageConstants;
import bingo.svmlight.SVMModelFrame;
public class CommonTasks
{
   private static SessionBuffer session;
   private static LanguageConstants lc;
   private static CommonTasks me;
   private final static String classLinksQuery = "SELECT L.Target, b.depth, b.SvmScore, "
   + "L.Link_ID "
   + "FROM Links L, bingodocuments b "
   + "WHERE b.ID = L.ID AND b.Class_ID = ? "
   + "ORDER BY b.SvmScore DESC";
   private final static String bingoDocsIDQuery = "SELECT ID FROM bingodocuments ORDER BY ID ASC";
   private final static String countBingoDocsQuery = "SELECT COUNT(*) FROM bingodocuments";
   private CommonTasks()
   {
      session = SessionBuffer.getInstance();
      lc = LanguageConstants.getInstance();
   }
   public static CommonTasks getInstance()
   {
      if (me == null)
         me = new CommonTasks();
      return me;
   }
   public void modelFromDB()
   {
      boolean ok = true;
      if (ok)
         ok = pullLatestModel();
      else
         return;
      if (ok)
         ok = initCrawlerFromDB();
      else
         return;
   }
   public void modelCrawlFromDB()
   {
      boolean ok = true;
      if (ok)
         ok = pullLatestModel();
      else
         return;
      if (ok)
         ok = initCrawlerFromDB();
      else
         return;
      if (ok)
         ok = readCrawlLinksFromDB();
   }
   public boolean initCrawlerFromDB()
   {
      BINGOCrawler crawler = session.getCrawler();
      if (crawler == null)
         return false;
      crawler.reset();
      DBInterface dbinterface = DBInterface.getInstance();
      if (!dbinterface.isValid())
         return false;
      
      long[] visited = dbinterface.getVisitedLinks();
      if (visited == null || visited.length <= 0)
         return true;
      for (int i = 0; i < visited.length; i++)
         crawler.queue.addVisitedPath(visited[i]);
      session.log(visited.length + " URL locks put to crawler", 3);
      visited = null;
      
      long[][] parsed = dbinterface.getParsedLinks();
      if (parsed == null || parsed.length <= 0)
         return true;
      for (int i = 0; i < parsed.length; i++)
         crawler.queue.addAlreadyParsed(parsed[i][0], parsed[i][1]);
      session.log(crawler.queue.parsedPath.size() + " content locks put to crawler", 3);
      visited = null;
      
      dbinterface.close();
      dbinterface = null;
      crawler = null;
      return true;
   }
   public boolean readCrawlLinksFromDB()
   {
      int linksProKlasse = 500;
      BINGOCrawler crawler = session.getCrawler();
      if (crawler == null)
         return false;
      DBInterface dbinterface = DBInterface.getInstance();
      if (!dbinterface.isValid())
         return false;
      BingoTreeNode rootNode = session.getRootNode();
      if (rootNode == null)
         return false;
      NumberInDialog nid = new NumberInDialog(lc.get("bDLinksPerClass"),
      session.numDF);
      nid.setVisible(true);
      if (nid.canceled)
         return false;
      else
         linksProKlasse = nid.getNumber();
      nid = null;
      try
      {
         Connection conn = dbinterface.conn;
         Statement stmt = conn.createStatement();
         ResultSet rset = stmt.executeQuery(countBingoDocsQuery);
         rset.next();
         int numCrawlDocs = rset.getInt(1);
         rset.close();
         rset = null;
         long[] visited = new long[numCrawlDocs];
         rset = stmt.executeQuery(bingoDocsIDQuery);
         int i = 0;
         while (rset.next())
         {
            visited[i] = rset.getLong(1);
            i++;
         }
         rset.close();
         rset = null;
         stmt.close();
         stmt = null;
         session.log("Visited URLs: " + i, 3);
         PreparedStatement pstmt = conn.prepareStatement(classLinksQuery);
         Iterator it = rootNode.allClasses.values().iterator();
         while (it.hasNext())
         {
            String[] linkStrings = new String[linksProKlasse];
            int[] linkDepths = new int[linksProKlasse];
            double[] linkPrios = new double[linksProKlasse];
            int count = 0;
            BingoTreeNode node = (BingoTreeNode) it.next();
            int nodeID = node.ID;
            pstmt.setInt(1, nodeID);
            rset = pstmt.executeQuery();
            while (rset.next() && count < linksProKlasse)
            {
               int pos = Arrays.binarySearch(visited, rset.getLong(4));
               if (pos >= 0)
                  continue;
               linkStrings[count] = rset.getString(1);
               linkDepths[count] = rset.getInt(2);
               linkPrios[count] = rset.getDouble(3);
               count++;
            }
            rset.close();
            rset = null;
            session.log("Links: " + count + " -> " + node.label, 3);
            for (int j = 0; j < count; j++)
            {
               URL tempURL = UrlVerifier.checkAbsoluteLink(linkStrings[j]);
               
               if (tempURL != null)
               {
                  url2resolve tempU = UrlVerifier.verifyURL(tempURL);
                  tempURL = null;
                  
                  if (tempU != null && ! tempU.isBad())
                  {
                     tempU.setPrio((float) 2.0);
                     crawler.addLink(tempU);
                  }
                  tempU = null;
               }
               tempURL = null;
            }
            
            linkStrings = null;
            linkDepths = null;
            linkPrios = null;
            node = null;
         }
         pstmt.close();
         pstmt = null;
         dbinterface.close();
         conn = null;
         dbinterface = null;
         visited = null;
      } catch (Exception e)
      {
         session.log(e);
         return false;
      }
      if (dbinterface != null)
      {
         dbinterface.close();
         dbinterface = null;
      }
      crawler = null;
      rootNode = null;
      return true;
   }
   public boolean modelFromBookmarks()
   {
      boolean ok = true;
      if (ok)
         ok = importBookmarks();
      else
         return ok;
      
      if (ok)
         ok = crawlBookmarks();
      else
         return ok;
      
      if (ok)
         ok = elementsToDB();
      else
         return ok;
      
      if (ok)
         ok = indexCreate();
      else
         return ok;
      if (ok)
         ok = buildModel();
      else
         return ok;
      
      if (ok)
         ok = initCrawlerFromDB();
      else
         return ok;
      
      if (ok)
         ok = indexDrop();
      else
         return ok;
      
      return ok;
   }
   
   public boolean modelFromRetrain()
   {
      
      boolean ok = true;
      
      if (ok)
         ok = indexCreate();
      else
         return ok;
      
      if (ok)
         ok = elementsFromDB();
      else
         return ok;
      
      if (ok)
         ok = buildModel();
      else return ok;
      
      if (ok)
         ok = initCrawlerFromDB();
      else return ok;
      
      if (ok)
         ok = indexDrop();
      else
         return ok;
      
      return ok;
   }
   
   public boolean buildModel()
   {
      boolean ok = true;
      
      if (ok)
         ok = berechneDF();
      else
         return ok;
      if (ok)
         ok = berechneMI();
      else
         return ok;
      if (ok)
         ok = berechneSVM();
      else
         return ok;
      if (ok)
         ok = pushModel();
      else
         return ok;
      
      return ok;
   }
   
   public static synchronized boolean elementsFromDB()
   {
      SessionBuffer session = SessionBuffer.getInstance();
      boolean ok = false;
      
      if (session == null && !session.dbConnect)
      {
         session.log("DB Connection missing", 1);
         return ok;
      }
      
      DBInterface dbi = DBInterface.getInstance();
      if (dbi == null || !dbi.isValid())
      {
         session.log("Invalid DB Connection", 1);
         return ok;
      }
      
      BingoTreeNode root = session.getRootNode();
      if (root == null || root.getChildCount() <= 0)
      {
         dbi.pullLatestModel();
         root = session.getRootNode();
         if (root == null)
         {
            session.log("BINGO root node missing", 1);
            return ok;
         }
      }
      
      HashMap allClasses = root.allClasses;
      if (allClasses == null || allClasses.size() < 1)
      {
         session.log("Topic list in BingoTreeNode missing", 1);
         return ok;
      }
      
      root.removeTrainingData();
      
      Iterator it = allClasses.entrySet().iterator();
      while (it.hasNext())
      {
         Entry entry = (Entry)it.next();
         
         String className = (String)entry.getKey();
         if (className == null)
            continue;
         
         BingoTreeNode node = (BingoTreeNode)entry.getValue();
         if (node == null)
            continue;
         
         session.log("Process Topic: " + className + " (" + node.ID + ")", 10);
         ArrayList train = dbi.pullTrainingDocuments(node);
         ArrayList junk = dbi.pullJunkDocuments(node);
         
         if (train == null)
            node.documents = new ArrayList(0);
         else
            node.documents = train;
         
         if (junk == null)
            node.junk = new ArrayList(0);
         else
            node.junk = junk;
         
         session.log("POS: " + node.documents.size() + " + Junk: " + node.junk.size(), 50);
         
         train=null;
         junk=null;
         node=null;
         entry=null;
      }
      it=null;
      allClasses=null;
      
      session=null;
      dbi.close();
      dbi=null;
      
      ok = true;
      return ok;
   }
   
   public boolean importBookmarks()
   {
      // Bookmark-File importieren
      session.log("Lese Bookmarks ein...", 4);
      FileDialog fileDialog = new FileDialog(BINGODesktop.getInstance(), lc.get("openBookmarks"), FileDialog.LOAD);
      fileDialog.setVisible(true);
      if (fileDialog.getFile() == null)
         return false;
      String file = fileDialog.getDirectory() + fileDialog.getFile();
      
      fileDialog = null;
      
                /*String file = System.getProperty("user.dir") + "\\bingo\\data\\bookmarks.html";
                System.out.println("Achtung: Lese nur " + file + " ein!!!");*/
      BookmarkParser bp = new BookmarkParser(file);
      BingoTreeNode newRootNode = bp.parseBookmarks();
      file = null;
      bp = null;
      if (newRootNode == null)
         return false;
      session.setRootNode(newRootNode);
      
      session.log("Bookmarks eingelesen!", 4);
      return true;
   }
   public boolean crawlBookmarks()
   {
      ArrayList a = session.getRootNode().getPathDocuments();
                /*
                 * System.out.println("###################################################");
                 * System.out.println("a.size(): " + a.size()); for (int i=0; i
                 * <a.size(); i++){ System.out.println("--begin--"); BingoDocument b =
                 * (BingoDocument)a.get(i); System.out.println("URL: " + b.getUrl());
                 * System.out.println("Depth: " + b.getDepth());
                 * System.out.println("ClassifiedNode: " + b.getClassifiedNode());
                 * System.out.println("RootNode: " + b.getRootNode());
                 * System.out.println("Node.getLeafCount(): " +
                 * b.getNode().getLeafCount()); System.out.println("--end--");
                 */
      FeatureDialog fd = new FeatureDialog(BINGODesktop.getInstance(),new Vector(a));
      fd.setVisible(true);
      return true;
   }
   public boolean elementsToDB()
   {
      if (!session.dbConnect)
         return false;
      if (session == null || session.getRootNode() == null)
         return false;
      DBInterface dbinterface = DBInterface.getInstance();
      if (!dbinterface.isValid())
         return false;
      dbinterface.pushClasses(session.getRootNode());
      ArrayList allDocs = session.getRootNode().getPathDocuments();
      Iterator it = allDocs.iterator();
      while (it.hasNext())
      {
         BingoDocument bd = (BingoDocument) it.next();
         //dbinterface.pushTrainingDocument(bd);
         dbinterface.pushLink(bd);
         dbinterface.pushNested(bd);
         bd = null;
      }
      allDocs = null;
      it = null;
      dbinterface.commit();
      dbinterface.close();
      dbinterface = null;
      return true;
   }
   public boolean berechneDF()
   {
      NumberInDialog nid = new NumberInDialog(lc.get("bDDFTitle"),
      session.numDF);
      nid.setVisible(true);
      if (nid.canceled)
         return false;
      else
         new berechne_df(nid.getNumber()).go();
      session.setFeaturesPerClass(nid.getNumber());
      return true;
   }
   public boolean berechneMI()
   {
      NumberInDialog nid = new NumberInDialog(lc.get("bDMITitle"),
      session.numMI);
      nid.setVisible(true);
      session.setFeaturesPerClass(nid.getNumber());
      new MiManager(nid.getNumber()).go();
      try
      {
         new BingoFeatureReader().importClassFeatures();
      } catch (Exception e1)
      {
         session.log(e1);
         return false;
      }
      return true;
   }
   public boolean berechneSVM()
   {
      new SVMModelFrame(BINGODesktop.getInstance()).setVisible(true);
      return true;
   }
   public boolean indexCreate()
   {
      DBInterface dbinterface = DBInterface.getInstance();
      if (dbinterface == null || !dbinterface.isValid())
         return false;
      dbinterface.createIndex();
      dbinterface.close();
      dbinterface = null;
      return true;
   }
   public boolean indexDrop()
   {
      DBInterface dbinterface = DBInterface.getInstance();
      if (dbinterface == null || !dbinterface.isValid())
         return false;
      dbinterface.dropIndex();
      dbinterface.close();
      dbinterface = null;
      return true;
   }
   public boolean pushModel()
   {
      DBInterface dbinterface = DBInterface.getInstance();
      if (dbinterface == null || !dbinterface.isValid())
         return false;
      try
      {
         dbinterface.pushModel(0);
      } catch (Exception e)
      {
         session.log(e);
         return false;
      }
      dbinterface.commit();
      dbinterface.close();
      dbinterface = null;
      return true;
   }
   public boolean pullLatestModel()
   {
      DBInterface dbinterface = DBInterface.getInstance();
      if (dbinterface == null || !dbinterface.isValid())
         return false;
      dbinterface.pullLatestModel();
      dbinterface.close();
      dbinterface = null;
      if (session.getRootNode() == null)
         return false;
      return true;
   }
}
