package bingo.util;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Enumeration;
import java.util.Vector;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;
import java.util.zip.ZipOutputStream;


/**
 *  Die Klasse dient dem direkten Lesen bzw. Schreiben von serialisierbaren
 *  Objekten aus bzw. in eine Datei als ObjectIn/OutputStream. Es bestehen
 *  entsprechende Methoden zur Bearbeitung einzelner Objekte sowie fr Vektoren
 *  von Objekten.
 *
 * @author     T
 * @created    15. Juni 2001
 */
public class FileInterface
{
   
   /**
    *  Liest ein serialisiertes Objekt aus einem File.
    *
    * @param  file  Zu lesendes File.
    * @return       gelesenes Object.
    */
   public static Object readObject( String file )
   {
      Object object = null;
      
      try
      {
         FileInputStream fis = new FileInputStream( file );
         ObjectInputStream ois = new ObjectInputStream( fis );
         
         object = ois.readObject();
         ois.close();
         fis.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
      return object;
   }
   
   /**
    *  Liest ein serialisiertes Objekt aus einem Zip-File.
    *
    * @param  file  Zu lesendes File.
    * @return       gelesenes Object.
    */
   public static Object readZippedObject( String file )
   {
      Object object = null;
      
      try
      {
         FileInputStream fis = new FileInputStream( file );
         ZipInputStream zis = new ZipInputStream( fis );
         
         zis.getNextEntry();
         
         ObjectInputStream ois = new ObjectInputStream( zis );
         
         object = ois.readObject();
         
         zis.closeEntry();
         zis.close();
         ois.close();
         fis.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
      return object;
   }
   
   public static Object readZippedObjectFromStream( InputStream is )
   {
      Object object = null;
      
      try
      {
         ZipInputStream zis = new ZipInputStream( is );
         
         zis.getNextEntry();
         
         ObjectInputStream ois = new ObjectInputStream( zis );
         
         object = ois.readObject();
         
         zis.closeEntry();
         zis.close();
         ois.close();
         is.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
      return object;
   }
   /**
    *  Schreibt ein einzelnes Objekt in ein File.
    *
    * @param  object  Zu schreibendes Object.
    * @param  file    File, in das geschrieben werden soll.
    */
   public static void writeObject( Object object, String file )
   {
      try
      {
         FileOutputStream fos = new FileOutputStream( file );
         ObjectOutputStream oos = new ObjectOutputStream( fos );
         
         oos.writeObject( object );
         oos.flush();
         oos.close();
         fos.flush();
         fos.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
   }
   
   
   /**
    *  Schreibt ein einzelnes Objekt in ein Zip-File.
    *
    * @param  object  Zu schreibendes Object.
    * @param  file    File, in das geschrieben werden soll.
    */
   public static void writeZippedObject( Object object, String file )
   {
      try
      {
         FileOutputStream fos = new FileOutputStream( file );
         ZipOutputStream zos = new ZipOutputStream( fos );
         
         zos.putNextEntry( new ZipEntry( file ) );
         
         ObjectOutputStream oos = new ObjectOutputStream( zos );
         
         oos.writeObject( object );
         oos.flush();
         
         zos.closeEntry();
         zos.finish();
         zos.flush();
         zos.close();
         fos.flush();
         fos.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
   }
   
   
   /**
    *  Liest einen Vector von serialisiertes Objekten aus einem File. Vermeidet
    *  eine Casting-Exception bei Vektoren von spezialisierten Objekten.
    *
    * @param  file  Zu lesendes File.
    * @return       Gelesener Vector.
    */
   public static Vector readVector( String file )
   {
      Integer SIZE;
      Object object;
      Vector vector = new Vector();
      
      try
      {
         FileInputStream fis = new FileInputStream( file );
         ObjectInputStream ois = new ObjectInputStream( fis );
         
         SIZE = (Integer) ois.readObject();
         
         for ( int i = 0; i < SIZE.intValue(); i++ )
         {
            object = ois.readObject();
            vector.addElement( object );
         }
         
         ois.close();
         fis.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
      return vector;
   }
   
   
   /**
    *  Liest einen Vector als Folge von serialisierten Objekten aus einem
    *  Zip-File. Vermeidet eine Casting-Exception bei Vektoren von
    *  spezialisierten Objekten.
    *
    * @param  file  Zu lesendes File.
    * @return       Gelesener Vector.
    */
   public static Vector readZippedVector( String file )
   {
      Integer SIZE;
      Object object;
      Vector vector = new Vector();
      
      try
      {
         FileInputStream fis = new FileInputStream( file );
         ZipInputStream zis = new ZipInputStream( fis );
         
         zis.getNextEntry();
         
         ObjectInputStream ois = new ObjectInputStream( zis );
         
         SIZE = (Integer) ois.readObject();
         
         for ( int i = 0; i < SIZE.intValue(); i++ )
         {
            object = ois.readObject();
            vector.addElement( object );
         }
         
         ois.close();
         zis.close();
         fis.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
      return vector;
   }
   
   
   public static Vector readZippedVectorFromInputStream( InputStream in )
   {
      Integer SIZE;
      Object object;
      Vector vector = new Vector();
      
      try
      {
         ZipInputStream zis = new ZipInputStream( in );
         
         zis.getNextEntry();
         
         ObjectInputStream ois = new ObjectInputStream( zis );
         
         SIZE = (Integer) ois.readObject();
         
         for ( int i = 0; i < SIZE.intValue(); i++ )
         {
            object = ois.readObject();
            vector.addElement( object );
         }
         
         ois.close();
         zis.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
      return vector;
   }
   
   
   /**
    *  Schreibt einen Vector von Objekten in ein File. Vermeidet eine
    *  Casting-Exception bei Vektoren von spezialisierten Objekten.
    *
    * @param  vector  Zu schreibender Vector.
    * @param  file    File, in das geschrieben werden soll.
    */
   public static void writeVector( Vector vector, String file )
   {
      Integer SIZE = new Integer( vector.size() );
      
      try
      {
         FileOutputStream fos = new FileOutputStream( file );
         ObjectOutputStream oos = new ObjectOutputStream( fos );
         
         oos.writeObject( SIZE );
         
         for ( Enumeration enum = vector.elements(); enum.hasMoreElements();  )
         {
            oos.writeObject( enum.nextElement() );
         }
         
         oos.flush();
         oos.close();
         fos.flush();
         fos.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
   }
   
   
   /**
    *  Schreibt einen Vector von Objekten in ein Zip-File. Vermeidet eine
    *  Casting-Exception bei Vektoren von spezialisierten Objekten.
    *
    * @param  vector  Zu schreibender Vector.
    * @param  file    File, in das geschrieben werden soll.
    */
   public static void writeZippedVector( Vector vector, String file )
   {
      Integer SIZE = new Integer( vector.size() );
      
      try
      {
         FileOutputStream fos = new FileOutputStream( file );
         ZipOutputStream zos = new ZipOutputStream( fos );
         
         zos.putNextEntry( new ZipEntry( file ) );
         
         ObjectOutputStream oos = new ObjectOutputStream( zos );
         
         oos.writeObject( SIZE );
         
         for ( Enumeration enum = vector.elements(); enum.hasMoreElements();  )
         {
            oos.writeObject( enum.nextElement() );
         }
         oos.flush();
         
         zos.closeEntry();
         zos.finish();
         zos.flush();
         zos.close();
         fos.flush();
         fos.close();
      }
      catch ( Exception any )
      {
         SessionBuffer.log(any);
      }
   }
}

