package bingo.util;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.FileDialog;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JTextField;

import bingo.crawler.URL_Crawler;
import bingo.db.ReadModelDialog;

/**
 *  Dialog zur Ausgabe einer Fehlermeldung whrend des Klassifizierens bzw.
 *  Modellierens.
 *
 * @author     T
 * @created    15. Juni 2001
 */
public class NoModelDialog extends JDialog implements ActionListener
{
  
  private SessionBuffer session = SessionBuffer.getInstance();
  private Object[] domain_cb1 =
  {"Abbruch", "Lese IDF/Feature-Tabelle", "Lese SVM-Modell aus Datei", "Lese SVM-Modell aus DB"};
  private JComboBox cb1 = new JComboBox( domain_cb1 );
  private JButton ok = new JButton( "Ok" );
  private JTextField label = new JTextField();
  private JFrame parentFrame;
  private String errorMsg;
  /**
   *  Konstruktor fr das NoModelDialog Objekt
   *
   * @param  parentFrame  Der Start-Frame.
   * @param  title        Der Titel.
   * @param  errorMsg     Die Fehlermeldung.
   * @param  crawler      Description of the Parameter
   */
  public NoModelDialog( JFrame parentFrame, String title, String errorMsg, URL_Crawler crawler )
  {
    super( parentFrame, title, true );
    this.parentFrame = parentFrame;
    this.errorMsg = errorMsg;
    //this.crawler = crawler;
    this.setSize( 400, 115 );
    
    Container content = this.getContentPane();
    
    content.setLayout( new BorderLayout() );
    
    cb1.setBackground( Color.white );
    
    label.setText( errorMsg );
    label.setFont( new Font( "Serif", Font.BOLD, 15 ) );
    label.setBackground( Color.white );
    label.setForeground( Color.red );
    label.setEditable( false );
    
    JPanel panel1 = new JPanel();
    JPanel panel2 = new JPanel();
    
    panel1.setLayout( new FlowLayout() );
    //panel1.add( pBar );
    
    panel2.setLayout( new FlowLayout() );
    panel2.add( cb1 );
    panel2.add( ok );
    
    content.add( label, BorderLayout.CENTER );
    content.add( panel2, BorderLayout.SOUTH );
    
    if ( errorMsg.indexOf( "SVM" ) > -1 )
    {
      cb1.setSelectedIndex( 2 );
    } else if ( errorMsg.indexOf( "IDF" ) > -1 )
    {
      cb1.setSelectedIndex( 1 );
    } else
    {
      cb1.setSelectedIndex( 0 );
    }
    
    ok.addActionListener( this );
    
    setLocationRelativeTo( parentFrame );
    
    if ( crawler != null )
    {
      crawler.stop();
    }
    
    session.bShowParserDialog = false;
  }
  
  
  /**
   *  berwacht den Button.
   *
   * @param  e  Das Event.
   */
  public void actionPerformed( ActionEvent e )
  {
    String cmd = e.getActionCommand();
    
    if ( cmd.equals( "Ok" ) )
    {
      if ( cb1.getSelectedIndex() == 0 )
      {
        super.dispose();
      } 
      /*
      else if ( cb1.getSelectedIndex() == 1 )
      {
      
      }*/
      else if ( cb1.getSelectedIndex() == 2 )
      {
        try
        {
          FileDialog fileDialog = new FileDialog( parentFrame, "Modelldatei lesen", FileDialog.LOAD );
          
          fileDialog.setVisible( true );
          if ( fileDialog.getFile() == null )
          {
            return;
          }
          String sFileName = fileDialog.getDirectory() + fileDialog.getFile();
          BingoTreeNode node = (BingoTreeNode)FileInterface.readZippedObject( sFileName );
           
          if ( node != null )
          {
            session.setRootNode( node );
          } 
          else
          {
            throw new NullPointerException( "Keine gltige Modelldatei!" );
          }
        }
        catch ( Exception any )
        {
          this.errorMsg = any.toString().trim();
          label.setText( errorMsg );
          return;
        }
        super.dispose();
      } else if ( cb1.getSelectedIndex() == 3 )
      {
        new ReadModelDialog( parentFrame ).setVisible( true );
        super.dispose();
      }
    }
  }
  
}

