package bingo.util;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JTextField;

import bingo.base.FeatureDialog;

/**
 *  Die Klasse dient dazu, eine Fehlermeldung anzuzeigen, falls beim Parsen von
 *  Dokumenten etwas schief luft.
 *
 * @author     T
 * @created    15. Juni 2001
 */
public class ParserDialog extends JDialog implements ActionListener {

  private SessionBuffer session = SessionBuffer.getInstance();
  private JButton ok = new JButton( "OK" );
  private JButton ok_all = new JButton( "OK (alle)" );
  private JButton cancel = new JButton( "Abbruch" );
  private JTextField label1 = new JTextField();
  private JTextField label2 = new JTextField();
  private JPanel panel1 = new JPanel();


  /**
   *  Konstruktor fr das ParserDialog Objekt
   *
   * @param  parentFrame  Der Start-Dialog.
   * @param  title        Der Titel.
   * @param  modal
   * @param  sURL         Die BaseURL.
   * @param  errorMsg     Die Fehlermeldung.
   */
  public ParserDialog( JDialog parentFrame, String title, boolean modal, String sURL, String errorMsg ) {
    super( parentFrame, title, modal );
    buildUp( sURL, errorMsg );
    setLocationRelativeTo( parentFrame );
  }


  /**
   *  Konstruktor fr das ParserDialog Objekt
   *
   * @param  parentFrame  Der Start-Frame.
   * @param  title        Der Titel.
   * @param  modal
   * @param  sURL         Die BaseURL.
   * @param  errorMsg     Die Fehlermeldung.
   */
  public ParserDialog( JFrame parentFrame, String title, boolean modal, String sURL, String errorMsg ) {
    super( parentFrame, title, modal );
    buildUp( sURL, errorMsg );
    setLocationRelativeTo( parentFrame );
  }


  private void buildUp( String sURL, String errorMsg ) {
    setSize( 450, 130 );

    Container content = this.getContentPane();

    content.setLayout( new BorderLayout() );

    label1.setText( sURL );
    label1.setFont( new Font( "Serif", Font.ITALIC, 15 ) );
    label1.setBackground( Color.white );
    label1.setForeground( Color.black );
    label1.setEditable( false );

    label2.setText( errorMsg );
    label2.setFont( new Font( "Serif", Font.BOLD, 15 ) );
    label2.setBackground( Color.white );
    label2.setForeground( Color.red );
    label2.setEditable( false );

    panel1.setLayout( new FlowLayout() );
    panel1.add( ok );
    panel1.add( ok_all );
    panel1.add( cancel );

    content.add( label1, BorderLayout.NORTH );
    content.add( label2, BorderLayout.CENTER );
    content.add( panel1, BorderLayout.SOUTH );

    ok.addActionListener( this );
    ok_all.addActionListener( this );
    cancel.addActionListener( this );
  }


  /**
   *  berwacht die Buttons.
   *
   * @param  e  Das Event.
   */
  public void actionPerformed( ActionEvent e ) {
    String cmd = e.getActionCommand();

    if ( cmd.equals( "OK" ) ) {
      super.dispose();
    } else if ( cmd.equals( "OK (alle)" ) ) {
      session.setShowParserDialog( false );
      FeatureDialog.closeDialogs();
      super.dispose();
    } else if ( cmd.equals( "Abbruch" ) ) {
      session.setParserBreak( true );
      super.dispose();
    }
  }
}

