package bingo.util;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.*;

import javax.swing.JButton;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import bingo.crawler.frontier.*;
import bingo.language.LanguageConstants;
import bingo.crawler.handler.parser.TextItem;

/**
 *  Die Klasse dient dem Parsen von HTML-Dokumenten unter Verwendung des
 *  HTMLEditorKit.ParserCallback. Optional werden die Stammformen gebildet bzw.
 *  Stoppwrter gestrichen. Diese Klasse hat eine optionale graphische Ausgabe.
 *
 * @author     T
 * @created    15. Juni 2001
 */
public class ParserFrame extends JInternalFrame implements ActionListener, Runnable
{
   
   private static SessionBuffer session = SessionBuffer.getInstance();
   private static LanguageConstants lc = LanguageConstants.getInstance();
   
   private BingoDocument bd;
   private JScrollPane parserPane;
   private JScrollPane previewPane;
   private JScrollPane sysinfoPane;
   private JButton ok = new JButton( "OK" );
   private JButton browser = new JButton( lc.get( "browserButton" ) );
   private JTextArea parserTextArea = new JTextArea();
   private JTextArea previewTextArea = new JTextArea();
   private JTextArea sysinfoTextArea = new JTextArea();
   private JLabel label1 = new JLabel( " Preview " );
   private JLabel label2 = new JLabel( " Parser Ansicht " );
   private JLabel label3 = new JLabel( " Sysinfo " );
   
   public ParserFrame( BingoDocument bd )
   {
      super( "Quellansicht",
      true,
      //resizable
      true,
      //closable
      true,
      //maximizable
      true );
      //iconifiable
      
      this.bd = bd;
      
      BINGODesktop.getInstance().add( this );
      setTitle( "BingoDocument: " + bd.getUrl() );
      setSize( 700, 600 );
      setVisible( true );
      
      Container content = getContentPane();
      
      content.setLayout( new BorderLayout() );
      
      parserTextArea.setEditable( false );
      parserTextArea.setLineWrap( true );
      previewTextArea.setEditable( false );
      previewTextArea.setLineWrap( true );
      sysinfoTextArea.setEditable( false );
      sysinfoTextArea.setLineWrap( true );
      
      parserPane = new JScrollPane( parserTextArea );
      previewPane = new JScrollPane( previewTextArea );
      sysinfoPane = new JScrollPane( sysinfoTextArea );
      
      JSplitPane splitPane = new JSplitPane( JSplitPane.VERTICAL_SPLIT );
      JSplitPane megasplitPane = new JSplitPane( JSplitPane.VERTICAL_SPLIT );
      
      splitPane.setDividerSize( 2 );
      splitPane.setContinuousLayout( true );
      splitPane.setTopComponent( previewPane );
      splitPane.setBottomComponent( parserPane );
      splitPane.setDividerLocation( 200 );
      
      megasplitPane.setDividerSize( 4 );
      megasplitPane.setContinuousLayout( true );
      megasplitPane.setDividerLocation( 260 );
      
      label1.setFont( new Font( "Serif", Font.ITALIC, 18 ) );
      label2.setFont( new Font( "Serif", Font.ITALIC, 18 ) );
      
      megasplitPane.setTopComponent( sysinfoPane );
      megasplitPane.setBottomComponent( splitPane );
      
      content.add( megasplitPane, BorderLayout.CENTER );
      content.add( ok, BorderLayout.SOUTH );
      content.add( browser, BorderLayout.NORTH );
      
      setContentPane( content );
      ok.addActionListener( this );
      browser.addActionListener( this );
      
      new Thread( this ).start();
   }
   
   
   public void run()
   {
      displayURL( bd );
   }
   
   
   private void displayURL( BingoDocument b )
   {
      try
      {
         if (b == null)
            return;
         
         url2resolve u = b.getUrl2r();
         
         StringBuffer tempBuffer = new StringBuffer(1000);
         
         tempBuffer.append("BingoDocument summary: \n\n");
         
         tempBuffer.append("URL: ");
         
         if (u != null && u.getURL() != null)
         {
            tempBuffer.append(u.getURL().toString());
         }
         else
         {
            if (b.getUrl() != null)
               tempBuffer.append(b.getUrl());
            else
               tempBuffer.append("unknown");
         }
         
         tempBuffer.append("\n");
         
         tempBuffer.append("ID: ");
         tempBuffer.append(b.getId());
         tempBuffer.append(",  ");
         
         tempBuffer.append("Parent ID: ");
         tempBuffer.append(b.getParentID());
         tempBuffer.append(",  ");
         
         tempBuffer.append("Response code: ");
         tempBuffer.append(b.getResponseCode());
         tempBuffer.append(",  ");
         
         tempBuffer.append("Depth: ");
         tempBuffer.append(b.getDepth());
         tempBuffer.append(",  ");
         
         tempBuffer.append("Lastmod: ");
         tempBuffer.append(new Date(b.getLastModified()).toString());
         
         tempBuffer.append("\n");
         
         tempBuffer.append("BingoTreeNode: ");
         if (b.getNode() == null)
            tempBuffer.append("NULL");
         else
            tempBuffer.append(b.getNode().label);
         tempBuffer.append(",  ");
         
         tempBuffer.append("Class distance: ");
         tempBuffer.append(b.getClassDistance());
         tempBuffer.append(",  ");
         
         tempBuffer.append("\n");
         
         tempBuffer.append("Terms: ");
         if (b.getText() != null)
            tempBuffer.append(b.getText().size());
         else
            tempBuffer.append("none");
         
         tempBuffer.append(",  ");
         
         tempBuffer.append("Features: ");
         if (b.getFeatureVector() != null)
            tempBuffer.append(b.getFeatureVector().size());
         else
            tempBuffer.append("none");
         
         tempBuffer.append(",  ");
         
         tempBuffer.append("Links: ");
         if (b.getLinks() != null)
            tempBuffer.append(b.getLinks().size());
         else
            tempBuffer.append("none");
         
         tempBuffer.append("\n");
         
         tempBuffer.append("SVM: ");
         if (b.getSvmRank() != null)
            tempBuffer.append(b.getSvmRank().doubleValue());
         else
            tempBuffer.append("unknown");
         
         tempBuffer.append("\n");
         
         tempBuffer.append("Status: ");
         tempBuffer.append(b.getStatus());
         
         tempBuffer.append("\n");
         
         tempBuffer.append("Content Type: ");
         if (b.getContentType() == null)
            tempBuffer.append("unknown");
         else
            tempBuffer.append(b.getContentType());
         tempBuffer.append(",  ");
         
         tempBuffer.append("Encoding: ");
         if (b.getContentEncoding() == null)
            tempBuffer.append("unknown");
         else
            tempBuffer.append(b.getContentEncoding());
         tempBuffer.append(",  ");
         
         tempBuffer.append("Content length: ");
         tempBuffer.append(b.getContentLength());
         tempBuffer.append(",  ");
         
         tempBuffer.append("Bulk size: ");
         tempBuffer.append(b.getBulkSize());
         tempBuffer.append("\n");
         
         tempBuffer.append("Signatures: \n");
         tempBuffer.append("MD5 of URL: ");
         tempBuffer.append(b.getId());
         tempBuffer.append("\n");
         tempBuffer.append("MD5 of Content: ");
         tempBuffer.append(b.getMD5());
         
         tempBuffer.append("\n\n");
         
         tempBuffer.append("Url2Resolve summary: \n\n");
         
         if (u == null)
         {
            tempBuffer.append(" - is NULL \n");
         }
         else
         {
            tempBuffer.append("Valid: ");
            tempBuffer.append(u.isBad());
            tempBuffer.append("\n");
            
            tempBuffer.append("IP: ");
            if (u.IP != null)
               tempBuffer.append(u.IP.toString());
            else
               tempBuffer.append("unknown");
            tempBuffer.append(",   ");
            
            tempBuffer.append("Host: ");
            if (u.getHost() != null)
               tempBuffer.append(u.getHost().toString());
            else
               tempBuffer.append("unknown");
            tempBuffer.append(",   ");
            
            tempBuffer.append("Port: ");
            tempBuffer.append(u.getPort());
            tempBuffer.append("\n");
            
            tempBuffer.append("Prio: ");
            tempBuffer.append(u.getPrio());
            tempBuffer.append("\n");
         }
         u = null;
         
         sysinfoTextArea.setText(tempBuffer.toString());
         tempBuffer = null;
         
         if (b.getPreview() != null)
         {
            StringBuffer sb = new StringBuffer(500);
            sb.append("Title:\n\n");
            sb.append(b.getTitle());
            sb.append("\n\nPreview:\n\n");
            sb.append(b.getPreview());
            previewTextArea.setText( sb.toString() );
            sb = null;
         }
         else
            previewTextArea.setText( "No preview" );
         
         if (b.getText() == null || b.getText().size() <= 0)
            parserTextArea.setText("No Features");
         else
         {
            StringBuffer buffer = new StringBuffer(500);
            Iterator it = b.getText().iterator();
            while ( it.hasNext() && buffer.length() <= 500)
            {
               buffer.append( ( (TextItem) it.next()).stem + " - " );
            }
            it=null;
            
            parserTextArea.setText(buffer.toString());
            buffer = null;
         }
         
      }
      catch ( Exception e )
      {
         session.log(e);
      }
   }
   
   
   /**
    *  berwacht den Button.
    *
    * @param  e  Das Event.
    */
   public void actionPerformed( ActionEvent e )
   {
      String cmd = e.getActionCommand();
      
      if ( cmd.equals( "OK" ) )
      {
         dispose();
      }
      else if ( cmd == ( lc.get( "browserButton" ) ) && bd != null)
      {
         String target = null;
         if (bd.getUrl2r() != null && bd.getUrl2r().getURL() != null)
         {
            target = bd.getUrl2r().getURL().toString();
         }
         else
         {
            target = "http://" + bd.getUrl();
         }
         
         if (target != null)
         try
         {
            bingo.util.BrowserLauncher.openURL(target);
         }
         catch (Exception ble)
         {  }
         
         target = null;
      }
      
   }
   
}

