/*
 Copyright (C) 2002 MySQL AB
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */
/**
 * MysqlDefs contains many values that are needed for communication
 * with the MySQL server.
 *
 *
 * @author     Mark Matthews <mmatthew@worldserver.com>
 * @version    $Id: MysqlDefs.java,v 1.1.1.1 2004/01/12 13:31:05 kaster Exp $
 */
package com.mysql.jdbc;

import java.sql.Types;

final class MysqlDefs {
  //
  // Constants defined from mysql
  //
  // DB Operations
  final static int SLEEP = 0;
  final static int QUIT = 1;
  final static int INIT_DB = 2;
  final static int QUERY = 3;
  final static int FIELD_LIST = 4;
  final static int CREATE_DB = 5;
  final static int DROP_DB = 6;
  final static int RELOAD = 7;
  final static int SHUTDOWN = 8;
  final static int STATISTICS = 9;
  final static int PROCESS_INFO = 10;
  final static int CONNECT = 11;
  final static int PROCESS_KILL = 12;
  final static int DEBUG = 13;
  final static int PING = 14;
  final static int TIME = 15;
  final static int DELAYED_INSERT = 16;
  final static int CHANGE_USER = 17;
  final static int COM_BINLOG_DUMP = 18;
  final static int COM_TABLE_DUMP = 19;
  final static int COM_CONNECT_OUT = 20;
  final static int COM_REGISTER_SLAVE = 21;
  final static int COM_PREPARE = 22;
  final static int COM_EXECUTE = 23;
  final static int COM_LONG_DATA = 24;

  // Data Types
  final static int FIELD_TYPE_DECIMAL = 0;
  final static int FIELD_TYPE_TINY = 1;
  final static int FIELD_TYPE_SHORT = 2;
  final static int FIELD_TYPE_LONG = 3;
  final static int FIELD_TYPE_FLOAT = 4;
  final static int FIELD_TYPE_DOUBLE = 5;
  final static int FIELD_TYPE_NULL = 6;
  final static int FIELD_TYPE_TIMESTAMP = 7;
  final static int FIELD_TYPE_LONGLONG = 8;
  final static int FIELD_TYPE_INT24 = 9;
  final static int FIELD_TYPE_DATE = 10;
  final static int FIELD_TYPE_TIME = 11;
  final static int FIELD_TYPE_DATETIME = 12;

  // Newer data types
  final static int FIELD_TYPE_YEAR = 13;
  final static int FIELD_TYPE_NEWDATE = 14;
  final static int FIELD_TYPE_ENUM = 247;
  final static int FIELD_TYPE_SET = 248;

  // Older data types
  final static int FIELD_TYPE_TINY_BLOB = 249;
  final static int FIELD_TYPE_MEDIUM_BLOB = 250;
  final static int FIELD_TYPE_LONG_BLOB = 251;
  final static int FIELD_TYPE_BLOB = 252;
  final static int FIELD_TYPE_VAR_STRING = 253;
  final static int FIELD_TYPE_STRING = 254;

  // Limitations
  final static int MAX_ROWS = 50000000;// From the MySQL FAQ


  static int mysqlToJavaType( int mysql_type ) {

    int sql_type;

    switch ( mysql_type ) {

        case MysqlDefs.FIELD_TYPE_DECIMAL:
          sql_type = Types.DECIMAL;

          break;
        case MysqlDefs.FIELD_TYPE_TINY:
          sql_type = Types.TINYINT;

          break;
        case MysqlDefs.FIELD_TYPE_SHORT:
          sql_type = Types.SMALLINT;

          break;
        case MysqlDefs.FIELD_TYPE_LONG:
          sql_type = Types.INTEGER;

          break;
        case MysqlDefs.FIELD_TYPE_FLOAT:
          sql_type = Types.FLOAT;

          break;
        case MysqlDefs.FIELD_TYPE_DOUBLE:
          sql_type = Types.DOUBLE;

          break;
        case MysqlDefs.FIELD_TYPE_NULL:
          sql_type = Types.NULL;

          break;
        case MysqlDefs.FIELD_TYPE_TIMESTAMP:
          sql_type = Types.TIMESTAMP;

          break;
        case MysqlDefs.FIELD_TYPE_LONGLONG:
          sql_type = Types.BIGINT;

          break;
        case MysqlDefs.FIELD_TYPE_INT24:
          sql_type = Types.INTEGER;

          break;
        case MysqlDefs.FIELD_TYPE_DATE:
          sql_type = Types.DATE;

          break;
        case MysqlDefs.FIELD_TYPE_TIME:
          sql_type = Types.TIME;

          break;
        case MysqlDefs.FIELD_TYPE_DATETIME:
          sql_type = Types.TIMESTAMP;

          break;
        case MysqlDefs.FIELD_TYPE_YEAR:
          sql_type = Types.DATE;

          break;
        case MysqlDefs.FIELD_TYPE_NEWDATE:
          sql_type = Types.DATE;

          break;
        case MysqlDefs.FIELD_TYPE_ENUM:
          sql_type = Types.CHAR;

          break;
        case MysqlDefs.FIELD_TYPE_SET:
          sql_type = Types.CHAR;

          break;
        case MysqlDefs.FIELD_TYPE_TINY_BLOB:
          sql_type = Types.VARBINARY;

          break;
        case MysqlDefs.FIELD_TYPE_MEDIUM_BLOB:
          sql_type = Types.LONGVARBINARY;

          break;
        case MysqlDefs.FIELD_TYPE_LONG_BLOB:
          sql_type = Types.LONGVARBINARY;

          break;
        case MysqlDefs.FIELD_TYPE_BLOB:
          sql_type = Types.LONGVARBINARY;

          break;
        case MysqlDefs.FIELD_TYPE_VAR_STRING:
          sql_type = Types.VARCHAR;

          break;
        case MysqlDefs.FIELD_TYPE_STRING:
          sql_type = Types.CHAR;

          break;
        default:
          sql_type = Types.VARCHAR;
    }

    return sql_type;
  }


  static int mysqlToJavaType( String MySQLType ) {
    MySQLType = MySQLType.toUpperCase();

    if ( MySQLType.equals( "TINYINT" ) ) {

      return java.sql.Types.TINYINT;
    } else if ( MySQLType.equals( "SMALLINT" ) ) {

      return java.sql.Types.SMALLINT;
    } else if ( MySQLType.equals( "MEDIUMINT" ) ) {

      return java.sql.Types.SMALLINT;
    } else if ( MySQLType.equals( "INT" ) ) {

      return java.sql.Types.INTEGER;
    } else if ( MySQLType.equals( "INTEGER" ) ) {

      return java.sql.Types.INTEGER;
    } else if ( MySQLType.equals( "BIGINT" ) ) {

      return java.sql.Types.BIGINT;
    } else if ( MySQLType.equals( "INT24" ) ) {

      return java.sql.Types.BIGINT;
    } else if ( MySQLType.equals( "REAL" ) ) {

      return java.sql.Types.REAL;
    } else if ( MySQLType.equals( "FLOAT" ) ) {

      return java.sql.Types.FLOAT;
    } else if ( MySQLType.equals( "DECIMAL" ) ) {

      return java.sql.Types.DECIMAL;
    } else if ( MySQLType.equals( "NUMERIC" ) ) {

      return java.sql.Types.NUMERIC;
    } else if ( MySQLType.equals( "DOUBLE" ) ) {

      return java.sql.Types.DOUBLE;
    } else if ( MySQLType.equals( "CHAR" ) ) {

      return java.sql.Types.CHAR;
    } else if ( MySQLType.equals( "VARCHAR" ) ) {

      return java.sql.Types.VARCHAR;
    } else if ( MySQLType.equals( "DATE" ) ) {

      return java.sql.Types.DATE;
    } else if ( MySQLType.equals( "TIME" ) ) {

      return java.sql.Types.TIME;
    } else if ( MySQLType.equals( "TIMESTAMP" ) ) {

      return java.sql.Types.TIMESTAMP;
    } else if ( MySQLType.equals( "DATETIME" ) ) {

      return java.sql.Types.TIMESTAMP;
    } else if ( MySQLType.equals( "TINYBLOB" ) ) {

      return java.sql.Types.BINARY;
    } else if ( MySQLType.equals( "BLOB" ) ) {

      return java.sql.Types.VARBINARY;
    } else if ( MySQLType.equals( "MEDIUMBLOB" ) ) {

      return java.sql.Types.VARBINARY;
    } else if ( MySQLType.equals( "LONGBLOB" ) ) {

      return java.sql.Types.LONGVARBINARY;
    } else if ( MySQLType.equals( "TINYTEXT" ) ) {

      return java.sql.Types.VARCHAR;
    } else if ( MySQLType.equals( "TEXT" ) ) {

      return java.sql.Types.LONGVARCHAR;
    } else if ( MySQLType.equals( "MEDIUMTEXT" ) ) {

      return java.sql.Types.LONGVARCHAR;
    } else if ( MySQLType.equals( "ENUM" ) ) {

      return java.sql.Types.CHAR;
    } else if ( MySQLType.equals( "SET" ) ) {

      return java.sql.Types.CHAR;
    }

    // Punt
    return java.sql.Types.OTHER;
  }
}
