/*
 Copyright (C) 2002 MySQL AB
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */
package com.mysql.jdbc;

import java.sql.SQLException;

/**
 * This interface abstracts away how row data is accessed by
 * the result set. It is meant to allow a static implementation
 * (Current version), and a streaming one.
 *
 * @author     martin
 * @created    19. August 2003
 */
public interface RowData {

  /**
   * Returns true if another row exsists.
   *
   * @return                   Description of the Return Value
   * @exception  SQLException  Description of the Exception
   */
  public boolean hasNext()
       throws SQLException;


  /**
   * Returns the next row.
   *
   * @return                   Description of the Return Value
   * @exception  SQLException  Description of the Exception
   */
  public byte[][] next()
       throws SQLException;


  /**
   * Returns if iteration has not occured yet.
   *
   * @return                   The beforeFirst value
   * @exception  SQLException  Description of the Exception
   */
  public boolean isBeforeFirst()
       throws SQLException;


  /**
   * Returns true if we got the last element.
   *
   * @return                   The afterLast value
   * @exception  SQLException  Description of the Exception
   */
  public boolean isAfterLast()
       throws SQLException;


  /**
   * Moves to before first.
   *
   * @exception  SQLException  Description of the Exception
   */
  public void beforeFirst()
       throws SQLException;


  /**
   * Moves to after last.
   *
   * @exception  SQLException  Description of the Exception
   */
  public void afterLast()
       throws SQLException;


  /**
   * Moves to before last so next el is the last el.
   *
   * @exception  SQLException  Description of the Exception
   */
  public void beforeLast()
       throws SQLException;


  /**
   * DOCUMENT ME!
   *
   *
   * @return                DOCUMENT ME!
   * @throws  SQLException  DOCUMENT ME!
   */
  public boolean isFirst()
       throws SQLException;


  /**
   * DOCUMENT ME!
   *
   *
   * @return                DOCUMENT ME!
   * @throws  SQLException  DOCUMENT ME!
   */
  public boolean isLast()
       throws SQLException;


  /**
   * We're done.
   *
   * @exception  SQLException  Description of the Exception
   */
  public void close()
       throws SQLException;


  /**
   * DOCUMENT ME!
   *
   *
   * @return                DOCUMENT ME!
   * @throws  SQLException  DOCUMENT ME!
   */
  public int getCurrentRowNumber()
       throws SQLException;


  /**
   * DOCUMENT ME!
   *
   *
   * @param  index          DOCUMENT ME!
   * @throws  SQLException  DOCUMENT ME!
   */
  public void setCurrentRow( int index )
       throws SQLException;


  /**
   * DOCUMENT ME!
   *
   *
   * @param  rows           DOCUMENT ME!
   * @throws  SQLException  DOCUMENT ME!
   */
  public void moveRowRelative( int rows )
       throws SQLException;


  /**
   * Returns true if the result set is dynamic.
   *
   * This means that move back and move forward won't work
   * because we do not hold on to the records.
   *
   * @return                   The dynamic value
   * @exception  SQLException  Description of the Exception
   */
  public boolean isDynamic()
       throws SQLException;


  /**
   * Only works on non dynamic result sets.
   *
   * @return                   Description of the Return Value
   * @exception  SQLException  Description of the Exception
   */
  public int size()
       throws SQLException;


  /**
   * Has no records.
   *
   * @return                   The empty value
   * @exception  SQLException  Description of the Exception
   */
  public boolean isEmpty()
       throws SQLException;


  /**
   * Only works on non dynamic result sets.
   *
   * @param  index             Description of the Parameter
   * @return                   The at value
   * @exception  SQLException  Description of the Exception
   */
  public byte[][] getAt( int index )
       throws SQLException;


  /**
   * DOCUMENT ME!
   *
   *
   * @param  row            DOCUMENT ME!
   * @throws  SQLException  DOCUMENT ME!
   */
  public void addRow( byte[][] row )
       throws SQLException;


  /**
   * DOCUMENT ME!
   *
   *
   * @param  index          DOCUMENT ME!
   * @throws  SQLException  DOCUMENT ME!
   */
  public void removeRow( int index )
       throws SQLException;
}
