/*
 *  WordNet-Java
 *
 *  Copyright 1998 by Oliver Steele.  You can use this software freely so long as you preserve
 *  the copyright notice and this restriction, and label your changes.
 */
package edu.brandeis.cs.steele.wn;

/**
 *  A <code>PointerTarget</code> is the source or target of a <code>Pointer</code>
 *  . The target of a semantic <code>PointerTarget</code> is a <code>Synset</code>
 *  ; the target of a lexical <code>PointerTarget</code> is a <code>Word</code>.
 *
 * @author     Oliver Steele, steele@cs.brandeis.edu
 * @created    14. August 2002
 * @see        Pointer
 * @see        Synset
 * @see        Word
 * @version    1.0
 */
public interface PointerTarget {
  public POS getPOS();


  /**
   *  Return a description of the target. For a <code>Word</code>, this is it's
   *  lemma; for a <code>Synset</code>, it's the concatenated lemma's of its
   *  <code>Word</code>s.
   *
   * @return    The description value
   */
  public String getDescription();


  /**
   *  Return the long description of the target. This is its description,
   *  appended by, if it exists, a dash and it's gloss.
   *
   * @return    The longDescription value
   */
  public String getLongDescription();


  /**
   *  Return the outgoing <code>Pointer</code>s from the target -- those <code>Pointer</code>
   *  s that have this object as their source.
   *
   * @return    The pointers value
   */
  public Pointer[] getPointers();


  /**
   *  Return the outgoing <code>Pointer</code>s of type <var>type</var> .
   *
   * @param  type  Description of the Parameter
   * @return       The pointers value
   */
  public Pointer[] getPointers( PointerType type );


  /**
   *  Return the targets of the outgoing <code>Pointer</code>s.
   *
   * @return    The targets value
   */
  public PointerTarget[] getTargets();


  /**
   *  Return the targets of the outgoing <code>Pointer</code>s that have type
   *  <var>type</var> .
   *
   * @param  type  Description of the Parameter
   * @return       The targets value
   */
  public PointerTarget[] getTargets( PointerType type );
}
