/*
 *  WebSPHINX web crawling toolkit
 *  Copyright (C) 1998,1999 Carnegie Mellon University
 *
 *  This library is free software; you can redistribute it
 *  and/or modify it under the terms of the GNU Library
 *  General Public License as published by the Free Software
 *  Foundation, version 2.
 *
 *  WebSPHINX homepage: http://www.cs.cmu.edu/~rcm/websphinx/
 */
package graph;

import java.awt.Dimension;
import java.awt.Event;
import java.awt.Frame;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.image.MemoryImageSource;
import java.text.NumberFormat;
import java.util.Hashtable;
import java.util.Locale;

import bingo.crawler.BINGOCrawler;
import bingo.crawler.CrawlEvent;
import bingo.crawler.CrawlListener;
import bingo.crawler.LinkEvent;
import bingo.crawler.LinkListener;
import bingo.util.BingoDocument;
import bingo.util.SessionBuffer;

public class WebGraph extends GraphLayout implements CrawlListener, LinkListener
{
   
   private SessionBuffer session = SessionBuffer.getInstance();
   
   private Hashtable linkTable = new Hashtable();
   
   // Filtering of a node's outgoing links
   
   public final static int NO_LINKS = 0;
   // Show no outgoing links
   
   public final static int RETRIEVED_LINKS = 1;
   // Show only links that crawler started to retrieve
   
   public final static int WALKED_LINKS = 2;
   // Show RETRIEVED_LINKS, plus links queued for retrieval
   
   public final static int TREE_LINKS = 3;
   // Show WALKED_LINKS, plus links skipped by walk()
   
   public final static int ALL_LINKS = 4;
   // Node rendering
   
   public final static int ICON = 0;
   // Show an icon
   
   public final static int TITLE = 1;
   // Show page title (or URL if not downloaded)
   
   public final static int ABSOLUTE_URL = 2;
   // Show absolute URL
   
   public final static int RELATIVE_URL = 3;
   
   static int linkWidth = 17;
   static int linkHeight = 17;
   static int[] linkData =
   {
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xff343464, 0xff343464, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464, 0xff343464,
      0xffd4d4fc, 0xffc4c4c4, 0xff343464, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464,
      0xff343464, 0xff343464, 0xff343464, 0xffd4d4fc, 0xffd4d4fc, 0xffc4c4c4,
      0xff6464cc, 0xff343464, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464, 0xffccccfc, 0xff848484,
      0xff343464, 0xffd4d4fc, 0xff848484, 0xff848484, 0xff6464cc, 0xff343464,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xff343464, 0xff343464, 0xffd4d4fc, 0xffc4c4c4, 0xff343464, 0xffd4d4fc,
      0xffc4c4c4, 0xff6464cc, 0xff6464cc, 0xff343464, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464,
      0xff343464, 0xffd4d4fc, 0xffc4c4c4, 0xff343464, 0xffd4d4fc, 0xffc4c4c4,
      0xff6464cc, 0xff6464cc, 0xff343464, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464, 0xffccccfc, 0xffd4d4fc,
      0xffc4c4c4, 0xff343464, 0xff343464, 0xff6464cc, 0xff6464cc, 0xff343464,
      0xff343464, 0xff343464, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xff343464, 0xffd4d4fc, 0xffc4c4c4, 0xffc4c4c4, 0xff343464,
      0xfffcfcfc, 0xfffcfcfc, 0xff343464, 0xff343464, 0xffd4d4fc, 0xffccccfc,
      0xff343464, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464,
      0xffccccfc, 0xffc4c4c4, 0xff343464, 0xff343464, 0xfffcfcfc, 0xfffcfcfc,
      0xff040404, 0xff343464, 0xffccccfc, 0xffc4c4c4, 0xffc4c4c4, 0xff343464,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464, 0xff040404,
      0xff343464, 0xff343464, 0xff343464, 0xfffcfcfc, 0xff040404, 0xff343464,
      0xffccccfc, 0xffc4c4c4, 0xff343464, 0xff343464, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464, 0xff343464, 0xffd4d4fc,
      0xffc4c4c4, 0xffc4c4c4, 0xff343464, 0xff343464, 0xffccccfc, 0xffc4c4c4,
      0xff343464, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xff343464, 0xffccccfc, 0xff040404, 0xff6464cc,
      0xff6464cc, 0xff343464, 0xffccccfc, 0xff040404, 0xff343464, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xff343464, 0xffccccfc, 0xff040404, 0xff6464cc, 0xff6464cc,
      0xff343464, 0xffccccfc, 0xff040404, 0xff343464, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464,
      0xffccccfc, 0xffc4c4c4, 0xff6464cc, 0xff343464, 0xff343464, 0xffccccfc,
      0xffc4c4c4, 0xff343464, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464, 0xff040404,
      0xff6464cc, 0xff343464, 0xff343464, 0xff343464, 0xff343464, 0xff343464,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343464, 0xff343464, 0xff343464,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc
   };
   
   static int pageWidth = 22;
   static int pageHeight = 26;
   static int[] pageData =
   {
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xff343434, 0xff343434,
      0xff343434, 0xff343434, 0xff343434, 0xff343434, 0xff343434, 0xff343434,
      0xff343434, 0xff343434, 0xff343434, 0xff343434, 0xff343434, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xff343434, 0xfffcfcfc, 0xff343434, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff343434,
      0xfffcfcfc, 0xfffcfcfc, 0xff343434, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc,
      0xfffcfcfc, 0xff343434, 0xff343434, 0xff343434, 0xff343434, 0xff343434,
      0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xff343434, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xff343434, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xff343434,
      0xfffccc34, 0xfffccc34, 0xff64ccfc, 0xff64ccfc, 0xff64ccfc, 0xfffcfcfc,
      0xfffcfcfc, 0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xff343434, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xff343434, 0xfffccc34, 0xff64ccfc,
      0xff64ccfc, 0xff64ccfc, 0xff64ccfc, 0xfffcfcfc, 0xfffcfcfc, 0xff040404,
      0xff343434, 0xff040404, 0xff343434, 0xff040404, 0xff343434, 0xff040404,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc,
      0xfffcfcfc, 0xff343434, 0xff64ccfc, 0xff64ccfc, 0xff64ccfc, 0xff64ccfc,
      0xff64ccfc, 0xfffcfcfc, 0xfffcfcfc, 0xff040404, 0xff040404, 0xff040404,
      0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xff343434,
      0xff9c6434, 0xff9c6434, 0xff9c6434, 0xff9c6434, 0xff9c6434, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c,
      0xff9c9c9c, 0xff040404, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xff343434, 0xff9c6434, 0xff9c6434,
      0xff9c6434, 0xff9c6434, 0xff9c6434, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c, 0xff040404,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xff9c9c9c, 0xff040404, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xff9c9c9c, 0xff040404, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xff040404, 0xff040404, 0xff040404,
      0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404,
      0xff040404, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff9c9c9c, 0xff040404,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xff9c9c9c, 0xff040404, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xff040404,
      0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404,
      0xff040404, 0xff040404, 0xff040404, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xff9c9c9c, 0xff040404, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff9c9c9c, 0xff040404,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffc3434, 0xfffcfcfc, 0xff040404,
      0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xff9c9c9c, 0xff040404, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xff9c9c9c, 0xff040404, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffc3434,
      0xfffcfcfc, 0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404,
      0xff040404, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xff9c9c9c, 0xff040404,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xff9c9c9c, 0xff040404, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xff343434, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc, 0xfffcfcfc,
      0xff9c9c9c, 0xff040404, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xff343434, 0xff343434, 0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c,
      0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c,
      0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c, 0xff9c9c9c, 0xff343434, 0xff040404,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404,
      0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xff040404,
      0xff040404, 0xff040404, 0xff040404, 0xff040404, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff, 0xffffff,
      0xffffff, 0xffffff
   };
   
   static int retrieveWidth = 8;
   static int retrieveHeight = 8;
   static int[] retrieveData =
   {
      0xffffff, 0xffffff, 0xff008000, 0xff008000, 0xff008000, 0xff008000,
      0xffffff, 0xffffff, 0xffffff, 0xff008000, 0xff008000, 0xff008000,
      0xff008000, 0xff008000, 0xff008000, 0xffffff, 0xff008000, 0xff008000,
      0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000,
      0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000,
      0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000,
      0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000,
      0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000,
      0xffffff, 0xff008000, 0xff008000, 0xff008000, 0xff008000, 0xff008000,
      0xff008000, 0xffffff, 0xffffff, 0xffffff, 0xff008000, 0xff008000,
      0xff008000, 0xff008000, 0xffffff, 0xffffff
   };
   
   static int errorWidth = 8;
   static int errorHeight = 8;
   static int[] errorData =
   {
      0xfffc0404, 0xfffc0404, 0xfcfcfc, 0xfcfcfc, 0xfcfcfc, 0xfcfcfc,
      0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfcfcfc,
      0xfcfcfc, 0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfcfcfc, 0xfffc0404,
      0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfcfcfc,
      0xfcfcfc, 0xfcfcfc, 0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfffc0404,
      0xfcfcfc, 0xfcfcfc, 0xfcfcfc, 0xfcfcfc, 0xfffc0404, 0xfffc0404,
      0xfffc0404, 0xfffc0404, 0xfcfcfc, 0xfcfcfc, 0xfcfcfc, 0xfffc0404,
      0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfcfcfc,
      0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfcfcfc, 0xfcfcfc, 0xfffc0404,
      0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfffc0404, 0xfcfcfc, 0xfcfcfc,
      0xfcfcfc, 0xfcfcfc, 0xfffc0404, 0xfffc0404
   };
   
   Hashtable links = new Hashtable();
   // Show TREE_LINKS, plus links to already-visited pages
   
   int defaultFilter = ALL_LINKS;
   // Show URL relative to parent
   
   int defaultRendering = ICON;
   
   Hashtable icons;
   // maps String (CrawlEvent name or user-defined icon name) to Image
   
   private static NumberFormat numberFormat = NumberFormat.getInstance( Locale.US );
   
   public Image pageIcon;
   public Image linkIcon;
   public Image retrievingIcon;
   public Image errorIcon;
   
   
   // maps Link -> WebNode (for root links) or WebEdge (for internal links)
   
   /**
    *  Make a WebGraph.
    */
   public WebGraph()
   {
      icons = new Hashtable(50);
      
      setPageIcon( defaultPageIcon );
      setLinkIcon( defaultLinkIcon );
      setRetrievingIcon( defaultRetrievingIcon );
      setErrorIcon( defaultErrorIcon );
      
      // Weitere Standard-Icons
      // setIcon( "ROOT/Semistructured Data/XML/", Toolkit.getDefaultToolkit().getImage( "C:/MYICON.GIF" ) );
      //setIcon( "ROOT/Database Core Technology/", Toolkit.getDefaultToolkit().getImage( bingo.util.SessionBuffer.userDir + "/bingo/data/Icons/database.gif" ) );
      
      numberFormat.setMaximumFractionDigits( 3 );
   }
   
   
   /**
    *  Create a new Frame containing a WebGraph connected to a crawler.
    *
    * @param  crawler  Description of the Parameter
    * @return          Description of the Return Value
    */
   public static Frame monitor( BINGOCrawler crawler )
   {
      Frame win = new ClosableFrame();
      
      WebGraph g = new WebGraph();
      
      crawler.addCrawlListener( g );
      crawler.addLinkListener( g );
      g.setNodeCharge( 1000 );
      g.setRestLength( 50 );
      
      win.add( "Center", g );
      win.pack();
      win.show();
      
      return win;
   }
   
   
    /*static String[] getTip( BingoDocument bd ) {
        Vector result = new Vector();
     
        String exception = bd.getLabel( "exception" );
     
        if ( exception != null && exception.length() > 0 ) {
            result.addElement( "*** " + exception );
        }
     
        String anchor = bd.toText();
     
        if ( anchor != null && anchor.length() > 0 ) {
            result.addElement( anchor );
        }
     
        String url = link.getURL().toString();
     
        if ( url != null && url.length() > 0 ) {
            result.addElement( url );
        }
     
        String labels = link.getObjectLabels();
     
        if ( labels != null && labels.length() > 0 ) {
            result.addElement( labels );
        }
     
        String[] tip = new String[result.size()];
     
        result.copyInto( tip );
        return tip;
    }*/
   
   
/*    static String[] getTip( BingoDocument bd ) {
        Vector result = new Vector();
 
        String title = bd.getTitle();
 
        if ( title != null && title.length() > 0 ) {
            result.addElement( title );
        }
 
        String url = bd.getUrl().toString();
 
        if ( url != null && url.length() > 0 ) {
            result.addElement( url );
        }
 
        String labels = bd.getObjectLabels();
 
        if ( labels == null || labels.length() == 0 ) {
            labels = page.getOrigin().getObjectLabels();
            if ( labels != null && labels.length() > 0 ) {
                result.addElement( labels );
            }
        }
 
        String sClass = "CLASS: " + page.getOrigin().getClassString() + " [" + numberFormat.format( page.getSVMRank() ) + "]";
 
        if ( sClass != null && sClass.length() > 0 ) {
            result.addElement( sClass );
        }
 
        String[] tip = new String[result.size()];
 
        result.copyInto( tip );
        return tip;
    }*/
   
   
   /**
    *  Set the default rendering of a node.
    *
    * @param  rendering  The new DefaultRendering value
    */
   public void setDefaultRendering( int rendering )
   {
      defaultRendering = rendering;
   }
   
   
   /**
    *  Set the default filter for links.
    *
    * @param  filter  The new DefaultFilter value
    */
   public synchronized void setDefaultFilter( int filter )
   {
      defaultFilter = filter;
   }
   
   
   /**
    *  Map a name to an icon.
    *
    * @param  name  Name of icon.
    * @param  icon  Icon image. If null, mapping is deleted.
    */
   public synchronized void setIcon( String name, Image icon )
   {
      if (name == null || name.length() <= 0)
         return ;
      
      if ( icon != null )
         icons.put( name, icon );
      else
         icons.remove( name );
      
   }
   
   
   /**
    *  Set the default icon used for pages.
    *
    * @param  icon  Icon image. If null, mapping is deleted.
    */
   public synchronized void setPageIcon( Image icon )
   {
      pageIcon = icon;
      setIcon( LinkEvent.eventName[LinkEvent.VISITED], icon );
   }
   
   
   /**
    *  Set the default icon used for links.
    *
    * @param  icon  Icon image. If null, mapping is deleted.
    */
   public synchronized void setLinkIcon( Image icon )
   {
      linkIcon = icon;
      setIcon( LinkEvent.eventName[LinkEvent.QUEUED], icon );
      setIcon( LinkEvent.eventName[LinkEvent.TOO_DEEP], icon );
      setIcon( LinkEvent.eventName[LinkEvent.ALREADY_VISITED], icon );
      setIcon( LinkEvent.eventName[LinkEvent.SKIPPED], icon );
   }
   
   
   /**
    *  Set the default icon used for requests in progress.
    *
    * @param  icon  Icon image. If null, mapping is deleted.
    */
   public synchronized void setRetrievingIcon( Image icon )
   {
      retrievingIcon = icon;
      setIcon( LinkEvent.eventName[LinkEvent.RETRIEVING], icon );
      setIcon( LinkEvent.eventName[LinkEvent.DOWNLOADED], icon );
   }
   
   
   /**
    *  Set the default icon used for downloaded pages.
    *
    * @param  icon  Icon image. If null, mapping is deleted.
    */
   public synchronized void setDownloadedIcon( Image icon )
   {
      pageIcon = icon;
      setIcon( LinkEvent.eventName[LinkEvent.DOWNLOADED], icon );
   }
   
   
   /**
    *  Set the default icon used for failed requests.
    *
    * @param  icon  Icon image. If null, mapping is deleted.
    */
   public synchronized void setErrorIcon( Image icon )
   {
      errorIcon = icon;
      setIcon( LinkEvent.eventName[LinkEvent.ERROR], icon );
   }
   
   
   /**
    *  Gets the SelectedLink attribute of the WebGraph object
    *
    * @return    The SelectedLink value
    */
   public BingoDocument getSelectedLink()
   {
      WebNode n = (WebNode) getSelectedNode();
      
      if ( n != null )
      {
         return n.link;
      }
      
      WebEdge e = (WebEdge) getSelectedEdge();
      
      if ( e != null )
      {
         return e.link;
      }
      
      return null;
   }
   
   
   /**
    *  Get a named icon.
    *
    * @param  name  Name of icon.
    * @return       icon associated with the name, or null if name unknown.
    */
   public synchronized Image getIcon( String name )
   {
      return (Image) icons.get( name );
   }
   
   
   /**
    *  Clear the graph display.
    */
   public synchronized void clear()
   {
      linkTable.clear();
      links.clear();
      super.clear();
   }
   
   
   /**
    *  Notify that the crawler started.
    *
    * @param  event  Description of the Parameter
    */
   public void started( CrawlEvent event )
   { }
   
   
   /**
    *  Notify that the crawler has stopped.
    *
    * @param  event  Description of the Parameter
    */
   public void stopped( CrawlEvent event )
   { }
   
   
   /**
    *  Notify that the crawler's state was cleared.
    *
    * @param  event  Description of the Parameter
    */
   public void cleared( CrawlEvent event )
   {
      clear();
   }
   
   
   /**
    *  Notify that the crawler has timed out
    *
    * @param  event  Description of the Parameter
    */
   public void timedOut( CrawlEvent event )
   { }
   
   
   /**
    *  Notify that the crawler is paused
    *
    * @param  event  Description of the Parameter
    */
   public void paused( CrawlEvent event )
   { }
   
   
   /**
    *  Update all the links that the crawler reached from this link. Any
    *  reachable links not present in the graph are added.
    *
    * @param  links  Description of Parameter
    */
   public void updateClosure( BingoDocument[] links )
   {
      if ( links == null )
      {
         return;
      }
      for ( int i = 0; i < links.length; ++i )
      {
         BingoDocument link = links[i];
         int status = link.getEventStatus();
         
         if ( status == LinkEvent.NONE )
         {
            continue;
         }
         
         update( link );
         
         if ( status == LinkEvent.DOWNLOADED || status == LinkEvent.VISITED )
         {
            //Page page = link.getPage();
            
            if ( link != null )
            {
               BingoDocument[] linkarray = (BingoDocument[])link.getLinks().toArray();
               updateClosure( linkarray );
            }
         }
      }
   }
   
   
   /**
    *  Update the edge and node associated with a link. If the link is not
    *  present in the graph, it is added.
    *
    * @param  link  Description of the Parameter
    */
   public synchronized void update( BingoDocument link )
   {
      if (!renew())
      {
         
         Object obj = links.get( link );
         
         if ( obj == null )
         {
            add( link );
         } 
         else if ( obj instanceof WebEdge )
         {
            WebEdge e = (WebEdge) obj;
            
            //update( e );
            update( (WebNode) e.to );
         } 
         else
         {
            // obj instanceof WebNode
            update( (WebNode) obj );
         }
      } else
      {
         clear();
      }
      repaint();
   }
   
   
   public boolean handleEvent( Event event )
   {
      if ( event.id == Event.MOUSE_DOWN && event.clickCount == 2 )
      {
         doubleClick( event.x, event.y );
      } else
      {
         return super.handleEvent( event );
      }
      return true;
   }
   
   
   // Change the filter of a node
   synchronized void setLinkFilter( WebNode node, int filter )
   {
      if ( filter == node.filter )
      {
         return;
      }
      
      //Page page = node.link.getPage();
      BingoDocument bd = node.link;
      if ( bd != null )
      {
         BingoDocument[] linkarray = (BingoDocument[])bd.getLinks().toArray();
         
         if ( filter < node.filter )
         {
            // new mode is more restrictive; delete undesired edges
            for ( int j = 0; j < linkarray.length; ++j )
            {
               if ( !shouldDisplay( filter, linkarray[j].getEventStatus() ) )
               {
                  WebEdge edge = (WebEdge) links.get( linkarray[j] );
                  
                  if ( edge != null )
                  {
                     removeNode( (WebNode) edge.to );
                     removeEdge( edge );
                     links.remove( linkarray[j] );
                  }
               }
            }
         } else if ( filter > node.filter )
         {
         // new mode is less restrictive; add edges
            for ( int j = 0; j < linkarray.length; ++j )
            {
               update( linkarray[j] );
               // update() will check shouldDisplay()
            }
         }
      }
      
      node.filter = filter;
   }
   
   
   // Change the filter of ALL nodes
   synchronized void setLinkFilter( int filter )
   {
      defaultFilter = filter;
      
      Graph graph = getGraph();
      
      for ( int i = 0; i < graph.sizeNodes; ++i )
      {
         WebNode n = (WebNode) graph.nodes[i];
         
         setLinkFilter( n, filter );
      }
   }
   
   
   // Change the rendering of a node
   void setNodeRendering( WebNode n, int r )
   {
      n.rendering = r;
      update( n );
      
      repaint();
   }
   
   
   // Change the rendering of ALL nodes
   synchronized void setNodeRendering( int r )
   {
      defaultRendering = r;
      
      Graph graph = getGraph();
      
      for ( int i = 0; i < graph.sizeNodes; ++i )
      {
         WebNode n = (WebNode) graph.nodes[i];
         
         n.rendering = r;
         update( n );
      }
      
      changedGraph();
   }
   
   
   // check whether we want to display a link with this status
   boolean shouldDisplay( int filter, int status )
   {
      switch ( status )
      {
         case LinkEvent.QUEUED:
         case LinkEvent.TOO_DEEP:
            return ( filter > RETRIEVED_LINKS );
         case LinkEvent.SKIPPED:
            return ( filter > WALKED_LINKS );
         case LinkEvent.ALREADY_VISITED:
            return ( filter > TREE_LINKS );
         case LinkEvent.RETRIEVING:
         case LinkEvent.DOWNLOADED:
         case LinkEvent.VISITED:
         case LinkEvent.ERROR:
            return true;
         default:
            return false;
      }
   }
   
   
   synchronized void add( BingoDocument link )
   {
      WebNode n = new WebNode( link, defaultFilter, defaultRendering );
      WebNode parent = findParent( link );
      
      //System.out.println(">>>>>>>>>>>>>>>>>>>>Link>>>>>>>" + link);
      //System.out.println(">>>>>>>>>>>>>>>>>>>>Parent>>>>>>>" + parent);
      if ( parent == null )
      {
         links.put( link, n );
         update( n );
         addNode( n );
         
         if ( getGraph().sizeNodes == 1 )
         {
            // root node of first tree -- put it at the origin and fix it
            n.fixed = true;
            placeNodeOnGraph( n, 0, 0 );
         } else
         {
            // root node of an additional tree -- drop it randomly
            // within window
            
            Dimension d = size();
            
            int x = (int) ( Math.random() * d.width );
            int y = (int) ( Math.random() * d.height );
            
            placeNodeOnScreen( n, x, y );
         }
      } else
      {
         // check if parent's filter allows this link to be displayed
         if ( !shouldDisplay( parent.filter, link.getEventStatus() ) )
         {
            return;
         }
         
         // hang the node off its parent in a random direction
         double len = getRestLength();
         double angle = Math.random() * 2 * Math.PI;
         double x = parent.x + len * Math.cos( angle );
         double y = parent.y + len * Math.sin( angle );
         
         update( n );
         addNode( n );
         placeNodeOnGraph( n, x, y );
         
         WebEdge e = new WebEdge( link, parent, n );
         
         links.put( link, e );
         //update( e );
         addEdge( e );
      }
   }
   
    /*
    void update( WebEdge e ) {
        e.color = Colors.parseColor( e.link.getLabel( "Workbench.color" ) );
        e.thick = e.link.hasLabel( "Workbench.thick" );
    }*/
   
   
   void update( WebNode n )
   {
      BingoDocument b = n.link;
      int status = LinkEvent.NONE; 
      
      if (b != null) status = b.getEventStatus();
      if ( b == null )
      {
         // not downloaded yet
         switch ( n.rendering )
         {
            case ICON:
               n.name = "";
               n.icon = getIcon( LinkEvent.eventName[status] );
               break;
            case TITLE:
            case ABSOLUTE_URL:
               n.name = b.getUrl().toString();
               n.icon = null;
               break;
            case RELATIVE_URL:
            {
               BingoDocument origin = getOrigin( b );
               
               if ( origin != null )
               {
                  n.name = BingoDocument.relativeTo( origin.getUrl(), b.getUrl() );
               } else
               {
                  n.name = b.getUrl().toString();
               }
               n.icon = null;
               break;
            }
         }
      } else
      {
         switch ( n.rendering )
         {
            case ICON:
               n.name = b.toDescription();
               n.icon = getIcon( n.link );
               break;
            case TITLE:
               n.name = b.getTitle();
               if ( n.name == null )
               {
                  n.name = "[" + b.getUrl().toString() + "]";
               }
               n.icon = null;
               break;
            case ABSOLUTE_URL:
               n.name = b.getUrl().toString();
               n.icon = null;
               break;
            case RELATIVE_URL:
            {
               BingoDocument origin = getOrigin( b );
               
               if ( origin != null )
               {
                  n.name = BingoDocument.relativeTo( origin.getUrl(),
                  b.getUrl() );
               } else
               {
                  n.name = b.getUrl().toString();
               }
               n.icon = null;
               break;
            }
         }
            /*n.color = Colors.parseColor( page.getLabel( "Workbench.color" ) );
            n.scale = page.getNumericLabel( "Workbench.size",
            new Integer( 1 ) ).doubleValue();*/
      }
      
      if ( n.icon == null )
      {
         n.icon = defaultRetrievingIcon;
            /*FontMetrics fm = getFontMetrics();
            n.width = fm.stringWidth( n.name ) + 10;
            n.height = fm.getHeight() + 4;*/
      } else
      {
         n.width = (int) ( n.icon.getWidth( this ) * n.scale );
         n.height = (int) ( n.icon.getHeight( this ) * n.scale );
      }
   }
   
   
   WebEdge findEdge( BingoDocument doc )
   {
      if ( doc == null )
      {
         return null;
      }
      Object obj = links.get( doc );
      
      if ( obj instanceof WebEdge )
      {
         return (WebEdge) obj;
      }
      else
      {
         return null;
      }
   }
   
   WebNode findNode( BingoDocument nemo )
   {
      if ( nemo == null )
      {
         return null;
      }
      Object obj = links.get( nemo ); // he! i found nemo :o)
      
      if ( obj instanceof WebEdge )
      {
         return (WebNode) ( (WebEdge) obj ).to;
      } else
      {
         return (WebNode) obj;
      }
   }
   
   
   WebNode findParent( BingoDocument nemo )
   {
      if ( nemo == null )
      {
         return null;
      }
      
      BingoDocument origin = getOrigin( nemo );
      
      return findNode( origin );
   }
   
   
   void doubleClick( int x, int y )
   {
      Object over = pick( x, y );
      
      if ( over != null )
      {
         BingoDocument doc;
         
         if ( over instanceof WebNode )
         {
            doc = ( (WebNode) over ).link;
         } else
         {
            doc = ( (WebEdge) over ).link;
         }
      }
   }
   
   
   public void removeIcons()
   {
      icons.clear();
   }
   
   
   public static Image defaultPageIcon;
   public static Image defaultLinkIcon;
   public static Image defaultRetrievingIcon;
   public static Image defaultErrorIcon;
   
   static
   {
      Toolkit tkit = Toolkit.getDefaultToolkit();
      defaultPageIcon = tkit.createImage(
      new MemoryImageSource( pageWidth, pageHeight,
      pageData, 0, pageWidth ) );
      defaultLinkIcon = tkit.createImage(
      new MemoryImageSource( linkWidth, linkHeight,
      linkData, 0, linkWidth ) );
      defaultRetrievingIcon = tkit.createImage(
      new MemoryImageSource( retrieveWidth, retrieveHeight,
      retrieveData, 0, retrieveWidth ) );
      defaultErrorIcon = tkit.createImage(
      new MemoryImageSource( errorWidth, errorHeight,
      errorData, 0, errorWidth ) );
   }
   
   
   /**
    *  Notify that a crawling event has occured.
    *
    * @param  event  Description of the Parameter
    */
   public void crawled( LinkEvent event )
   {
      if ( event.getID() == LinkEvent.RETRIEVING )  // retrieving kommt immer zuerst
      {
         linkTable.put( new Long(event.getLink().getId()), event.getLink() );
      }
      update( event.getLink() );
      
      try
      {
         Thread.sleep( 80 );
      }
      catch ( InterruptedException e )
      {}
      
   }
   
   private BingoDocument getOrigin( BingoDocument l )
   {
      BingoDocument link = (BingoDocument) linkTable.get( new Long(l.getParentID()) );
      return link;
   }
   
   private Image getIcon( BingoDocument link )
   {
      Image icon = null;
      
      if (link.getEventStatus() == LinkEvent.VISITED )
      {
         icon = getIcon( link.getClassString() );
         if (icon != null)
           return icon;
      }
      
     icon = getIcon( LinkEvent.eventName[link.getEventStatus()] );
     if (icon == null)
     {
       System.out.println ("NAME: " + LinkEvent.eventName[link.getEventStatus()]);
       System.out.println ("ICON: " + icon);
     }   
     return icon;
   }
   
   private boolean renew()
   {
      return linkTable.size() > SessionBuffer.maxVisiblePages;
   }
}

