package multivalent;

import java.awt.Color;
import java.awt.Rectangle;
import java.awt.Graphics2D;
import java.util.Map;
import java.util.Observer;
import java.util.Observable;

import multivalent.std.SyncTimerTask;


/**
	Blinking cursor.
	Could be Mark subclass, but as Behavior we can remap.
	Maybe rename "Cursor" since not a Mark subclass.

	@version $Revision: 1.4 $ $Date: 2003/06/02 05:02:35 $
*/
public class CursorMark extends Behavior implements /*EventListener,*/ Observer {
  /**
	Announce cursor has been moved to new point.
	<p><tt>"setCursor"</tt>: <tt>arg=</tt> {@link CursorMark} <var>this</var>.
  */
  public static final String MSG_SET = "setCursor";

  static SyncTimerTask stt = new multivalent.std.SyncTimerTask();
  static {
	Multivalent.getInstance().getTimer().schedule(stt, 1000, 500);
  }


  static boolean viz_=true;
  //boolean insert_=true;	// insert or replace
  Mark posn_ = new Mark(null,-1, this);


  public Mark getMark() { return posn_; }
  // should be that only Browser can build one?

  public boolean isSet() { return posn_.isSet(); }

  public void move(Mark m) { if (m==null) move(null,-1); else move(m.leaf, m.offset); }

  public void move(Leaf n, int offset) {
	if (posn_.isSet()) repaint(100);   // unshow old
	posn_.move(n,offset);
	moved(50L);
  }

  public void move(int delta) {    // handle jumping from leaf to leaf here
	if (posn_.isSet()) repaint(100);
	posn_.move(delta);
	moved(10L);
  }

  private void moved(long ms) {
	if (posn_.isSet()) {
		//protected boolean knownviz_=false;	// when cursor moves, scroll as necessary to show it
		// ShowHeaders moves...
		//Browser br = getBrowser();
		//Document doc = getStart().leaf.getDocument();
		//if (doc!=br.getCurDocument()) br.setCurDocument(doc);
		viz_=true;
		repaint(10);
		getBrowser().eventq(MSG_SET, this);
	}
  }


  public void update(Observable o, Object arg) {
	if (o == stt.getObservable()) {
		viz_ = ((Boolean)arg).booleanValue();
		repaint(100);
	}
  }


  public boolean paintAfter(Context cx, Node start) {
	if (/*insert_ &&*/ viz_) {
		Rectangle bbox = start.bbox;
		Graphics2D g = cx.g;
		g.setColor(Color.BLACK);
		//g.drawLine(cx.x,bbox.y, cx.x,bbox.y+start.baseline);
		g.drawLine((int)cx.x,0, (int)cx.x,start.bbox.height/*start.baseline*/);
		//xoff_ = cx.x-bbox.x; => accurate if painted, but may be offscreen
//System.out.println("height = "+start.baseline+", @ "+cx.x+", y=0.."+start.bbox.height); //xoff_);
	}
	return false;
  }

  public int getPriority() { return ContextListener.PRIORITY_SPAN + ContextListener.LITTLE; }


  public void restore(ESISNode n, Map<String,Object> attr, Layer layer) {
	super.restore(n,attr, layer);
	stt.getObservable().addObserver(this);
  }

  public void destroy() {
	move(null);
	stt.getObservable().deleteObserver(this);
	super.destroy();
  }

  public void repaint() { repaint(0); }
  public void repaint(long ms) {
	if (isSet()) posn_.leaf.repaint(ms);
  }

  //public void remove() { move(null); }     // don't remove from layer

  /** Remove self when referenced document is closed. */
  public boolean semanticEventAfter(SemanticEvent se, String msg) {
	if (Document.MSG_CLOSE==msg && isSet() && getMark().leaf.getDocument()==se.getArg()) move(null);
	return false;
  }
}
