package multivalent.gui;

import java.awt.*;
import java.awt.geom.Ellipse2D;
import java.util.Map;

import multivalent.*;


/**
	Radiobox widget: only on in associated Radiogroup can be selected at the same time.
	Usable as button and menu item.

<!--
	<p>To do: replace VRadiogroup with key into Document globals so other behaviors can play too.
-->

	@version $Revision: 1.4 $ $Date: 2002/01/27 02:02:16 $
*/
public class VRadiobox extends VButton {
  /** Length of side of radiobox diamond. */
  static final int LENGTH=8;
  static final int GAP = LENGTH+6;
  static final Insets ZEROGAP = new Insets(0,GAP,0,0);	// top,left,bottom,right
  //static int[] xpts={LENGTH/2,LENGTH,LENGTH/2,0}, ypts={0,LENGTH/2,LENGTH,LENGTH/2};	  // LENGTH/2,0, LENGTH,LENGTH/2, LENGTH/2,LENGTH, 0,LENGTH/2,
  //static Polygon diamond_ = new Polygon(xpts, ypts, xpts.length);
  static final Ellipse2D.Double diamond_ = new Ellipse2D.Double(0,0, LENGTH,LENGTH);

  static Insets lastold_=INSETS_ZERO, lastnew_=ZEROGAP;

  VRadiogroup group_;


  public VRadiobox(String name,Map<String,Object> attr, INode parent, VRadiogroup group) {
	super(name,attr, parent);
	setRadiogroup(group);
  }

  public boolean getState() { return (group_!=null && group_.getActive()==this); }
  public void setState(boolean b) { if (b!=getState() && group_!=null) group_.setActive(b?this:null); }

  public VRadiogroup getRadiogroup() { return group_; }
  public void setRadiogroup(VRadiogroup group) { setState(false); group_=group; }

  /** Ensure that padding on left is large enough for checkbox. */
  public boolean formatNode(int width,int height, Context cx) {
	int lp=padding.left, dp=0;
	if (lp<GAP) {
		if (padding==INSETS_ZERO) padding=ZEROGAP;
		else if (padding==lastold_) padding=lastnew_;
		else {
			lastold_=padding;
			padding = lastnew_ = new Insets(padding.top,GAP,padding.bottom,padding.right);
//System.out.println("new radio padding = "+padding);
		}
		dp = lp-GAP;
	}

	super.formatNode(width+dp,height, cx);

	//if (!cx.elide && bbox.height<LENGTH) { baseline += LENGTH - bbox.height; bbox.height = LENGTH; }
	if (!cx.elide && bbox.height<LENGTH) bbox.height = baseline = LENGTH;

	return false;
  }


  /** Draw radiobox too. */
  public void paintNode(Rectangle docclip, Context cx) {
	super.paintNode(docclip,cx);

	int dx=(GAP-LENGTH)/2-GAP, dy=(bbox.height-padding.top-padding.bottom-border.top-border.bottom - LENGTH)/2;   // not: baseline - LENGTH -2

	Graphics2D g = cx.g;
	g.translate(dx,dy);
	//g.setColor(cx.foreground); if (getState()) g.fillPolygon(diamond_); else g.drawPolygon(diamond_);
	g.setColor(cx.foreground); if (getState()) g.fill(diamond_); else g.draw(diamond_);
	//g.setColor(cx.foreground); if (getState()) g.fillOval(x,y, LENGTH,LENGTH); else g.drawOval(x,y,LENGTH,LENGTH);	// looks bad in outline at small sizes
	g.translate(-dx,-dy);
  }

  /** Change state, then call associated VScript, if any. */
  public void invoke() {
	setState(true);
	super.invoke();
  }
}
