package schema;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import java.util.*;
import java.net.*;
import java.io.*;
import bingo.util.Account;

public class LoginDialog extends JDialog implements ActionListener, ItemListener
{
   String[] users;
   String[] passes;
   String[] hosts;
   String[] sids;
   String[] connects;
   
   TreeSet accounts;
   
   KeyAdapter keyAdapter =
   new KeyAdapter()
   {
      public void keyPressed( KeyEvent e )
      {
         if ( e.getKeyCode() == KeyEvent.VK_ENTER )
         {
            ok();
         }
         else if ( e.getKeyCode() == KeyEvent.VK_ESCAPE )
         {
            cancel();
         }
      }
   };
   
   private JTextField hostField;
   private JTextField sidField;
   private JTextField userField;
   private JPasswordField passwordField;
   private JTextField adminuserField;
   private JPasswordField adminpasswordField;
   
   private JComboBox connectBox;
   
   
   private String host;
   private String sid;
   private String user;
   private String adminuser ;
   private String password;
   private String adminpassword ;
   private boolean isOK;
   private JButton ok;
   private JButton cancel;
   
   //private static String delim = System.getProperty( "file.separator" );
   public LoginDialog()
   {
      super( new JFrame(), "BINGO DB-Schema", true );
      setSize( 290, 300 );
      Container content = this.getContentPane();
      content.setLayout( new BorderLayout() );
      
      Container cont = new Container();
      cont.setLayout( new GridLayout( 1, 1, 5, 5 ) );
      
      JPanel panel = createFieldPanel();
      cont.add( panel, BorderLayout.CENTER );
      
      panel = new JPanel();
      panel.setBorder( new EmptyBorder( 6, 0, 6, 12 ) );
      
      ok = new JButton( "Create" );
      ok.addActionListener( this );
      ok.addKeyListener( keyAdapter );
      panel.add( ok );
      cancel = new JButton( "Cancel" );
      cancel.addActionListener( this );
      panel.add( cancel );
      
      content.add( cont, BorderLayout.CENTER );
      content.add( panel, BorderLayout.SOUTH );
      
      show();
      userField.selectAll();
   }
   
   public boolean isOK()
   {
      return isOK;
   }
   
   public String getHost()
   {
      return host;
   }
   
   public String getSID()
   {
      return sid;
   }
   
   public String getUser()
   {
      return user;
   }
   
   public String getAdminUser()
   {
      return adminuser;
   }
   
   public String getPassword()
   {
      return password;
   }
   
   public String getAdminPassword()
   {
      return adminpassword;
   }
   
   public void ok()
   {
      user = userField.getText().trim();
      adminuser = adminuserField.getText().trim();
      password = new String( passwordField.getPassword()).trim();
      adminpassword = new String( adminpasswordField.getPassword() ).trim();
      host = hostField.getText().trim();
      sid = sidField.getText().trim();
      
      boolean allesKlar = true;
      if (user == null || adminuser == null || password == null || adminpassword == null || host == null || sid == null)
         allesKlar = false;
      else if (user.length() * adminuser.length() * password.length() * adminpassword.length() * host.length() * sid.length() == 0)
         allesKlar = false;
      
      if (allesKlar)
      {
         isOK = true;
         this.hide();
         this.setEnabled(false);
      }
   }
   
   public void cancel()
   {
      isOK = false;
      hide();
      System.out.println("CANCEL");
      System.exit(0);
   }
   
   public void actionPerformed( ActionEvent evt )
   {
      Object source = evt.getSource();
      if ( source == ok )
      {
         ok();
      }
      else if ( source == cancel )
      {
         cancel();
      }
   }
   
   // erstellt Eingabefelder
   private JPanel createFieldPanel()
   {
      accounts = readAccounts();
      
      users = new String[accounts.size()];
      passes = new String[accounts.size()];
      hosts = new String[accounts.size()];
      sids = new String[accounts.size()];
      connects = new String[accounts.size()];
      
      int i = 0;
      
      Iterator it = accounts.iterator();
      while(it.hasNext())
      {
         Account ac = (Account) it.next();
         
         users[i] = ac.user;
         passes[i] = ac.pass;
         hosts[i] = ac.host;
         sids[i] = ac.nsid;
         connects[i] = users[i] + "@" + sids[i];
         i++;
      }
      
      connectBox = new JComboBox( connects );
      connectBox.addKeyListener( keyAdapter );
      connectBox.addItemListener( this );
      
      //=================================================
      
      JPanel container = new JPanel( new BorderLayout() );
      GridLayout layout = new GridLayout( 6, 2, 10, 10 );
      JPanel panel = new JPanel( layout );
      
      JLabel label = new JLabel( "     User", SwingConstants.LEFT );
      panel.add( label );
      
      userField = new JTextField( "", 30 );
      userField.addKeyListener( keyAdapter );
      panel.add( userField );
      
      label = new JLabel( "     Password", SwingConstants.LEFT );
      panel.add( label );
      
      passwordField = new JPasswordField( "", 30 );
      passwordField.addKeyListener( keyAdapter );
      panel.add( passwordField );
      
      label = new JLabel( "     Admin", SwingConstants.LEFT );
      panel.add( label );
      
      adminuserField = new JTextField( "", 30 );
      adminuserField.addKeyListener( keyAdapter );
      panel.add( adminuserField );
      
      label = new JLabel( "     Password", SwingConstants.LEFT );
      panel.add( label );
      
      adminpasswordField = new JPasswordField( "", 30 );
      adminpasswordField.addKeyListener( keyAdapter );
      panel.add( adminpasswordField );
      
      label = new JLabel( "     Host", SwingConstants.LEFT );
      panel.add( label );
      
      hostField = new JTextField( "", 30 );
      hostField.addKeyListener( keyAdapter );
      panel.add( hostField );
      
      label = new JLabel( "     Service Name", SwingConstants.LEFT );
      panel.add( label );
      
      sidField = new JTextField( "", 30 );
      sidField.addKeyListener( keyAdapter );
      panel.add( sidField );
      
      container.add( connectBox, BorderLayout.NORTH );
      container.add( panel, BorderLayout.SOUTH );
      
      return container;
   }
   
   public void itemStateChanged( ItemEvent evt )
   {
      sidField.setText( sids[connectBox.getSelectedIndex()] );
      passwordField.setText( passes[connectBox.getSelectedIndex()] );
      userField.setText( users[connectBox.getSelectedIndex()] );
      hostField.setText( hosts[connectBox.getSelectedIndex()] );
      adminuserField.setText ("sys");
   }
   
   public static TreeSet readAccounts()
   {
      StringBuffer fileBuffer = new StringBuffer(200);
      
      fileBuffer.append(System.getProperty("user.dir"));
      fileBuffer.append(System.getProperty("file.separator"));
      fileBuffer.append("bingo");
      fileBuffer.append(System.getProperty("file.separator"));
      fileBuffer.append("data");
      fileBuffer.append(System.getProperty("file.separator"));
      fileBuffer.append("accounts.dat");
      
      TreeSet result = new TreeSet();
      
      try
      {
         BufferedReader bf = new BufferedReader( new FileReader( fileBuffer.toString() ) );
         String line = bf.readLine();
         
         while ( line != null )
         {
            if (line.trim().length() > 0)
            {
     //          System.out.println ("Line: '" + line.trim() + "'");                              
               Account ac = new Account();
               StringTokenizer st = new StringTokenizer( line, ";" );
               
               ac.user = ( st.hasMoreTokens() ? st.nextToken() : "" ).trim();
               ac.pass = ( st.hasMoreTokens() ? st.nextToken() : "" ).trim();
               ac.host = ( st.hasMoreTokens() ? st.nextToken() : "-" ).trim();
               ac.nsid = ( st.hasMoreTokens() ? st.nextToken() : "-" ).trim();
               result.add( ac );
            }
            line = bf.readLine();
         }
         bf.close();
      }
      catch ( Exception e )
      {
         e.printStackTrace();
      }
      
      Account ac = new Account();
      
      ac.user = "";
      ac.pass = "";
      ac.host = "";
      ac.nsid = "";
      result.add( ac );
      
      return result;
   }
   
}