package schema;

import java.sql.DriverManager;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Properties;

import com.mysql.jdbc.Driver;

/**
 *  Die Klasse dient als Zwischenspeicher fr diverse Objekte, auf die das
 *  System zur Laufzeit zugreift. Zudem werden systemweit gltige Konstanten
 *  definiert. Darber hinaus ist das Objekt zustndig fr den Aufbau und
 *  Abschluss der Oracle-Datenbankverbindung inklusive der Benutzer und
 *  Login-Verwaltung. Da zur Laufzeit nur genau eine Instanz dieses Objekts
 *  existieren darf, ist es als Singelton implementiert.
 */

public class SessionBufferMysql
{
   public static Hashtable stopWords;
   public static com.mysql.jdbc.Driver mysqlDriver = null;
   
   //Systemweite Einstellungen
   
   public static int maxProcessingThreads = 20;
   public static int maxStorageThreads = 10;
   
   public static int batchSize = 100;
   public static int chunkSize = 512*1024;
   
   public static boolean dbConnect = false;
   
   public static String sHost;
   public static String sServiceName;
   public static String userDir = null;
   public static String dataDir = null;
   public static String defaultUser;
   public static String defaultUserPasswd;
   public static String adminUser;
   public static String adminUserPasswd;
   
   public static Connection mysqlConnection = null;
   
   // Singelton
   private static SessionBufferMysql sessionBuffer = null;
   
   protected SessionBufferMysql()
   {
      String s = System.getProperty( "file.separator" );
      if ( userDir == null )
         this.userDir = System.getProperty( "user.dir" );
      this.dataDir = userDir + s + "sources";
   }
   
   public boolean isStopword( String s )
   {
      return ( stopWords.containsKey(s) );
   }
   
   public Enumeration getStopwords()
   {
      return stopWords.elements();
   }
   
   // =====================================================================
   
   public static synchronized Connection getMysqlConnection()
   {
      Connection oc = null;
      try
      {
         if ( mysqlDriver == null )
         {
            mysqlDriver = new com.mysql.jdbc.Driver();
            DriverManager.registerDriver( mysqlDriver );
         }
         
         String url = "jdbc:mysql://" + sHost + "/" + sServiceName;
         
         try
         {
            oc = DriverManager.getConnection( url, defaultUser, defaultUserPasswd );
            oc.setAutoCommit( false );
            dbConnect = true;
         } 
         catch (Exception e)
         {
            e.printStackTrace();
         }
         
         oc.setAutoCommit( false );
         dbConnect = true;
         
      } catch ( Exception q )
      {
         q.printStackTrace();
         dbConnect = false;
      }
      return oc;
   }
   
   public static synchronized Connection getAdminMysqlConnection()
   {
      Connection oc = null;
      try
      {
         if ( mysqlDriver == null )
         {
            mysqlDriver = new com.mysql.jdbc.Driver();
            DriverManager.registerDriver( mysqlDriver );
         }
         
         String url = "jdbc:mysql://" + sHost + "/" + sServiceName;
         
         try
         {
            oc = DriverManager.getConnection( url, adminUser, adminUserPasswd );
            oc.setAutoCommit( false );
            dbConnect = true;
         }
         catch (Exception e)
         {
            
         }
         
         oc.setAutoCommit( false );
         dbConnect = true;
         
      } catch ( Exception q )
      {
         q.printStackTrace();
         dbConnect = false;
      }
      return oc;
   }

/**
 *  Methode zum Aufruf der (einzigen) Instanz der Klasse. Es wird keine
 *  Datenbankverbindung aufgebaut.
 *
 *@return    Die Instanz.
 */
public static synchronized SessionBufferMysql getInstance()
{
   if ( sessionBuffer == null )
   {
      sessionBuffer = new SessionBufferMysql();
   }
   return sessionBuffer;
}


/**
 *  Methode zum Aufruf der (einzigen) Instanz der Klasse. Es wird eine
 *  Datenbankverbindung zu dem spezififizierten Oracle-Host aufgebaut. Dieser
 *  Verbindungsaufbau ist nur einmal zur Laufzeit erlaubt.
 *
 *@param  host         Der Host.
 *@param  serviceName  Der ServiceName.
 *@return              Die Instanz.
 */
public static SessionBufferMysql getInstance( String user, String pass, String host, String serviceName )
{
   defaultUser = user;
   defaultUserPasswd = pass;
   sHost = host;
   sServiceName = serviceName;
   
   if ( sessionBuffer == null )
   {
      sessionBuffer = new SessionBufferMysql();
   }
   //    if ( !dbConnect ) {
   //      oracleConnection = getOracleConnection();
   //      if ( oracleConnection == null ) {
   //        System.exit( 0 );
   //      }
   //      dbinterface = new myDBInterface();
   //    }
   return sessionBuffer;
}


/**
 *  Methode zum Aufruf der (einzigen) Instanz der Klasse. Es wird angegeben,
 *  ob eine Datenbankverbindung zu einem vorher spezififizierten Oracle-Host
 *  aufgebaut werden soll. Dieser Verbindungsaufbau ist nur einmal zur
 *  Laufzeit erlaubt.
 */
public static SessionBufferMysql getInstance( boolean connect )
{
   if ( !dbConnect && connect )
   {
      LoginDialog lDialog = new LoginDialog();
      if ( lDialog.isOK() )
      {
         defaultUser = lDialog.getUser();
         defaultUserPasswd = lDialog.getPassword();
         adminUser = lDialog.getAdminUser();
         adminUserPasswd = lDialog.getAdminPassword();
         
         sHost = lDialog.getHost();
         sServiceName = lDialog.getSID();
         
         lDialog = null;
      }
      else
      {
         System.exit(0);
      }
   }
   
   if ( sessionBuffer == null )
   {
      sessionBuffer = new SessionBufferMysql();
   }
   return sessionBuffer;
}

public static int getBatchSize()
{
   return batchSize;
}

public static int getChunkSize()
{
   return chunkSize;
}
}