--- ========================= 
--- DB table: AnchorTerms 
--- C-Schema: AnchorTerms 
--- ========================= 

CREATE TABLE AnchorTerms (
    Term               varchar2(200)  NOT NULL,
    ID                 integer        DEFAULT 0 NOT NULL
)
;

CREATE TABLE Duplicates (
id           number NOT NULL,
link_id      number NOT NULL,
MD5          number Default 0
)
;

CREATE TABLE BingoDocuments (
    ID                 	number        NOT NULL,
    URL                	varchar2(3000) NOT NULL,
    Expiration         	number         ,
    Hostname           	varchar2(255)  ,
    IP                 	number         DEFAULT 0,
    Hashcode          	number        ,
    Port               	integer        DEFAULT 80 ,
    Filename           	varchar2(3000) ,
    Lastmod            	date           default SYSDATE,
    mime               	varchar2(300)  default 'text/html' NOT NULL ,
    Responsecode       	integer        DEFAULT 200 ,
    Contentencoding    	varchar2(300)  DEFAULT 'text/html' ,
    Title   		varchar2(255),
    status             	char           ,
    depth              	integer        DEFAULT 0 NOT NULL,
    origin             	integer        DEFAULT -1 ,
    HubScore           	float          DEFAULT 0 ,
    AuthScore          	float          DEFAULT 0 ,
    PageRank           	number          DEFAULT 0 ,
    SvmScore           	float          DEFAULT 0 ,
    class_id           	integer        ,
    Clicks             	number         DEFAULT 0 ,
    POSevaluation      	number         DEFAULT 0 ,
    NEGevaluation      	number         DEFAULT 0 ,
    DIFFevaluation      number         default 0,
    md5                 number         default 0,
    PRIMARY KEY (ID) 
)
;


--- ================================================
--- DB Table: hasForms
--- C-Schema: hasForms
--- Tabelle zum Abspeichern von Formularmarkierungen
--- ================================================

CREATE TABLE hasForms (
	ID	integer		not null,	
	PRIMARY KEY (ID)
);


--- ========================= 
--- DB table: ArchiveFiles 
--- C-Schema: ArchiveFiles 
--- ========================= 

CREATE TABLE ArchiveFiles (
    title              varchar2(200),
    preview            varchar2(500),
    Source             BLOB           DEFAULT empty_blob() ,
    ID                 integer        NOT NULL
);

CREATE UNIQUE INDEX ArchiveFiles_IndexID
    ON ArchiveFiles (ID)
;

--- ========================= 
--- DB table: ClassFeatures 
--- C-Schema: ClassFeatures 
--- ========================= 

CREATE TABLE ClassFeatures (
    term               varchar2(200)  NOT NULL,
    MI                 number         DEFAULT 0 ,
    IG                 number         DEFAULT 0 ,
    DF                 number         DEFAULT 0 ,
    IDF                number         DEFAULT 0 ,
    Class_ID           integer        NOT NULL,
    evaluation         integer        default 1
)
;

CREATE INDEX Classfeatures_term_index
    ON Classfeatures (term)
;

CREATE INDEX Classfeatures_ClassID_index
    ON Classfeatures (Class_ID)
;



--- ========================= 
--- DB table: Classes 
--- C-Schema: Classes 
--- ========================= 

CREATE TABLE Classes (
    Class_ID           integer        NOT NULL,
    Name               varchar2(500)  NOT NULL,
    PRIMARY KEY (Class_ID) 
)
;

--- indices 

CREATE UNIQUE INDEX Classes_Name
    ON Classes (Name)
;




--- ========================= 
--- DB table: Classifier 
--- C-Schema: Classifier 
--- ========================= 

CREATE TABLE Classifier (
    id                 integer        primary key,
    timestamp          DATE           DEFAULT sysdate,
    classifier         BLOB           DEFAULT empty_blob() ,
    Step               integer        DEFAULT 0 ,
    Type               integer        DEFAULT -1 
)
;

CREATE INDEX Classifier_Index_Timestamp
    ON Classifier (timestamp)
;


--- ========================= 
--- DB table: ClusterElements 
--- C-Schema: ClusterElements 
--- ========================= 

CREATE TABLE ClusterElements (
    Cluster_id         integer        NOT NULL,
    ID                 integer        NOT NULL,
    FOREIGN KEY (ID) REFERENCES BingoDocuments
      ON DELETE CASCADE
)
;




--- ========================= 
--- DB table: Clusters 
--- C-Schema: Clusters 
--- ========================= 

CREATE TABLE Clusters (
    Cluster_id         integer        NOT NULL,
    Annotation         varchar2(300)  ,
    Medoid             integer        NOT NULL,
    PRIMARY KEY (Cluster_id) ,
    FOREIGN KEY (Medoid) REFERENCES BingoDocuments
      ON DELETE CASCADE
)
;


--- ========================= 
--- DB table: DocumentFeatures 
--- C-Schema: DocumentFeatures 
--- ========================= 

CREATE TABLE DocumentFeatures (
    Term               varchar2(200)  NOT NULL,
    RTF                number         DEFAULT 0 ,
    TF                 number         DEFAULT 0 ,
    TF_IDF             number         DEFAULT 0 ,
    ID                 integer        NOT NULL
)
;


--- ========================= 
--- DB table: DocumentUnion 
--- C-Schema: DocumentUnion 
--- ========================= 

CREATE TABLE DocumentUnion (
    term               varchar2(200)  NOT NULL,
    RTF                float          DEFAULT 0 ,
    TF                 float          DEFAULT 0 ,
    TF_IDF             float          DEFAULT 0 ,
    ID                 integer        NOT NULL,
    FOREIGN KEY (ID) REFERENCES BingoDocuments
      ON DELETE CASCADE
)
;

--- ========================= 
--- DB table: Feedback 
--- C-Schema: Feedback 
--- ========================= 

CREATE TABLE Feedback (
    iteration          integer        DEFAULT 0 ,
    result             integer        DEFAULT -1 ,
    class_old          integer        NOT NULL,
    class_new          integer        DEFAULT -1 ,
    ID                 integer        NOT NULL
)
;

--- ========================= 
--- DB table: FeedbackArchive 
--- C-Schema: FeedbackArchive 
--- ========================= 

CREATE TABLE FeedbackArchive (
    summary            CLOB           DEFAULT empty_clob() ,
    class              integer        NOT NULL,
    ID                 integer        NOT NULL
)
;

--- ========================= 
--- DB table: GoogleHubs 
--- C-Schema: GoogleHubs 
--- ========================= 

CREATE TABLE GoogleHubs (
    source             CLOB           DEFAULT empty_clob() ,
    lastmod            date           DEFAULT sysdate ,
    query              varchar2(1000) ,
    ID                 integer        NOT NULL
)
;

CREATE UNIQUE INDEX GoogleHubs_IndexID
    ON GoogleHubs (ID)
;

--- ========================= 
--- DB table: GoogleRelated 
--- C-Schema: GoogleRelated 
--- ========================= 

CREATE TABLE GoogleRelated (
    source             CLOB           DEFAULT empty_clob() ,
    lastmod            date           DEFAULT sysdate ,
    query              varchar2(1000) ,
    ID                 integer        NOT NULL
)
;

--- indices 

CREATE UNIQUE INDEX GoogleRelated_IndexID
    ON GoogleRelated (ID)
;




--- ========================= 
--- DB table: Links 
--- C-Schema: Links 
--- ========================= 

CREATE TABLE Links (
    Target             varchar2(3000) NOT NULL,
    Link_ID            integer        DEFAULT null ,
    Hostname           varchar2(255)  NOT NULL,
    ID                 integer        
)
;

--- ========================= 
--- DB table: MimeTypes 
--- C-Schema: MimeTypes 
--- ========================= 

CREATE TABLE MimeTypes (
    mime               varchar2(300)  NOT NULL,
    application        varchar2(200)  ,
    extension          varchar2(10)   ,
    maxsize            float          DEFAULT 1048576 ,
    handler            varchar2(200)  NOT NULL,
    PRIMARY KEY (mime) 
)
;




--- ========================= 
--- DB table: PairFeatures 
--- C-Schema: PairFeatures 
--- ========================= 

CREATE TABLE PairFeatures (
    pair               varchar2(400)  NOT NULL,
    RPF                float          DEFAULT 0 ,
    PF                 float          DEFAULT 0 ,
    PF_IPF             float          DEFAULT 0 ,
    ID                 integer        NOT NULL,
    FOREIGN KEY (ID) REFERENCES BingoDocuments
      ON DELETE CASCADE
)
;

--- indices 

CREATE INDEX PairFeatures_IndexID
    ON PairFeatures (ID)
;

CREATE INDEX PairFeatures_IndexPair
    ON PairFeatures (pair)
;




--- ========================= 
--- DB table: Preferences 
--- C-Schema: Preferences 
--- ========================= 

CREATE TABLE Preferences (
    lastmod            date           DEFAULT sysdate ,
    Value              varchar2(100)  NOT NULL,
    Name               varchar2(300)  NOT NULL
)
;




--- ========================= 
--- DB table: Stopwords 
--- C-Schema: Stopwords 
--- ========================= 

CREATE TABLE Stopwords (
    Term               varchar2(200)  NOT NULL
)
;




--- ========================= 
--- DB table: TempDocuments 
--- C-Schema: TempDocuments 
--- ========================= 

CREATE TABLE TempDocuments (
    Temp_ID            varchar2(10)   ,
    ID                 varchar2(10)   ,
    Class_ID           varchar2(10)   ,
    Prop_Class_ID      varchar2(10)   ,
    Status             char           
)
;




--- ========================= 
--- DB table: TempQuery 
--- C-Schema: TempQuery 
--- ========================= 

CREATE TABLE TempQuery (
    SessionID          varchar2(255)  NOT NULL,
    ANDquery           varchar2(3000) ,
    ORquery            varchar2(3000) ,
    PRIMARY KEY (SessionID) 
)
;




--- ========================= 
--- DB table: TermPairs 
--- C-Schema: TermPairs 
--- ========================= 

CREATE TABLE TermPairs (
    Term_A             varchar2(200)  NOT NULL,
    Term_B             varchar2(200)  NOT NULL,
    Feature            varchar2(200)  NOT NULL,
    Distance           number         DEFAULT 0 ,
    ID                 integer        NOT NULL,
    FOREIGN KEY (ID) REFERENCES BingoDocuments
      ON DELETE CASCADE
)
;

--- indices 

CREATE INDEX TermPairs_IndexID
    ON TermPairs (ID)
;

CREATE INDEX TermPairs_IndexFeature
    ON TermPairs (Feature)
;


--- ========================= 
--- DB table: Terms 
--- C-Schema: Terms 
--- ========================= 

CREATE TABLE Terms (
    Term              varchar2(200)  NOT NULL,
    Stem              varchar2(200)  , 
    Pos               number         DEFAULT 0 ,
    ID                integer        NOT NULL
)
;

--- foreign keys 

ALTER TABLE ClusterElements 
    ADD FOREIGN KEY (Cluster_id) REFERENCES Clusters
      ON DELETE CASCADE
;

--- Views

create view documents as
	select b.url, b.id, c.name as class, b.status, b.authscore as authority, 
		   b.hubscore as hub, b.pagerank as pagerank, b.svmscore as svm, 
	  	   b.hostname as host, b.mime
	from BingoDocuments b, classes c 
	where c.class_id = b.class_id
;

create table start_urls
(
link varchar2(1000)
);

create table feedback_options
(
URL_SELECT                                VARCHAR2(15) NOT NULL ,
OTHERS_SELECT                             VARCHAR2(10) NOT NULL ,
SCORE_SELECT                              VARCHAR2(15) NOT NULL ,
TOP_SELECT                                VARCHAR2(15) NOT NULL
);

commit;

--- end of SQL 
