--- ========================= 
--- DB table: AnchorTerms 
--- C-Schema: AnchorTerms 
--- ========================= 

CREATE TABLE AnchorTerms (
    Term               varchar(200)  NOT NULL,
    ID                 bigint        DEFAULT 0 NOT NULL
)
;

CREATE TABLE BingoDocuments (
    ID                 	bigint        NOT NULL,
    URL                	text           NOT NULL,
    Expiration         	bigint         ,
    Hostname           	varchar(255)  ,
    IP                 	bigint         DEFAULT 0,
    Hashcode          	bigint        ,
    Port               	integer        DEFAULT 80 ,
    Filename           	text ,
    Lastmod            	date         , 
    mime               	varchar(255)  default 'text/html' NOT NULL ,
    Responsecode       	integer        DEFAULT 200 ,
    Contentencoding    	varchar(255)  DEFAULT 'text/html' ,
    Title   		varchar(255),
    status             	char           ,
    depth              	integer        DEFAULT 0 NOT NULL,
    origin             	integer        DEFAULT -1 ,
    HubScore           	double          DEFAULT 0 ,
    AuthScore          	double          DEFAULT 0 ,
    PageRank           	double          DEFAULT 0 ,
    SvmScore           	double          DEFAULT 0 ,
    class_id           	bigint        ,
    Clicks             	bigint         DEFAULT 0 ,
    POSevaluation      	bigint         DEFAULT 0 ,
    NEGevaluation      	bigint         DEFAULT 0 ,
    DIFFevaluation      bigint         default 0,
    md5                 bigint         fefault 0,
    PRIMARY KEY (ID) 
)
;


--- ================================================
--- DB Table: hasForms
--- C-Schema: hasForms
--- Tabelle zum Abspeichern von Formularmarkierungen
--- ================================================

CREATE TABLE hasForms (
	ID	bigint		not null,	
	PRIMARY KEY (ID)
);


--- ========================= 
--- DB table: ArchiveFiles 
--- C-Schema: ArchiveFiles 
--- ========================= 

CREATE TABLE ArchiveFiles (
    title              varchar(200),
    preview            text, 
    Source             BLOB           ,
    ID                 bigint        NOT NULL
);

CREATE UNIQUE INDEX ArchiveFiles_IndexID
    ON ArchiveFiles (ID)
;

--- ========================= 
--- DB table: ClassFeatures 
--- C-Schema: ClassFeatures 
--- ========================= 

CREATE TABLE ClassFeatures (
    term               varchar(200)  NOT NULL,
    MI                 double         DEFAULT 0 ,
    IG                 double         DEFAULT 0 ,
    DF                 double         DEFAULT 0 ,
    IDF                double         DEFAULT 0 ,
    Class_ID           bigint        NOT NULL,
    evaluation         integer        default 1
)
;

CREATE INDEX Classfeatures_term_index
    ON Classfeatures (term)
;

CREATE INDEX Classfeatures_ClassID_index
    ON Classfeatures (Class_ID)
;



--- ========================= 
--- DB table: Classes 
--- C-Schema: Classes 
--- ========================= 

CREATE TABLE Classes (
    Class_ID           bigint        NOT NULL,
    Name               text           NOT NULL,
    PRIMARY KEY (Class_ID) 
)
;

--- indices 

CREATE UNIQUE INDEX Classes_Name
    ON Classes (Name(255))
;




--- ========================= 
--- DB table: Classifier 
--- C-Schema: Classifier 
--- ========================= 

CREATE TABLE Classifier (
    id                 bigint        primary key,
    timestamp          DATE           ,
    classifier         BLOB           ,
    Step               integer        DEFAULT 0 ,
    Type               integer        DEFAULT -1 
)
;

CREATE INDEX Classifier_Index_Timestamp
    ON Classifier (timestamp)
;


--- ========================= 
--- DB table: ClusterElements 
--- C-Schema: ClusterElements 
--- ========================= 

CREATE TABLE ClusterElements (
    Cluster_id         bigint        NOT NULL,
    ID                 bigint        NOT NULL,
    FOREIGN KEY (ID) REFERENCES BingoDocuments
      ON DELETE CASCADE
)
;




--- ========================= 
--- DB table: Clusters 
--- C-Schema: Clusters 
--- ========================= 

CREATE TABLE Clusters (
    Cluster_id         bigint        NOT NULL,
    Annotation         varchar(255)  ,
    Medoid             bigint        NOT NULL,
    PRIMARY KEY (Cluster_id) ,
    FOREIGN KEY (Medoid) REFERENCES BingoDocuments
      ON DELETE CASCADE
)
;


--- ========================= 
--- DB table: DocumentFeatures 
--- C-Schema: DocumentFeatures 
--- ========================= 

CREATE TABLE DocumentFeatures (
    Term               varchar(200)  NOT NULL,
    RTF                double         DEFAULT 0 ,
    TF                 double         DEFAULT 0 ,
    TF_IDF             double         DEFAULT 0 ,
    ID                 bigint        NOT NULL
)
;


--- ========================= 
--- DB table: DocumentUnion 
--- C-Schema: DocumentUnion 
--- ========================= 

CREATE TABLE DocumentUnion (
    term               varchar(200)  NOT NULL,
    RTF                double          DEFAULT 0 ,
    TF                 double          DEFAULT 0 ,
    TF_IDF             double          DEFAULT 0 ,
    ID                 bigint        NOT NULL,
    FOREIGN KEY (ID) REFERENCES BingoDocuments
      ON DELETE CASCADE
)
;

--- ========================= 
--- DB table: Feedback 
--- C-Schema: Feedback 
--- ========================= 

CREATE TABLE Feedback (
    iteration          bigint        DEFAULT 0 ,
    result             bigint        DEFAULT -1 ,
    class_old          bigint        NOT NULL,
    class_new          bigint        DEFAULT -1 ,
    ID                 bigint        NOT NULL
)
;

--- ========================= 
--- DB table: FeedbackArchive 
--- C-Schema: FeedbackArchive 
--- ========================= 

CREATE TABLE FeedbackArchive (
    summary            text ,
    class              bigint        NOT NULL,
    ID                 bigint        NOT NULL
)
;

--- ========================= 
--- DB table: GoogleHubs 
--- C-Schema: GoogleHubs 
--- ========================= 

CREATE TABLE GoogleHubs (
    source             text  ,
    lastmod            date ,
    query              text ,
    ID                 bigint        NOT NULL
)
;

CREATE UNIQUE INDEX GoogleHubs_IndexID
    ON GoogleHubs (ID)
;

--- ========================= 
--- DB table: GoogleRelated 
--- C-Schema: GoogleRelated 
--- ========================= 

CREATE TABLE GoogleRelated (
    source             text,
    lastmod            date,
    query              text ,
    ID                 bigint        NOT NULL
)
;

--- indices 

CREATE UNIQUE INDEX GoogleRelated_IndexID
    ON GoogleRelated (ID)
;




--- ========================= 
--- DB table: Links 
--- C-Schema: Links 
--- ========================= 

CREATE TABLE Links (
    Target             text          NOT NULL,
    Link_ID            bigint        DEFAULT null ,
    Hostname           varchar(255)  NOT NULL,
    ID                 bigint        
)
;

--- ========================= 
--- DB table: MimeTypes 
--- C-Schema: MimeTypes 
--- ========================= 

CREATE TABLE MimeTypes (
    mime               varchar(255)  NOT NULL,
    application        varchar(200)  ,
    extension          varchar(10)   ,
    maxsize            bigint          DEFAULT 1048576 ,
    handler            varchar(200)  NOT NULL,
    PRIMARY KEY (mime) 
)
;


--- ========================= 
--- DB table: Preferences 
--- C-Schema: Preferences 
--- ========================= 

CREATE TABLE Preferences (
    lastmod            date       ,
    Value              varchar(100)  NOT NULL,
    Name               varchar(255)  NOT NULL
)
;




--- ========================= 
--- DB table: Stopwords 
--- C-Schema: Stopwords 
--- ========================= 

CREATE TABLE Stopwords (
    Term               varchar(200)  NOT NULL
)
;




--- ========================= 
--- DB table: TempDocuments 
--- C-Schema: TempDocuments 
--- ========================= 

CREATE TABLE TempDocuments (
    Temp_ID            varchar(10)   ,
    ID                 varchar(10)   ,
    Class_ID           varchar(10)   ,
    Prop_Class_ID      varchar(10)   ,
    Status             char           
)
;




--- ========================= 
--- DB table: TempQuery 
--- C-Schema: TempQuery 
--- ========================= 

CREATE TABLE TempQuery (
    SessionID          varchar(255)  NOT NULL,
    ANDquery           text ,
    ORquery            text ,
    PRIMARY KEY (SessionID) 
)
;




--- ========================= 
--- DB table: TermPairs 
--- C-Schema: TermPairs 
--- ========================= 

CREATE TABLE TermPairs (
    Term_A             varchar(200)  NOT NULL,
    Term_B             varchar(200)  NOT NULL,
    Feature            varchar(200)  NOT NULL,
    Distance           bigint         DEFAULT 0 ,
    ID                 bigint        NOT NULL,
    FOREIGN KEY (ID) REFERENCES BingoDocuments
      ON DELETE CASCADE
)
;

--- indices 

CREATE INDEX TermPairs_IndexID
    ON TermPairs (ID)
;

CREATE INDEX TermPairs_IndexFeature
    ON TermPairs (Feature)
;


--- ========================= 
--- DB table: Terms 
--- C-Schema: Terms 
--- ========================= 

CREATE TABLE Terms (
    Term              varchar(200)  NOT NULL,
    Pos               bigint        DEFAULT 0 ,
    ID                bigint        NOT NULL
)
;

--- foreign keys 

ALTER TABLE ClusterElements 
    ADD FOREIGN KEY (Cluster_id) REFERENCES Clusters
      ON DELETE CASCADE
;

--- Views

create view documents as
	select b.url, b.id, c.name as class, b.status, b.authscore as authority, 
		   b.hubscore as hub, b.pagerank as pagerank, b.svmscore as svm, 
	  	   b.hostname as host, b.mime
	from BingoDocuments b, classes c 
	where c.class_id = b.class_id
;

create table start_urls
(
link text
);

create table feedback_options
(
URL_SELECT                                varchar(15) NOT NULL ,
OTHERS_SELECT                             varchar(10) NOT NULL ,
SCORE_SELECT                              varchar(15) NOT NULL ,
TOP_SELECT                                varchar(15) NOT NULL
);

commit;

--- end of SQL 
