package jodd.db.pool;

import java.sql.Connection;
import java.sql.SQLException;

/**
 * A database connection pool interface.
 */
public interface ConnectionPool {

	/**
	 * Initialize pool with current object state.
	 *
	 * @exception SQLException
	 */
	public void init() throws SQLException;
	

	/**
	 * Get one free connection from the connection pool.
	 *
	 * @return Connection object representing free database connection
	 * @exception SQLException
	 */
	public Connection getConnection() throws SQLException;

	/**
	 * Returns connection to connection pool.
	 *
	 * @param conn   database connection
	 */
	public void freeConnection(Connection conn);


	/**
	 * Close pool when it is not needed anymore.
	 */
	public void close();

}

