package jodd.servlet.tags.imports;

import javax.servlet.jsp.JspTagException;
import javax.servlet.jsp.tagext.BodyContent;
import javax.servlet.jsp.tagext.BodyTagSupport;

/**
 * Parameters for import tag.
 *
 * @see Import
 */
public class ImportParam extends BodyTagSupport {

	private String name = "";
	/**
	 * Sets parameter name (required).
	 *
	 * @param v      parameter name
	 */
	public void setName(String v) {
		name = v;
	}
	/**
	 * Returns parameter name.
	 *
	 * @return parameter name
	 */
	protected String getName() {
		return name;
	}


	private String value = "";
	/**
	 * Sets value name.
	 *
	 * @param v      value name
	 */
	public void setValue(String v) {
		value = v;
	}
	/**
	 * Returns parameter value.
	 *
	 * @return parameter value
	 */
	public String getValue() {
		return value;
	}

	private String persist = "";

	/**
	 * Set persistence flag for a parameter.
	 *
	 * @param v      parameter persistance
	 */
	public void setPersist(String v) {
		persist = v;
	}
	/**
	 * Returns parameter persistance.
	 *
	 * @return parameter persistence
	 */
	public String getPersist() {
		return persist;
	}
	
	
	private Import parent = null;
	
	private boolean isPersistent = false;

	public int doStartTag() {
		parent = (Import) findAncestorWithClass(this, Import.class);
		if (parent == null) {
			return SKIP_BODY;
		}
		if ((persist != null) && (persist.equals("true"))) {
			isPersistent = true;
		}

		if ((value == null) || (value.length() == 0)) {
			return EVAL_BODY_AGAIN;
		} else {
			pageContext.getRequest().setAttribute(name, value);
			if (isPersistent == false) {
				parent.addAttributeName(name);
			}
			return SKIP_BODY;
		}
	}


	public int doAfterBody() throws JspTagException {
		BodyContent body = getBodyContent();
		value = body.getString();
		pageContext.getRequest().setAttribute(name, value);
		if (isPersistent == false) {
			parent.addAttributeName(name);
		}
		return SKIP_BODY;
	}

	public int doEndTag() {
		return EVAL_PAGE;
	}

}

