package jodd.servlet.tags.imports;

import java.io.IOException;
import java.io.PrintWriter;

import javax.servlet.ServletOutputStream;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpServletResponseWrapper;
import javax.servlet.jsp.JspWriter;

/**
 * Simple servlet response wrapper that may be used for tags.
 */
public class ImportResponseWrapper extends HttpServletResponseWrapper {

	private PrintWriter writer;

	/**
	 * Constructs the wrapper class. Usually, wrapper objects are created using
	 * <code>pageContext.getResponse()</code> and <code>pageContext.getOut()</code>.
	 *
	 * @param response  usually pageContext.getResponse()
	 * @param jspWriter usual pageContext.getOut()
	 */
	public ImportResponseWrapper(ServletResponse response, JspWriter jspWriter) {
		super((HttpServletResponse)response);
		writer = new PrintWriter(jspWriter);
	}

	/**
	 * Returns writer created from JspWriter.
	 *
	 * @return writer that writes to jsp output
	 * @exception IOException
	 */
	public PrintWriter getWriter() throws IOException {
		return writer;
	}

	/**
	 * it doesn't do anything since this wraper has to return only writer.
	 *
	 * @return nothing, an exception is thrown
	 * @exception IOException
	 */
	public ServletOutputStream getOutputStream() throws IOException {
		throw new IllegalStateException();
	}
}
