package stored;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import bingo.util.BingoTreeNode;
import bingo.util.SessionBuffer;

/**
 *  Berechnet die DF- und IDF-Werte der Terme in der Feature-Tabelle.
 *
 * @author     martin
 * @created    14. August 2002
 */
public class berechne_df
{
  public Vector myAssistants;
  private SessionBuffer session;
  private int max_anzahl_terme;
  private Connection conn;
  
  
  /*
   *  Diese Klasse kann ber die Kommandozeile gestartet werden
   */
  public static void main( String[] args )
  {
    if ( args == null || args.length == 0 )
    {
      System.out.println( "Usage: java berechne_df <number of terms>" );
    } else
    {
      new berechne_df( Integer.parseInt( args[0] ) ).go();
    }
  }
  
  
  /**
   *  Initialisiert die Berechnung.
   *
   * @param  anzahl  Description of the Parameter
   */
  public berechne_df( int anzahl )
  {
    max_anzahl_terme = anzahl;
    session = SessionBuffer.getInstance();
    conn = session.getDBConnection();
  }
  
  
  public void go()
  {
    try
    {
      myAssistants = new Vector();
      
      session = SessionBuffer.getInstance();
      
      Statement stmt = conn.createStatement();
      // Nur statement, keine eigene Connection
      stmt.executeQuery( "DELETE FROM ClassFeatures" );
      stmt.executeQuery( "DELETE FROM Classes" );
      
      PreparedStatement pstmt = conn.prepareStatement( "insert into classes (name, class_id) values (?, ?)" );
      
      BingoTreeNode root = session.getRootNode();
      Collection allClasses = root.allClasses.values();
      
      Iterator it = allClasses.iterator();
      while ( it.hasNext() )
      {
        BingoTreeNode current = ((BingoTreeNode) it.next());
        String myClass = current.label;
        
        System.out.println( myClass );
        pstmt.setString( 1, myClass );
        pstmt.setInt( 2, current.ID );
        pstmt.addBatch();
      }
      it = null;
      
      pstmt.executeBatch();
      stmt.executeQuery( "COMMIT" );
      stmt.close();
      pstmt.close();
      
      starteThreads( allClasses );
      
      while ( myAssistants.size() > 0 )
      {
        synchronized ( myAssistants )
        {
          try
          {
            if ( myAssistants.size() > 0 )
              myAssistants.wait();
          }
          catch ( Exception e )
          {
            e.printStackTrace();
          }
        }
      }
    }
    catch ( Exception e )
    {
      e.printStackTrace();
    }
  }
  
  
  public void starteThreads( Collection all )
  {
    Iterator it = all.iterator();
    while ( it.hasNext() )
    {
      BingoTreeNode current = (BingoTreeNode)it.next();
      DfThread assi = new DfThread( this, current, max_anzahl_terme );
      
      synchronized ( myAssistants )
      {
        myAssistants.add( assi );
      }
      new Thread( assi ).start();
    }
    it = null;
  }
}

//======================================================================0

class DfThread implements Runnable
{
  private berechne_df parent;
  private BingoTreeNode myClass;
  private int ID;
  private Connection conn;
  private Statement stmt;
  private ResultSet rset;
  private SessionBuffer session;
  private int howMany;
  
  
  public DfThread( berechne_df p, BingoTreeNode c, int h )
  {
    parent = p;
    myClass = c;
    ID = myClass.ID;
    session = SessionBuffer.getInstance();
    howMany = h;
  }
  
  public void run()
  {
    int batchSize = session.getBatchSize();
    ResultSet rset;
    String query;
    
    try
    {
      conn = session.getOracleConnection();
      stmt = conn.createStatement();
      query = "SELECT count(distinct ID) AS N FROM DocumentFeatures WHERE ID IN "
      + "  (SELECT ID FROM bingodocuments WHERE status in ( 'B', 'M' ) and Class_ID IN "
      + "    (SELECT Class_ID FROM Classes WHERE Name LIKE '" + myClass.label + "%')) ";
      
            //System.out.println (query);
      
      rset = stmt.executeQuery( query );
      rset.next();
      
      int N = rset.getInt( "N" );
      
      query = "SELECT term, count(distinct ID) AS DF FROM DocumentFeatures WHERE ID IN "
      + "  (SELECT ID FROM bingodocuments WHERE status in ( 'B', 'M' ) and Class_ID IN "
      + "    (SELECT Class_ID FROM Classes WHERE Name LIKE '" + myClass.label + "%')) "
      + " GROUP BY term ORDER BY DF DESC";
      rset = stmt.executeQuery( query );
      
      PreparedStatement pstmt = conn.prepareStatement( "INSERT INTO ClassFeatures (Term, DF, IDF, Class_ID) VALUES (?, ?, ?, ?)" );
      
      if ( myClass.isRoot() )
        howMany = Integer.MAX_VALUE;
      
      int added = 0;
      
      for ( added = 0; rset.next() && added < howMany; added++ )
      {
        
        pstmt.setString( 1, rset.getString( "TERM" ) );
        pstmt.setInt( 2, rset.getInt( "DF" ) );
        pstmt.setDouble( 3, (double) N / (double)rset.getInt( "DF" ) );
        pstmt.setInt( 4, myClass.ID );
        pstmt.addBatch();
        if ( added % batchSize == 0 )
          pstmt.executeBatch();
      }
      pstmt.executeBatch();
      pstmt.close();
      stmt.executeQuery( "COMMIT" );
      System.out.println( "DF/IDF for CLASS: " + myClass.label + " (" + added + ")" );
      rset.close();
      stmt.close();
      conn.close();
    }
    catch ( Exception e )
    {
      e.printStackTrace();
    }
    
    synchronized ( parent.myAssistants )
    {
      parent.myAssistants.remove( this );
      parent.myAssistants.notifyAll();
    }
  }
}
