package stored.linkanal;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

import bingo.util.SessionBuffer;

// Revision 1.4 Datenbankbasiert
// a) Set von URL mit entfernten Zero-outdegree-nodes (ZON)
// b) komplettes URL-Set
// Hits berechnet fuer:                                    a),b)
// PageRank berechnet fuer:
//    1) ohne personalisierung                             a),b)
//    2) uniformer per-Verteilung innerhalb Topic          a),b)
//    3) per-Verteilung innerhalb Topic abhaengig von SVM  a),b)
//    4) per-Verteilung bestimmt durch BlockRank           a),b)
//




/**
 * Hauptklasse auf der alle Methoden aufsetzen.
 * <p>
 * ZON in der Dokumentation bezeichnet Webseiten, die keine Links enthalten.
 *
 * @author Michael Koenig
 * @version 1.4, 08/21/03
 */
public class WebRanker
{
   private Webnode[] graph;
   private int urlcount;
   private int linkcount;
   private Connection oc;
   private int batchSize = 100; //??!!
   private int zeroctr;
   private static SessionBuffer session;
   
   /**
    * Konstruktor der Hauptklasse, der die Datenbank-Connection erzeugt
    *
    */
   public WebRanker()
   // WICHTIG: Hier Zugriffsdaten fr DB aendern !!!!
   {
      session = SessionBuffer.getInstance();
      oc = session.getDBConnection();
      batchSize = session.getBatchSize();
/*
     try
     {
       Class.forName("oracle.jdbc.driver.OracleDriver");
       String user = "handwerk1";
       String pwd = "00handwerk00";
       String host = "134.96.246.1";
       String service ="handwerk.world";
       String jdbcURL = "jdbc:oracle:thin:@";
       String db = "(DESCRIPTION=(ADDRESS_LIST=(ADDRESS=(PROTOCOL=TCP)(HOST="+host+")(PORT=1521)))(CONNECT_DATA=(SERVICE_NAME="+service+")(server=dedicated)))";
       oc = (Connection)DriverManager.getConnection(jdbcURL+db,user,pwd);
       oc.setAutoCommit( false );
     }
     catch (Exception e){e.printStackTrace();}
 */
      
   }
   
   /**
    * Objekt zur Speicherung der relevanten Informationen zu einer URL
    *
    */
   public class Webnode
   {
      long id;
      int in_deg = 0;
      int out_deg = 0;
      int nummer;
      ArrayList link_to;
      ArrayList link_from;
      
      /**
       * Konstruktor fuer Webnode-Objekte
       *
       * @param bez die ID der Webseite
       * @param in  die Anzahl der Seiten die zu uns linken
       * @param out die Anzahl der Seiten die von uns verlinkt werden
       * @param nr  die laufende Nummer der Webseite in unserer Architektur
       */
      public Webnode(long bez, int in, int out, int nr)
      {
         id = bez;
         in_deg=in;
         out_deg=out;
         link_from = new ArrayList(in);
         link_to = new ArrayList(out);
         nummer = nr;
      }
      
      /**
       * Konstruktor fuer Webnode-Objekte
       *
       * @param bez die ID der Webseite
       * @param in  die Anzahl der Seiten die zu uns linken
       * @param out die Anzahl der Seiten die von uns verlinkt werden
       */
      public Webnode(long bez, int in, int out)
      {
         id = bez;
         in_deg=in;
         out_deg=out;
         link_from = new ArrayList(in);
         link_to = new ArrayList(out);
      }
   }
   
   /**
    * Objekt zur Speicherung des Personalisierungsvektors bei BlockRank
    *
    */
   public class BlockPer
   {
      Integer cid;
      float value;
      // Da die Werte der wirklich vorhandenen Klassen spaeter normiert werden,
      // muessen die Summe der values eines Arrays nicht 1 sein. Eine cid sollte
      // pro Array jedoch nur einmal vorkommen, fehlende werden mit 0 aufgefuellt
      // falls keine der cids hier im BlockRank vorkommen sollte, wird eine
      // uniforme Verteilung angenommen
      
      /**
       * Konstruktor fuer BlockPer-Objekte
       *
       * @param id  Nummer der Blatt-Klasse in der Datenbank
       * @param val Wert der der Klasse zur Personalisierung zugeordnet wird
       */
      public BlockPer(int id, float val)
      {
         cid = new Integer(id);
         value = val;
      }
   }
   
   
   /**
    * Erzeugt aus dem ID-String die eindeute ID-Nummer der Webseite
    *
    * @param str ID-String der Webseite
    * @return ID-Nummer der Webseite
    */
   public Long parseIDObj(String str)
   {
      StringTokenizer st = new StringTokenizer(str, "-G");
      long dir = Long.parseLong(st.nextToken());
      long fil = Long.parseLong(st.nextToken());
      long doc = Long.parseLong(st.nextToken());
      long id = (dir << 40) + (fil << 32) + doc;
      return (new Long(id));
   }
   
   /**
    * Erzeugt aus dem ID-String die eindeute ID-Nummer der Webseite
    *
    * @param str ID-String der Webseite
    * @return ID-Nummer der Webseite
    */
   public long parseID(String str)
   {
      StringTokenizer st = new StringTokenizer(str, "-G");
      long dir = Long.parseLong(st.nextToken());
      long fil = Long.parseLong(st.nextToken());
      long doc = Long.parseLong(st.nextToken());
      long id = (dir << 40) + (fil << 32) + doc;
      return id;
   }
   
   
   
   //===================================================================================
   //                          GraphCreator
   
   /**
    * Klasse zur Erzeugung der fuer die Ranking-Algorithmen benoetigten Datenstrukturen
    * aus den Datenbank-Informationen. Es werden keine Seiten entfernt.
    */
   public class GraphCreator
   {
      HashMap graphtemp;
      
      //---------------prescan-----------------
      
      /**
       * Diese interne Funktion initialisiert die Basis-Datenstruktur.
       * Dabei werden keine Seiten entfernt.
       */
      private void prescan()
      {
         ResultSet rset;
         
         HashMap resultn, resultv;
         Webnode n;
         Iterator it;
         Integer tmp;
         Long nodeID;
         long id;
         int in,out,init;
         Statement stmt;
         
         resultv = new HashMap();
         resultn = new HashMap();
         try
         {
            stmt = oc.createStatement();
            rset = stmt.executeQuery("SELECT ID, count(distinct LINK_ID) FROM Links GROUP BY ID");
            // rset = stmt.executeQuery("SELECT DOCID, count(distinct LINKID) FROM Links GROUP BY DOCID");
            //      Liste: long [ID des Documents in DB], int [Anzahl der Seiten auf die ID verlinkt]
            while (rset.next())
            {
               resultn.put(new Long(rset.getLong(1)), new Integer(rset.getInt(2)));
            }
            // Bestimme die Anzahl der Nachfolger jeder Seite
            rset.close(); rset = null;
            
            rset = stmt.executeQuery("SELECT LINK_ID, count(distinct ID) FROM Links GROUP BY LINK_ID");
            //rset = stmt.executeQuery("SELECT LINKID, count(distinct DOCID) FROM Links GROUP BY LINKID");
            // Liste: long [ID des Documents in DB], int [Anzahl der Seiten die auf ID linken]
            while (rset.next())
            {
               resultv.put(new Long(rset.getLong(1)), new Integer(rset.getInt(2)));
            }
            // bestimme die Anzahl der Vorgaenger jeder Seite
            rset.close(); rset = null;
            stmt.close(); stmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         HashSet hs = new HashSet();
         hs.addAll(resultn.keySet());
         hs.addAll(resultv.keySet());
         init = (int) (Math.ceil((hs.size()+1)/0.75));
         graphtemp = new HashMap(init);
         // Wir intialisieren Graph ideal
         it = hs.iterator();
         while (it.hasNext())
         {
            urlcount++;
            // bestimme fortlaufende Nummerierung der Nodes
            nodeID = (Long) (it.next());
            // hole dir die naechste Node-ID als Long-Object
            tmp = (Integer)resultn.get(nodeID);
            // hole dir die Anzahl der Nachfolger
            if (tmp == null)
            { out = 0;}
            else
            { out = tmp.intValue();}
            // speichere sie in out
            tmp = (Integer)resultv.get(nodeID);
            // hole dir die Anzahl der Vorgaenger
            if (tmp == null)
            {in = 0;}
            else
            {in = tmp.intValue();}
            // speichere sie in in
            id = nodeID.longValue();
            // bestimme die ID der node als long
            n = new Webnode(id,in,out,urlcount);
            // erstelle Webnode mit diesen Infos
            graphtemp.put(nodeID, n);
            // haenge sie in den arry
         }
         hs = null;
         resultn = null;
         resultv = null;
         n = null;
      }
      
      //-----------importLinks--------------------
      
      /**
       * Diese interne Funktion uebertraegt die Link-Informationen in die Datenstruktur.
       * Dabei werden keine Seiten entfernt.
       */
      private void importLinks()
      {
         Webnode n1,n2;
         
         try
         {
            Statement stmt = oc.createStatement();
            ResultSet rset = stmt.executeQuery("SELECT DISTINCT ID, LINK_ID FROM Links");
            // ResultSet rset = stmt.executeQuery("SELECT DISTINCT DOCID, LINKID FROM Links");
            // Liste: long [ID des Documents in DB], long [ID des Documents auf das der Link zeigt in DB]
            while (rset.next())
            {
               n1 = (Webnode) (graphtemp.get(new Long(rset.getLong(1))));
               // besorge dir die Webnode von der ein Link ausgeht
               n2 = (Webnode) (graphtemp.get(new Long(rset.getLong(2))));
               // nimm die Webnode die verlinkt wird
               (n1.link_to).add(new Integer(n2.nummer));
               // haenge die 2.Url in die Folge von Seiten auf die wir linken
               (n2.link_from).add(new Integer(n1.nummer));
               // haenge die 1.Url in die Folge von Seiten von denen wir Links haben
               linkcount++;
            }
            rset.close(); rset = null;
            stmt.close(); stmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
      }
      
      /**
       * Diese interne Funktion aendert die Speicherung der Nodes, um die Berechnung
       * spaeter effizienter zu gestalten. Es werden keine Seiten entfernt.
       */
      private void convert()
      {
         Iterator it;
         Webnode node;
         graph = new Webnode[urlcount];
         // wir initialisieren den Array fuer den Graphen
         it = (graphtemp.values()).iterator();
         while (it.hasNext())
         {
            node = (Webnode) (it.next());
            if (node.out_deg == 0)
            {zeroctr++;}
            // merke Anzahl Nodes mit outdeg 0
            graph[(node.nummer)-1] = node;
            it.remove();
         }
         graphtemp = null;
      }
      
      /**
       * Diese interne Funktion fuehrt die Erstellung der Datenstruktur durch.
       * Es werden keine Seiten entfernt.
       */
      public void constructGraph()
      {
         zeroctr = 0;
         urlcount = 0;
         linkcount = 0;
         System.out.println("Start");
         prescan();
         System.out.println("begin Import");
         importLinks();
         System.out.println("begin convert");
         convert();
      }
   }
   
   //===================================================================================
   //                          GraphCreator PAGERANK
   
   /**
    * Klasse zur Erzeugung der fuer die Ranking-Algorithmen benoetigten Datenstrukturen
    * aus den Datenbank-Informationen
    * ZONs werden iterativ entfernt
    */
   public class GraphCreator_zd
   {
      HashMap graphtemp;
      HashSet zerotest;
      
      //---------------prescan_zd-----------------
      
      /**
       * Diese interne Funktion initialisiert die Basis-Datenstruktur.
       * ZONs werden iterativ entfernt, die 1.Iteration erfolgt hier.
       */
      private void prescan_zd()
      {
         ResultSet rset;
         HashMap resultn, resultv;
         Webnode n;
         Iterator it;
         Integer tmp;
         Long nodeID;
         long id;
         int in,out,init;
         Statement stmt;
         
         resultv = new HashMap();
         resultn = new HashMap();
         try
         {
            stmt = oc.createStatement();
            rset = stmt.executeQuery("SELECT ID, count(distinct LINK_ID) FROM Links GROUP BY ID");
            //        rset = stmt.executeQuery ("SELECT DOCID, count(distinct LINKID) FROM Links GROUP BY DOCID");
            // Liste: long [ID des Documents in DB], int [Anzahl der Seiten auf die ID verlinkt]
            while (rset.next())
            {
               resultn.put(new Long(rset.getLong(1)), new Integer(rset.getInt(2)));
            }
            // Bestimme die Anzahl der Nachfolger jeder Seite
            rset.close(); rset = null;
            
            rset = stmt.executeQuery("SELECT LINK_ID, count(distinct ID) FROM Links GROUP BY LINK_ID");
            // rset = stmt.executeQuery("SELECT LINKID, count(distinct DOCID) FROM Links GROUP BY LINKID");
            // Liste: long [ID des Documents in DB], int [Anzahl der Seiten die zu ID links haben]
            while (rset.next())
            {
               resultv.put(new Long(rset.getLong(1)), new Integer(rset.getInt(2)));
            }
            // bestimme die Anzahl der Vorgaenger jeder Seite
            rset.close(); rset = null;
            stmt.close(); stmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         HashSet hs = new HashSet();
         hs.addAll(resultn.keySet());
         hs.addAll(resultv.keySet());
         init = (int) (Math.ceil((hs.size()+1)/0.75));
         graphtemp = new HashMap(init);
         // Wir intialisieren Graph ideal
         it = hs.iterator();
         while (it.hasNext())
         {
            urlcount++;
            // bestimme fortlaufende Nummerierung der Nodes
            nodeID = (Long) (it.next());
            // hole dir die naechste Node-ID als Long-Object
            tmp = (Integer)resultn.get(nodeID);
            // hole dir die Anzahl der Nachfolger
            if (tmp == null)
            { out = 0;}
            else
            { out = tmp.intValue();}
            // speichere sie in out
            tmp = (Integer)resultv.get(nodeID);
            // hole dir die Anzahl der Vorgaenger
            if (tmp == null)
            {in = 0;}
            else
            {in = tmp.intValue();}
            // speichere sie in in
            id = nodeID.longValue();
            // bestimme die ID der node als long
            n = new Webnode(id,in,out,urlcount);
            // erstelle Webnode mit diesen Infos
            graphtemp.put(nodeID, n);
            // haenge sie in den arry
            if (out==0)
            {
               zeroctr++;
               zerotest.add(new Long(id));
            }
         }
         hs = null;
         resultn = null;
         resultv = null;
         n = null;
      }
      
      //-----------importLinks_zd--------------------
      
      /**
       * Diese interne Funktion uebertraegt die Link-Informationen in die Datenstruktur.
       * ZONs werden iterativ entfernt.
       */
      private void importLinks_zd()
      {
         Webnode n1,n2;
         Long dummy, dummy2;
         
         try
         {
            Statement stmt = oc.createStatement();
            ResultSet rset = stmt.executeQuery("SELECT DISTINCT ID, LINK_ID FROM Links");
            // ResultSet rset = stmt.executeQuery("SELECT DISTINCT DOCID, LINKID FROM Links");
            // Liste: long [ID des Documents in DB], long [ID des Documents auf das der Link zeigt in DB]
            while (rset.next())
            {
               dummy = new Long(rset.getLong(1));
               dummy2 = new Long(rset.getLong(2));
               if (!(zerotest.contains(dummy2)))
               {
                  linkcount++;
                  n1 = (Webnode) (graphtemp.get(dummy));
                  // besorge dir die Webnode von der ein Link ausgeht
                  n2 = (Webnode) (graphtemp.get(dummy2));
                  // nimm die Webnode die verlinkt wird
                  (n1.link_to).add(dummy2);
                  // haenge die 2.Url in die Folge von Seiten auf die wir linken
                  (n2.link_from).add(dummy);
                  // haenge die 1.Url in die Folge von Seiten von denen wir Links haben
               }
               else
               {
                  n1 = (Webnode) (graphtemp.get(dummy));
                  (n1.out_deg)--;
                  //andernfalls reduzieren wir den outdeg um 1 der Node;
                  (n1.link_to).trimToSize();
               }
            }
            rset.close(); rset = null;
            stmt.close(); stmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         n1 = null;
         n2 = null;
         zerotest=null;
      }
      
      /**
       * Diese interne Funktion entfernt iterativ alle Webseiten ohne Links.
       */
      private void graphtrim_zd()
      {
         Iterator it;
         Webnode node,node2;
         ArrayList al;
         int i;
         Long id,id2;
         
         while (zeroctr!=0)
         {
            zeroctr =0;
            zerotest = new HashSet();
            it = (graphtemp.values()).iterator();
            while (it.hasNext())
            {
               //  wir durchlaufen alle Nodes
               node = (Webnode) (it.next());
               if (node.out_deg==0)
                  // bei einer zu entfernenden Node
               {
                  zeroctr++;
                  zerotest.add(new Long(node.id));
                  //wir fuegen alle Elemente die zu entfernen sind in HashSet zerotest
               }
            }
            it = zerotest.iterator();
            while (it.hasNext())
            {
               // wir durchlaufen alle Nodes die 0 sind
               id = (Long) (it.next());
               node = (Webnode) (graphtemp.get(id));
               al = node.link_from;
               graphtemp.remove(id);
               //wir holen uns den Vektor von URLs die zu uns linken
               for (i=0; i<al.size();i++)
               {
                  id2 = (Long) (al.get(i));
                  node2 = (Webnode) (graphtemp.get(id2));
                  (node2.link_to).remove(id);
                  (node2.out_deg)--;
               }
            }
            node = null;
            al = null;
            zerotest=null;
         }
      }
      
      /**
       * Diese interne Funktion aendert die Speicherung der Nodes um die Berechnung
       * spaeter effizienter zu gestalten. ZONs werden iterativ entfernt.
       */
      private void convert_zd()
      {
         Iterator it;
         ArrayList al;
         int ctr=0;
         Webnode node,newnode;
         linkcount = 0;
         int i;
         Integer num;
         Long id;
         HashMap test;
         
         test = new HashMap();
         urlcount=graphtemp.size();
         graph = new Webnode[urlcount];
         // wir initialisieren den Array fuer den Graphen
         it = (graphtemp.values()).iterator();
         while (it.hasNext())
         {
            ctr++;
            node = (Webnode) (it.next());
            node.nummer = ctr;
            //nummeriere die Nodes
            test.put(new Long(node.id), new Integer(ctr));
         }
         //nummeriere die Nodes
         it = (graphtemp.values()).iterator();
         while (it.hasNext())
            //durchlaufe alle Nodes wieder
         {
            node = (Webnode) (it.next());
            newnode = new Webnode(node.id,node.in_deg,node.out_deg,node.nummer);
            // erzeuge neue Node
            al=node.link_to;
            for (i=0; i<al.size();i++)
            {
               id = (Long) (al.get(i));
               linkcount++;
               (newnode.link_to).add((Integer) test.get(id));
            }
            // schreibe die Array-Position anstelle der ID in die neue link_to - Liste
            al=node.link_from;
            for (i=0; i<al.size();i++)
            {
               id = (Long) (al.get(i));
               (newnode.link_from).add((Integer) test.get(id));
            }
            // schreibe die Array-Position anstelle der ID in die neue link_from - Liste
            graph[(newnode.nummer)-1] = newnode;
            // haenge die neue Node in den Graph-array
            it.remove();
            // entferne die alte aus der webliste
            node=null;
         }
         graphtemp = null;
         node = null;
         newnode = null;
         al = null;
         test = null;
      }
      
      /**
       * Diese interne Funktion fuehrt die Erstellung der Datenstruktur durch.
       * ZONs werden dabei iterativ entfernt.
       */
      public void constructGraph_zd()
      {
         urlcount = 0;
         linkcount = 0;
         zeroctr=0;
         zerotest = new HashSet();
         prescan_zd();
         importLinks_zd();
         graphtrim_zd();
         convert_zd();
      }
      
   }
   
   //===================================================================================
   //                          BlockRankCalculator
   
   /**
    * Diese Klasse stellt alle Methoden, die zur Berechnung des PageRank mit Hilfe
    * von topic-orientiertem BlockRank benoetigt werden.
    */
   public class BlockRankCalculator
   {
      float[] ranktemp,values,rank,per,init,zeroinit;
      float alpha = (float) 0.2;
      float nalpha = 1-alpha;
      float zerosum;
      int[] point, index;
      int[] Bpoint, Bindex, Bzeroind; //alle fuer BR nicht aendern !!
      float[] localPR;		  //nicht aendern !! nach Berechnung
      float[] Bvalues;                //Bvalues fuer BR, nicht aendern
      int topicctr;
      HashMap topictest;
      int[] topic;
      
      //-------------------------------------------------------------------
      
      /**
       * Dummy-Implementierung zum Speichern des BlockRanks in die Datenbank.
       * Die Anpassung an das vorhandene Schema sollte hier vorgenommen werden.
       *
       * @param num Nummer des Personalisierungs-Vektors im Stack
       */
      public void storeBR(int num)
      {
         
/*
     int count=0;
     try
     {
// !!! HIER muesste irgendwie die Nummer der aktuellen Personaliserung verwendet werden !!!!
       PreparedStatement pstmt = oc.prepareStatement("UPDATE BingoDocuments SET PageRank=? Where ID=?");
       for (int i=0; i<graph.length; i++)
       {
         pstmt.setDouble(1, (double)rank[i]);
         pstmt.setLong(2, ((Webnode) (graph[i])).id);
         pstmt.addBatch();
         if (++count % batchSize == 0){pstmt.executeBatch();}
       }
       pstmt.executeBatch();
       pstmt.close();
       oc.commit();
     }
     catch (SQLException e){e.printStackTrace();}
 */
/*
     Webnode node;
     int i;
     try
     {
       FileWriter f1 = new FileWriter("BlockRank_"+anz+".dat);
       for (i=0;i<urlcount;i++)
       {
         node = (Webnode) (graph[i]);
         f1.write(node.id+": "+rank[i]+"\n");
       }
       f1.close();
     }
     catch  (IOException e) {System.out.println("Error writing BlockRank-Result");}
 */
      }
      
      //--------------------------------------------------------------------
      //  Volle Version BlockRank
      //--------------------------------------------------------------------
      
      /**
       * Diese interne Funktion berechnet den lokalen PageRank von URLs
       * und erstellt die Blockmatrix. Sie ist verwendbar mit allen Webseiten-Sets.
       *
       * @param th   Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       */
      public void prepareBlockRank(int th, int iter)
      {
         int i,j,temp, topicfrom, topicto,id,laufnr,cellctr;
         Webnode node;
         ArrayList al;
         float[] prsum,valpertopic;
         float value;
         Statement stmt;
         ResultSet rset;
         HashMap topicnr;
         String quer;
         Integer dummy;
         int[] zerotemp;
         float[][] matrix;
         Integer top;
         
         //-----Start Calculation PageRank-----
         per = new float[urlcount];
         value = (float) 1/urlcount;
         // fuer den ersten Durchlauf gehen wir von einem uniformen Per-Vektor aus (alpha)*1/url
         for (i=0; i<urlcount;i++)
         {per[i]=value;}
         init = per;
         calcPRper(th,iter,false);
         // false = wir initialisieren ZONs mit 1/urlcount uniform
         per = null;
         init = null;
         // nach ausfuehren befindet sich PageRank in globaler Variable rank
         //-----Bestimmung der Topic der URL und der Anzahl verschiedener Topics------
         temp = (int) (Math.ceil((urlcount+1)/0.75));
         topicnr = new HashMap(temp);
         topictest = new HashMap();
         topic = new int[urlcount];
         try
         {
            stmt = oc.createStatement();
            quer = "SELECT DISTINCT ID,class_id FROM bingodocuments";
            // Liste long [ID des Documents in DB], int [Nummer der Klasse des Dokuments]
            rset = stmt.executeQuery(quer);
            while (rset.next())
            {
               topicnr.put(new Long(rset.getLong(1)),new Integer(rset.getInt(2)));
               // haenge alle IDs und ihr Topicnr in Hashmap
            }
            rset.close(); rset = null;
            stmt.close(); stmt = null;
            
         }
         catch (SQLException e)
         {e.printStackTrace();}
         topicctr = 0;
         for (i=0; i<urlcount;i++)
         {
            node = (Webnode) (graph[i]);
            top = (Integer)(topicnr.get(new Long(node.id)));
            if (topictest.containsKey(top))
            {
               topic[i] = ((Integer) (topictest.get(top))).intValue();
               // wenn Topic schon laufende Nr hat hole sie
            }
            else
            {
               dummy = new Integer(topicctr);
               topictest.put(top,dummy);
               // haenge den DB Code der Topic und die interne Nr ins Hashmap
               topic[i] = topicctr;
               topicctr++;
            }
         }
         topicnr = null;
         //-----Normalisiere den PageRank fuer die Topics
         prsum = new float[topicctr];
         valpertopic = new float[topicctr];
         for (i=0;i<topicctr;i++)
         {valpertopic[i]= (float) 0;}
         for (i=0; i<topicctr;i++)
         {prsum[i]=0;}
         for (i=0; i<urlcount;i++)
         {
            prsum[topic[i]] += rank[i];
            //addiere die PageRanks fuer die Topics auf
            valpertopic[topic[i]] += 1;
            //zaehle die Anzahl der URL per topic
         }
         for (i =0; i<topicctr;i++)
         {valpertopic[i]=valpertopic[i]/urlcount;}
         // wir berechnen die Anteil der Urls einer topic an Gesamtmenge URLs
         localPR = new float[urlcount];
         for (i=0;i<urlcount;i++)
         {
            localPR[i] = rank[i] / prsum[topic[i]];
            // wir dividieren den Rank durch den GesamtRank der Topic und normalisieren damit
            // DivZero kann nicht vorkommen, da nur topics die auch URL haben also auch PR
         }
         prsum = null;
         rank = null;
         // ---> localPR enthaelt localen PageRank innerhalb topic
         //----- Initialisiere BlockMatrix mit 0
         matrix = new float[topicctr][topicctr];
         for (i=0;i<topicctr;i++)
         {
            for (j=0;j<topicctr;j++)
            {matrix[i][j]=0;}
         }
         //-----Erstellung der Matrix------
         for (i=0;i<urlcount;i++)
            //durchlaufe alle Nodes
         {
            node = (Webnode) (graph[i]);
            topicfrom = topic[i];
            //hole dir das topic der Node
            if (node.out_deg !=0)
            {
               value = (float) 1/node.out_deg;
               al = node.link_to;
               // hole dir die Liste der Seiten auf die wir linken
               for (j=0; j<al.size(); j++)
                  // durchlaufe die ArrayList
               {
                  id = ((Integer) (al.get(j))).intValue()-1;
                  topicto = topic[id];
                  // bestimme die Topic der Seite zu der wir linken
                  matrix[topicto][topicfrom] += (localPR[i] * value);
                  //erhoehe den entsprechenden Matrixeintrag um den mit unserem lokalenPR gewichteten outdeg
               }
            }
            else
               // wenn wir eine Node mit outdeg 0 haben
            {
               for (j=0;j<topicctr;j++)
               {
                  matrix[j][topicfrom] += (localPR[i] * valpertopic[j]);
                  //geben wir an jede Topic Rank im Wert von unserem lokalen PR * Anzahl URL in Topic/Gesamtanzahl URL
               }
            }
         }
         valpertopic = null;
         node = null;
         al = null;
         //-----Zaehle die Anzahl MatrixElemente != 0
         zerotemp = new int[topicctr];
         laufnr = 0;
         cellctr = 0;
         for (i=0;i<topicctr;i++)
         {
            temp=0;
            for (j=0;j<topicctr;j++)
            {
               if (matrix[j][i]!=0)
               {cellctr++;temp++;}
            }
            if (temp==0)
            {zerotemp[laufnr]=i;laufnr++;}
            // wenn unsere Spalte keinen Wert != 0 enthaelt, merken wir uns den index
         }
         Bzeroind = new int[laufnr];
         for (i=0;i<laufnr;i++)
         {Bzeroind[i]=zerotemp[i];}
         // wir speichern die Indexe der ZON-Spalten im Bzeroind-Array
         zerotemp = null;
         //-----Konvertierung der Matrix in RCS-Format
         laufnr = 0;
         Bpoint = new int[topicctr+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         Bindex = new int[cellctr];
         Bvalues = new float[cellctr];
         for (i=0; i<topicctr; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            Bpoint[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            for (j=0; j<topicctr; j++)
               // durchlaufe alle Spalten
            {
               value = matrix[i][j];
               if (value !=0)
               {
                  Bvalues[laufnr]=value;
                  // Trage den Wert im Values Array CRS ein
                  Bindex[laufnr]=j;
                  // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
                  laufnr++;
                  //erhoehe die laufnr;
               }
            }
         }
         Bpoint[topicctr]=laufnr;
         matrix = null;
      }
      
      
      //--------------------------------------------------------------------
      
      /**
       * Diese interne Funktion berechnet den personalisierten PageRank
       * der Topics und verwendet diesen zur Berechnung des personalisierten
       * PageRank der Webseiten. Die Funktion kann mehrfach nach prepareBlockRank
       * aufgerufen werden und ist verwendbar mit allen Webseiten-Sets.
       * <p>
       * Die Variablen Bpoint, Bvalues, Bindex, Bzeroind und localPR
       * duerfen nicht mehr veraendert werden !
       *
       * @param th    Anzahl der Threads die verwendet werden soll
       * @param iter  Anzahl der Iterationen die der Algorithmus durchlaufen soll
       * @param perob Array von BlockPer Objekten mit Werten zur Personalisierung
       */
      public void calcFinal(int th, int iter, BlockPer[] perob)
      {
         float[] pers;
         float sum;
         int i, topicanz,temp;
         BlockPer pero;
         
         sum = 0;
         topicanz = topictest.size();
         pers = new float[topicanz];
         for (i=0; i<topicanz;i++)
         {pers[i] = (float) 0;}
         for (i=0;i<perob.length;i++)
            // wir durchlaufen den Array von BlockPer-Objekten
         {
            pero = (BlockPer)	perob[i];
            if (topictest.containsKey(pero.cid))
               // wenn die Klasse auch in unserem BlockRank vorkommt
            {
               temp = ((Integer) (topictest.get(pero.cid))).intValue();
               // holen wir die laufende Nummer der Klasse
               pers[temp]=pero.value;
               // und setzen unseren pers Wert an entsprechender Stelle auf den Wert
               sum = sum + pero.value;
               // und berechnen die gesamt-Summe aller verwendeten Per-Werte
            }
         }
         if (sum==0)
         {for (i=0; i<topicanz;i++)
          {pers[i] = (float) 1/topicanz;} }
         // wenn keiner der angegebenen Klassen in unserem BlockRank ist initialisieren wir uniform
         else
         {for (i=0; i<topicanz;i++)
          {pers[i] = pers[i]/sum;}}
         // ansonsten normalisieren wir die Werte auf 1
         per = (float[]) pers.clone();
         calcBRper(th, iter, true);
         // Berechne den PageRank der einzelnen Topics (unter Verwendung Per zum auffuellen
         // zero-outdeg nodes) (false => uniforme link-verteilung von diesen Topics)
         init = new float[urlcount];
         for (i=0; i<urlcount;i++)
         {init[i] = rank[topic[i]] * localPR[i];}
         //  berechne estimated global PR Seite aus Rank[Topic] * local PR Seite als INIT
         per = new float[urlcount];
         for (i=0; i<urlcount;i++)
         {per[i] = pers[topic[i]] * localPR[i];}
         //  berechne per-Wert URL als per-Wert topic * localPR der seite
         calcPRper(th, iter, true);
         // Berechne den wirklichen BlockRank jeder Seite (unter Verwendung Per zum auffuellen
         // zero-outdeg nodes) (false => uniforme link-verteilung von diesen Nodes)
      }
      
      
      //--------------------------------------------------------------------
      
      /**
       * Klasse von Hilfsthreads der Methode calcBRper
       */
      private class CalccBR implements Runnable
      {
         int srow; //Zeile zum starten
         int erow; //Zeile zum enden
         boolean mod;
         
         /**
          * Konstruktor des Hilfsthread der Methode calcBRper
          *
          * @param nr   laufende Nummer des Threads
          * @param l    Anzahl der zu berechnenden Zeilen
          * @param mode true = verwende Pers-Vektor fuer ZONs, false = verwende uniforme Verteilung
          */
         public CalccBR(int nr, int l,boolean mode)
         {
            srow = nr*l;
            erow = srow+l;
            mod = mode;
         }
         
         /**
          * Hilfsthread der Methode calcBRper
          */
         public void run()
         {
            int i,j;
            float temp;
            
            for (i=srow; i<erow && i<topicctr; i++)
               // fuer alle Zeilen Matrix in unserem Zeilenbereich
            {
               temp=0;
               for (j=Bpoint[i];j<Bpoint[i+1]; j++)
                  //fuer j = value index aller Werte der Zeile
               {
                  temp = temp + Bvalues[j]*rank[Bindex[j]];
                  // die Addition: Element Matrix * Rank(Spalte des Werts)
               }
               if (mod)
               {
                  ranktemp[i]= nalpha * (temp + (zeroinit[i]*zerosum));
                  // Im fall einer Auffuellung von Zero-Nodes mit dem per-Vektor multiplizieren wir die
                  // Summe der Ranks der Zero-out-deg Nodes mit dem Per-Wert der Topic
               }
               else
               {ranktemp[i] = nalpha * (temp + zerosum);}
               // Im falle keine exakte Berechnung haben alle Spalten mit 0-outdeg Wert 1/Anzahl topic
               // und zerosum wurde bereits damit multipliziert
            }
         }
      }
      
      
      /**
       * Diese interne Funktion berechnet den personalisierten PageRank
       * der BlockMatrix aus den Topics.
       * Dazu werden die Variablen Bvalues, Bpoint, Bindex, Bzeroind und Per benoetigt.
       * Das Ergebnis wird in der Variable Rank gespeichert.
       * Die Methode kann mit allen Webseiten-Sets verwendet werden.
       *
       * @param th   Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       * @param mode true = verwende Pers-Vektor fuer ZONs, false = verwende uniforme Verteilung
       */
      public void calcBRper(int th, int iter, boolean mode)
      {
         int maxiter = iter;
         float temp;
         int laenge;
         int threads = th;
         Thread[] t;
         int i,j;
         
         temp = (float) (1 / topicctr);
         zeroinit = new float[topicctr];
         rank = new float[topicctr];
         for (i=0; i<topicctr; i++)
         {
            rank[i]=temp;
            // initialisiere Rank-Vektor mit 1/Anzahl Topics
            zeroinit[i] = per[i];
            per[i]= per[i]*alpha;
            // multipliziere den per-Vektor mit alpha
         }
         
         laenge = (int) Math.ceil(topicctr / (double) threads);
         // bestimme Partitionierung fuer Threads
         
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            zerosum = 0;
            for (j=0; j<Bzeroind.length; j++)
            {zerosum = zerosum + rank[Bzeroind[j]];}
            if (!mode)
            {zerosum = zerosum / topicctr;}
            // fuer den Fall nicht-exakt Initialisierung Zero-outdeg-Nodes mit 1/Anzahl Topics
            ranktemp = new float[topicctr];
            try
            {
               
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new CalccBR(j,laenge,mode));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<topicctr;j++)
            {ranktemp[j] = ranktemp[j]+ per[j];}
            // addiere den Per-Vektor
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
/*   float x=0;
     for (i=0;i<topicctr;i++){x=x+rank[i];}
     System.out.println(x); 			*/
         
         
      }
      
      //--------------------------------------------------------------------
      
      /**
       * Klasse von Hilfsthreads der Methode calcPRper
       */
      private class CalccPR implements Runnable
      {
         int srow; //Zeile zum starten
         int erow; //Zeile zum enden
         boolean mod;
         
         /**
          * Konstruktor des Hilfsthread der Methode calcBRper
          *
          * @param nr   laufende Nummer des Threads
          * @param l    Anzahl der zu berechnenden Zeilen
          * @param mode true = verwende Pers-Vektor fuer ZONs, false = verwende uniforme Verteilung
          */
         public CalccPR(int nr, int l, boolean mode)
         {
            srow = nr*l;
            erow = srow+l;
            mod = mode;
         }
         
         /**
          * Hilfsthread der Methode calcBRper
          */
         public void run()
         {
            int i,j;
            float temp;
            
            for (i=srow; i<erow && i<urlcount; i++)
               // fuer alle Zeilen Matrix in unserem Zeilenbereich
            {
               temp=0;
               for (j=point[i];j<point[i+1]; j++)
                  //fuer j = value index aller Werte der Zeile
               {
                  temp = temp + values[j]*rank[index[j]];
                  // die Addition: Element Matrix * Rank(Spalte des Werts)
               }
               if (mod)
               {
                  ranktemp[i]= nalpha * (temp + (zeroinit[i]*zerosum));
                  // Im fall einer Auffuellung von Zero-Nodes mit dem per-Vektor multiplizieren wir die
                  // Summe der Ranks der Zero-out-deg Nodes mit dem Per-Wert der Url
               }
               else
               {ranktemp[i] = nalpha * (temp + zerosum);}
               // Im falle keine exakte Berechnung haben alle Spalten mit 0-outdeg Wert 1/urlcount
               // und zerosum wurde bereits damit multipliziert
            }
         }
      }
      
      /**
       * Diese interne Funktion berechnet den (personalisierten) PageRank.
       * Dazu werden die Variablen Init und Per benoetigt.
       * Das Ergebnis wird in der Variable Rank gespeichert.
       * Die Methode ist mit allen Webseiten-Sets verwendbar.
       *
       * @param th   Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       * @param mode true = verwende Pers-Vektor fuer ZONs, false = verwende uniforme Verteilung
       */
      public void calcPRper(int th, int iter, boolean mode)
      {
         int maxiter = iter;
         float temp;
         int laenge;
         int threads = th;
         Thread[] t;
         Iterator it;
         Webnode node;
         int i,j;
         ArrayList al;
         int id1, id2;
         int[] outde,zeroind;
         int laufnr,topctr;
         
         
         outde = new int[urlcount];
         zeroind = new int[zeroctr];
         zeroinit = new float[urlcount];
         rank = init;
         
         j=0;
         for (i=0; i<urlcount; i++)
         {
            node = (Webnode) (graph[i]);
            outde[i] = node.out_deg;
            // erstelle Array der in Feld i den Outdeg von URL Nr i+1 hat
            if (node.out_deg==0)
            {zeroind[j]=i; j++;}
            zeroinit[i]=per[i];
            per[i]=per[i]*alpha;
         }
         if (!mode)
         {zeroinit = null;}
         // im falle einer nicht-exakten Berechnung brauchen wir zeroinit nicht
         laufnr = 0;
         point = new int[urlcount+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         index = new int[linkcount];
         values = new float[linkcount];
         for (i=0; i<urlcount; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            point[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            node = (Webnode) (graph[i]);
            al = node.link_from;
            // hole Vektor mit Liste aller Seiten-ID die auf uns verlinken
            java.util.Collections.sort(al);
            //wir sortieren die ArrayList aufsteigend
            for (j=0; j<al.size(); j++)
               // durchlaufe den Vektor
            {
               id2 = ((Integer) (al.get(j))).intValue()-1;
               // hol dir die Matrix-ID id2 der URL die auf uns zeigt
               temp = (float) 1 / (outde[id2]);
               // bestimme den Wert den uns die Seite gibt
               values[laufnr]=temp;
               // Trage den Wert im Values Array CRS ein
               index[laufnr]=id2;
               // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
               laufnr++;
               //erhoehe die laufnr;
            }
         }
         point[urlcount]=laufnr;
         // speichere die erste nicht mehr benoetigte Stelle in das letzte Feld des Point array
         outde = null;
         node = null;
         al = null;
         laenge = (int) Math.ceil(urlcount / (double) threads);
         // bestimme Partitionierung fuer Threads
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            zerosum = 0;
            for (j=0; j<zeroctr; j++)
            {zerosum = zerosum + rank[zeroind[j]];}
            if (!mode)
            {zerosum = zerosum / urlcount;}
            // fuer den Fall nicht-exakt Initialisierung Zero-outdeg-Nodes mit 1/Anzahl URLs
            ranktemp = new float[urlcount];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new CalccPR(j,laenge,mode));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<urlcount;j++)
            {ranktemp[j] = ranktemp[j] + per[j];}
            // addiere den Per-Vektor
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         values = null;
         index = null;
         point = null;
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
/*     float x=0;
     for (i=0;i<urlcount;i++){x=x+rank[i];}
     System.out.println(x); 			*/
      }
      
      //----------------------------------------------------------
      // ZD-VERSION
      //----------------------------------------------------------
      
      
      //-------------------------------------------------------------------
      
      /**
       * Diese interne Funktion berechnet den lokalen PageRank von URLs
       * und erstellt die Blockmatrix. Sie ist nur mit Webseiten-Sets ohne ZONs
       * verwendbar.
       *
       * @param th   Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       */
      public void prepareBlockRank_zd(int th, int iter)
      {
         int i,j,temp, topicfrom, topicto,id,laufnr,cellctr;
         Webnode node;
         ArrayList al;
         float[] prsum;
         float value;
         Statement stmt;
         ResultSet rset;
         HashMap topicnr;
         String quer;
         Integer dummy;
         float[][] matrix;
         Integer top;
         
         //-----Start Calculation PageRank-----
         per = new float[urlcount];
         value = (float) 1/urlcount;
         // fuer den ersten Durchlauf gehen wir von einem uniformen Per-Vektor aus (alpha)*1/url
         for (i=0; i<urlcount;i++)
         {per[i]=value;}
         init = per;
         calcPRper_zd(th,iter);
         per = null;
         init = null;
         // nach ausfuehren befindet sich PageRank in globaler Variable rank
         //-----Bestimmung der Topic der URL und der Anzahl verschiedener Topics------
         temp = (int) (Math.ceil((urlcount+1)/0.75));
         topicnr = new HashMap(temp);
         topictest = new HashMap();
         topic = new int[urlcount];
         try
         {
            stmt = oc.createStatement();
            quer = "SELECT DISTINCT ID,class_id FROM bingodocuments";
            // Liste long [ID des Documents in DB], int [Nummer der Klasse des Dokuments]
            rset = stmt.executeQuery(quer);
            while (rset.next())
            {
               topicnr.put(new Long(rset.getLong(1)),new Integer(rset.getInt(2)));
               // haenge alle IDs und ihre DB-ID in Hashmap
            }
            rset.close(); rset = null;
            stmt.close(); stmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         
         topicctr = 0;
         for (i=0; i<urlcount;i++)
         {
            node = (Webnode) (graph[i]);
            top = (Integer)(topicnr.get(new Long(node.id)));
            if (topictest.containsKey(top))
            {
               topic[i] = ((Integer) (topictest.get(top))).intValue();
               // hole dir fortlaufende interne Nr der Topic;
            }
            else
            {
               dummy = new Integer(topicctr);
               topictest.put(top,dummy);
               // haenge den DB Code der Topic und die interne Nr ins Hashmap
               topic[i] = topicctr;
               topicctr++;
            }
         }
         topicnr = null;
         //-----Normalisiere den PageRank fuer die Topics
         prsum = new float[topicctr];
         for (i=0; i<topicctr;i++)
         {prsum[i]=0;}
         for (i=0; i<urlcount;i++)
         {
            prsum[topic[i]] += rank[i];
            //addiere die PageRanks fuer die Topics auf
         }
         localPR = new float[urlcount];
         for (i=0;i<urlcount;i++)
         {
            localPR[i] = rank[i] / prsum[topic[i]];
            // wir definieren den Rank durch den GesamtRank der Topic und normalisieren damit
            // DivZero kann nicht vorkommen, da nur topics die auch PR haben
         }
         prsum = null;
         rank = null;
         // ---> localPR enthaelt localen PageRank innerhalb topic
         //----- Initialisiere BlockMatrix mit 0
         matrix = new float[topicctr][topicctr];
         for (i=0;i<topicctr;i++)
         {
            for (j=0;j<topicctr;j++)
            {
               matrix[i][j]=0;
            }
         }
         //-----Erstellung der Matrix------
         for (i=0;i<urlcount;i++)
            //durchlaufe alle Nodes
         {
            node = (Webnode) (graph[i]);
            topicfrom = topic[i];
            //hole dir das topic der Node
            value = (float) 1/node.out_deg;
            // Division kein Problem hier da nur Nodes mit Outdeg != 0
            al = node.link_to;
            // hole dir die Liste der Seiten auf die wir linken
            for (j=0; j<al.size(); j++)
               // durchlaufe die ArrayList
            {
               id = ((Integer) (al.get(j))).intValue()-1;
               topicto = topic[id];
               // bestimme die Topic der Seite zu der wir linken
               matrix[topicto][topicfrom] += (localPR[i] * value);
               //erhoehe den entsprechenden Matrixeintrag um den mit unserem lokalenPR gewichteten outdeg
            }
         }
         node = null;
         al = null;
         //-----Zaehle die Anzahl MatrixElemente != 0
         cellctr = 0;
         for (i=0;i<topicctr;i++)
         {
            for (j=0;j<topicctr;j++)
            {
               if (matrix[i][j]!=0)
               {cellctr++;}
            }
         }
         //-----Konvertierung der Matrix in RCS-Format
         laufnr = 0;
         Bpoint = new int[topicctr+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         Bindex = new int[cellctr];
         Bvalues = new float[cellctr];
         for (i=0; i<topicctr; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            Bpoint[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            for (j=0; j<topicctr; j++)
               // durchlaufe alle Spalten
            {
               value = matrix[i][j];
               if (value !=0)
               {
                  Bvalues[laufnr]=value;
                  // Trage den Wert im Values Array CRS ein
                  Bindex[laufnr]=j;
                  // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
                  laufnr++;
                  //erhoehe die laufnr;
               }
            }
         }
         Bpoint[topicctr]=laufnr;
         matrix = null;
      }
      
      //-------------------------------------------------------------------
      
      /**
       * Diese interne Funktion berechnet den personalisierten PageRank
       * der Topics und verwendet diesen zur Berechnung des personalisierten
       * PageRank der Webseiten.
       * Die Funktion kann mehrfach nach prepareBlockRank aufgerufen werden
       * unt ist nur mit Webseiten-Sets ohne ZONs verwendbar.
       * <p>
       * Die Variablen Bpoint, Bvalues, Bindex, Bzeroind und localPR
       * duerfen nicht mehr veraendert werden
       *
       * @param th   Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       * @param perob Array von BlockPer Objekten mit Werten zur Personalisierung
       */
      public void calcFinal_zd(int th, int iter, BlockPer[] perob)
      {
         float[] pers;
         float sum;
         int i, topicanz,temp;
         BlockPer pero;
         
         sum = 0;
         topicanz = topictest.size();
         pers = new float[topicanz];
         for (i=0; i<topicanz;i++)
         {pers[i] = (float) 0;}
         for (i=0;i<perob.length;i++)
            // wir durchlaufen den Array von BlockPer-Objekten
         {
            pero = (BlockPer)	perob[i];
            if (topictest.containsKey(pero.cid))
               // wenn die Klasse auch in unserem BlockRank vorkommt
            {
               temp = ((Integer) (topictest.get(pero.cid))).intValue();
               // holen wir die laufende Nummer der Klasse
               pers[temp]=pero.value;
               // und setzen unseren pers Wert an entsprechender Stelle auf den Wert
               sum = sum + pero.value;
               // und berechnen die gesamt-Summe aller verwendeten Per-Werte
            }
         }
         if (sum==0)
         {for (i=0; i<topicanz;i++)
          {pers[i] = (float) 1/topicanz;} }
         // wenn keiner der angegebenen Klassen in unserem BlockRank ist initialisieren wir uniform
         else
         {for (i=0; i<topicanz;i++)
          {pers[i] = pers[i]/sum;}}
         // ansonsten normalisieren wir die Werte auf 1
         per = (float[]) pers.clone();
         calcBRper_zd(th, iter);
         // Berechne den PageRank der einzelnen Topics
         init = new float[urlcount];
         for (i=0; i<urlcount;i++)
         {init[i] = rank[topic[i]] * localPR[i];}
         //  berechne estimated global PR Seite aus Rank[Topic] * local PR Seite als INIT
         per = new float[urlcount];
         for (i=0; i<urlcount;i++)
         {per[i] = pers[topic[i]] * localPR[i];}
         //  berechne per-Wert URL als per-Wert topic * localPR der seite
         calcPRper_zd(th, iter);
         // Berechne den wirklichen BlockRank jeder Seite
      }
      
      //---------------------------------------------------------------
      
      /**
       * Klasse von Hilfsthreads der Methode calcBRper_zd
       */
      private class CalccBR_zd implements Runnable
      {
         int srow; //Zeile zum starten
         int erow; //Zeile zum enden
         
         /**
          * Konstruktor des Hilfsthread der Methode calcBRper_zd
          *
          * @param nr laufende Nummer des Threads
          * @param l  Anzahl der zu berechnenden Zeilen
          */
         public CalccBR_zd(int nr, int l)
         {
            srow = nr*l;
            erow = srow+l;
         }
         
         /**
          * Hilfsthread der Methode calcBRper_zd
          */
         public void run()
         {
            int i,j;
            float temp;
            
            for (i=srow; i<erow && i<topicctr; i++)
               // fuer alle Zeilen Matrix in unserem Zeilenbereich
            {
               temp=0;
               for (j=Bpoint[i];j<Bpoint[i+1]; j++)
                  //fuer j = value index aller Werte der Zeile
               {
                  temp = temp + Bvalues[j]*rank[Bindex[j]];
                  // die Addition: Element Matrix * Rank(Spalte des Werts)
               }
               ranktemp[i]= (nalpha * temp);
               // wir multiplizieren unseren Vektor mit (1-alpha)
            }
         }
      }
      
      /**
       * Diese interne Funktion berechnet den personalisierten PageRank
       * der BlockMatrix aus den Topics.
       * Dazu werden die Variablen Bvalues, Bpoint, Bindex, Bzeroind und Per benoetigt.
       * Das Ergebnis wird in der Variable Rank gespeichert.
       * Die Methode ist nur verwendbar mit Webseiten-Sets ohne ZONs.
       *
       * @param th   Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       */
      public void calcBRper_zd(int th, int iter)
      {
         int maxiter = iter;
         float temp;
         int laenge;
         int threads = th;
         Thread[] t;
         int i,j;
         
         temp = (float) (1 / topicctr);
         rank = new float[topicctr];
         for (i=0; i<topicctr; i++)
         {
            rank[i]=temp;
            // initialisiere Rank-Vektor mit 1/Anzahl Topics
            per[i]= per[i]*alpha;
            // multipliziere den per-Vektor mit alpha
         }
         
         laenge = (int) Math.ceil(topicctr / (double) threads);
         // bestimme Partitionierung fuer Threads
         
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            ranktemp = new float[topicctr];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new CalccBR_zd(j,laenge));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<topicctr;j++)
            {ranktemp[j] = ranktemp[j]+ per[j];}
            // addiere den Per-Vektor
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
/*     float x=0;
     for (i=0;i<topicctr;i++){x=x+rank[i];}
     System.out.println(x); 			*/
         
      }
      
      //---------------------------------------------------------------
      
      /**
       * Klasse von Hilfsthreads der Methode calcPRper_zd
       */
      private class CalccPR_zd implements Runnable
      {
         int srow; //Zeile zum starten
         int erow; //Zeile zum enden
         
         /**
          * Konstruktor des Hilfsthread der Methode calcPRper_zd
          *
          * @param nr laufende Nummer des Threads
          * @param l  Anzahl der zu berechnenden Zeilen
          */
         public CalccPR_zd(int nr, int l)
         {
            srow = nr*l;
            erow = srow+l;
         }
         
         /**
          * Hilfsthread der Methode calcPRper_zd
          */
         public void run()
         {
            int i,j;
            float temp;
            
            for (i=srow; i<erow && i<urlcount; i++)
               // fuer alle Zeilen Matrix in unserem Zeilenbereich
            {
               temp=0;
               for (j=point[i];j<point[i+1]; j++)
                  //fuer j = value index aller Werte der Zeile
               {
                  temp = temp + values[j]*rank[index[j]];
                  // die Addition: Element Matrix * Rank(Spalte des Werts)
               }
               ranktemp[i]= (nalpha * temp);
               // wir multiplizieren unseren Vektor mit (1-alpha)
            }
         }
      }
      
      
      /**
       * Diese interne Funktion berechnet den (personalisierten) PageRank.
       * Dazu werden die Variablen Init und Per benoetigt.
       * Das Ergebnis wird in der Variable Rank gespeichert.
       * Die Methode ist nur verwendbar mit Webseiten-Sets ohne ZONs.
       *
       * @param th Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       */
      public void calcPRper_zd(int th, int iter)
      {
         int maxiter = iter;
         float temp;
         int laenge;
         int threads = th;
         Thread[] t;
         Iterator it;
         Webnode node;
         int i,j;
         ArrayList al;
         int id1, id2;
         int[] outde;
         int laufnr;
         
         outde = new int[urlcount];
         rank = init;
         // initialisiere den Rank-Vektor mit init
         for (i=0; i<urlcount; i++)
         {
            node = (Webnode) (graph[i]);
            outde[i] = node.out_deg;
            per[i] = per[i]*alpha;
            // multipliziere den Per-Wert mit alpha
         }
         // erstelle Array der in Feld i den Outdeg von URL Nr i+1 hat
         laufnr = 0;
         point = new int[urlcount+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         index = new int[linkcount];
         values = new float[linkcount];
         for (i=0; i<urlcount; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            point[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            node = (Webnode) (graph[i]);
            al = node.link_from;
            // hole Vektor mit Liste aller Seiten-ID die auf uns verlinken
            java.util.Collections.sort(al);
            //wir sortieren die ArrayList aufsteigend
            for (j=0; j<al.size(); j++)
               // durchlaufe den Vektor
            {
               id2 = ((Integer) (al.get(j))).intValue()-1;
               // hol dir die Matrix-ID id2 der URL die auf uns zeigt
               temp = (float) 1 / (outde[id2]);
               // bestimme den Wert den uns die Seite gibt  (outde>0 da link auf uns)
               values[laufnr]=temp;
               // Trage den Wert im Values Array CRS ein
               index[laufnr]=id2;
               // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
               laufnr++;
               //erhoehe die laufnr;
            }
         }
         point[urlcount]=laufnr;
         // speichere die erste nicht mehr benoetigte Stelle in das letzte Feld des Point array
         outde = null;
         node = null;
         al = null;
         
         laenge = (int) Math.ceil(urlcount / (double) threads);
         // bestimme Partitionierung fuer Threads
         
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            ranktemp = new float[urlcount];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new CalccPR_zd(j,laenge));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<urlcount;j++)
            {ranktemp[j] = ranktemp[j]+ per[j];}
            // addiere den Per-Vektor
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         values = null;
         index = null;
         point = null;
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
/*     float x=0;
     for (i=0;i<urlcount;i++){x=x+rank[i];}
     System.out.println(x); 			*/
         
      }
      
      
      
   }   //Ende Class BlockRankCalculator
   
   //===================================================================================
   //                          PageRankCalculator
   
   /**
    * Diese Klasse stellt alle Methoden, die zur Berechnung des PageRank benoetigt
    * werden, ausser durch Berechnung des BlockRank.
    */
   public class PageRankCalculator
   {
      float[] ranktemp,values,rank;
      float alpha = (float) 0.2;
      float nalpha = 1-alpha;
      int[] point, index;
      int[] zeroind;               //fuer Version ohne zero-drop
      float zerosum;               //fuer Version ohne zero-drop
      float zerotopic;             //fuer personalized Version ohne zero-drop
      HashSet itopic;              //fuer personalized Version ohne zero-drop
      float[] init;                //fuer personalized Version ohne zero-drop
      
      
      //-----------------------------------------------------------------------------------
      //                          Page Rank  multithreaded uses Compressed Row Storage
      //
      //  double[] values : hier wird fuer jeder Wert der Tabelle eingetragen,
      //                    sortiert nach Reihen und innerhalb nach Spalten
      //                    die Anzahl der Eintraege = Anzahl Zeilen in Links_ID, index 0..i-1
      //  int[] index     : index[i] enthaelt die Spaltenzahl von values[i]
      //                    mit Spalten 0..n-1; Groesse das Arrays wie values
      //  int[] point     : point[i] enthaelt den Index des value-array-Elements
      //                    das den ersten nichtnegativen Wert in Zeile i hat
      //                    die Groesse ist hier (Anzahl Zeilen = Anzahl Url)
      //                    +1 (der Index den der naechste Wert in values haette) =  #LINKS i
      
      
      //---------------------------------------------------------------
      
      /**
       * Klasse von Hilfsthreads der Methode calcPageRankmt
       */
      private class Calcc implements Runnable
      {
         int srow; //Zeile zum starten
         int erow; //Zeile zum enden
         
         /**
          * Konstruktor des Hilfsthread der Methode calcPageRankmt
          *
          * @param nr laufende Nummer des Threads
          * @param l  Anzahl der zu berechnenden Zeilen
          */
         public Calcc(int nr, int l)
         {
            srow = nr*l;
            erow = srow+l;
         }
         
         /**
          * Hilfsthread der Methode calcPageRankmt
          */
         public void run()
         {
            int i,j;
            float temp;
            
            for (i=srow; i<erow && i<urlcount; i++)
               // fuer alle Zeilen Matrix in unserem Zeilenbereich
            {
               temp=0;
               for (j=point[i];j<point[i+1]; j++)
                  //fuer j = value index aller Werte der Zeile
               {
                  temp = temp + values[j]*rank[index[j]];
                  // die Addition: Element Matrix * Rank(Spalte des Werts)
               }
               ranktemp[i]= nalpha * (temp + zerosum);
               // Rank = 1-alpha (rank Spalten mit non-0-outdeg+ Spalten mit 0-outdeg) //!!!!!
            }
         }
      }
      
      /**
       * Klasse von Hilfsthreads der Methode calcPageRankper
       */
      private class Calccper implements Runnable
      {
         int srow; //Zeile zum starten
         int erow; //Zeile zum enden
         
         /**
          * Konstruktor des Hilfsthread der Methode calcPageRankper
          *
          * @param nr laufende Nummer des Threads
          * @param l  Anzahl der zu berechnenden Zeilen
          */
         public Calccper(int nr, int l)
         {
            srow = nr*l;
            erow = srow+l;
         }
         
         /**
          * Hilfsthread der Methode calcPageRankper
          */
         public void run()
         {
            int i,j;
            float temp;
            
            for (i=srow; i<erow && i<urlcount; i++)
               // fuer alle Zeilen Matrix in unserem Zeilenbereich
            {
               temp=0;
               for (j=point[i];j<point[i+1]; j++)
                  //fuer j = value index aller Werte der Zeile
               {
                  temp = temp + values[j]*rank[index[j]];
                  // die Addition: Element Matrix * Rank(Spalte des Werts)
               }
               if (itopic.contains(new Integer(i)))
               {ranktemp[i]= nalpha * (temp + zerotopic);}
               // wenn die Seite eine topicSeite ist  Rank = 1-alpha  *
               // (rank Spalten mit non-0-outdeg + Spalten mit 0-outdeg [alle pertopic als Wert])
               else
               {ranktemp[i] = nalpha * (temp + zerosum);}
               // Rank = 1-alpha (rank Spalten mit non-0-outdeg + Spalten mit 0-outdeg [alle pernormal])
            }
         }
      }
      
      /**
       * Klasse von Hilfsthreads der Methode calcPageRanksvm
       */
      private class Calccsvm implements Runnable
      {
         int srow; //Zeile zum starten
         int erow; //Zeile zum enden
         boolean mod;
         
         /**
          * Konstruktor des Hilfsthread der Methode calcPageRanksvm
          *
          * @param nr   laufende Nummer des Threads
          * @param l    Anzahl der zu berechnenden Zeilen
          * @param mode true = verwende Pers-Vektor fuer ZONs, false = verwende uniforme Verteilung
          */
         public Calccsvm(int nr, int l, boolean mode)
         {
            srow = nr*l;
            erow = srow+l;
            mod = mode;
         }
         
         /**
          * Hilfsthread der Methode calcPageRanksvm
          */
         public void run()
         {
            int i,j;
            float temp;
            
            for (i=srow; i<erow && i<urlcount; i++)
               // fuer alle Zeilen Matrix in unserem Zeilenbereich
            {
               temp=0;
               for (j=point[i];j<point[i+1]; j++)
                  //fuer j = value index aller Werte der Zeile
               {
                  temp = temp + values[j]*rank[index[j]];
                  // die Addition: Element Matrix * Rank(Spalte des Werts)
               }
               if (mod)
               {
                  if (itopic.contains(new Integer(i)))
                  {ranktemp[i]= nalpha * (temp + (init[i]*zerosum));}
                  // die Werte in unserer Zeile lassen sich durch den Wert einer non-topic-Seite*SVM berechen
                  // SVM liegt zwischen 1 und 2, 1 entspricht non-topic page
                  else
                  {ranktemp[i] = nalpha * (temp + zerosum);}
                  // Rank = 1-alpha (rank Spalten mit non-0-outdeg + Spalten mit 0-outdeg [alle 1/Svmsum])
               }
               else
               {ranktemp[i] = nalpha * (temp + zerosum);}
               // Im falle keine exakte Berechnung haben alle Spalten mit 0-outdeg Wert 1/urlcount
               // und zerosum wurde bereits damit multipliziert
            }
         }
      }
      
      /**
       * Klasse von Hilfsthreads der Methoden auf Sets ohne ZONs
       */
      private class Calcc_zd implements Runnable
      {
         int srow; //Zeile zum starten
         int erow; //Zeile zum enden
         
         /**
          * Konstruktor des Hilfsthread der Methoden auf Sets ohne ZONs
          *
          * @param nr laufende Nummer des Threads
          * @param l  Anzahl der zu berechnenden Zeilen
          */
         public Calcc_zd(int nr, int l)
         {
            srow = nr*l;
            erow = srow+l;
         }
         
         /**
          * Hilfsthread der Methoden auf Sets ohne ZONs
          */
         public void run()
         {
            int i,j;
            float temp;
            
            for (i=srow; i<erow && i<urlcount; i++)
               // fuer alle Zeilen Matrix in unserem Zeilenbereich
            {
               temp=0;
               for (j=point[i];j<point[i+1]; j++)
                  //fuer j = value index aller Werte der Zeile
               {
                  temp = temp + values[j]*rank[index[j]];
                  // die Addition: Element Matrix * Rank(Spalte des Werts)
               }
               ranktemp[i]= (nalpha * temp);
               // wir multiplizieren unseren Vektor mit (1-alpha) und addieren den Wert
               // des Person-Vektors..dieser ist immer gleich da uniform
            }
         }
      }
      
      //---------------------------------------------------------------
      
      
      /**
       * Dummy-Implementierung zum Speichern des PageRanks in die Datenbank
       * Anpassung an das vorhandene Schema hier vornehmen
       */
      public void storePR()
      {
         
         
         int count=0;
         try
         {
            PreparedStatement pstmt = oc.prepareStatement("UPDATE bingodocuments SET PageRank=? Where ID=?");
            for (int i=0; i<graph.length; i++)
            {
               pstmt.setDouble(1, (double)rank[i]);
               pstmt.setLong(2, ((Webnode) (graph[i])).id);
               pstmt.addBatch();
               if (++count % batchSize == 0)
               {pstmt.executeBatch();}
            }
            pstmt.executeBatch();
            pstmt.close();
            oc.commit();
            pstmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         
/*
     Webnode node;
     int i;
 
     try
     {
       FileWriter f1 = new FileWriter("PageRank.dat");
       for (i=0;i<urlcount;i++)
       {
         node = (Webnode) (graph[i]);
         f1.write(node.id+": "+rank[i]+"\n");
       }
       f1.close();
     }
     catch  (IOException e) {System.out.println("Error writing PageRank-Result");}
 */
      }
      
      //---------------------------------------------------------------
      
      /**
       * Diese interne Funktion berechnet den PageRank.
       * Sie ist verwendbar mit allen Webseiten-Sets.
       *
       * @param th   Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       */
      public void calcPageRankmt(int th, int iter)
      {
         int maxiter = iter;
         float temp;
         int laenge;
         int threads = th;
         Thread[] t;
         Iterator it;
         Webnode node;
         int i,j;
         ArrayList al;
         int id1, id2;
         int[] outde;
         int laufnr;
         
         float per = (float) (alpha / urlcount);
         
         zeroind = new int[zeroctr];
         j=0;
         outde = new int[urlcount];
         for (i=0; i<urlcount; i++)
         {
            node = (Webnode) (graph[i]);
            if (node.out_deg==0)
            {zeroind[j]=i; j++;}
            // Merke dir die SpaltenNr mit outdeg 0
            outde[i] = node.out_deg;
         }
         // erstelle Array der in Feld i den Outdeg von URL Nr i+1 hat
         laufnr = 0;
         point = new int[urlcount+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         index = new int[linkcount];
         values = new float[linkcount];
         rank = new float[urlcount];
         temp = (float) 1 / urlcount;
         for (i=0; i<urlcount; i++)
         {rank[i]=temp;}
         // initialisiere Rank-Vektor mit 1/Anzahl URL
         for (i=0; i<urlcount; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            point[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            
            node = (Webnode) (graph[i]);
            al = node.link_from;
            // hole Vektor mit Liste aller Seiten-ID die auf uns verlinken
            java.util.Collections.sort(al);
            //wir sortieren die ArrayList aufsteigend
            for (j=0; j<al.size(); j++)
               // durchlaufe den Vektor
            {
               id2 = ((Integer) (al.get(j))).intValue()-1;
               // hol dir die Matrix-ID id2 der URL die auf uns zeigt
               temp = (float) 1 / (outde[id2]);
               // bestimme den Wert den uns die Seite gibt
               values[laufnr]=temp;
               // Trage den Wert im Values Array CRS ein
               index[laufnr]=id2;
               // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
               laufnr++;
               //erhoehe die laufnr;
            }
         }
         point[urlcount]=laufnr;
         // speichere die erste nicht mehr benoetigte Stelle in das letzte Feld des Point array
         outde = null;
         node = null;
         al = null;
         
         laenge = (int) Math.ceil(urlcount / (double) threads);
         // bestimme Partitionierung fuer Threads
         
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            zerosum = 0;
            for (j=0; j<zeroctr; j++)
            {zerosum = zerosum + rank[zeroind[j]];}
            // wir summieren die PageRanks aller Spalten mit outdeg 0 auf
            zerosum = zerosum / urlcount;
            // wir multiplizieren die Summe mit 1/n
            ranktemp = new float[urlcount];   //new
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new Calcc(j,laenge));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<urlcount;j++)
            {ranktemp[j] = ranktemp[j] + per;}
            // addiere den Per-Vektor (alle Eintraege selben Wert da uniform)
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         
         values = null;
         index = null;
         point = null;
         zeroind = null;
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
/*     float x=0;
     for (i=0;i<urlcount;i++){x=x+rank[i];}
     System.out.println(x);                      */
         
      }
      
      //---------------------------------------------------------------
      
      
      /**
       * Diese interne Funktion berechnet den personalisierten PageRank fuer eine Klasse.
       * Dabei ist die Verteilung innerhalb der Klasse uniform.
       * Sie ist verwendbar mit allen Webseiten-Sets.
       *
       * @param th     Anzahl der Threads die verwendet werden soll
       * @param iter   Anzahl der Iterationen die der Algorithmus durchlaufen soll
       * @param cname  Name der Klasse fuer die der PageRank berechnet werden soll
       * @param factor Wert mit dem Pages der Klasse bewertet werden, muss groesser 1 sein
       */
      public void calcPageRankper(int th, int iter, String cname, float factor)
      {
         int maxiter = iter;
         float temp, pernormal, pertopic, sum;
         int laenge;
         int threads = th;
         Thread[] t;
         Iterator it;
         Webnode node;
         int i,j;
         ArrayList al;
         int id1, id2;
         int[] outde;
         int laufnr;
         float[] per;
         Statement stmt;
         ResultSet rset;
         HashSet topic;
         String quer;
         
         
         per = new float[urlcount];
         zeroind = new int[zeroctr];
         
         topic = new HashSet();
         try
         {
            stmt = oc.createStatement();
            quer = "SELECT DISTINCT B.ID FROM Classes A, bingodocuments B WHERE A.Class_ID = B.class_id AND A.Name LIKE '"+cname+"%'";
            // Liste long [IDs aller Dokumente deren Klassenname mit cname beginnt]
            rset = stmt.executeQuery(quer);
            while (rset.next())
            {
               topic.add(new Long(rset.getLong(1)));
               // haenge die IDs der URLs die zur Topic gehoeren in ein HashSet
            }
            rset.close(); rset = null;
            stmt.close(); stmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         
         sum = urlcount + (factor-1) * topic.size();
         // die Gesamtsumme = summe aller Urls *1 + Summe topic URLs * (Factor-1)
         pernormal = alpha/sum;
         // fuer non-topic URLS Per = alpha * 1/sum
         pertopic = alpha * (factor/sum);
         // fuer topic URLS alpha * factor/sum mit factor >1
         
         j=0;
         outde = new int[urlcount];
         itopic = new HashSet();
         for (i=0; i<urlcount; i++)
         {
            node = (Webnode) (graph[i]);
            if (node.out_deg==0)
            {zeroind[j]=i; j++;}
            // Merke dir die SpaltenNr mit outdeg 0
            outde[i] = node.out_deg;
            // erstelle Array der in Feld i den Outdeg von URL Nr i+1 hat
            if (topic.contains(new Long(node.id)))
               // page gehoert zu topic
            {
               per[i]=pertopic;
               itopic.add(new Integer(i));
               // haenge die Integer-Nummer der Seiten die zur Topic gehoeren ins HashSet itopic
            }
            else
            { per[i]=pernormal;}
            // out of topic pages
         }
         topic = null;
         
         laufnr = 0;
         point = new int[urlcount+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         index = new int[linkcount];
         values = new float[linkcount];
         rank = new float[urlcount];
         temp = (float) 1 / urlcount;
         for (i=0; i<urlcount; i++)
         {rank[i]=temp;}
         // initialisiere Rank-Vektor mit 1/Anzahl URL
         for (i=0; i<urlcount; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            point[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            
            node = (Webnode) (graph[i]);
            al = node.link_from;
            // hole Vektor mit Liste aller Seiten-ID die auf uns verlinken
            java.util.Collections.sort(al);
            //wir sortieren die ArrayList aufsteigend
            for (j=0; j<al.size(); j++)
               // durchlaufe den Vektor
            {
               id2 = ((Integer) (al.get(j))).intValue()-1;
               // hol dir die Matrix-ID id2 der URL die auf uns zeigt
               temp = (float) 1 / (outde[id2]);
               // bestimme den Wert den uns die Seite gibt
               values[laufnr]=temp;
               // Trage den Wert im Values Array CRS ein
               index[laufnr]=id2;
               // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
               laufnr++;
               //erhoehe die laufnr;
            }
         }
         point[urlcount]=laufnr;
         // speichere die erste nicht mehr benoetigte Stelle in das letzte Feld des Point array
         outde = null;
         node = null;
         al = null;
         
         laenge = (int) Math.ceil(urlcount / (double) threads);
         // bestimme Partitionierung fuer Threads
         
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            zerosum = 0;
            for (j=0; j<zeroctr; j++)
            {zerosum = zerosum + rank[zeroind[j]];}
            // wir summieren die PageRanks aller Spalten mit outdeg 0 auf
            zerotopic = zerosum * (factor / sum);
            // im Falle einer Topic Seite haetten alle Eintraege den Wert Faktor/Gesamtsumme
            zerosum = zerosum /sum;
            // im Falle das die Seite keine Topic Seite ist haben alle Eintraege den Wert 1/Gesamtsumme
            ranktemp = new float[urlcount];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new Calccper(j,laenge));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<urlcount;j++)
            {ranktemp[j] = ranktemp[j] + per[j];}
            // addiere den Per-Vektor
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         
         itopic = null;
         values = null;
         per = null;
         index = null;
         point = null;
         zeroind = null;
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
/*     float x=0;
     for (i=0;i<urlcount;i++){x=x+rank[i];}
     System.out.println(x);                      */
         
      }
      
      //---------------------------------------------------------------
      
      /**
       * Diese interne Funktion berechnet den personalisierten PageRank fuer eine Klasse.
       * Dabei ist die Verteilung innerhalb der Klasse vom SVM-Wert abhaengig.
       * Sie ist verwendbar mit allen Webseiten-Sets.
       *
       * @param th    Anzahl der Threads die verwendet werden soll
       * @param iter  Anzahl der Iterationen die der Algorithmus durchlaufen soll
       * @param cname Name der Klasse fuer die der PageRank berechnet werden soll
       * @param mode  true = verwende Pers-Vektor fuer ZONs, false = verwende uniforme Verteilung
       */
      public void calcPageRanksvm(int th, int iter, String cname, boolean mode)
      {
         int maxiter = iter;
         float temp;
         int laenge;
         int threads = th;
         Thread[] t;
         Iterator it;
         Webnode node;
         int i,j;
         ArrayList al;
         int id1, id2;
         int[] outde;
         int laufnr,topctr;
         float[] per;
         Statement stmt;
         ResultSet rset;
         String quer;
         Long nodeid;
         float svmval, svmsum,svmmax;
         HashMap svm;
         
         per = new float[urlcount];
         init = new float[urlcount];
         outde = new int[urlcount];
         zeroind = new int[zeroctr];
         
         
         svm = new HashMap();
         topctr=0;
         svmmax=0;
         svmsum = 0;
         try
         {
            stmt = oc.createStatement();
            quer = "SELECT B.ID,B.SvmScore FROM Classes A, bingodocuments B WHERE A.Class_ID = B.class_id AND A.Name LIKE '"+cname+"%'";
            // Liste long [IDs aller Dokumente deren Klassenname mit cname beginnt], float [SVM-Score des Dokuments]
            rset = stmt.executeQuery(quer);
            while (rset.next())
            {
               svmval = rset.getFloat(2);
               if (svmval <= 0)
               {svmval = 0;}
               // Abfangen von Negativen Werten da diese Probleme machen
               // negativer Wert => wie Nicht-Topic URL
               else
               {
                  if (svmval > svmmax)
                  {svmmax = svmval;}
                  // wir bestimmen den groessten SVM-Wert
               }
               svm.put(new Long(rset.getLong(1)), new Float(svmval));
               // pushe die IDs mit dem SVM Wert in die HashMap
            }
            rset.close(); rset = null;
            stmt.close(); stmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         
         j=0;
         itopic = new HashSet();
         for (i=0; i<urlcount; i++)
         {
            node = (Webnode) (graph[i]);
            outde[i] = node.out_deg;
            // erstelle Array der in Feld i den Outdeg von URL Nr i+1 hat
            if (node.out_deg==0)
            {zeroind[j]=i; j++;}
            nodeid = new Long(node.id);
            if (svm.containsKey(nodeid))
               // page gehoert zu topic => SVM beruecksichtigen
            {
               itopic.add(new Integer(i));
               svmval = ((Float) (svm.get(nodeid))).floatValue();
               if (svmmax==0)
               {svmval = 1;} else
               {svmval = 1 + (svmval/svmmax);}
               // wir dividieren den SVM durch sein Maximum und addieren 1 -> Werte zwischen 1 und 2
               init[i]=svmval;
               svmsum = svmsum + svmval;
               topctr++;
            }
            else
            {
               init[i]=1;
            }
            // out of topic pages = 1
         }
         if (!mode)
         {itopic = null;}
         // im falle einer nicht-exakten Berechnung brauchen wir i-topic nicht
         svm = null;
         svmsum = svmsum + (urlcount - topctr);
         //berechne die Gesamtsumme zur normalisierung aus der Summe der SVM und der Summe der
         // out of topic pages bewertet mit 1
         for (i=0; i<urlcount; i++)
         {
            per[i] = alpha*(init[i]/svmsum);
            // teile die SVM-Werte durch die Gesamtsumme und multipliziere mit alpha
         }
         
         laufnr = 0;
         point = new int[urlcount+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         index = new int[linkcount];
         values = new float[linkcount];
         rank = new float[urlcount];
         temp = (float) 1 / urlcount;
         for (i=0; i<urlcount; i++)
         {rank[i]=temp;}
         // initialisiere Rank-Vektor mit 1/Anzahl URL
         for (i=0; i<urlcount; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            point[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            node = (Webnode) (graph[i]);
            al = node.link_from;
            // hole Vektor mit Liste aller Seiten-ID die auf uns verlinken
            java.util.Collections.sort(al);
            //wir sortieren die ArrayList aufsteigend
            for (j=0; j<al.size(); j++)
               // durchlaufe den Vektor
            {
               id2 = ((Integer) (al.get(j))).intValue()-1;
               // hol dir die Matrix-ID id2 der URL die auf uns zeigt
               temp = (float) 1 / (outde[id2]);
               // bestimme den Wert den uns die Seite gibt
               values[laufnr]=temp;
               // Trage den Wert im Values Array CRS ein
               index[laufnr]=id2;
               // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
               laufnr++;
               //erhoehe die laufnr;
            }
         }
         point[urlcount]=laufnr;
         // speichere die erste nicht mehr benoetigte Stelle in das letzte Feld des Point array
         outde = null;
         node = null;
         al = null;
         
         laenge = (int) Math.ceil(urlcount / (double) threads);
         // bestimme Partitionierung fuer Threads
         
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            zerosum = 0;
            for (j=0; j<zeroctr; j++)
            {zerosum = zerosum + rank[zeroind[j]];}
            if (mode)
            {zerosum = zerosum / svmsum;}
            // fuer den Fall Init mit per Non-topic-Pages = Zerosum * (1/svmsum)
            else
            {zerosum = zerosum / urlcount;}
            // fuer den Fall nicht-exakt Initialisierung Zero-outdeg-Nodes mit 1/Anzahl URLs
            ranktemp = new float[urlcount];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new Calccsvm(j,laenge,mode));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<urlcount;j++)
            {ranktemp[j] = ranktemp[j] + per[j];}
            // addiere den Per-Vektor
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         values = null;
         index = null;
         point = null;
         itopic = null;
         init = null;
         per = null;
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
/*     float x=0;
     for (i=0;i<urlcount;i++){x=x+rank[i];}
     System.out.println(x); 			*/
         
      }
      
      
      
      //---------------------------------------------------------------
      /**
       * Diese interne Funktion berechnet den PageRank.
       * Sie ist nur verwendbar mit Webseiten-Sets ohne ZONs.
       *
       * @param th   Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       */
      public void calcPageRankmt_zd(int th, int iter)
      {
         int maxiter = iter;
         float temp;
         int laenge;
         int threads = th;
         Thread[] t;
         Iterator it;
         Webnode node;
         int i,j;
         ArrayList al;
         int id1, id2;
         int[] outde;
         int laufnr;
         float per = (float) (alpha / urlcount);
         
         outde = new int[urlcount];
         for (i=0; i<urlcount; i++)
         {
            node = (Webnode) (graph[i]);
            outde[i] = node.out_deg;
         }
         // erstelle Array der in Feld i den Outdeg von URL Nr i+1 hat
         laufnr = 0;
         point = new int[urlcount+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         index = new int[linkcount];
         values = new float[linkcount];
         rank = new float[urlcount];
         temp = (float) 1 / urlcount;
         for (i=0; i<urlcount; i++)
         {rank[i]=temp;}
         // initialisiere Rank-Vektor mit 1/Anzahl URL
         for (i=0; i<urlcount; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            point[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            node = (Webnode) (graph[i]);
            al = node.link_from;
            // hole Vektor mit Liste aller Seiten-ID die auf uns verlinken
            java.util.Collections.sort(al);
            //wir sortieren die ArrayList aufsteigend
            for (j=0; j<al.size(); j++)
               // durchlaufe den Vektor
            {
               id2 = ((Integer) (al.get(j))).intValue()-1;
               // hol dir die Matrix-ID id2 der URL die auf uns zeigt
               temp = (float) 1 / (outde[id2]);
               // bestimme den Wert den uns die Seite gibt
               values[laufnr]=temp;
               // Trage den Wert im Values Array CRS ein
               index[laufnr]=id2;
               // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
               laufnr++;
               //erhoehe die laufnr;
            }
         }
         point[urlcount]=laufnr;
         // speichere die erste nicht mehr benoetigte Stelle in das letzte Feld des Point array
         outde = null;
         node = null;
         al = null;
         
         laenge = (int) Math.ceil(urlcount / (double) threads);
         // bestimme Partitionierung fuer Threads
         
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            ranktemp = new float[urlcount];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new Calcc_zd(j,laenge));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<urlcount;j++)
            {ranktemp[j] = ranktemp[j]+per;}
            // addiere den Per-Vektor (alle Eintraege selben Wert da uniform)
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         values = null;
         index = null;
         point = null;
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
/*     float x=0;
     for (i=0;i<urlcount;i++){x=x+rank[i];}
     System.out.println(x); 			*/
         
      }
      
      //---------------------------------------------------------------
      
      /**
       * Diese interne Funktion berechnet den personalisierten PageRank fuer eine Klasse.
       * Dabei ist die Verteilung innerhalb der Klasse uniform.
       * Sie ist nur verwendbar mit Webseiten-Sets ohne ZONs.
       *
       * @param th     Anzahl der Threads die verwendet werden soll
       * @param iter   Anzahl der Iterationen die der Algorithmus durchlaufen soll
       * @param cname  Name der Klasse fuer die der PageRank berechnet werden soll
       * @param factor Wert mit dem Pages der Klasse bewertet werden, muss groesser 1 sein
       */
      public void calcPageRankper_zd(int th, int iter, String cname, float factor)
      {
         int maxiter = iter;
         float temp, pertopic, sum;
         int laenge;
         int threads = th;
         Thread[] t;
         Iterator it;
         Webnode node;
         int i,j;
         ArrayList al;
         int id1, id2;
         int[] outde;
         int laufnr,topctr;
         float[] per;
         Statement stmt;
         ResultSet rset;
         HashSet topic;
         String quer;
         
         per = new float[urlcount];
         outde = new int[urlcount];
         
         topic = new HashSet();
         topctr=0;
         try
         {
            stmt = oc.createStatement();
            quer = "SELECT DISTINCT B.ID FROM Classes A, bingodocuments B WHERE A.Class_ID = B.class_id AND A.Name LIKE '"+cname+"%'";
            // Liste long [IDs aller Dokumente deren Klassenname mit cname beginnt]
            rset = stmt.executeQuery(quer);
            while (rset.next())
            {
               topic.add(new Long(rset.getLong(1)));
               // haenge alle IDs die zu der Topic gehoeren in ein Hashset
            }
            rset.close(); rset = null;
            stmt.close(); stmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         
         pertopic = alpha * factor;
         // per fuer topic sites = alpha * (factor>1)
         for (i=0; i<urlcount; i++)
            // durchlaufe alle Nodes
         {
            node = (Webnode) (graph[i]);
            outde[i] = node.out_deg;
            // erstelle Array der in Feld i den Outdeg von URL Nr i+1 hat
            if (topic.contains(new Long(node.id)))
            {
               per[i]=pertopic;
               topctr++;
            }
            // wenn die betrachtete page die topic hat merke dir den Per-Wert fuer Topic Pages
            else
            { per[i]=alpha;}
            // sonst den fuer out of topic pages  = alpha*1
         }
         topic = null;
         
         sum = urlcount + (factor-1) * topctr;
         // die Gesamtsumme = summe aller Urls *1 + Summe topic URLs * (Factor-1)
         for (i=0; i<urlcount; i++)
         { per[i] = per[i]/sum; }
         // teile die Per-Werte durch die Gesamtsumme
         laufnr = 0;
         point = new int[urlcount+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         index = new int[linkcount];
         values = new float[linkcount];
         rank = new float[urlcount];
         temp = (float) 1 / urlcount;
         for (i=0; i<urlcount; i++)
         {rank[i]=temp;}
         // initialisiere Rank-Vektor mit 1/Anzahl URL
         for (i=0; i<urlcount; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            point[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            node = (Webnode) (graph[i]);
            al = node.link_from;
            // hole Vektor mit Liste aller Seiten-ID die auf uns verlinken
            java.util.Collections.sort(al);
            //wir sortieren die ArrayList aufsteigend
            for (j=0; j<al.size(); j++)
               // durchlaufe den Vektor
            {
               id2 = ((Integer) (al.get(j))).intValue()-1;
               // hol dir die Matrix-ID id2 der URL die auf uns zeigt
               temp = (float) 1 / (outde[id2]);
               // bestimme den Wert den uns die Seite gibt
               values[laufnr]=temp;
               // Trage den Wert im Values Array CRS ein
               index[laufnr]=id2;
               // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
               laufnr++;
               //erhoehe die laufnr;
            }
         }
         point[urlcount]=laufnr;
         // speichere die erste nicht mehr benoetigte Stelle in das letzte Feld des Point array
         outde = null;
         node = null;
         al = null;
         
         laenge = (int) Math.ceil(urlcount / (double) threads);
         // bestimme Partitionierung fuer Threads
         
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            ranktemp = new float[urlcount];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new Calcc_zd(j,laenge));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<urlcount;j++)
            {ranktemp[j] = ranktemp[j] + per[j];}
            // addiere den Per-Vektor
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         values = null;
         per = null;
         index = null;
         point = null;
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
/*     float x=0;
     for (i=0;i<urlcount;i++){x=x+rank[i];}
     System.out.println(x); 			*/
         
      }
      
      //---------------------------------------------------------------
      
      /**
       * Diese interne Funktion berechnet den personalisierten PageRank fuer eine Klasse.
       * Dabei ist die Verteilung innerhalb der Klasse vom SVM-Wert abhaengig.
       * Sie ist nur verwendbar mit Webseiten-Sets ohne ZONs.
       *
       * @param th    Anzahl der Threads die verwendet werden soll
       * @param iter  Anzahl der Iterationen die der Algorithmus durchlaufen soll
       * @param cname Name der Klasse fuer die der PageRank berechnet werden soll
       * @param mode  true = verwende Pers-Vektor fuer ZONs, false = uniforme Verteilung
       */
      public void calcPageRanksvm_zd(int th, int iter, String cname)
      {
         int maxiter = iter;
         float temp;
         int laenge;
         int threads = th;
         Thread[] t;
         Iterator it;
         Webnode node;
         int i,j;
         ArrayList al;
         int id1, id2;
         int[] outde;
         int laufnr,topctr;
         float[] per;
         Statement stmt;
         ResultSet rset;
         String quer;
         Long nodeid;
         float svmval, svmsum,svmmax;
         HashMap svm;
         
         per = new float[urlcount];
         outde = new int[urlcount];
         
         svm = new HashMap();
         topctr=0;
         svmmax=0;
         svmsum = 0;
         try
         {
            stmt = oc.createStatement();
            quer = "SELECT B.ID,B.SvmScore FROM Classes A, bingodocuments B WHERE A.Class_ID = B.class_id AND A.Name LIKE '"+cname+"%'";
            // Liste long [IDs aller Dokumente deren Klassenname mit cname beginnt], float [SVM-Score des Dokuments]
            rset = stmt.executeQuery(quer);
            while (rset.next())
            {
               svmval = rset.getFloat(2);
               if (svmval <= 0)
               {svmval = 0;}
               // Abfangen von Negativen Werten da diese Probleme machen
               // negativer Wert => wie Nicht-Topic URL
               else
               {
                  if (svmval > svmmax)
                  {svmmax = svmval;}
                  // wir bestimmen den groessten SVM-Wert
               }
               svm.put(new Long(rset.getLong(1)), new Float(svmval));
               // pushe die IDs mit dem SVM Wert in die HashMap
            }
            rset.close(); rset = null;
            stmt.close(); stmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         
         
         for (i=0; i<urlcount; i++)
         {
            node = (Webnode) (graph[i]);
            outde[i] = node.out_deg;
            // erstelle Array der in Feld i den Outdeg von URL Nr i+1 hat
            nodeid = new Long(node.id);
            if (svm.containsKey(nodeid))
               // page gehoert zu topic => SVM beruecksichtigen
            {
               svmval = ((Float) (svm.get(nodeid))).floatValue();
               if (svmmax == 0)
               {svmval = 1;} else
               {svmval = 1 + (svmval/svmmax);}
               // wir dividieren den SVM durch sein Maximum und addieren 1 -> Werte zwischen 1 und 2
               svmsum = svmsum + svmval;
               per[i] = alpha * svmval;
               topctr++;
            }
            else
            {per[i] = alpha;}
            // out of topic pages = 1
         }
         svm = null;
         svmsum = svmsum + (urlcount - topctr);
         //berechne die Gesamtsumme zur normalisierung aus der Summe der SVM und der Summe der
         // out of topic pages bewertet mit 1
         for (i=0; i<urlcount; i++)
         { per[i] = per[i]/svmsum; }
         // teile die Per-Werte durch die Gesamtsumme
         
         laufnr = 0;
         point = new int[urlcount+1];
         //+1 da spaeter check point[i+1] mit i = 0..n-1
         index = new int[linkcount];
         values = new float[linkcount];
         rank = new float[urlcount];
         temp = (float) 1 / urlcount;
         for (i=0; i<urlcount; i++)
         {rank[i]=temp;}
         // initialisiere Rank-Vektor mit 1/Anzahl URL
         for (i=0; i<urlcount; i++)
            // jeder durchlauf entspricht einer Zeile der Matrix
         {
            point[i]=laufnr;
            //speichere den Index den das erste element einer Neuen zeile im Value array hat (0..n-1)
            node = (Webnode) (graph[i]);
            al = node.link_from;
            // hole Vektor mit Liste aller Seiten-ID die auf uns verlinken
            java.util.Collections.sort(al);
            //wir sortieren die ArrayList aufsteigend
            for (j=0; j<al.size(); j++)
               // durchlaufe den Vektor
            {
               id2 = ((Integer) (al.get(j))).intValue()-1;
               // hol dir die Matrix-ID id2 der URL die auf uns zeigt
               temp = (float) 1 / (outde[id2]);
               // bestimme den Wert den uns die Seite gibt
               values[laufnr]=temp;
               // Trage den Wert im Values Array CRS ein
               index[laufnr]=id2;
               // die Nummer der Spalte (= Nummer URL von der wir bekommen) ! (0...n-1)..Skript 1..n
               laufnr++;
               //erhoehe die laufnr;
            }
         }
         point[urlcount]=laufnr;
         // speichere die erste nicht mehr benoetigte Stelle in das letzte Feld des Point array
         outde = null;
         node = null;
         al = null;
         
         laenge = (int) Math.ceil(urlcount / (double) threads);
         // bestimme Partitionierung fuer Threads
         
         for (i=0;i<maxiter;i++)
            //   Iteration
         {
            ranktemp = new float[urlcount];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j]= new Thread(new Calcc_zd(j,laenge));
                  t[j].start();
                  // Berechne Matrix x Rankvektor multi-threaded
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            {ie.printStackTrace();}
            for (j=0;j<urlcount;j++)
            {ranktemp[j] = ranktemp[j] + per[j];}
            // addiere den Per-Vektor
            rank = ranktemp;
            t=null;
         } //Ende Iteration
         values = null;
         index = null;
         point = null;
         per = null;
         ranktemp = null;
         
         //     !!!TestAREA!!! 3 Zeilen
 /*    float x=0;
     for (i=0;i<urlcount;i++){x=x+rank[i];}
     System.out.println(x); 			*/
         
      }
      
      
   }    //Ende Class PageRankCalculator
   
   
   
   
   //===================================================================================
   //                          HITSCalculator
   //===================================================================================
   
   
   /**
    * Diese Klasse beinhaltet Methoden zur Berechnungen von HITS
    */
   public class HITSCalculator
   {
      private float[] auth, hub;
      private float[] helpauth, helphub;
      
      //-----------------------------------------------------------------------------------
      //                          HITS multithreaded
      
      //----------mtAuth Class----------
      
      /**
       * Klasse von Hilfsthreads der Methode calcHITSmt zur Berechnung der Auth-Score
       */
      private class mtAuth implements Runnable
      {
         private int start;
         private int ende;
         private int nummer;
         
         /**
          * Konstruktor des Hilfsthread der Methode calcHITSmt
          *
          * @param nr laufende Nummer des Threads
          * @param l  Anzahl der zu berechnenden Zeilen
          */
         public mtAuth(int nr, int l)
         {
            nummer = nr;
            start = nr*l;
            ende = start+l;
         }
         
         /**
          * Hilfsthread der Methode calcHITSmt
          */
         public void run()
         {
            int i,j,dummy,dummy2;
            Webnode node;
            ArrayList al;
            float value, authsc;
            float authall=0;
            
            for (i= start; i<ende && i<graph.length; i++)
               // solange wir uns in unserem Bereich befinden und im gueltigen Bereich des nrmap Arrays
            {
               authsc = 0;
               node = (Webnode) (graph[i]);
               dummy2 = (node.nummer)-1;
               al = node.link_from;
               // hole vektor mit allen Pages mit Links auf uns
               for (j=0; j<al.size(); j++)
               {
                  dummy = ((Integer) (al.get(j))).intValue()-1;
                  value = hub[dummy];
                  // hole dir die Hubscore dieser Seite
                  authsc = authsc+value;
                  // und addiere sie zu unserer authscore
                  authall = authall+value;
               }
               auth[dummy2]=authsc;
               node = null;
               al = null;
               // und schreiben die neue Authscore in unseren auth array
            }
            helpauth[nummer]=authall;
            // speichere gesammelte auth dieses Threads in Array an Stelle ThreadNr
         }
      }
      
      //----------mtHub Class----------
      
      /**
       * Klasse von Hilfsthreads der Methode calcHITSmt zur Berechnung der Hub-Score
       */
      private class mtHub implements Runnable
      {
         private int start;
         private int ende;
         private int nummer;
         
         /**
          * Konstruktor des Hilfsthread der Methode calcHITSmt
          *
          * @param nr laufende Nummer des Threads
          * @param l  Anzahl der zu berechnenden Zeilen
          */
         public mtHub(int nr, int l)
         {
            nummer = nr;
            start = nr*l;
            ende = start+l;
         }
         
         /**
          * Hilfsthread der Methode calcHITSmt
          */
         public void run()
         {
            int i,j,dummy,dummy2;
            Webnode node;
            ArrayList al;
            float value,hubsc;
            float huball=0;
            
            for (i= start; i<ende && i<graph.length; i++)
               // solange wir uns in unserem Bereich befinden und im gueltigen Bereich des nrmap Arrays
            {
               hubsc = 0;
               node = (Webnode) (graph[i]);
               dummy2 = (node.nummer)-1;
               al = node.link_to;
               // hole vektor mit allen Pages mit Links von uns auf andere
               for (j=0; j<al.size(); j++)
               {
                  dummy = ((Integer) (al.get(j))).intValue()-1;
                  value = auth[dummy];
                  // hole dir die Authscore dieser Seite
                  hubsc = hubsc+value;
                  // und addiere sie zu unserer Hubscore
                  huball = huball+value;
               }
               hub[dummy2] = hubsc;
               // und schreiben die neue Hubscore in unseren hub array
               node = null;
               al = null;
            }
            helphub[nummer]=huball;
            // speichere gesammelte auth dieses Threads in Array an Stelle ThreadNr
         }
      }
      
      //----------storeHITS----------
      
      /**
       * Methode zur Speicherung der berechneten Auth- und Hub-Score.
       */
      public void storeHITS()
      {
         
         int count=0;
         try
         {
            PreparedStatement pstmt = oc.prepareStatement("UPDATE bingodocuments SET AuthScore=?, HubScore=? WHERE ID=?");
            for (int i=0; i<graph.length; i++)
            {
               pstmt.setDouble(1, (double)auth[i]);
               pstmt.setDouble(2, (double)hub[i]);
               pstmt.setLong(3, ((Webnode) (graph[i])).id);
               pstmt.addBatch();
               if (++count % batchSize == 0)
               {pstmt.executeBatch();}
            }
            pstmt.executeBatch();
            pstmt.close();
            oc.commit();
            pstmt = null;
         }
         catch (SQLException e)
         {e.printStackTrace();}
         
/*
     float testa = 0;
     float testh = 0;
     Webnode node;
     int idnr;
     float a,h;
     try
     {
       FileWriter f1 = new FileWriter("HITS.dat");
       for (i=0;i<urlcount;i++)
       {
          node = (Webnode) (graph[i]);
          idnr = (node.nummer)-1;
          h = hub[idnr];
          a = auth[idnr];
          testa = testa + a;
          testh = testh + h;
          f1.write(node.id+"   Auth: "+a+"   Hub: "+h+"\n");
       }
       f1.write("\n");
       f1.write("Summe Auth: "+testa+"   Summe Hub: "+testh+"\n");
       f1.close();
     }
     catch  (IOException e) {System.out.println("Error writing HITS-Result");}
 */
      }
      
      
      //----------Multithread-Calculator----------
      
      /**
       * Berechnet die Auth- und Hubscore einer Menge von Webseiten.
       *
       * @param th   Anzahl der Threads die verwendet werden soll
       * @param iter Anzahl der Iterationen die der Algorithmus durchlaufen soll
       */
      public void calcHITSmt(int th, int iter)
      {
         float authinit = 1;
         float hubinit = 1;
         int laenge;
         int threads = th;
         Thread[] t;
         int i,j;
         float authsum;
         float hubsum;
         
         int maxiter = iter;
         laenge = (int) (Math.ceil(graph.length / (double) threads));
         
         auth = new float[urlcount];
         hub = new float[urlcount];
         for (i=0; i<auth.length; i++)
         {auth[i]= authinit;}
         for (i=0; i<hub.length; i++)
         {hub[i]= hubinit;}
         // 2 Arrays mit Hubscore bzw Authscore wurden erstellt und mit 1 initialisiert
         for (i=0;i<maxiter;i++)
            // iterative Berechnung
         {
            //-----Auth-Part-----
            helpauth = new float[threads];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j] = new Thread(new mtAuth(j,laenge));
                  t[j].start();
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            { ie.printStackTrace();}
            //-----Hub-Part-----
            helphub = new float[threads];
            try
            {
               t = new Thread[threads];
               for (j=0; j<threads; j++)
               {
                  t[j] = new Thread(new mtHub(j,laenge));
                  t[j].start();
               }
               for (j=0; j<threads; j++)
               {t[j].join();}
            }
            catch (InterruptedException ie)
            { ie.printStackTrace();}
            //-----Normalisierungspart
            authsum=0;
            hubsum =0;
            for (j=0; j<threads; j++)
            {
               authsum = authsum + helpauth[j];
               hubsum  = hubsum  + helphub[j];
               // berechne die Gesamtsumme der auth und hub
            }
            if (authsum==0)
            {for (j=0; j<auth.length; j++)
             {auth[j]= 0;}}
            else
            {for (j=0; j<auth.length; j++)
             {auth[j]= auth[j]/authsum;}}
            if (hubsum==0)
            {for (j=0; j<hub.length; j++)
             {hub[j]= 0;}}
            else
            {for (j=0; j<hub.length; j++)
             {hub[j]= hub[j]/hubsum;}}
            // fuehrt die Normaliseriung durch
            helpauth = null;
            helphub  = null;
            t = null;
         }  	// ENDE Iteration
      }
      
   }  // ende Class HITSCalculator
   
   //===================================================================================
   //                          Caller
   
   /**
    * Diese oeffentliche Methode schliesst die DB-Connection und entfernt die Datenstruktur.
    */
   public void cleanup()
   {
      graph = null;
      try
      {oc.close();}
      catch (Exception e)
      {e.printStackTrace();}
   }
   
   /**
    * Diese oeffentliche Methode erzeugt die Datenstruktur fuer ein Set mit allen Webseiten.
    */
   public void createGraph()
   {
      GraphCreator h = new GraphCreator();
      h.constructGraph();
      System.out.println("Structure created consisting of "+urlcount+" URLs and "+linkcount+" Links");
   }
   
   /**
    * Diese oeffentliche Methode berechnet die Auth- und Hubscore.
    * Sie ist anwendbar auf alle Sets von Webseiten.
    *
    * @param it  Anzahl der Iterationen die der Algorithmus durchlaufen soll
    * @param thr Anzahl der Threads die verwendet werden soll
    */
   public void calculateHITS(int it, int thr)
   {
      HITSCalculator h = new HITSCalculator();
      h.calcHITSmt(thr,it);
      h.storeHITS();
      h.auth = null;
      h.hub = null;
   }
   
   /**
    * Diese oeffentliche Methode berechnet den PageRank.
    * Sie ist anwendbar auf alle Sets von Webseiten.
    *
    * @param it  Anzahl der Iterationen die der Algorithmus durchlaufen soll
    * @param thr Anzahl der Threads die verwendet werden soll
    */
   public void calculatePageRank(int it, int thr)
   {
      PageRankCalculator h = new PageRankCalculator();
      h.calcPageRankmt(thr,it);
      h.storePR();
      h.rank = null;
   }
   
   /**
    * Diese oeffentliche Methode berechnet den personalisierten PageRank fuer eine Klasse.
    * Dabei ist die Verteilung innerhalb der Klasse uniform.
    * Sie ist anwendbar auf alle Sets von Webseiten.
    *
    * @param it        Anzahl der Iterationen die der Algorithmus durchlaufen soll
    * @param thr       Anzahl der Threads die verwendet werden soll
    * @param classname Name der Klasse fuer die die Berechnung durchgefuehrt werden soll
    * @param factor    Wert mit dem Pages der Klasse bewertet werden, muss >1 sein
    */
   public void calculatePageRankper(int it, int thr, String classname, float factor)
   {
      PageRankCalculator h = new PageRankCalculator();
      h.calcPageRankper(thr,it,classname,factor);
      h.storePR();
      h.rank = null;
   }
   
   /**
    * Diese oeffentliche Methode berechnet den personalisierten PageRank fuer eine Klasse.
    * Dabei ist die Verteilung innerhalb der Topic vom SVM abhaengig.
    * Sie ist anwendbar auf alle Sets von Webseiten.
    *
    * @param it        Anzahl der Iterationen die der Algorithmus durchlaufen soll
    * @param thr       Anzahl der Threads die verwendet werden soll
    * @param classname Name der Klasse fuer die die Berechnung durchgefuehrt werden soll
    */
   public void calculatePageRankSVM(int it, int thr, String classname)
   {
      PageRankCalculator h = new PageRankCalculator();
      h.calcPageRanksvm(thr,it,classname,true);
      h.storePR();
      h.rank = null;
   }
   
   /**
    * Diese oeffentliche Methode berechnet den personalisierten PageRank ueber
    * die Berechnung des BlockRanks der einzelnen Klassen.
    * Sie ist anwendbar auf alle Sets von Webseiten.
    *
    * @param it   Anzahl der Iterationen die der Algorithmus durchlaufen soll
    * @param thr  Anzahl der Threads die verwendet werden soll
    * @param pers Array von BlockPersonalisierungs-Arrays
    */
   public void calculateBlockRank(int it, int thr, BlockPer[][] pers)
   {
      int i;
      BlockRankCalculator h = new BlockRankCalculator();
      h.prepareBlockRank(thr, it);
      for (i=0;i<pers.length;i++)
      {
         h.calcFinal(thr,it,pers[i]);
         h.storeBR(i);
         h.rank = null;
      }
      h.topictest = null;
   }
   
   //------------------------------------------------------
   
   /**
    * Diese oeffentliche Methode erzeugt die Datenstruktur fuer ein Set von Webseiten,
    * wobei ZONs entfernt werden.
    */
   public void createGraph_zd()
   {
      GraphCreator_zd h = new GraphCreator_zd();
      h.constructGraph_zd();
      System.out.println("ZON-free Structure created consisting of "+urlcount+" URLs and "+linkcount+" Links");
   }
   
   /**
    * Diese oeffentliche Methode berechnet den PageRank.
    * Sie ist anwendbar auf Sets von Webseiten bei denen ZONs entfernt wurden.
    *
    * @param it  Anzahl der Iterationen die der Algorithmus durchlaufen soll
    * @param thr Anzahl der Threads die verwendet werden soll
    */
   public void calculatePageRank_zd(int it, int thr)
   {
      PageRankCalculator h = new PageRankCalculator();
      h.calcPageRankmt_zd(thr,it);
      h.storePR();
      h.rank = null;
   }
   
   /**
    * Diese oeffentliche Methode berechnet den personalisierten PageRank fuer eine Klasse.
    * Dabei ist die Verteilung innerhalb der Klasse uniform.
    * Sie ist anwendbar auf Sets von Webseiten bei denen ZONs entfernt wurden.
    *
    * @param it        Anzahl der Iterationen die der Algorithmus durchlaufen soll
    * @param thr       Anzahl der Threads die verwendet werden soll
    * @param classname Name der Klasse fuer die die Berechnung durchgefuehrt werden soll
    * @param factor    Wert mit dem Pages der Klasse bewertet werden, muss >1 sein
    */
   public void calculatePageRankper_zd(int it, int thr, String classname, float factor)
   {
      PageRankCalculator h = new PageRankCalculator();
      h.calcPageRankper_zd(thr,it,classname,factor);
      h.storePR();
      h.rank = null;
   }
   
   /**
    * Diese oeffentliche Methode berechnet den personalisierten PageRank fuer eine Klasse.
    * Dabei ist die Verteilung innerhalb der Topic vom SVM abhaengig.
    * Sie ist anwendbar auf Sets von Webseiten bei denen ZONs entfernt wurden.
    *
    * @param it        Anzahl der Iterationen die der Algorithmus durchlaufen soll
    * @param thr       Anzahl der Threads die verwendet werden soll
    * @param classname Name der Klasse fuer die die Berechnung durchgefuehrt werden soll
    */
   public void calculatePageRankSVM_zd(int it, int thr, String classname)
   {
      PageRankCalculator h = new PageRankCalculator();
      h.calcPageRanksvm_zd(thr,it,classname);
      h.storePR();
      h.rank = null;
   }
   
   /**
    * Diese oeffentliche Methode berechnet den personalisierten PageRank ueber
    * die Berechnung des BlockRanks der einzelnen Klassen.
    * Sie ist anwendbar auf Sets von Webseiten bei denen ZONs entfernt wurden.
    *
    * @param it   Anzahl der Iterationen die der Algorithmus durchlaufen soll
    * @param thr  Anzahl der Threads die verwendet werden soll
    * @param pers Array von BlockPersonalisierungs-Arrays
    */
   public void calculateBlockRank_zd(int it, int thr, BlockPer[][] pers)
   {
      int i;
      BlockRankCalculator h = new BlockRankCalculator();
      h.prepareBlockRank_zd(thr, it);
      for (i=0;i<pers.length;i++)
      {
         h.calcFinal_zd(thr,it,pers[i]);
         h.storeBR(i);
         h.rank = null;
      }
      h.topictest = null;
   }
   
   //===================================================================================
   //                          MAIN
   
   
   public static void main(String[] argument)
   {
      session = SessionBuffer.getInstance(true);
      
      long start,end;
      double x;
      WebRanker a = new WebRanker();
      
      start = System.currentTimeMillis();
      a.createGraph();
      end = System.currentTimeMillis()-start;
      x = end/1000;
      System.out.println("Graph created in "+x+" seconds");
      start = System.currentTimeMillis();
      a.calculateHITS(50,1);
      //   1.Parameter Anzahl Iterationen, 2.Anzahl Threads
      end = System.currentTimeMillis()-start;
      x = end/1000;
      System.out.println("HITS calculated in "+x+" seconds");
      
      start = System.currentTimeMillis();
      a.createGraph_zd();
      end = System.currentTimeMillis()-start;
      x = end/1000;
      System.out.println("Graph PR created in "+x+" seconds");
      start = System.currentTimeMillis();
      a.calculatePageRank_zd(50,1);
      //int iteration, int threads
      end = System.currentTimeMillis()-start;
      x = end/1000;
      System.out.println("PageRank calculated in "+x+" seconds");
   }
   
}