package stored.mi;

import java.sql.Connection;
import java.sql.ResultSet;
import java.sql.Statement;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Vector;

import bingo.util.BingoTreeNode;
import bingo.util.SessionBuffer;

public class MiManager
{
  SessionBuffer session;
  MiTreeNode root;
  Hashtable allClasses;
  Hashtable sortedClasses;
  Vector tasks;
  Vector results;
  Connection conn;
  int maxThreads = 30;
  int perClass = 300;
  
  public Vector myAssistants;
  
  
  public MiManager( int anzahl )
  {
    perClass = anzahl;
    try
    {
      session = SessionBuffer.getInstance();
      conn = session.getOracleConnection();
      allClasses = new Hashtable();
      sortedClasses = new Hashtable();
      myAssistants = new Vector();
      tasks = new Vector();
      results = new Vector();
    }
    catch ( Exception e )
    {
      e.printStackTrace();
    }
  }
  
  
  public void go()
  {
    
    long d1 = ( new java.util.Date() ).getTime();
    
    root = buildTree();
    
    einlesen();
    tasks = root.getTasks();
    System.out.println( "Tasks: " + tasks.size() );
    
    //===========================
    
    MiWorkerThread mwt;
    
    for ( int i = 1; i <= maxThreads; i++ )
    {
      mwt = new MiWorkerThread( this );
      myAssistants.add( mwt );
      new Thread( mwt ).start();
    }
    
    MiStore mst = new MiStore( this );
    
    myAssistants.add( mst );
    new Thread( mst ).start();
    
    //===========================
    
    do
    {
      synchronized ( myAssistants )
      {
        try
        {
          if ( myAssistants.size() > 0 )
          {
            myAssistants.wait();
          }
        }
        catch ( Exception e )
        {
          e.printStackTrace();
        }
      }
    } while ( myAssistants.size() > 0 );
    
    long d2 = ( new java.util.Date() ).getTime() - d1;
    
    System.out.println( "MI preparation using " + maxThreads + " Threads took " + d2 );
    
    try
    {
      conn.close();
    }
    catch ( Exception e )
    {
      e.printStackTrace();
    }
  }
  
  
  public synchronized void addResult( MiTask res )
  {
    results.add( res );
  }
  
  
  public synchronized int getNumAssistants()
  {
    return myAssistants.size();
  }
  
  
  public synchronized int getNumTasks()
  {
    return tasks.size();
  }
  
  
  public synchronized int getNumResults()
  {
    return results.size();
  }
  
  
  public synchronized Vector getResults()
  {
    Vector antwort = null;
    
    if ( getNumResults() > 0 )
    {
      antwort = results;
      results = new Vector();
    }
    return antwort;
  }
  
  
  public synchronized MiTask getTask()
  {
    MiTask antwort = null;
    
    if ( getNumTasks() > 0 )
    {
      antwort = (MiTask) tasks.elementAt( 0 );
      tasks.remove( antwort );
    }
    return antwort;
  }
  
  
  //==========================================================================
  
  public MiTreeNode buildTree()
  {
    BingoTreeNode rootNode = session.getRootNode();
    
    MiTreeNode root = new MiTreeNode( rootNode.label, rootNode.ID );
    
    root.setVorgaenger( null );
    for ( int i = 0; i < rootNode.getChildCount(); i++ )
    {
      root.addNachfolger( buildTree( (BingoTreeNode) rootNode.getChildAt( i ), root ) );
    }
    
    allClasses.put( new Integer( rootNode.ID ), root );
    return root;
  }
  
  
  public MiTreeNode buildTree( BingoTreeNode actNode, MiTreeNode parent )
  {
    String name = actNode.label;
    System.out.println( name );
    int id = actNode.ID;
    
    MiTreeNode antwort = new MiTreeNode( name, id );
    
    antwort.setVorgaenger( parent );
    for ( int i = 0; i < actNode.getChildCount(); i++ )
      antwort.addNachfolger( buildTree( (BingoTreeNode) actNode.getChildAt( i ), antwort ) );
    
    allClasses.put( new Integer( id ), antwort );
    return antwort;
  }
  
  
  //==========================================================================
  
  public void einlesen()
  {
    HashSet workTerms = new HashSet();
    int eingelesen = 0;
    int rootID = session.getRootNode().ID;
    
    try
    {
      Statement stmt = conn.createStatement();
      ResultSet rset = stmt.executeQuery( "SELECT Term, class_id FROM ClassFeatures where class_id != " + String.valueOf( rootID ) );
      
      while ( rset.next() )
      {
        workTerms.add( rset.getString( 1 ) );
        
        MiTreeNode mtn = (MiTreeNode) allClasses.get( new Integer( rset.getInt( 2 ) ) );
        
        mtn.addTerm( rset.getString( 1 ) );
      }
      
      rset = stmt.executeQuery( "SELECT t.Term, b.Class_ID, b.ID " +
      							"FROM DocumentFeatures t, bingodocuments b, classes cl " +
      							"WHERE b.class_id = cl.class_id " +
      							"AND t.id = b.id AND b.STATUS IN ('B','M') " +
      							"AND b.class_id != " + String.valueOf( rootID ) + " " +
      							"ORDER BY b.class_id, b.id" );
      
      int actClass = 0;
      long actDoc = 0;
      
      rset.next();
      
      String actTerm = rset.getString( 1 );
      int prevClass = rset.getInt( 2 );
      long prevDoc = rset.getLong( 3 );
      HashSet actHash = new HashSet();
      
      if ( workTerms.contains( actTerm ) )
      {
        actHash.add( actTerm );
      }
      MiTreeNode actNode = (MiTreeNode) allClasses.get( new Integer( prevClass ) );
      
      while ( rset.next() )
      {
        
        eingelesen++;
        
        actDoc = rset.getLong( 3 );
        if ( actDoc != prevDoc )
        {
          if ( actNode != null )
          {
            actNode.addDocument( prevDoc, actHash );
          }
          actHash = new HashSet();
        }
        
        actClass = rset.getInt( 2 );
        if ( actClass != prevClass )
        {
          actNode = (MiTreeNode) allClasses.get( new Integer( actClass ) );
        }
        
        actTerm = rset.getString( 1 );
        
        if ( workTerms.contains( actTerm ) )
        {
          actHash.add( rset.getString( 1 ) );
        }
        
        prevDoc = actDoc;
        prevClass = actClass;
      }
      actNode.addDocument( prevDoc, actHash );
      rset.close();
      stmt.close();
      System.out.println( "Reading: " + eingelesen );
      conn.close();
    }
    catch ( Exception e )
    {
      e.printStackTrace();
    }
  }
  
  //==========================================================================
  
}

