package stored.mi;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.util.Enumeration;
import java.util.Random;
import java.util.Vector;

import bingo.db.DBInterface;
import bingo.util.SessionBuffer;

public class MiStore implements Runnable {
  SessionBuffer session;
  MiManager parent = null;
  DBInterface dbinterface;
  Connection conn;
  PreparedStatement pstmt;
  Vector results;
  int batchSize;
  int done;
  Random rd;


  public MiStore( MiManager p ) {
    session = SessionBuffer.getInstance();
    batchSize = session.getBatchSize();

    dbinterface = DBInterface.getInstance();

    done = 0;
    this.parent = p;
    conn = dbinterface.conn;

    rd = new Random();
  }


  public void run()
  {
    try {
      pstmt = conn.prepareStatement( "UPDATE ClassFeatures SET MI=? WHERE Term=? and Class_ID=?" );

      while ( parent.getNumAssistants() > 1 || parent.getNumTasks() > 0 || parent.getNumResults() > 0 ) {
        results = parent.getResults();
        if ( results == null || results.size() == 0 ) {
          try {
            Thread.sleep( 100 );
          }
          catch ( Exception e ) {
            e.printStackTrace();
          }
        } else {
          Enumeration enum = results.elements();

          while ( enum.hasMoreElements() ) {
            MiTask res = (MiTask) enum.nextElement();

            pstmt.setDouble( 1, res.getMI() + ( rd.nextDouble() / 10000 ) );
            // small randomization
            pstmt.setString( 2, res.getTerm() );
            pstmt.setInt( 3, res.getID() );
            pstmt.addBatch();
            if ( ++done % batchSize == 0 ) {
              pstmt.executeBatch();
            }
          }
          pstmt.executeBatch();
        }
      }
      pstmt.close();
      conn.commit();
      conn.close();
    }
    catch ( Exception e ) {
      e.printStackTrace();
    }

    synchronized ( parent.myAssistants ) {
      parent.myAssistants.remove( this );
      parent.myAssistants.notifyAll();
    }
    return;
  }
}
