package stored.mi;

import java.util.HashSet;
import java.util.Iterator;

public class MiWorkerThread implements Runnable {
  MiManager parent;
  MiTask task;
  MiTreeNode node;
  String term;


  public MiWorkerThread( MiManager p ) {
    this.parent = p;
    task = null;
    node = null;
    term = null;
  }


  public void run() {
    task = parent.getTask();
    while ( task != null ) {
      node = task.getNode();
      term = task.getTerm();

      performTask();

      parent.addResult( task );
      task = parent.getTask();
    }

    synchronized ( parent.myAssistants ) {
      parent.myAssistants.remove( this );
      parent.myAssistants.notifyAll();
    }
  }


  public int getNumDocuments( MiTreeNode workNode ) {
    if ( workNode.numGesamt >= 0 ) {
      return workNode.numGesamt;
    }

    int antwort = workNode.numDocuments;

    Iterator it = workNode.nachfolger.iterator();

    while ( it.hasNext() ) {
      antwort += getNumDocuments( (MiTreeNode) it.next() );
    }

    workNode.numGesamt = antwort;

    return antwort;
  }


  public int getNumDocuments( String term, MiTreeNode workNode ) {
    Iterator it;
    int antwort = 0;
    Object num;

    num = workNode.numPositive.get( term );
    if ( num != null ) {
      return ( (Integer) num ).intValue();
    }

    it = workNode.documents.values().iterator();
    while ( it.hasNext() ) {
      if ( ( (HashSet) it.next() ).contains( term ) ) {
        antwort++;
      }
    }

    it = workNode.nachfolger.iterator();
    while ( it.hasNext() ) {
      antwort += getNumDocuments( term, (MiTreeNode) it.next() );
    }

    synchronized ( workNode.numPositive ) {
      workNode.numPositive.put( term, new Integer( antwort ) );
    }

    return antwort;
  }


  public int getNumNegativeDocuments( String term, MiTreeNode workNode ) {
    Iterator it;
    int antwort = 0;
    Object num;

    num = workNode.numNegative.get( term );
    if ( num != null ) {
      return ( (Integer) num ).intValue();
    }

    it = workNode.documents.values().iterator();
    while ( it.hasNext() ) {
      if ( !( (HashSet) it.next() ).contains( term ) ) {
        antwort++;
      }
    }

    it = workNode.nachfolger.iterator();
    while ( it.hasNext() ) {
      antwort += getNumNegativeDocuments( term, (MiTreeNode) it.next() );
    }

    synchronized ( workNode.numNegative ) {
      workNode.numPositive.put( term, new Integer( antwort ) );
    }

    return antwort;
  }


  public void performTask() {
    double A;
    double N;
    double B;
    double C;
    double antwort;
    double antwort1;
    double antwort2;
    Iterator it;

    String term = task.getTerm();
    MiTreeNode actNode = task.getNode();
    MiTreeNode parentNode = actNode.getVorgaenger();
    MiTreeNode tmpNode = null;

    // A = Anzahl der Dokumente in Klasse, die den term enthalten
    A = (double) getNumDocuments( term, actNode );

    // B = Anzahl der Dokumente in Nachbarn, die den Term enthalten
    B = (double) 0.0;

    it = parentNode.nachfolger.iterator();
    while ( it.hasNext() ) {
      tmpNode = (MiTreeNode) it.next();
      if ( tmpNode != actNode ) {
        B += (double) getNumDocuments( term, tmpNode );
      }
    }

    // C = Anzahl Dokumente in Klasse ohne term
    C = (double) getNumDocuments( actNode ) - A;

    // N = Gesamtzahl der Dokumente in Klasse und Nachbarn
    N = (double) 0.0;
    it = parentNode.nachfolger.iterator();
    while ( it.hasNext() ) {
      N += (double) getNumDocuments( (MiTreeNode) it.next() );
    }

    // Positive Connection
    antwort1 = A * N / ( ( A + B ) * ( A + C ) );
    antwort1 = Math.log( antwort1 ) / Math.log( 2 );
    antwort1 = ( A / N ) * antwort1;

    //Negative Connection
    if ( B > 0 && ( N > ( A + C ) ) ) {
      antwort2 = B * N / ( ( A + B ) * ( N - A - C ) );
      antwort2 = Math.log( antwort2 ) / Math.log( 2 );
      antwort2 = ( B / N ) * antwort2;
    } else {
      antwort2 = 0;
    }

    antwort = antwort1 + antwort2;

    task.setMI( antwort );
  }

}
