package jodd.bean;

import java.math.BigDecimal;
import java.sql.Date;
import java.sql.Time;
import java.sql.Timestamp;
import java.util.HashMap;

import jodd.bean.modifiers.TrimStrings;
import junit.framework.TestCase;

public class TestBeanUtil extends TestCase {

	public void testSetPropertyNumbers() {
		FooBean fb = new FooBean();

		// Integer
		boolean valid;
		String propName = "fooInteger";
		valid = BeanUtil.setProperty(fb, propName, new Integer(1));
		assertTrue(valid);
		assertEquals(1, fb.getFooInteger().intValue());
		valid = BeanUtil.setProperty(fb, propName, null);
		assertTrue(valid);
		assertNull(fb.getFooInteger());
		valid = BeanUtil.setProperty(fb, propName, "2");			// valid string
		assertTrue(valid);
		assertEquals(2, fb.getFooInteger().intValue());
		valid = BeanUtil.setProperty(fb, propName, "x");			// invalid string - value stays the same
		assertFalse(valid);
		assertEquals(2, fb.getFooInteger().intValue());

		// int
		propName = "fooint";
		valid = BeanUtil.setProperty(fb, propName, new Integer(1));
		assertTrue(valid);
		assertEquals(1, fb.getFooint());
		valid = BeanUtil.setProperty(fb, propName, null);			// null is not an int
		assertFalse(valid);
		assertEquals(1, fb.getFooint());
		valid = BeanUtil.setProperty(fb, propName, "2");			// valid string
		assertTrue(valid);
		assertEquals(2, fb.getFooint());
		valid = BeanUtil.setProperty(fb, propName, "w");			// invalid string
		assertFalse(valid);
		assertEquals(2, fb.getFooint());

		// Long
		propName = "fooLong";
		valid = BeanUtil.setProperty(fb, propName, new Long(1));
		assertTrue(valid);
		assertEquals(1L, fb.getFooLong().longValue());
		valid = BeanUtil.setProperty(fb, propName, new Integer(3));
		assertTrue(valid);
		assertEquals(3L, fb.getFooLong().longValue());
		valid = BeanUtil.setProperty(fb, propName, null);
		assertTrue(valid);
		assertNull(fb.getFooLong());
		valid = BeanUtil.setProperty(fb, propName, "2");			// valid string
		assertTrue(valid);
		assertEquals(2L, fb.getFooLong().longValue());
		valid = BeanUtil.setProperty(fb, propName, "x");			// invalid string - value stays the same
		assertFalse(valid);
		assertEquals(2L, fb.getFooLong().longValue());

		// long
		propName = "foolong";
		valid = BeanUtil.setProperty(fb, propName, new Long(1));
		assertTrue(valid);
		assertEquals(1L, fb.getFoolong());
		valid = BeanUtil.setProperty(fb, propName, new Integer(3));
		assertTrue(valid);
		assertEquals(3L, fb.getFoolong());
		valid = BeanUtil.setProperty(fb, propName, null);			// null is not a long
		assertFalse(valid);
		assertEquals(3L, fb.getFoolong());
		valid = BeanUtil.setProperty(fb, propName, "2");			// valid string
		assertTrue(valid);
		assertEquals(2L, fb.getFoolong());
		valid = BeanUtil.setProperty(fb, propName, "w");			// invalid string
		assertFalse(valid);
		assertEquals(2L, fb.getFoolong());

		// Byte
		propName = "fooByte";
		valid = BeanUtil.setProperty(fb, propName, new Byte((byte) 1));
		assertTrue(valid);
		assertEquals(1, fb.getFooByte().byteValue());
		valid = BeanUtil.setProperty(fb, propName, new Integer(3));
		assertTrue(valid);
		assertEquals(3, fb.getFooByte().byteValue());
		valid = BeanUtil.setProperty(fb, propName, new Integer(257));
		assertTrue(valid);
		assertEquals(1, fb.getFooByte().byteValue());				// lower byte of 257
		valid = BeanUtil.setProperty(fb, propName, null);
		assertTrue(valid);
		assertNull(fb.getFooByte());
		valid = BeanUtil.setProperty(fb, propName, "2");			// valid string
		assertTrue(valid);
		assertEquals(2, fb.getFooByte().byteValue());
		valid = BeanUtil.setProperty(fb, propName, "x");			// invalid string - value stays the same
		assertFalse(valid);
		assertEquals(2, fb.getFooByte().byteValue());

		// byte
		propName = "foobyte";
		valid = BeanUtil.setProperty(fb, propName, new Byte((byte) 1));
		assertTrue(valid);
		assertEquals(1, fb.getFoobyte());
		valid = BeanUtil.setProperty(fb, propName, new Integer(3));
		assertTrue(valid);
		assertEquals(3, fb.getFoobyte());
		valid = BeanUtil.setProperty(fb, propName, new Integer(257));
		assertTrue(valid);
		assertEquals(1, fb.getFoobyte());
		valid = BeanUtil.setProperty(fb, propName, null);			// null is not a byte
		assertFalse(valid);
		assertEquals(1, fb.getFoobyte());
		valid = BeanUtil.setProperty(fb, propName, "2");			// valid string
		assertTrue(valid);
		assertEquals(2, fb.getFoobyte());
		valid = BeanUtil.setProperty(fb, propName, "x");			// invalid string - value stays the same
		assertFalse(valid);
		assertEquals(2, fb.getFoobyte());

		// Boolean
		propName = "fooBoolean";
		valid = BeanUtil.setProperty(fb, propName, Boolean.TRUE);
		assertTrue(valid);
		assertEquals(true, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, new Boolean(false));
		assertTrue(valid);
		assertEquals(false, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, null);
		assertTrue(valid);
		assertNull(fb.getFooBoolean());
		valid = BeanUtil.setProperty(fb, propName, "yes");
		assertTrue(valid);
		assertEquals(true, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, "y");
		assertTrue(valid);
		assertEquals(true, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, "true");
		assertTrue(valid);
		assertEquals(true, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, "on");
		assertTrue(valid);
		assertEquals(true, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, "1");
		assertTrue(valid);
		assertEquals(true, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, "no");
		assertTrue(valid);
		assertEquals(false, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, "n");
		assertTrue(valid);
		assertEquals(false, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, "false");
		assertTrue(valid);
		assertEquals(false, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, "off");
		assertTrue(valid);
		assertEquals(false, fb.getFooBoolean().booleanValue());
		valid = BeanUtil.setProperty(fb, propName, "0");
		assertTrue(valid);
		assertEquals(false, fb.getFooBoolean().booleanValue());

		// boolean
		propName = "fooboolean";
		valid = BeanUtil.setProperty(fb, propName, Boolean.TRUE);
		assertTrue(valid);
		assertEquals(true, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, new Boolean(false));
		assertTrue(valid);
		assertEquals(false, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, null);
		assertFalse(valid);
		assertEquals(false, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "yes");
		assertTrue(valid);
		assertEquals(true, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "y");
		assertTrue(valid);
		assertEquals(true, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "true");
		assertTrue(valid);
		assertEquals(true, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "on");
		assertTrue(valid);
		assertEquals(true, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "1");
		assertTrue(valid);
		assertEquals(true, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "no");
		assertTrue(valid);
		assertEquals(false, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "n");
		assertTrue(valid);
		assertEquals(false, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "false");
		assertTrue(valid);
		assertEquals(false, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "off");
		assertTrue(valid);
		assertEquals(false, fb.getFooboolean());
		valid = BeanUtil.setProperty(fb, propName, "0");
		assertTrue(valid);
		assertEquals(false, fb.getFooboolean());

		// Float
		propName = "fooFloat";
		valid = BeanUtil.setProperty(fb, propName, new Float(1.1));
		assertTrue(valid);
		assertEquals(1.1, fb.getFooFloat().floatValue(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, new Integer(3));
		assertTrue(valid);
		assertEquals(3.0, fb.getFooFloat().floatValue(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, null);
		assertTrue(valid);
		assertNull(fb.getFooFloat());
		valid = BeanUtil.setProperty(fb, propName, "2.2");			// valid string
		assertTrue(valid);
		assertEquals(2.2, fb.getFooFloat().floatValue(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, "x");			// invalid string - value stays the same
		assertFalse(valid);
		assertEquals(2.2, fb.getFooFloat().floatValue(), 0.0005);

		// float
		propName = "foofloat";
		valid = BeanUtil.setProperty(fb, propName, new Float(1.1));
		assertTrue(valid);
		assertEquals(1.1, fb.getFoofloat(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, new Integer(3));
		assertTrue(valid);
		assertEquals(3.0, fb.getFoofloat(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, null);			// null is not a long
		assertFalse(valid);
		assertEquals(3.0, fb.getFoofloat(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, "2.2");			// valid string
		assertTrue(valid);
		assertEquals(2.2, fb.getFoofloat(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, "w");			// invalid string
		assertFalse(valid);
		assertEquals(2.2, fb.getFoofloat(), 0.0005);


		// Double
		propName = "fooDouble";
		valid = BeanUtil.setProperty(fb, propName, new Double(1.1));
		assertTrue(valid);
		assertEquals(1.1, fb.getFooDouble().doubleValue(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, new Integer(3));
		assertTrue(valid);
		assertEquals(3.0, fb.getFooDouble().doubleValue(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, null);
		assertTrue(valid);
		assertNull(fb.getFooDouble());
		valid = BeanUtil.setProperty(fb, propName, "2.2");			// valid string
		assertTrue(valid);
		assertEquals(2.2, fb.getFooDouble().doubleValue(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, "x");			// invalid string - value stays the same
		assertFalse(valid);
		assertEquals(2.2, fb.getFooDouble().doubleValue(), 0.0005);

		// double
		propName = "foodouble";
		valid = BeanUtil.setProperty(fb, propName, new Double(1.1));
		assertTrue(valid);
		assertEquals(1.1, fb.getFoodouble(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, new Integer(3));
		assertTrue(valid);
		assertEquals(3.0, fb.getFoodouble(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, null);			// null is not a long
		assertFalse(valid);
		assertEquals(3.0, fb.getFoodouble(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, "2.2");			// valid string
		assertTrue(valid);
		assertEquals(2.2, fb.getFoodouble(), 0.0005);
		valid = BeanUtil.setProperty(fb, propName, "w");			// invalid string
		assertFalse(valid);
		assertEquals(2.2, fb.getFoodouble(), 0.0005);
	}


	public void testSetPropertySql() {
		FooBean2 fb = new FooBean2();

		String propName = "fooTimestamp";
		Timestamp ts = new Timestamp(101, 0, 17, 1, 2, 3, 4);
		boolean valid = BeanUtil.setProperty(fb, propName, ts);
		assertTrue(valid);
		assertEquals("2001-01-17 01:02:03.000000004", fb.getFooTimestamp().toString());

		propName = "fooTime";
		Time t = new Time(17, 13, 15);
		valid = BeanUtil.setProperty(fb, propName, t);
		assertTrue(valid);
		assertEquals("17:13:15", fb.getFooTime().toString());

		propName = "fooDate";
		Date d = new Date(101, 1, 17);
		valid = BeanUtil.setProperty(fb, propName, d);
		assertTrue(valid);
		assertEquals("2001-02-17", fb.getFooDate().toString());
	}


	public void testSetPropertyMath() {
		FooBean2 fb = new FooBean2();

		String propName = "fooBigDecimal";
		boolean valid = BeanUtil.setProperty(fb, propName, new BigDecimal(1.2));
		assertTrue(valid);
		assertEquals(1.2, fb.getFooBigDecimal().doubleValue(), 0.0005);
	}

	public void testSetPropertyString() {
		FooBean fb = new FooBean();

		// String
		String propName = "fooString";
		boolean valid = BeanUtil.setProperty(fb, propName, "string");
		assertTrue(valid);
		assertEquals("string", fb.getFooString());
		valid = BeanUtil.setProperty(fb, propName, null);
		assertTrue(valid);
		assertNull(fb.getFooString());

		// String array
		propName = "fooStringA";
		String[] sa = new String[] {"one", "two", "three"};
		valid = BeanUtil.setProperty(fb, propName, sa);
		assertTrue(valid);
		assertEquals("one", fb.getFooStringA()[0]);
		assertEquals("two", fb.getFooStringA()[1]);
		assertEquals("three", fb.getFooStringA()[2]);

		valid = BeanUtil.setProperty(fb, propName, "just a string");
		assertTrue(valid);
		sa = (String[]) BeanUtil.getProperty(fb, propName);
		assertEquals(1, sa.length);
		assertEquals("just a string", sa[0]);

		// Character
		propName = "fooCharacter";
		valid = BeanUtil.setProperty(fb, propName, new Character('a'));
		assertTrue(valid);
		assertEquals('a', fb.getFooCharacter().charValue());
		valid = BeanUtil.setProperty(fb, propName, new String("123"));
		assertTrue(valid);
		assertEquals('1', fb.getFooCharacter().charValue());
		valid = BeanUtil.setProperty(fb, propName, new Integer(789));
		assertTrue(valid);
		assertEquals('7', fb.getFooCharacter().charValue());

		// char
		propName = "foochar";
		valid = BeanUtil.setProperty(fb, propName, new Character('a'));
		assertTrue(valid);
		assertEquals('a', fb.getFoochar());
		valid = BeanUtil.setProperty(fb, propName, new String("123"));
		assertTrue(valid);
		assertEquals('1', fb.getFoochar());
		valid = BeanUtil.setProperty(fb, propName, new Integer(789));
		assertTrue(valid);
		assertEquals('7', fb.getFoochar());
	}


	public void	testLoaders() {
		HashMap map = new HashMap();
		
		map.put("fooInteger", new Integer(1));
		map.put("fooint", new Integer(2));
		map.put("fooLong", new Long(3));
		map.put("foolong", new Long(4));
		map.put("fooByte", new Byte((byte)5));
		map.put("foobyte", new Byte((byte)6));
		map.put("fooCharacter", new Character('7'));
		map.put("foochar", new Character('8'));
		map.put("fooBoolean", new Boolean(true));
		map.put("fooboolean", new Boolean(false));
		map.put("fooFloat", new Float(9.0));
		map.put("foofloat", new Float(10.0));
		map.put("fooDouble", new Double(11.0));
		map.put("foodouble", new Double(12.0));
		map.put("fooString", new String("13"));
		map.put("fooStringA", new String[]{"14", "15"});

		FooBean fb = new FooBean();
		BeanUtil.load(fb, map);

		assertEquals(1, fb.getFooInteger().intValue());
		assertEquals(2, fb.getFooint());
		assertEquals(3, fb.getFooLong().longValue());
		assertEquals(4, fb.getFoolong());
		assertEquals(5, fb.getFooByte().byteValue());
		assertEquals(6, fb.getFoobyte());
		assertEquals('7', fb.getFooCharacter().charValue());
		assertEquals('8', fb.getFoochar());
		assertTrue(fb.getFooBoolean().booleanValue());
		assertFalse(fb.getFooboolean());
		assertEquals(9.0, fb.getFooFloat().floatValue(), 0.005);
		assertEquals(10.0, fb.getFoofloat(), 0.005);
		assertEquals(11.0, fb.getFooDouble().doubleValue(), 0.005);
		assertEquals(12.0, fb.getFoodouble(), 0.005);
		assertEquals("13", fb.getFooString());
		assertEquals("14", fb.getFooStringA()[0]);
		assertEquals("15", fb.getFooStringA()[1]);
	}

	public void	testForEach() {
		FooBean fb = new FooBean();
		fb.setFooString("   xxx   ");
		fb.setFooStringA(new String[] {"   xxx   ", "  yy  ", " z "});
		BeanUtil.forEachProperty(fb, new TrimStrings());
		assertEquals("xxx", fb.getFooString());
		assertEquals("xxx", fb.getFooStringA()[0]);
		assertEquals("yy", fb.getFooStringA()[1]);
		assertEquals("z", fb.getFooStringA()[2]);
	}


	public void testGet() {
		FooBean fb = new FooBean();
		fb.setFooInteger(new Integer(101));
		fb.setFooint(102);
		fb.setFooLong(new Long(103));
		fb.setFoolong(104);
		fb.setFooByte(new Byte((byte) 105));
		fb.setFoobyte((byte) 106);
		fb.setFooCharacter(new Character('7'));
		fb.setFoochar('8');
		fb.setFooBoolean(new Boolean(true));
		fb.setFooboolean(false);
		fb.setFooFloat(new Float(109.0));
		fb.setFoofloat((float)110.0);
		fb.setFooDouble(new Double(111.0));
		fb.setFoodouble(112.0);
		fb.setFooString("113");
		fb.setFooStringA(new String[] {"114", "115"} );

		Integer v =	(Integer) BeanUtil.getProperty(fb, "FooInteger");
		assertEquals(101, v.intValue());
		v = (Integer) BeanUtil.getProperty(fb, "Fooint");
		assertEquals(102, v.intValue());
		Long vl = (Long) BeanUtil.getProperty(fb, "FooLong");
		assertEquals(103, vl.longValue());
		vl = (Long) BeanUtil.getProperty(fb, "Foolong");
		assertEquals(104, vl.longValue());
		Byte vb = (Byte) BeanUtil.getProperty(fb, "FooByte");
		assertEquals(105, vb.intValue());
		vb = (Byte) BeanUtil.getProperty(fb, "Foobyte");
		assertEquals(106, vb.intValue());
		Character c = (Character) BeanUtil.getProperty(fb, "FooCharacter");
		assertEquals('7', c.charValue());
		c = (Character) BeanUtil.getProperty(fb, "Foochar");
		assertEquals('8', c.charValue());
		Boolean b = (Boolean) BeanUtil.getProperty(fb, "FooBoolean");
		assertTrue(b.booleanValue());
		b = (Boolean) BeanUtil.getProperty(fb, "Fooboolean");
		assertFalse(b.booleanValue());
		Float f = (Float) BeanUtil.getProperty(fb, "FooFloat");
		assertEquals(109.0, f.floatValue(), 0.005);
		f = (Float) BeanUtil.getProperty(fb, "Foofloat");
		assertEquals(110.0, f.floatValue(), 0.005);
		Double d = (Double) BeanUtil.getProperty(fb, "FooDouble");
		assertEquals(111.0, d.doubleValue(), 0.005);
		d = (Double) BeanUtil.getProperty(fb, "Foodouble");
		assertEquals(112.0, d.doubleValue(), 0.005);
		String s = (String) BeanUtil.getProperty(fb, "FooString");
		assertEquals("113", s);
		String[] sa = (String[]) BeanUtil.getProperty(fb, "FooStringA");
		assertEquals(2, sa.length);
		assertEquals("114", sa[0]);
		assertEquals("115", sa[1]);

	}


	public void testCopy() {
		FooBean fb = new FooBean();
		fb.setFooInteger(new Integer(201));
		fb.setFooint(202);
		fb.setFooLong(new Long(203));
		fb.setFoolong(204);
		fb.setFooByte(new Byte((byte) 95));
		fb.setFoobyte((byte) 96);
		fb.setFooCharacter(new Character('7'));
		fb.setFoochar('8');
		fb.setFooBoolean(new Boolean(true));
		fb.setFooboolean(false);
		fb.setFooFloat(new Float(209.0));
		fb.setFoofloat((float)210.0);
		fb.setFooDouble(new Double(211.0));
		fb.setFoodouble(212.0);
		fb.setFooString("213");
		fb.setFooStringA(new String[] {"214", "215"} );

		FooBean dest = new FooBean();
		BeanUtil.copyProperties(fb, dest);

		Integer v =	(Integer) BeanUtil.getProperty(dest, "FooInteger");
		assertEquals(201, v.intValue());
		v = (Integer) BeanUtil.getProperty(dest, "Fooint");
		assertEquals(202, v.intValue());
		Long vl = (Long) BeanUtil.getProperty(dest, "FooLong");
		assertEquals(203, vl.longValue());
		vl = (Long) BeanUtil.getProperty(dest, "Foolong");
		assertEquals(204, vl.longValue());
		Byte vb = (Byte) BeanUtil.getProperty(dest, "FooByte");
		assertEquals(95, vb.intValue());
		vb = (Byte) BeanUtil.getProperty(dest, "Foobyte");
		assertEquals(96, vb.intValue());
		Character c = (Character) BeanUtil.getProperty(dest, "FooCharacter");
		assertEquals('7', c.charValue());
		c = (Character) BeanUtil.getProperty(dest, "Foochar");
		assertEquals('8', c.charValue());
		Boolean b = (Boolean) BeanUtil.getProperty(dest, "FooBoolean");
		assertTrue(b.booleanValue());
		b = (Boolean) BeanUtil.getProperty(dest, "Fooboolean");
		assertFalse(b.booleanValue());
		Float f = (Float) BeanUtil.getProperty(dest, "FooFloat");
		assertEquals(209.0, f.floatValue(), 0.005);
		f = (Float) BeanUtil.getProperty(dest, "Foofloat");
		assertEquals(210.0, f.floatValue(), 0.005);
		Double d = (Double) BeanUtil.getProperty(dest, "FooDouble");
		assertEquals(211.0, d.doubleValue(), 0.005);
		d = (Double) BeanUtil.getProperty(dest, "Foodouble");
		assertEquals(212.0, d.doubleValue(), 0.005);
		String s = (String) BeanUtil.getProperty(dest, "FooString");
		assertEquals("213", s);
		String[] sa = (String[]) BeanUtil.getProperty(dest, "FooStringA");
		assertEquals(2, sa.length);
		assertEquals("214", sa[0]);
		assertEquals("215", sa[1]);


		FooBean empty = new FooBean();
		BeanUtil.copyProperties(empty, dest);
		v =	(Integer) BeanUtil.getProperty(dest, "FooInteger");
		assertNull(v);
		v = (Integer) BeanUtil.getProperty(dest, "Fooint");
		assertEquals(0, v.intValue());
		vl = (Long) BeanUtil.getProperty(dest, "FooLong");
		assertNull(vl);
		vl = (Long) BeanUtil.getProperty(dest, "Foolong");
		assertEquals(0, vl.longValue());
		vb = (Byte) BeanUtil.getProperty(dest, "FooByte");
		assertNull(vb);
		vb = (Byte) BeanUtil.getProperty(dest, "Foobyte");
		assertEquals(0, vb.byteValue());
		c = (Character) BeanUtil.getProperty(dest, "FooCharacter");
		assertNull(c);
		c = (Character) BeanUtil.getProperty(dest, "Foochar");
		assertEquals(0, c.charValue());
		b = (Boolean) BeanUtil.getProperty(dest, "FooBoolean");
		assertNull(b);
		b = (Boolean) BeanUtil.getProperty(dest, "Fooboolean");
		assertFalse(b.booleanValue());
		f = (Float) BeanUtil.getProperty(dest, "FooFloat");
		assertNull(f);
		f = (Float) BeanUtil.getProperty(dest, "Foofloat");
		assertEquals(0, f.floatValue(), 0.005);
		d = (Double) BeanUtil.getProperty(dest, "FooDouble");
		assertNull(d);
		d = (Double) BeanUtil.getProperty(dest, "Foodouble");
		assertEquals(0, d.doubleValue(), 0.005);
		s = (String) BeanUtil.getProperty(dest, "FooString");
		assertNull(s);
		sa = (String[]) BeanUtil.getProperty(dest, "FooStringA");
		assertNull(sa);
	}

}
