package jodd.file;

import junit.framework.TestCase;

public class TestFileUtil extends TestCase {

	public void testObjSerialization() {
		SimpleBean sb = new SimpleBean();
		assertEquals(3, sb.getFoo());
		try {
			FileUtil.writeObject("test/data/sb.dat", sb);
		} catch (Exception ex) {
			fail("FileUtil.writeObject " + ex.toString());
		}

		try {
			Object o = FileUtil.readObject("test/data/sb.dat");
			SimpleBean sb2 = (SimpleBean) o;
			assertEquals(3, sb2.getFoo());
		} catch (Exception ex) {
			fail("FileUtil.readObject " + ex.toString());
		}
	}


	public void testFileManipulation() {

		try {
			FileUtil.copy("test/data/sb.dat", "test/data/sb1.dat");
		} catch (Exception ex) {
			fail("FileUtil.copy " + ex.toString());
		}
	}


	public void testString() {
		String s = "This is a test file\nIt only has\nthree lines!!";

		try {
			FileUtil.writeString("test/data/test.txt", s);
		} catch(Exception ex) {
			fail("FileUtil.writeString " + ex.toString());
		}

		String s2 = null;
		try {
			s2 = FileUtil.readString("test/data/test.txt");
		} catch(Exception ex) {
			fail("FileUtil.readString " + ex.toString());
		}

		assertEquals(s, s2);

		// test unicode chars (i.e. greaer then 255)
		char buf[] = s.toCharArray();
		buf[0] = 256;
		s = new String(buf);

		try {
			FileUtil.writeString("test/data/test.txt", s);
		} catch(Exception ex) {
			fail("FileUtil.writeString " + ex.toString());
		}

		try {
			s2 = FileUtil.readString("test/data/test.txt");
		} catch(Exception ex) {
			fail("FileUtil.readString " + ex.toString());
		}

		assertEquals(s.substring(1), s2.substring(1));
		assertTrue(s.charAt(0) != s2.charAt(0));
	}

	public void testUnicodeString() {
		String s = "This is a test file\nIt only has\nthree lines!!";

		char buf[] = s.toCharArray();
		buf[0] = 256;
		s = new String(buf);

		try {
			FileUtil.writeString("test/data/test2.txt", s, "UTF-16");
		} catch(Exception ex) {
			fail("FileUtil.writeString " + ex.toString());
		}

		String s2 = null;

		try {
			s2 = FileUtil.readString("test/data/test2.txt", "UTF-16");
		} catch(Exception ex) {
			fail("FileUtil.readString " + ex.toString());
		}

		assertEquals(s, s2);

	}
}
