package jodd.servlet;

import junit.framework.TestCase;

public class TestHtmlUtil extends TestCase {

	public void testGetName() {
		String tag = "<select name=\"Marie\" >";
		String s = HtmlUtil.getAttribute(tag, "name");
		assertEquals("Marie", s);

		tag = "<select NAME='Marie' >";
		s = HtmlUtil.getAttribute(tag, "name");
		assertEquals("Marie", s);

		tag = "<select NaME=\"Marie\">";
		s = HtmlUtil.getAttribute(tag, "name");
		assertEquals("Marie", s);

		tag = "<select nake='Marie' >";
		s = HtmlUtil.getAttribute(tag, "name");
		assertNull(s);
		tag = "<select nake=\"Marie >";
		s = HtmlUtil.getAttribute(tag, "name");
		assertNull(s);

		tag = "<select name='   Marie \\\"the Hot\\\" >";
		s = HtmlUtil.getAttribute(tag, "name");
		assertNull(s);			// attribute not closed
		tag += "'>";
		s = HtmlUtil.getAttribute(tag, "name");
		assertEquals("   Marie \\\"the Hot\\\" >", s);
	}

	public void testTagName() {
		String body = "<html><   TaG atr='val'>";
		String tag = HtmlUtil.getTagName(body);
		assertEquals("html", tag);

		tag = HtmlUtil.getTagName(body, 1);
		assertNull(tag);

		tag = HtmlUtil.getTagName(body, 6);
		assertEquals("TaG", tag);

		body = "</html></    tag>";
		tag = HtmlUtil.getTagName(body);
		assertEquals("/html", tag);

		tag = HtmlUtil.getTagName(body, 7);
		assertEquals("/tag", tag);
	}



	public void testAddValue() {
		String tag = "<input name=\"xxx\">";
		String s = HtmlUtil.addAttribute(tag, "value", "qwerty");
		assertEquals("<input name=\"xxx\" value=\"qwerty\">" , s);

		s = HtmlUtil.addAttribute(tag, "value", "123\"");
		assertEquals("<input name=\"xxx\" value=\"123&quot;\">", s);

		s = HtmlUtil.addAttribute(tag, "value");
		assertEquals("<input name=\"xxx\" value>", s);


		tag = "<before><input name=\"xxx\">";
		s = HtmlUtil.addAttribute(tag, "value", "qwerty", 8);
		assertEquals("<input name=\"xxx\" value=\"qwerty\">" , s);

		s = HtmlUtil.addAttribute(tag, "value", "123\"", 8);
		assertEquals("<input name=\"xxx\" value=\"123&quot;\">", s);

		s = HtmlUtil.addAttribute(tag, "value", 8);
		assertEquals("<input name=\"xxx\" value>", s);

	}

}
