package jodd.servlet;

import java.util.HashMap;
import java.util.Map;

import junit.framework.TestCase;

public class TestServletUtil extends TestCase {

	public void testMakeUrlParams() {
		HashMap hm = new HashMap();
		hm.put("param1", "value1");
		assertEquals("param1=value1", ServletUtil.makeUrlParams(hm));
		
		hm.put("param2", "value2");
		String param = ServletUtil.makeUrlParams(hm);
		assertTrue(param.indexOf("param1=value1") != -1);
		assertTrue(param.indexOf("param2=value2") != -1);
		assertTrue(param.indexOf("&param") != -1);
		
		hm.put("param3", "value3");
		param = ServletUtil.makeUrlParams(hm);
		assertTrue(param.indexOf("param1=value1") != -1);
		assertTrue(param.indexOf("param2=value2") != -1);
		assertTrue(param.indexOf("param3=value3") != -1);
		int i = param.indexOf("&param");
		assertTrue(i != -1);
		assertTrue(param.indexOf("&param", i) != -1);

		hm.clear();
		hm.put("param1", "1/2?3&4");
		assertEquals("param1=1%2F2%3F3%264", ServletUtil.makeUrlParams(hm));
	}

	public void testGetUrlParams() {
		String url = "?param1=value1";
		Map map = ServletUtil.getUrlParams(url);
		assertEquals("value1", map.get("param1"));

		url = "foo.do?param1=value1";
		map = ServletUtil.getUrlParams(url);
		assertEquals("value1", map.get("param1"));
		assertEquals(1, map.size());

		url = "?param1=value1&param2=value2";
		map = ServletUtil.getUrlParams(url);
		assertEquals("value1", map.get("param1"));
		assertEquals("value2", map.get("param2"));
		assertEquals(2, map.size());

		url = "?param1=1%2F2%3F3%264&param2=value2&param3=value3";
		map = ServletUtil.getUrlParams(url);
		assertEquals("1/2?3&4", map.get("param1"));
		assertEquals("value2", map.get("param2"));
		assertEquals("value3", map.get("param3"));
		assertNull(map.get("param4"));
		assertEquals(3, map.size());


		url = "?param1=value1&";
		map = ServletUtil.getUrlParams(url);
		assertEquals("value1", map.get("param1"));
		assertEquals(1, map.size());
		
		url = "?param1=value1&&";
		map = ServletUtil.getUrlParams(url);
		assertEquals("value1", map.get("param1"));
		assertEquals(1, map.size());

		url = "?param1=value1&&&&&";
		map = ServletUtil.getUrlParams(url);
		assertEquals("value1", map.get("param1"));
		assertEquals(1, map.size());
		
		url = "?param1=";
		map = ServletUtil.getUrlParams(url);
		assertEquals("", map.get("param1"));
		assertEquals(1, map.size());

		url = "?param2=value2&param1=";
		map = ServletUtil.getUrlParams(url);
		assertEquals("", map.get("param1"));
		assertEquals("value2", map.get("param2"));
		assertEquals(2, map.size());

		url = "?param1=&param2=value2";
		map = ServletUtil.getUrlParams(url);
		assertEquals("", map.get("param1"));
		assertEquals("value2", map.get("param2"));
		assertEquals(2, map.size());

		url = "?param1";
		map = ServletUtil.getUrlParams(url);
		assertEquals("", map.get("param1"));
		assertEquals(1, map.size());

		url = "?param2=value2&param1";
		map = ServletUtil.getUrlParams(url);
		assertEquals("", map.get("param1"));
		assertEquals("value2", map.get("param2"));
		assertEquals(2, map.size());

		url = "?param1&param2=value2";
		map = ServletUtil.getUrlParams(url);
		assertEquals("", map.get("param1"));
		assertEquals("value2", map.get("param2"));
		assertEquals(2, map.size());

		url = "";
		map = ServletUtil.getUrlParams(url);
		assertEquals(0, map.size());

		url = "action.do?";
		map = ServletUtil.getUrlParams(url);
		assertEquals(0, map.size());

		url = "action.do";
		map = ServletUtil.getUrlParams(url);
		assertEquals(0, map.size());

		url = "action.do?&";
		map = ServletUtil.getUrlParams(url);
		assertEquals(0, map.size());

		url = "action.do?&param1=value1";
		map = ServletUtil.getUrlParams(url);
		assertEquals("value1", map.get("param1"));
		assertEquals(1, map.size());
	}

}
