import java.io.*;
import javax.servlet.*;
import javax.servlet.http.*;

import jodd.servlet.*;


public class MvcAction extends ActionServlet {

	/**
	 * Main entry point for the framework actions. No redirection and forwarding
	 * is allowed here - instead, a string is returned that represents
	 * 'mapping', i.e. key from the configuration file that defines where to
	 * forward or redirect after this method finishes.
	 *
	 * @param request
	 * @param response
	 *
	 * @return
	 * @exception IOException
	 * @exception ServletException
	 */
	public String doAction(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
		System.out.println("---------------------");
		System.out.println("FooAction.doAction()");
		
		String mapping = invokeAction(request, response);
		System.out.println("invoke result: " + mapping);

		return mapping;
	}


	/**
	 * This is an action handler that returns bad mapping.
	 *
	 * @param request
	 * @param response
	 *
	 * @return
	 * @exception IOException
	 * @exception ServletException
	 */
	public String bad(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
		System.out.println(">ACTION:bad");
		return "notok";
	}

	/**
	 * This is an action handler that returns good mapping. These action handlers are
	 * invoked by doAction() method and they usually returns a String, as if they
	 * were doActions itself, although it is not necessary.
	 *
	 * @param request
	 * @param response
	 *
	 * @return
	 * @exception IOException
	 * @exception ServletException
	 */
	public String good(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
		System.out.println(">ACTION:good");
		System.out.println("param: " + getActionParameter(request, "nnaammee"));
		System.out.println("action path: " + getActionPath(request));
		System.out.println("forward path: " + getActionForwardPath(request, "ok"));
		System.out.println("action type: " + getActionType(request));
		System.out.println("action method: " +getActionMethodName(request));
		return "ok";
	}


	/**
	 * Reloads XML configuration file of the controller without restarting the
	 * application.
	 *
	 * @param request
	 * @param response
	 *
	 * @exception IOException
	 * @exception ServletException
	 */
	public String reload(HttpServletRequest request, HttpServletResponse response) throws IOException, ServletException {
		System.out.println(">ACTION:reload");
		ActionController controller = getController();
		controller.reload();
		return "";

	}
}

