// Decompiled by DJ v3.5.5.77 Copyright 2003 Atanas Neshkov  Date: 24.03.2004 15:52:10
// Home Page : http://members.fortunecity.com/neshkov/dj.html  - Check often for new version!
// Decompiler options: packimports(3) 
// Source File Name:   QuadTreeAlgorithm.java

package gd;

import graph.*;

// Referenced classes of package gd:
//            GDAlgorithm

public class QuadTreeAlgorithm extends GDAlgorithm
{

    public QuadTreeAlgorithm(double d, double d1)
    {
        super.springConstant = d;
        super.nodeCharge = d1;
    }

    public synchronized double improveGraph(Graph graph)
    {
        if(graph.sizeNodes == 0)
            return 0.0D;
        if(previousX == null || previousX.length < graph.sizeNodes)
        {
            previousX = new double[graph.sizeNodes];
            previousY = new double[graph.sizeNodes];
            moveX = new double[graph.sizeNodes];
            moveY = new double[graph.sizeNodes];
        }
        for(int i = 0; i < graph.sizeNodes; i++)
        {
            previousX[i] = graph.nodes[i].x;
            previousY[i] = graph.nodes[i].y;
        }

        double d = energy(graph);
        double d1 = graph.barycenter.x;
        double d2 = graph.barycenter.y;
        improveBarycenter(graph.barycenter);
        for(int j = 0; j < graph.sizeNodes; j++)
        {
            moveX[j] = graph.nodes[j].x - previousX[j];
            moveY[j] = graph.nodes[j].y - previousY[j];
        }

        improveAll(graph, d);
        graph.center(d1, d2);
        return energy(graph) - d;
    }

    private synchronized void moveAll(Graph graph, double d)
    {
        for(int i = 0; i < graph.sizeNodes; i++)
            graph.placeNode(graph.nodes[i], previousX[i] + d * moveX[i], previousY[i] + d * moveY[i]);

    }

    private synchronized void improveAll(Graph graph, double d)
    {
        double d1 = energy(graph);
        moveAll(graph, -2D);
        double d2 = energy(graph);
        double d3 = (d1 + d2) - 2D * d;
        if(d3 == 0.0D)
        {
            moveAll(graph, 2D);
            return;
        }
        double d4 = (d1 - d2) / 2D;
        double d5 = -d4 / (2D * d3);
        moveAll(graph, d5 + 2D);
        if(energy(graph) >= d)
            moveAll(graph, 1.0D - d5);
    }

    public synchronized double energy(Graph graph)
    {
        return energy(graph.barycenter);
    }

    private synchronized double energy(Barycenter barycenter)
    {
        if(barycenter.NW == null)
            return localEnergy(barycenter);
        else
            return groupEnergy(barycenter, 0.0D, 0.0D) + energy(barycenter.NW) + energy(barycenter.NE) + energy(barycenter.SW) + energy(barycenter.SE);
    }

    private synchronized void improveBarycenter(Barycenter barycenter)
    {
        if(barycenter.sizeNodes == 0)
            return;
        Coords coords = groupForce(barycenter, 0.0D, 0.0D);
        double d = Math.sqrt(My.square(coords.x) + My.square(coords.y));
        if(d != 0.0D)
        {
            coords.x /= d;
            coords.y /= d;
            double d1 = groupEnergy(barycenter, 0.0D, 0.0D);
            double d2 = quadApprox(barycenter, d1, coords);
            if(d2 == 0.0D)
            {
                double d3 = coords.x;
                coords.x = coords.y;
                coords.y = -d3;
                d2 += quadApprox(barycenter, d1 + d2, coords);
                if(d2 == 0.0D)
                {
                    coords.x = Math.random();
                    coords.y = Math.random();
                    d2 += quadApprox(barycenter, d1 + d2, coords);
                }
            }
        }
        if(barycenter.NW != null)
        {
            improveBarycenter(barycenter.NW);
            improveBarycenter(barycenter.NE);
            improveBarycenter(barycenter.SW);
            improveBarycenter(barycenter.SE);
            return;
        } else
        {
            locallyImprove(barycenter);
            return;
        }
    }

    private synchronized Coords groupForce(Barycenter barycenter, double d, double d1)
    {
        if(barycenter.parent == null)
            return new Coords(0.0D, 0.0D);
        else
            return repulsionForce(barycenter, d, d1).add(springForce(barycenter, d, d1));
    }

    private synchronized double groupEnergy(Barycenter barycenter, double d, double d1)
    {
        if(barycenter.parent == null)
            return 0.0D;
        else
            return repulsionEnergy(barycenter, d, d1) + springEnergy(barycenter, d, d1);
    }

    private synchronized Coords springForce(Node node, Edge edge, double d, double d1)
    {
        if(!edge.from.placed || !edge.to.placed)
            return new Coords(0.0D, 0.0D);
        double d2 = edge.to.x - edge.from.x;
        double d3 = edge.to.y - edge.from.y;
        if(node == edge.from)
        {
            d2 -= d;
            d3 -= d1;
        } else
        {
            d2 += d;
            d3 += d1;
        }
        double d4 = Math.sqrt(My.square(d2) + My.square(d3));
        double d5 = d4 - edge.restLength;
        double d6 = (2D * super.springConstant * d5) / d4;
        if(node == edge.to)
            d6 = -d6;
        Coords coords = new Coords(d6 * d2, d6 * d3);
        if(edge.directed && d3 > edge.restLength)
            coords.y -= (2D * super.springConstant * (d3 - edge.restLength)) / 2D;
        return coords;
    }

    private synchronized Coords springForce(Barycenter barycenter, double d, double d1)
    {
        Coords coords = new Coords(0.0D, 0.0D);
        for(int i = 0; i < barycenter.sizeEdges; i++)
        {
            Edge edge = barycenter.edges[i];
            if(edge.from.barycenter == barycenter)
                coords.add(springForce(edge.from, edge, d, d1));
            else
                coords.add(springForce(edge.to, edge, d, d1));
        }

        return coords;
    }

    private synchronized double springEnergy(Edge edge, Node node, double d, double d1)
    {
        if(!edge.from.placed || !edge.to.placed)
            return 0.0D;
        double d2 = edge.to.x - edge.from.x;
        double d3 = edge.to.y - edge.from.y;
        if(node == edge.from)
        {
            d2 -= d;
            d3 -= d1;
        } else
        {
            d2 += d;
            d3 += d1;
        }
        double d4 = Math.sqrt(My.square(d2) + My.square(d3));
        double d5 = super.springConstant * My.square(d4 - edge.restLength);
        if(edge.directed && d3 > edge.restLength)
            d5 -= (super.springConstant * My.square(d3 - edge.restLength)) / 2D;
        return d5;
    }

    private synchronized double springEnergy(Barycenter barycenter, double d, double d1)
    {
        double d2 = 0.0D;
        new Coords(0.0D, 0.0D);
        for(int i = 0; i < barycenter.sizeEdges; i++)
        {
            Edge edge = barycenter.edges[i];
            if(edge.from.barycenter == barycenter)
                d2 += springEnergy(edge, edge.from, d, d1);
            else
                d2 += springEnergy(edge, edge.to, d, d1);
        }

        return d2;
    }

    private synchronized Coords repulsionForce(Barycenter barycenter, double d, double d1)
    {
        if(barycenter == null || barycenter.sizeNodes == 0 || barycenter.parent == null)
            return new Coords(0.0D, 0.0D);
        int i = barycenter.sizeNodes;
        int j = barycenter.parent.sizeNodes;
        if(i == j)
            return new Coords(0.0D, 0.0D);
        double d2 = (barycenter.parent.x * (double)j - barycenter.x * (double)i) / (double)(j - i);
        double d3 = (barycenter.parent.y * (double)j - barycenter.y * (double)i) / (double)(j - i);
        double d4 = (barycenter.x + d) - d2;
        double d5 = (barycenter.y + d1) - d3;
        if(d4 == 0.0D && d5 == 0.0D)
        {
            d4 = Math.random() - 0.5D;
            d5 = Math.random() - 0.5D;
        }
        double d6 = My.square((barycenter.width + barycenter.parent.width + 1.0D) * d4) + My.square((barycenter.height + barycenter.parent.height + 1.0D) * d5);
        double d7 = (My.square(super.nodeCharge) * (double)i * barycenter.degree * (double)j * barycenter.parent.degree) / Math.pow(d6, 1.5D);
        return (new Coords(d7 * d4, d7 * d5)).add(repulsionForce(barycenter.parent, (d * (double)i) / (double)j, (d1 * (double)i) / (double)j));
    }

    private synchronized double repulsionEnergy(Barycenter barycenter, double d, double d1)
    {
        if(barycenter == null || barycenter.sizeNodes == 0 || barycenter.parent == null)
            return 0.0D;
        int i = barycenter.sizeNodes;
        int j = barycenter.parent.sizeNodes;
        if(i == j)
            return 0.0D;
        double d2 = (barycenter.parent.x * (double)j - barycenter.x * (double)i) / (double)(j - i);
        double d3 = (barycenter.parent.y * (double)j - barycenter.y * (double)i) / (double)(j - i);
        double d4 = (barycenter.x + d) - d2;
        double d5 = (barycenter.y + d1) - d3;
        if(d4 == 0.0D && d5 == 0.0D)
        {
            d4 = Math.random() - 0.5D;
            d5 = Math.random() - 0.5D;
        }
        double d6 = My.square((barycenter.width + barycenter.parent.width + 1.0D) * d4) + My.square((barycenter.height + barycenter.parent.height + 1.0D) * d5);
        double d7 = My.square(super.nodeCharge) * (double)i * barycenter.degree * (double)j * barycenter.parent.degree;
        return d7 / Math.sqrt(d6) + repulsionEnergy(barycenter.parent, (d * (double)i) / (double)j, (d1 * (double)i) / (double)j);
    }

    private synchronized double quadApprox(Barycenter barycenter, double d, Coords coords)
    {
        double d1 = groupEnergy(barycenter, coords.x, coords.y);
        double d2 = groupEnergy(barycenter, -coords.x, -coords.y);
        double d3 = (d1 + d2) - 2D * d;
        double d4 = (d1 - d2) / 2D;
        double d5 = -d4 / (2D * d3);
        double d6 = groupEnergy(barycenter, d5 * coords.x, d5 * coords.y);
        if(d6 < d)
        {
            barycenter.translate(d5 * coords.x, d5 * coords.y);
            return d - d6;
        } else
        {
            return 0.0D;
        }
    }

    private synchronized double locallyImprove(Barycenter barycenter)
    {
        double d = localEnergy(barycenter);
        for(int i = 0; i < barycenter.sizeNodes; i++)
            locallyImprove(barycenter.nodes[i]);

        return localEnergy(barycenter) - d;
    }

    private synchronized double locallyImprove(Node node)
    {
        if(node.degree == 0 || node.fixed || !node.placed)
            return 0.0D;
        Coords coords = localForce(node, 0.0D, 0.0D);
        double d = Math.sqrt(My.square(coords.x) + My.square(coords.y));
        if(d == 0.0D)
            return 0.0D;
        coords.x /= d;
        coords.y /= d;
        double d1 = localEnergy(node, 0.0D, 0.0D);
        double d2 = quadApprox(node, d1, coords);
        if(d2 == 0.0D)
        {
            double d3 = coords.x;
            coords.x = coords.y;
            coords.y = -d3;
            d2 += quadApprox(node, d1 + d2, coords);
            if(d2 == 0.0D)
            {
                coords.x = Math.random();
                coords.y = Math.random();
                d2 += quadApprox(node, d1 + d2, coords);
            }
        }
        return d2;
    }

    private synchronized Coords localForce(Node node, double d, double d1)
    {
        Coords coords = new Coords(0.0D, 0.0D);
        for(int i = 0; i < node.degree; i++)
            coords.add(springForce(node, node.incidentEdges[i], d, d1));

        Barycenter barycenter = node.barycenter;
        double d2 = barycenter.sizeNodes;
        if(d2 > 1.0D)
        {
            double d3 = (barycenter.x * d2 - node.x) / (d2 - 1.0D);
            double d4 = (barycenter.y * d2 - node.y) / (d2 - 1.0D);
            double d5 = (barycenter.width * d2 - (double)node.width) / (d2 - 1.0D);
            double d6 = (barycenter.height * d2 - (double)node.height) / (d2 - 1.0D);
            double d7 = (barycenter.degree * d2 - (double)node.degree) / (d2 - 1.0D);
            double d8 = (node.x + d) - d3;
            double d9 = (node.y + d1) - d4;
            if(d8 == 0.0D && d9 == 0.0D)
            {
                d8 = Math.random() - 0.5D;
                d9 = Math.random() - 0.5D;
            }
            double d10 = My.square(((double)node.width + d5 + 1.0D) * d8) + My.square(((double)node.height + d6 + 1.0D) * d9);
            double d11 = (My.square(super.nodeCharge) * (d2 - 1.0D) * (double)node.degree * d7) / Math.pow(d10, 1.5D);
            coords.x += d11 * d8;
            coords.y += d11 * d9;
        }
        Coords coords1 = repulsionForce(node.barycenter, d / d2, d1 / d2);
        coords.x += coords1.x;
        coords.y += coords1.y;
        return coords;
    }

    private synchronized double localEnergy(Node node, double d, double d1)
    {
        double d2 = 0.0D;
        for(int i = 0; i < node.degree; i++)
            d2 += springEnergy(node.incidentEdges[i], node, d, d1);

        Barycenter barycenter = node.barycenter;
        double d3 = barycenter.sizeNodes;
        if(d3 > 1.0D)
        {
            double d4 = (barycenter.x * d3 - node.x) / (d3 - 1.0D);
            double d6 = (barycenter.y * d3 - node.y) / (d3 - 1.0D);
            double d7 = (barycenter.width * d3 - (double)node.width) / (d3 - 1.0D);
            double d8 = (barycenter.height * d3 - (double)node.height) / (d3 - 1.0D);
            double d9 = (barycenter.degree * d3 - (double)node.degree) / (d3 - 1.0D);
            double d10 = (node.x + d) - d4;
            double d11 = (node.y + d1) - d6;
            if(d10 == 0.0D && d11 == 0.0D)
                return 1.7976931348623156E+306D;
            double d12 = My.square(((double)node.width + d7 + 1.0D) * d10) + My.square(((double)node.height + d8 + 1.0D) * d11);
            double d13 = My.square(super.nodeCharge) * (d3 - 1.0D) * (double)node.degree * d9;
            d2 += d13 / Math.sqrt(d12);
        }
        double d5 = repulsionEnergy(node.barycenter, d / d3, d1 / d3);
        return d2 + d5;
    }

    private synchronized double localEnergy(Barycenter barycenter)
    {
        double d = 0.0D;
        for(int i = 0; i < barycenter.sizeNodes; i++)
            d += localEnergy(barycenter.nodes[i], 0.0D, 0.0D);

        return d;
    }

    private synchronized double quadApprox(Node node, double d, Coords coords)
    {
        double d1 = localEnergy(node, coords.x, coords.y);
        double d2 = localEnergy(node, -coords.x, -coords.y);
        double d3 = (d1 + d2) - 2D * d;
        double d4 = (d1 - d2) / 2D;
        double d5 = -d4 / (2D * d3);
        double d6 = localEnergy(node, d5 * coords.x, d5 * coords.y);
        if(d6 < d)
        {
            double d7 = node.x + d5 * coords.x;
            double d8 = node.y + d5 * coords.y;
            node.barycenter.moveNode(node, d7, d8);
            node.x = d7;
            node.y = d8;
            return d - d6;
        } else
        {
            return 0.0D;
        }
    }

    private double previousX[];
    private double previousY[];
    private double moveX[];
    private double moveY[];
}